/*
 * astro.h
 ******************************************************************************
 * Copyright (C) 2002  Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. http://www.gnu.org
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *******************************************************************************
 * Defines data structures to be used for the astronomical calculations
 */

#ifndef _ASTRO
#define _ASTRO

#include "types.h"
/*******************************************************************************
 * Constants
 */

#define ASH_HMS_to_deg(X,Y,Z)	(double)(((X) + (Y)/60.0 + (Z)/3600.0)*360.0/24.0)
#define ASH_HMS_to_rad(X,Y,Z)	AS_deg_to_rad(ASH_HMS_to_deg(X,Y,Z))

#define ASH_DMS_to_deg(X,Y,Z)	(double)(((X) < 0.0) ? \
					 ((X) - (Y)/60.0 - (Z)/3600.0) : \
					 ((X) + (Y)/60.0 + (Z)/3600.0))
#define ASH_DMS_to_rad(X,Y,Z)	AS_deg_to_rad(ASH_DMS_to_deg(X,Y,Z))

#define ASH_AZ_to_int(X)	((int)((X)*3600.0))
#define ASH_ALT_to_int(X)	((int)((X)*900.0))

#define SID_TO_SOLAR_CONSTANT (1.00273790935) // multiply sidereal time by this to get solar time

enum PLANET_NAMES
{
  MERCURY = 0,
  VENUS,
  EARTH,
  MARS,
  JUPITER,
  SATURN,
  URANUS,
  NEPTUNE,
  MAX_PLANETS // Pluto not yet supported
} ;

/*******************************************************************************
 * Date and time is stored in the following structure
 */

struct  _time_struct
{
    Ushort      year ;
    Ushort      month ; /* 0=JAN, 1=FEB, ... */
    Ushort      date ;
    Ushort      hour ;
    Ushort      minute ;
    Ushort      seconds ;
} ;

typedef struct _time_struct  AS_TIME_STRUCT ;

#define ASH_DAY_LENGTH (60 * 60 * 24)

/*******************************************************************************
 * The following structure holds values which will be used or calculated by the
 * AS_apparent_position routine.
 */

struct _as_postiion_struct
{
  AS_TIME_STRUCT	GMT ;		// greenwich mean time
  double		Grnwch_SidTime ;// in units of radians

  double		obj_ra ;	// RA in radians  (0 .. 2*PI)
  double		obj_dec ;	// DEC in radians (-PI/2 .. +PI/2)
  double		apa_ra ;	// RA in radians  (0 .. 2*PI)
  double		apa_dec ;	// DEC in radians (-PI/2 .. +PI/2)
  double		apa_distance ;	// in A.U.
  double		delta_ra ;	// Correction to be ADDED to ra
  double		delta_dec ;	// Correction to be ADDED to dec
  double		altitude ;	// +ve above horizon
  double		azimuth ;	// measured from south

  /*
   * The following elements are for internal calculations and carry results
   * from one routine to the next.
   */

  double		delta_psi ;	// calculated by AS_nutation
  double		delta_epsilon ; // calculated by AS_nutation
  double		epsilon0 ;	// calculated by AS_nutation
  double		lon_sun ;	// calculated by AS_lon_sun
  double		e ;		// eccentricity of earth's orbit
  double		pi ;		// longitude of perihelion of orbit

  double		JD ;		// Julian ephemeris time
  double		JDE ;		// Julian ephemeris time
  double		T ;		// Julian ephemeris time - cubed
  double		T2 ;		// Julian ephemeris time - squared
  double		T3 ;		// Julian ephemeris time - cubed
  double		T4 ;		// Julian ephemeris time - power 4
} ;

typedef struct _as_postiion_struct ASH_POSITION_STRUCT ;

/*******************************************************************************
 * The following are structures to hold algorithmic data to be used in various
 * calculations
 */

struct _table22A
{
  char  D ;
  char  M ;
  char  m ;
  char  F ;
  char  O ;
  long  psi_const ;
  short psi_T ;
  long  eps_const ;
  char  eps_T ;
} ;

struct Table47A
{
  char D ;
  char M ;
  char m ;
  char F ;
  long sigma_l ;
  long sigma_r ;
} ;
    
struct Table47B
{
  char D ;
  char M ;
  char m ;
  char F ;
  long sigma_b ;
} ;

struct VSOP_87 //Refer to chapter 32 & appendix III
{
  double A ;
  double B ;
  double C ;
} ;

typedef const struct VSOP_87 *VSOP_PTR ;
typedef const VSOP_PTR VSOP_ptr_list[6] ;

struct PLANET //Refer to chapter 32 & appendix III
{
  const char name[8] ; // Just long enough to hold the longest name + '\0'
  const VSOP_ptr_list L ; // holds pointers to 0..5 element lists
  const VSOP_ptr_list B ;
  const VSOP_ptr_list R ;
} ;

extern const struct _table22A Nutation[] ;
extern const struct Table47A Lon_And_Distance[] ;
extern const struct Table47B Lat[] ;
extern const struct PLANET Planets[] ;

extern double AS_longitude ;
extern double AS_lattitude ;

/******************************************************************************
 *
 * Templates for each of the public routines are defined below...
 */

extern	Ulong		AS_rad_to_intern_ra (const double x) ;
extern	long		AS_rad_to_intern_dec (const double x) ;
extern	double		AS_intern_ra_to_rad (const Ulong x) ;
extern	double		AS_intern_dec_to_rad (const long x) ;
extern	char *		AS_display_intern_RA (char *ptr, const Ulong ra, const int accurate) ;
extern	char *		AS_display_intern_DEC (char *ptr, long dec, const int accurate) ;

extern	double		AS_normalise_degrees (double angle) ;
extern	double		AS_normalise_rads (double angle) ;
extern	double		AS_deg_to_rad (double x) ;
extern	double		AS_signed_deg_to_rad (double x) ;
extern	double		AS_rad_to_deg (double x) ;
extern	double		AS_signed_rad_to_deg (double x) ;

extern	void		AS_fill_time_struct (AS_TIME_STRUCT *t, const Ushort year, const Ushort month, const Ushort date, const Ushort hour, const Ushort minute, const Ushort seconds) ;
extern	void		AS_initialise_struct (ASH_POSITION_STRUCT *pos) ;
extern	char *		AS_rad_to_DMS (char *ptr, double rad) ;
extern	char *		AS_rad_to_HMS (char *ptr, double rad) ;
extern	char *		AS_display_RA (char *ptr, const double ra) ;
extern	char *		AS_display_date (char *ptr, const AS_TIME_STRUCT *ts) ;

extern	double		AS_julian_day (const AS_TIME_STRUCT *bin_struct) ;
extern	void		AS_date_from_julian (const double julian_time, AS_TIME_STRUCT *bin_struct) ;

extern	void		AS_greenwich_sidereal_time (ASH_POSITION_STRUCT *pos) ;
extern	void		AS_time_functions (ASH_POSITION_STRUCT *pos) ;

extern	void		AS_precession (ASH_POSITION_STRUCT *pos) ;
extern	void		AS_nutation (ASH_POSITION_STRUCT *pos) ;
extern	void		AS_lon_sun (ASH_POSITION_STRUCT *pos) ;
extern	void		AS_aberration (ASH_POSITION_STRUCT *pos) ;
extern	void		AS_allow_refraction (ASH_POSITION_STRUCT *pos) ;
extern	void		AS_find_apparent_location (ASH_POSITION_STRUCT *pos) ;
extern	void		AS_find_moon_apparent_location (ASH_POSITION_STRUCT *pos) ;
extern	void		AS_find_planet_apparent_location (ASH_POSITION_STRUCT *pos, const Uchar planet) ;

#endif
