/*
 * data.c
 ******************************************************************************
 */

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include "types.h"
#include "data.h"
#include "tree.h"
#include "readline.h"
#include "stringfn.h"
#include "sort.h"
#include "astro.h"
#include "database.h"

#define COMPRESSED "COMP.dat"

#define LIMITING_STAR_MAG  6.6	// Ignore stars dimmer than this
#define LIMITING_MAGNITUDE 8.0	// Ignore any object dimmer than this

#define LIM_STAR_MAG	(30.0+(LIMITING_STAR_MAG * 10.0))
#define LIM_MAG		(30.0+(LIMITING_MAGNITUDE * 10.0))

// **** NOTE **** The following constants must be big enough to hold all
//		  the objects you load! If you start to parse more and
//		  more files, increase these sizes to avoid segmentation
//		  faults!!!
#define MAX_NGC     0x10000
#define MAX_IC      0x10000
#define MAX_UGC     0x10000
#define MAX_HR      0x10000
#define MAX_SAO     0x80000
#define MAX_PGC     0x10000
#define MAX_HD      0x80000
#define MAX_PPM     0x80000
#define MAX_MESSIER 200 /* should only be 110! */
#define MAX_MATCHES 20

const TRH_TREE *NGC [MAX_NGC] ;
const TRH_TREE *IC  [MAX_IC] ;
const TRH_TREE *UGC [MAX_UGC] ;
const TRH_TREE *HR  [MAX_HR] ;
const TRH_TREE *SAO [MAX_SAO] ;
const TRH_TREE *PGC [MAX_PGC] ;
const TRH_TREE *HD  [MAX_HD] ;
const TRH_TREE *PPM [MAX_PPM] ;
const TRH_TREE *mes [MAX_MESSIER] ;
NUM_ENTRIES num ;

Ushort type_boundary ;

char *key_str ; /* BIG KLUDGE. This will point where the key goes in next */
TRH_TREE dummy[4] ;
char common_name [BUFFER_MAX_SIZE] ;
char formal_name [BUFFER_MAX_SIZE] ;
char description [BUFFER_MAX_SIZE] ;
char spectrum    [BUFFER_MAX_SIZE] ;

/*****************************************************************************
 */

static const char *
next_field (const char *current)
{
  while (*current && (*current != ','))
    current++ ;

  if (*current == ',')
    current++ ;

  return (current) ;
}

static void
strip_space (char *start_of_string, char *end_of_string)
{
  char *s, *e ;

  *(end_of_string+1) = '\0' ;

  for (e=end_of_string ; (e >= start_of_string) && isspace(*e) ; e--)
    *e = '\0' ;

  for (s=e=start_of_string ; *e && isspace(*e) ; e++) ;

  if (s != e)
    for (; (*s++ = *e++) ;) ;
}

void
capitalisation (char *string)
{
  char *p ;

  strip_space (p = string, string+strlen(string)-1) ;

  if (islower(*p)) /* capitalise the first section */
    *p = toupper(*p) ;

  if (*p) for (p++ ; *p ; p++) /* tolower the rest of first section */
    if (isupper(*p))
      *p = tolower(*p) ;
}

static void
strip_garbage (char *ptr)
{
  char *x, *y ;

  for (y=x=ptr ; *x ; x++) /* remove garbage from within */
  {
    if (isspace(*x) && (x == ptr)) /* strip leading space */
      continue ;

    if (*x < 0) /* ignore crazy characters */
      continue ;

    if (*x == '"') /* ignore quotes */
      continue ;

    if (*x == ';') /* convert semicolons */
      *x = ',' ;

    if ((*x == 'N') && (*(x+1) == '/') && (*(x+2) == 'A'))
    {
      x += 2 ;
      continue ;
    }

    *y++ = *x ;
  }

  *y = '\0' ;
  x = &ptr [strlen(ptr)-1] ;

  for (; (x >= ptr) && isspace(*x) ; x--) ; /* remove trailing white space */
  *(x+1) = '\0' ;

  if (*x == '.') /* remove trailing full stops */
    *x-- = '\0' ;

  for (y=x=ptr ; *x ;) /* remove contiguous white space */
    if (isspace(*y++ = *x++))
      for (; *x && isspace(*x) ; x++) ;

  *y = '\0' ;
}

Ushort
DAT_classify_string (const char *string, const STRING_CLASS *fields)
{
  Ushort i ;
  const char *str, *tok, *p, *s ;

  /* The zero entry in the STRING structure is always the null entry */
  /* Start searching from the 2nd entry */

  for (i=1 ; fields[i].keywords != (char *)0 ; i++) /* check each possible category */
  {
    for (tok=fields[i].keywords ; *tok ; tok=next_field(tok)) /* & each equiv token */
    {
      for (str=string ; *str ; str++) /* scan along string looking for token */
      {
	for (p=tok,s=str ; *p && *s && (*p != ',') ; p++,s++) /* scan string looking for tok */
	{
	  if (!*p || (*p == ',') || (*p != tolower(*s)))
	    break ;
	}

	if ((!*p || (*p == ',')) && (!*s || isspace(*s))) /* A MATCH! */
	  return i ;
      }
    }
  }

  return 0 ;
}

static int
process_delimiters (char *work, char *pointers[MAX_POINTERS], const char delimiter)
{
  int i ;
  Uchar *p, *q ;

  for (i=0 ; i<MAX_POINTERS ; i++)
    pointers[i] = "" ;

 /* convert all field delimiters to terminators and make a list of pointers */

  for (i=0,p=q=work ; (i < MAX_POINTERS) && *p ; p++)
  {
    if ((*p < ' ') || (*p > 0x7f))
      *p = ' ' ;

    if (*p == delimiter)
    {
      *p = '\0' ;
      pointers[i++] = q ;
      q = p+1 ;
    }
  }

  *p = '\0' ;
  pointers[i++] = q ;
  q = p+1 ;
  return i ;
}

void
compress_spaces (char *dest, const char *src)
{
  char *q ;
  const char *p ;

  q = dest ;
  p = src ;

  for (*q='\0' ; *p ; p++)
  {
    if (isspace(*p) && (q == dest)) /* strip leading spaces */
      continue ;

    if (isspace(*p) && isspace(*(p+1))) /* avoid multiple spaces */
      continue ;

    *q++ = *p ; /* then copy the character */
    *q = '\0' ;

    if (isdigit(*p) && !isspace(*(p+1)) && !isdigit(*(p+1))) /* add space if we need to */
    {
      *q++ = ' ' ;
      *q = '\0' ;
      continue ;
    }

    if (!isdigit(*p) && !isspace(*p) && isdigit(*(p+1))) /* add space if we need to */
    {
      *q++ = ' ' ;
      *q = '\0' ;
      continue ;
    }
  }

  if ((strlen(dest) > 0) && (dest[strlen(dest) - 1] == ' '))
    dest [strlen(dest) - 1] = '\0' ;
}

/*****************************************************************************
 * The database is written in the following format.
 *	Fields are separated by commas.
 *	Any comma within a field is converted to 0x7f
 *
 * key, messier #, NGC #, common name, RA, DEC, type,
 *	apparent size, magnitude, constellation, season,
 *	description
 *
 * RA is expressed as an unsigned integer representing fortieths of a second
 * starting from 0h 0.0m
 *
 * DEC is expressed as a signed integer representing seconds north or south.
 * South is negative.
 */

void
DAT_convert_data_format (OBJECT_INFO *o, const TRH_TREE *p)
{
  memset (o, '\0', sizeof(OBJECT_INFO)) ;
  o->common_name = o->formal_name = o->description = o->spectrum = &o->buffer[0] ;

  // Firstly, set RA, DEC information
  o->ra = p->RA ;
  o->dec = p->DEC ;

  if (p->RA_apm)
  {
    o->bit_fields |= BIT_RA_APM ;
    o->ra_apm = p->RA_apm ;
  }

  if (p->DEC_apm)
  {
    o->bit_fields |= BIT_DEC_APM ;
    o->dec_apm = p->DEC_apm ;
  }

  if (p->messier)
  {
    o->bit_fields |= BIT_MESSIER ;
    o->messier = p->messier ;
  }

  if (p->NGC)
  {
    o->bit_fields |= BIT_NGC ;
    o->ngc = p->NGC ;
  }

  if (p->NGC_comp)
  {
    o->bit_fields |= BIT_NGC_COMP ;
    o->ngc_comp = p->NGC_comp ;
  }

  if (p->UGC)
  {
    o->bit_fields |= BIT_UGC ;
    o->ugc = p->UGC ;
  }

  if (p->UGC_comp)
  {
    o->bit_fields |= BIT_UGC_COMP ;
    o->ugc_comp = p->UGC_comp ;
  }

  if (p->HR)
  {
    o->bit_fields |= BIT_HR ;
    o->hr = p->HR ;
  }

  if (p->IC)
  {
    o->bit_fields |= BIT_IC ;
    o->ic = p->IC ;
  }

  if (p->IC_comp)
  {
    o->bit_fields |= BIT_IC_COMP ;
    o->ic_comp = p->IC_comp ;
  }

  if (p->SAO)
  {
    o->bit_fields |= BIT_SAO ;
    o->sao = p->SAO ;
  }

  if (p->PGC)
  {
    o->bit_fields |= BIT_PGC ;
    o->pgc = p->PGC ;
  }

  if (p->HD)
  {
    o->bit_fields |= BIT_HD ;
    o->hd = p->HD ;
  }

  if (p->PPM)
  {
    o->bit_fields |= BIT_PPM ;
    o->ppm = p->PPM ;
  }

  if (p->constellation)
  {
    o->bit_fields |= BIT_CONST ;
    o->constellation = p->constellation ;
  }

  if (p->LY)
  {
    o->bit_fields |= BIT_LY ;
    o->ly = p->LY ;
  }

  if (p->rating)
  {
    o->bit_fields |= BIT_RATING ;
    o->rating = p->rating ;
  }

  if (p->season)
  {
    o->bit_fields |= BIT_SEASON ;
    o->season = p->season ;
  }

  if (p->type)
  {
    o->bit_fields |= BIT_TYPE ;
    o->type = p->type ;
  }

  if (p->angle)
  {
    o->bit_fields |= BIT_ANGLE ;
    o->angle = p->angle ;
  }

  if (p->magnitude)
  {
    o->bit_fields |= BIT_MAG ;
    o->mag =p->magnitude ;
  }

  if (p->mag_diff)
  {
    o->bit_fields |= BIT_MAG_DIFF ;
    o->mag_diff = p->mag_diff ;
  }

  if (p->radius)
  {
    o->bit_fields |= BIT_RADIUS ;
    o->radius = p->radius ;
  }

  if (p->separation)
  {
    o->bit_fields |= BIT_SEP ;
    o->sep = p->separation ;
  }

  if (p->minor)
  {
    o->bit_fields |= BIT_MINOR ;
    o->minor = p->minor ;
  }

  if (p->major)
  {
    o->bit_fields |= BIT_MAJOR ;
    o->major = p->major ;
  }

  if (p->num_components)
  {
    o->bit_fields |= BIT_NUM_COMP ;
    o->num_comp = p->num_components ;
  }

  if (*(p->common_name) != '\0')
  {
    o->bit_fields |= BIT_COMMON_NAME ;
    strcpy (o->common_name, p->common_name) ;
    o->formal_name = o->description = o->spectrum = o->buffer + strlen(o->common_name) + 1 ;
  }

  if (*(p->formal_name) != '\0')
  {
    o->bit_fields |= BIT_FORMAL_NAME ;
    strcpy (o->formal_name, p->formal_name) ;
    o->description = o->spectrum = o->formal_name + strlen(o->formal_name) + 1 ;
  }

  if (*(p->description) != '\0')
  {
    o->bit_fields |= BIT_DESCRIPTION ;
    strcpy (o->description, p->description) ;
    o->spectrum = o->description + strlen(o->description) + 1 ;
  }

  if (*(p->spectrum) != '\0')
  {
    o->bit_fields |= BIT_SPECTRUM ;
    strcpy (o->spectrum, p->spectrum) ;
  }

  if (p->alignment)
    o->bit_fields |= BIT_ALIGNMENT ;
}

int
DAT_display_node (const TRH_TREE *node)
{
  OBJECT_INFO o ;

  if (node == TRH_NULL)
  {
    printf ("Trying to print a NULL node!\n") ;
    return 1 ;
  }

  printf ("\n---------\"%s\"---------\n", node->key) ;
  DAT_convert_data_format (&o, node) ;
  DB_display_object_info (&o, stdout) ;
  return 0 ;
}

void
display_list (TRH_TREE *array[], int num_items)
{
  int i ;

  for (i=0 ; i < num_items ; i++)
    if (DAT_display_node (array[i]))
      break ;
}

void
DAT_write_compressed_database (TRH_TREE *tree)
{
  FILE *f ;
  const TRH_TREE *p ;
  int i ;
  OBJECT_INFO o ;

  // Firstly, count the items we are going to write
  num.ngc = num.hr = num.ic = num.sao = num.pgc = num.ugc = num.hd = num.ppm
    = num.messier = num.align = 0 ;

  p = TR_find_nth (tree, 1) ;

  for (i=0; p != TRH_NULL ; p=p->next, i++)
  {
    if ((p->magnitude > LIM_MAG) && !(p->messier))
    {
      i-- ;
      continue ;
    }

    if (p->rating > 5)
    {
      i-- ;
      continue ;
    }

    if (p->NGC != 0)
      num.ngc++ ;

    if (p->IC != 0)
      num.ic++ ;

    if (p->UGC != 0)
      num.ugc++ ;

    if (p->HR != 0)
      num.hr++ ;

    if (p->SAO != 0)
      num.sao++ ;

    if (p->PGC != 0)
      num.pgc++ ;

    if (p->HD != 0)
      num.hd++ ;

    if (p->PPM != 0)
      num.ppm++ ;

    if (p->messier != 0)
      num.messier++ ;

    if (p->alignment != 0)
      num.align++ ;
  }

  // Now write the real database

  if ((f = fopen (COMPRESSED, "w")) == (FILE *)0)
  {
    fprintf (stderr, "Can't open " COMPRESSED " for writing\n") ;
    exit (1) ;
  }

  // Firstly, write the sizes of the database lists
  DB_write_header (f, &num) ;

  p = TR_find_nth (tree, 1) ;

  for (i=0; p != TRH_NULL ; p=p->next, i++)
  {
    if ((p->magnitude > LIM_MAG) && !(p->messier))
    {
      i-- ;
      continue ;
    }

    if (p->rating > 5)
    {
      i-- ;
      continue ;
    }

    DAT_convert_data_format (&o, p) ;
    DB_write_object_info (&o, f) ;
  }

  fclose (f) ;
  fprintf (stdout, "Wrote %d items to " COMPRESSED "\n", i) ;
}

/*****************************************************************************
 */

void
zero_structure (TRH_TREE *node)
{
  node->messier = 0 ;
  node->constellation = 0 ;
  node->season = 0 ;
  node->type = 0 ;
  node->alignment = 0 ;
  node->magnitude = 0 ;
  node->mag_diff = 0 ;
  node->num_components = 0 ;
  node->NGC_comp = 0 ;
  node->IC_comp = 0 ;
  node->UGC_comp = 0 ;
  node->rating = 0 ;
  node->angle = 0 ;
  node->major = 0 ;
  node->minor = 0 ;

  node->NGC = 0 ;
  node->HR = 0 ;
  node->IC = 0 ;
  node->RA = 0 ;
  node->DEC = 0 ;
  node->RA_apm = 0 ;
  node->DEC_apm = 0 ;
  node->separation = 0 ;
  node->radius = 0 ;

  node->SAO = 0 ;
  node->PGC = 0 ;
  node->UGC = 0 ;
  node->HD = 0 ;
  node->PPM = 0 ;
  node->LY = 0 ;

  node->common_name = TRH_USER_PTR(node) ;
  *node->common_name = '\0' ;

  node->formal_name =
    node->description = 
      node->spectrum = node->common_name ;

  if (node == &dummy[0])
  {
    key_str = &node->key[0] ;
    *key_str = '\0' ;
    *common_name = '\0' ;
    *formal_name = '\0' ;
    *description = '\0' ;
    *spectrum = '\0' ;

    node->common_name = common_name ;
    node->formal_name = formal_name ;
    node->description = description ;
    node->spectrum = spectrum ;
  }
}

void
copy_fixed_elements (TRH_TREE *to, const TRH_TREE *from)
{
  to->messier = from->messier ;
  to->constellation = from->constellation ;
  to->season = from->season ;
  to->type = from->type ;
  to->magnitude = from->magnitude ;
  to->mag_diff = from->mag_diff ;
  to->num_components = from->num_components ;
  to->NGC_comp = from->NGC_comp ;
  to->IC_comp = from->IC_comp ;
  to->UGC_comp = from->UGC_comp ;
  to->rating = from->rating ;
  to->angle = from->angle ;
  to->major = from->major ;
  to->minor = from->minor ;

  to->NGC = from->NGC ;
  to->HR = from->HR ;
  to->IC = from->IC ;
  to->RA = from->RA ;
  to->DEC = from->DEC ;
  to->RA_apm = from->RA_apm ;
  to->DEC_apm = from->DEC_apm ;
  to->alignment = from->alignment ;
  to->separation = from->separation ;
  to->radius = from->radius ;

  to->SAO = from->SAO ;
  to->PGC = from->PGC ;
  to->UGC = from->UGC ;
  to->HD = from->HD ;
  to->PPM = from->PPM ;
  to->LY = from->LY ;
}

void
copy_node (TRH_TREE *to, TRH_TREE *from)
{
  copy_fixed_elements (to, from) ;

  to->common_name = TRH_USER_PTR(to) ;
  to->formal_name = to->common_name + strlen(from->common_name) + 1 ;
  to->description = to->formal_name + strlen(from->formal_name) + 1 ;
  to->spectrum = to->description + strlen(from->description) + 1 ;

  strcpy (to->common_name, from->common_name) ;
  strcpy (to->formal_name, from->formal_name) ;
  strcpy (to->description, from->description) ;
  strcpy (to->spectrum, from->spectrum) ;
}

void
ccc (const TRH_TREE *x, const char c)
{
  if (strlen (x->description) > (BUFFER_MAX_SIZE-100))
  {
    fprintf (stderr, "Error @ %c. LEN=%d\n", c, strlen(x->description)) ;
    x->description[300] = '\0' ;
    DAT_display_node (x) ;
    exit (1) ;
  }
}

TRH_TREE *
merge_nodes (TRH_TREE *node1, TRH_TREE *node2, int disp_err)
{
  TRH_TREE *new ;
  int j ;
  char backup_common_name [BUFFER_MAX_SIZE] ;
  char backup_formal_name [BUFFER_MAX_SIZE] ;
  char backup_description [BUFFER_MAX_SIZE] ;
  char backup_spectrum    [BUFFER_MAX_SIZE] ;
  int error ;

  strcpy (backup_common_name, node1->common_name) ;
  strcpy (backup_formal_name, node1->formal_name) ;
  strcpy (backup_description, node1->description) ;
  strcpy (backup_spectrum, node1->spectrum) ;

  if (*(node1->common_name) == '\0')
    strcpy (backup_common_name, node2->common_name) ;

  if (*(node1->formal_name) == '\0')
    strcpy (backup_formal_name, node2->formal_name) ;

  if (*(node1->description) == '\0')
  {
    if (*(node2->description) != '\0')
      strcpy (backup_description, node2->description) ;
  }
  else if (*(node2->description) != '\0')
  {
    if (strcmp (node1->description, node2->description) != 0)
      STR_snaprintf (backup_description, BUFFER_MAX_SIZE, " %s", node2->description) ;
  }

  if (*node1->spectrum == '\0')
    strcpy (backup_spectrum, node2->spectrum) ;

  j = strlen (backup_common_name) + 1
    + strlen (backup_formal_name) + 1
    + strlen (backup_description) + 1
    + strlen (backup_spectrum) + 1 ;

  if (j > BUFFER_MAX_SIZE)
  {
    fprintf (stderr, "Error. length of strings in entry is too long (%d bytes)\n", j) ;
    fprintf (stderr, "Entry=<%s>\n", node1->key) ;
    *(node1->description) = *(node2->description) = '\0' ;
    DAT_display_node (node1) ;
    DAT_display_node (node2) ;
    exit (1) ;
  }

  new = TR_create_node (node1->key, j) ;
  copy_fixed_elements (new, node1) ;
  error = 0 ;

  if (new->RA == 0)
    new->RA = node2->RA ;
  else if ((node2->RA != 0) && (new->RA != node2->RA))
    error++ ;

  if (new->DEC == 0)
    new->DEC = node2->DEC ;
  else if ((node2->DEC != 0) && (new->DEC != node2->RA))
    error++ ;

  if (new->alignment == 0)
    new->alignment = node2->alignment ;

  if (new->RA_apm == 0)
    new->RA_apm = node2->RA_apm ;
  else if ((node2->RA_apm != 0) && (new->RA_apm != node2->RA_apm))
    error++ ;

  if (new->DEC_apm == 0)
    new->DEC_apm = node2->DEC_apm ;
  else if ((node2->DEC_apm != 0) && (new->DEC_apm != node2->DEC_apm))
    error++ ;

  if (new->messier == 0)
    new->messier = node2->messier ;
  else if ((node2->messier != 0) && (new->messier != node2->messier))
    error++ ;

  if (new->constellation == 0)
    new->constellation = node2->constellation ;
  else if ((node2->constellation != 0) && (new->constellation != node2->constellation))
    error++ ;

  if (new->season == 0)
    new->season = node2->season ;
  else if ((node2->season != 0) && (new->season != node2->season))
    error++ ;

  if (new->type == 0)
    new->type = node2->type ;
  else if ((new->type > type_boundary) && (node2->type != 0) && (node2->type < type_boundary))
    new->type = node2->type ;
  else if ((node2->type != 0) && (new->type != node2->type))
    error++ ;

  if (new->magnitude == 0)
    new->magnitude = node2->magnitude ;
  else if ((node2->magnitude != 0) && (new->magnitude != node2->magnitude))
    error++ ;

  if (new->mag_diff == 0)
    new->mag_diff = node2->mag_diff ;
  else if ((node2->mag_diff != 0) && (new->mag_diff != node2->mag_diff))
    error++ ;

  if (new->num_components == 0)
    new->num_components = node2->num_components ;
  else if ((node2->num_components != 0) && (new->num_components != node2->num_components))
    error++ ;

  if (new->NGC_comp == 0)
    new->NGC_comp = node2->NGC_comp ;
  else if ((node2->NGC_comp != 0) && (new->NGC_comp != node2->NGC_comp))
    error++ ;

  if (new->IC_comp == 0)
    new->IC_comp = node2->IC_comp ;
  else if ((node2->IC_comp != 0) && (new->IC_comp != node2->IC_comp))
    error++ ;

  if (new->UGC_comp == 0)
    new->UGC_comp = node2->UGC_comp ;
  else if ((node2->UGC_comp != 0) && (new->UGC_comp != node2->UGC_comp))
    error++ ;

  if (new->rating == 0)
    new->rating = node2->rating ;
  else if ((node2->rating != 0) && (new->rating != node2->rating))
    error++ ;

  if (new->angle == 0)
    new->angle = node2->angle ;
  else if ((node2->angle != 0) && (new->angle != node2->angle))
    error++ ;

  if (new->major == 0)
    new->major = node2->major ;
  else if ((node2->major != 0) && (new->major != node2->major))
    error++ ;

  if (new->minor == 0)
    new->minor = node2->minor ;
  else if ((node2->minor != 0) && (new->minor != node2->minor))
    error++ ;

  if (new->NGC == 0)
    new->NGC = node2->NGC ;
  else if ((node2->NGC != 0) && (new->NGC != node2->NGC))
    error++ ;

  if (new->HR == 0)
    new->HR = node2->HR ;
  else if ((node2->HR != 0) && (new->HR != node2->HR))
    error++ ;

  if (new->IC == 0)
    new->IC = node2->IC ;
  else if ((node2->IC != 0) && (new->IC != node2->IC))
    error++ ;

  if (new->separation == 0)
    new->separation = node2->separation ;
  else if ((node2->separation != 0) && (new->separation != node2->separation))
    error++ ;

  if (new->radius == 0)
    new->radius = node2->radius ;
  else if ((node2->radius != 0) && (new->radius != node2->radius))
    error++ ;

  if (new->SAO == 0)
    new->SAO = node2->SAO ;
  else if ((node2->SAO != 0) && (new->SAO != node2->SAO))
    error++ ;

  if (new->PGC == 0)
    new->PGC = node2->PGC ;
  else if ((node2->PGC != 0) && (new->PGC != node2->PGC))
    error++ ;

  if (new->UGC == 0)
    new->UGC = node2->UGC ;
  else if ((node2->UGC != 0) && (new->UGC != node2->UGC))
    error++ ;

  if (new->HD == 0)
    new->HD = node2->HD ;
  else if ((node2->HD != 0) && (new->HD != node2->HD))
    error++ ;

  if (new->PPM == 0)
    new->PPM = node2->PPM ;
  else if ((node2->PPM != 0) && (new->PPM != node2->PPM))
    error++ ;

  if (new->LY == 0)
    new->LY = node2->LY ;
  else if ((node2->LY != 0) && (new->LY != node2->LY))
    error++ ;

  /* now do some reclassification kludges */

  if ((new->type > type_boundary) &&
      (new->mag_diff || new->num_components || new->separation))
    new->type = DAT_classify_string ("multiple", DB_obj_type) ;

  new->common_name = TRH_USER_PTR(new) ;
  new->formal_name = new->common_name + strlen(backup_common_name) + 1 ;
  new->description = new->formal_name + strlen(backup_formal_name) + 1 ;
  new->spectrum    = new->description + strlen(backup_description) + 1 ;

  strcpy (new->common_name, backup_common_name) ;
  strcpy (new->formal_name, backup_formal_name) ;
  strcpy (new->description, backup_description) ;
  strcpy (new->spectrum, backup_spectrum) ;

  if (disp_err && error)
  {
    printf ("***********************************ERROR COUNT=%d\n", error) ;
    DAT_display_node (node1) ;
    DAT_display_node (node2) ;
  }

  return new ;
}

/******************************************************************************
 */

int
cmp_NGC (void *a, void *b)
{
  if (((TRH_TREE *)a)->NGC > ((TRH_TREE *)b)->NGC)
    return 1;

  if (((TRH_TREE *)a)->NGC < ((TRH_TREE *)b)->NGC)
    return -1;

  if (((TRH_TREE *)a)->NGC_comp > ((TRH_TREE *)b)->NGC_comp)
    return 1;

  if (((TRH_TREE *)a)->NGC_comp < ((TRH_TREE *)b)->NGC_comp)
    return -1;

  return 0;
}

int
cmp_IC (void *a, void *b)
{
  if (((TRH_TREE *)a)->IC > ((TRH_TREE *)b)->IC)
    return 1;

  if (((TRH_TREE *)a)->IC < ((TRH_TREE *)b)->IC)
    return -1;

  if (((TRH_TREE *)a)->IC_comp > ((TRH_TREE *)b)->IC_comp)
    return 1;

  if (((TRH_TREE *)a)->IC_comp < ((TRH_TREE *)b)->IC_comp)
    return -1;

  return 0;
}

int
cmp_UGC (void *a, void *b)
{
  if (((TRH_TREE *)a)->UGC > ((TRH_TREE *)b)->UGC)
    return 1;

  if (((TRH_TREE *)a)->UGC < ((TRH_TREE *)b)->UGC)
    return -1;

  if (((TRH_TREE *)a)->UGC_comp > ((TRH_TREE *)b)->UGC_comp)
    return 1;

  if (((TRH_TREE *)a)->UGC_comp < ((TRH_TREE *)b)->UGC_comp)
    return -1;

  return 0;
}

int
cmp_HR (void *a, void *b)
{
  if (((TRH_TREE *)a)->HR > ((TRH_TREE *)b)->HR)
    return 1;

  if (((TRH_TREE *)a)->HR < ((TRH_TREE *)b)->HR)
    return -1;

  return 0;
}

int
cmp_SAO (void *a, void *b)
{
  if (((TRH_TREE *)a)->SAO > ((TRH_TREE *)b)->SAO)
    return 1;

  if (((TRH_TREE *)a)->SAO < ((TRH_TREE *)b)->SAO)
    return -1;

  return 0;
}

int
cmp_PGC (void *a, void *b)
{
  if (((TRH_TREE *)a)->PGC > ((TRH_TREE *)b)->PGC)
    return 1;

  if (((TRH_TREE *)a)->PGC < ((TRH_TREE *)b)->PGC)
    return -1;

  return 0;
}

int
cmp_HD (void *a, void *b)
{
  if (((TRH_TREE *)a)->HD > ((TRH_TREE *)b)->HD)
    return 1;

  if (((TRH_TREE *)a)->HD < ((TRH_TREE *)b)->HD)
    return -1;

  return 0;
}

int
cmp_PPM (void *a, void *b)
{
  if (((TRH_TREE *)a)->PPM > ((TRH_TREE *)b)->PPM)
    return 1;

  if (((TRH_TREE *)a)->PPM < ((TRH_TREE *)b)->PPM)
    return -1;

  return 0;
}

int
cmp_mes (void *a, void *b)
{
  if (((TRH_TREE *)a)->messier > ((TRH_TREE *)b)->messier)
    return 1;

  if (((TRH_TREE *)a)->messier < ((TRH_TREE *)b)->messier)
    return -1;

  return 0;
}

void
build_cross_reference (const TRH_TREE *tree, const int disp, const int rebuild)
{
  const TRH_TREE *node ;

  if (tree == TRH_NULL)
  {
    printf ("Build Cross Reference: called with a null pointer!\n") ;
    return ;
  }

  if (rebuild)
  {
    num.ngc = num.hr = num.ic = num.sao = num.pgc = num.ugc = num.hd = num.ppm
      = num.messier = num.align = 0 ;

    node = TR_find_nth (tree, 1) ;

    while (node != TRH_NULL)
    {
      if (node->NGC != 0)
	NGC[num.ngc++] = node ;

      if (node->IC != 0)
	IC[num.ic++] = node ;

      if (node->UGC != 0)
	UGC[num.ugc++] = node ;

      if (node->HR != 0)
	HR[num.hr++] = node ;

      if (node->SAO != 0)
	SAO[num.sao++] = node ;

      if (node->PGC != 0)
	PGC[num.pgc++] = node ;

      if (node->HD != 0)
	HD[num.hd++] = node ;

      if (node->PPM != 0)
	PPM[num.ppm++] = node ;

      if (node->messier != 0)
	mes[num.messier++] = node ;

      if (node->alignment != 0)
	num.align++ ;

      node = node->next ;
    }
  }

  if (disp)
  {
    fprintf (stdout, "Current tallies: PPM=%lu, NGC=%lu, IC=%lu, UGC=%lu\n",
	   num.ppm, num.ngc, num.ic, num.ugc) ;
    fprintf (stdout, "\tHR=%lu, SAO=%lu, PGC=%lu, HD=%lu, messier=%lu\n",
	   num.hr, num.sao, num.pgc, num.hd, num.messier) ;
  }

  if (num.ngc > 0)
    SOR_qsort ((void *)NGC, cmp_NGC, 0, num.ngc-1) ;

  if (num.ic > 0)
    SOR_qsort ((void *)IC,  cmp_IC,  0, num.ic-1) ;

  if (num.ugc > 0)
    SOR_qsort ((void *)UGC, cmp_UGC, 0, num.ugc-1) ;

  if (num.hr > 0)
    SOR_qsort ((void *)HR,  cmp_HR,  0, num.hr-1) ;

  if (num.sao > 0)
    SOR_qsort ((void *)SAO, cmp_SAO, 0, num.sao-1) ;

  if (num.pgc > 0)
    SOR_qsort ((void *)PGC, cmp_PGC, 0, num.pgc-1) ;

  if (num.hd > 0)
    SOR_qsort ((void *)HD,  cmp_HD,  0, num.hd-1) ;

  if (num.ppm > 0)
    SOR_qsort ((void *)PPM,  cmp_PPM,  0, num.ppm-1) ;

  if (num.messier > 0)
    SOR_qsort ((void *)mes, cmp_mes, 0, num.messier-1) ;
}

int
process_cross_reference (TRH_TREE *node, TRH_TREE **tree, int disp_err, const int rec)
{
  Ulong ret ;
  TRH_TREE *new ;
  const TRH_TREE *old ;
  int i, j ;
  const TRH_TREE **list[MAX_MATCHES] ;
  int num_matches ;

  // The purpose of this routine is to take a newly defined node and search
  // all databases to see if there are any other matches. If so, merge the
  // nodes into one and rebuild all the cross references.

  if (node == TRH_NULL)
  {
    printf ("Process Cross Reference: called with a null pointer!\n") ;
    return -1 ;
  }

  // create a real node based upon the input (which is not a real node)
  i = strlen (node->common_name) + 1
    + strlen (node->formal_name) + 1
    + strlen (node->description) + 1
    + strlen (node->spectrum) + 1 ;

  new = TR_create_node (node->key, i) ;
  copy_node (new, node) ;
  node = new ;

  num_matches = 0 ;

  // search each known list to find first match
  if ((ret=SOR_search ((void *)NGC, cmp_NGC, num.ngc, (void *)node)) != SOR_FAIL)
    list[num_matches++] = &NGC[ret] ;

  if ((ret=SOR_search ((void *)HD, cmp_HD, num.hd, (void *)node)) != SOR_FAIL)
    list[num_matches++] = &HD[ret] ;

  if ((ret=SOR_search ((void *)PPM, cmp_PPM, num.ppm, (void *)node)) != SOR_FAIL)
    list[num_matches++] = &PPM[ret] ;

  if ((ret=SOR_search ((void *)SAO, cmp_SAO, num.sao, (void *)node)) != SOR_FAIL)
    list[num_matches++] = &SAO[ret] ;

  if ((ret=SOR_search ((void *)IC, cmp_IC, num.ic, (void *)node)) != SOR_FAIL)
    list[num_matches++] = &IC[ret] ;

  if ((ret=SOR_search ((void *)UGC, cmp_UGC, num.ugc, (void *)node)) != SOR_FAIL)
    list[num_matches++] = &UGC[ret] ;

  if ((ret=SOR_search ((void *)PGC, cmp_PGC, num.pgc, (void *)node)) != SOR_FAIL)
    list[num_matches++] = &PGC[ret] ;

  if ((ret=SOR_search ((void *)HR, cmp_HR, num.hr, (void *)node)) != SOR_FAIL)
    list[num_matches++] = &HR[ret] ;

  if ((ret=SOR_search ((void *)mes, cmp_mes, num.messier, (void *)node)) != SOR_FAIL)
    list[num_matches++] = &mes[ret] ;

  if (num_matches == 0)
  {
    // this next bit is why this programme runs so slowly - slow to add nodes
    TR_add_node (tree, node) ;
    build_cross_reference (*tree, 0, 1) ; // ouch
    return 0 ;
  }

  // num_matches == number of lists containing matches to <node>
  for (i=0 ; i < num_matches ; i++)
  {
    if (list[i] == NULL)
      continue ;

    old = *(list[i]) ;
    new = merge_nodes (old, node, disp_err) ; // keep key from old node

    for (j=i ; j < num_matches ; j++)
      if (*(list[j]) == old)
	list[j] = NULL ;

    // kill the old node, remember the new
    TR_delete_node (tree, old->key) ;
    free (node) ;
    node = new ;
  }

  TR_add_node (tree, node) ;
  build_cross_reference (*tree, 0, 1) ; // This is the call which slows everything down!
  return 0 ;
}


/*****************************************************************************
 * Routines to read RA
 */

int
ra_H_space_x_dot_y_M_S (const char *src, Ulong *ptr)
{
  int i, j, k ;
  float l ;
  double val ;

  i = j = k = 0 ;
  l = 0.0 ;

  if (sscanf (src, "%dh %fm %ds", &i, &l, &k) < 2)
  {
    printf ("ra_H_space_x_dot_y_M_S: %s\n", src) ;
    return -1 ;
  }

  if (*ptr)
    return 0 ;

  /* Either L will contain seconds or J, but not both */

  val = (i*3600.0 + l*60.0 + k) / ASH_DAY_LENGTH * 2.0 * M_PI;
  *ptr = AS_rad_to_intern_ra (val) ;
  return 0 ;
}

int
ra_h_space_min_space_sec (const char *src, Ulong *ptr)
{
  int i, j ;
  float l ;
  double val ;

  i = j = 0 ;
  l = 0.0 ;

  if (sscanf (src, "%d %d %f", &i, &j, &l) != 3)
  {
    printf ("\nra_h_space_min_space_sec: %s\n", src) ;
    return -1 ;
  }

  if (*ptr)
    return 0 ;

  val = (i*3600.0 + j*60.0 + l) / ASH_DAY_LENGTH * 2.0 * M_PI;
  *ptr = AS_rad_to_intern_ra (val) ;
  return 0 ;
}

int
ra_h_space_min_dot_sec (const char *src, Ulong *ptr)
{
  int i ;
  float l ;
  double val ;

  i = 0 ;
  l = 0.0 ;

  if (sscanf (src, "%d %f", &i, &l) != 2)
  {
    printf ("ra_h_space_min_dot_sec: %s\n", src) ;
    return -1 ;
  }

  if (*ptr)
    return 0 ;

  val = (i*3600.0 + l*60.0) / ASH_DAY_LENGTH * 2.0 * M_PI;
  *ptr = AS_rad_to_intern_ra (val) ;
  return 0 ;
}

/*****************************************************************************
 * Routines to read DEC (in tens of seconds)
 */

int
dec_sign_DMS (const char *src, long *ptr)
{
  int i, j ;
  int sign ;
  float l, m ;
  double val ;
  const char *p ;

  i = j = 0 ;
  l = m = 0.0 ;

  if (*src == '-')
  {
    p = src+1 ;
    sign = -1 ;
  }
  else if (*src == '+')
  {
    p = src+1 ;
    sign = 1 ;
  }
  else
  {
    p = src ;
    sign = 1 ;
  }

  if (sscanf (p, "%d %f %f", &i, &l, &m) < 2)
  {
    printf ("dec_sign_DMS: %s\n", src) ;
    return -1 ;
  }

  if (*ptr)
    return 0 ;

  val = sign * (i + l/60.0 + m/3600.0) / 90.0 * M_PI / 2.0 ;
  *ptr = AS_rad_to_intern_dec (val) ;
  return 0 ;
}

int
dec_deg_space_min_NS (const char *src, long *ptr)
{
  int i, j ;
  float l ;
  char *p ;
  char c ;
  double val ;

  i = j = 0 ;
  l = 0.0 ;
  c = ' ' ;

  p = strchr (src, '\'') ;

  if ((p == NULL) ||
      (sscanf (src, "%d %f", &i, &l) != 2))
  {
    printf ("(A)dec_deg_space_min_NS: %s\n", src) ;
    return -1 ;
  }

  if ((sscanf (p, "'%d %c", &j, &c) != 2) && (sscanf (p, "'%c", &c) != 1))
  {
    printf ("(B)dec_deg_space_min_NS: %s\n", src) ;
    return -1 ;
  }

  if (*ptr)
    return 0 ;

  val = (i + l/60.0 + j/3600.0) / 90.0 * M_PI / 2.0 ;

  if (c == 'S')
    val = -val ;

  *ptr = AS_rad_to_intern_dec (val) ;
  return 0 ;
}

/*****************************************************************************
 * Routines to read numbers
 */

int
uchar_a_dot_b (const char *src, Uchar *ptr, const Uchar multiplier, const int ismag)
{
  float l ;

  if (*src == '\0')
    return 0 ;

  l = 0.0 ;

  if (sscanf (src, "%f", &l) < 1)
  {
    printf ("uchar_a_dot_b: %s\n", src) ;
    return -1 ;
  }

  if (*ptr)
    return 0 ;

  if ((ismag == 2) && (l < -5.0))
    return 0 ;

  if ((ismag != 2) && (l < 0.0))
    return 0 ;

  if (ismag)
    l = 30.0 + (l * (float)multiplier) ;
  else
    l = l * (float)multiplier ;

  if (fmod(l,1.0) >= 0.5)
    *ptr = (Uchar)(l + 1.0) ;
  else
    *ptr = (Uchar)l ;

  return 0 ;
}

int
char_apm (const char *src, char *ptr, const int RA)
{
  float l ;
  char sign ;

  if (*src == '\0')
    return 0 ;

  l = 0.0 ;
  sign = 1 ;

  if (sscanf (src, "%f", &l) < 1)
  {
    printf ("char_apm: %s\n", src) ;
    return -1 ;
  }

  if (RA)
    l *= 300.0 ;
  else
    l *= 20.0 ;

  if (l >= 126.5)
  {
    printf ("\nchar_apm: Truncating %f to 127\n", l) ;
    l = 127.0 ;
  }
  else if (l <= -127.5)
  {
    printf ("\nchar_apm: Truncating %f to -128\n", l) ;
    l = -128.0 ;
  }

  if (l < 0.0)
  {
    l = -l ;
    sign = -1 ;
  }

  if (fmod(l,1.0) >= 0.5)
    *ptr = (char)(l + 1.0) * sign ;
  else
    *ptr = (char)l * sign ;

  return 0 ;
}

int
uchar_a_x_b (const char *src, Uchar *ptr1, Uchar *ptr2)
{
  float l ;
  char *p ;

  if (*src == '\0')
    return 0 ;

  l = 0.0 ;

  if (sscanf (src, "%f", &l) < 1)
  {
    printf ("uchar_a_x_b: %s\n", src) ;
    return -1 ;
  }

  if (*ptr1)
    return 0 ;

  if (l < 0.0)
    return 0 ;

  if (fmod(l,1.0) >= 0.5)
    *ptr1 = (Uchar)(l + 1.0) ;
  else
    *ptr1 = (Uchar)l ;

  /* we need to read the major and minor separately because scanf doesn't like
   * the x character in the middle of two %fs */

  p = strchr (src, 'x') ;

  if ((p != NULL) && (sscanf (p, "x%f", &l) == 1))
  {
    if (fmod(l,1.0) >= 0.5)
      *ptr2 = (Uchar)(l + 1.0) ;
    else
      *ptr2 = (Uchar)l ;
  }

  return 0 ;
}

int
ushort_a_dot_b (const char *src, Ushort *ptr, const Ushort multiplier, const int ismag)
{
  float l ;

  if (*src == '\0')
    return 0 ;

  l = 0.0 ;

  if (sscanf (src, "%f", &l) < 1)
  {
    printf ("ushort_a_dot_b: %s\n", src) ;
    return -1 ;
  }

  if (*ptr)
    return 0 ;

  if (l < 5.0)
    return 0 ;

  if (ismag)
    l = 30.0 + (l * (float)multiplier) ;
  else
    l = l * (float)multiplier ;

  if (fmod(l,1.0) >= 0.5)
    *ptr = (Ushort)(l + 1.0) ;
  else
    *ptr = (Ushort)l ;

  return 0 ;
}

int
ulong_a_dot_b (const char *src, Ulong *ptr, const Ulong multiplier, const int ismag)
{
  float l ;

  if (*src == '\0')
    return 0 ;

  l = 0.0 ;

  if (sscanf (src, "%f", &l) < 1)
  {
    printf ("ulong_a_dot_b: %s\n", src) ;
    return -1 ;
  }

  if (*ptr)
    return 0 ;

  if (l < 5.0)
    return 0 ;

  if (ismag)
    l = 30.0 + (l * (float)multiplier) ;
  else
    l = l * (float)multiplier ;

  if (fmod(l,1.0) >= 0.5)
    *ptr = (Ulong)(l + 1.0) ;
  else
    *ptr = (Ulong)l ;

  return 0 ;
}

int
char_a (const char *src, Uchar *ptr)
{
  if (*src == '\0')
    return 0 ;

  if (*ptr)
    return 0 ;

  if (isprint(*src))
    *ptr = *src ;

  return 0 ;
}

int
uchar_a (const char *src, Uchar *ptr)
{
  float l ;

  if (*src == '\0')
    return 0 ;

  l = 0.0 ;

  if (sscanf (src, "%f", &l) != 1)
  {
    printf ("uchar_a: %s\n", src) ;
    return -1 ;
  }

  if ((*ptr) || (l < 0.0))
    return 0 ;

  if (fmod(l,1.0) >= 0.5)
    *ptr = (Uchar)(l + 1.0) ;
  else
    *ptr = (Uchar)l ;

  return 0 ;
}

int
ushort_a (const char *src, Ushort *ptr)
{
  float l ;

  if (*src == '\0')
    return 0 ;

  l = 0.0 ;

  if (sscanf (src, "%f", &l) != 1)
  {
    printf ("ushort_a: %s\n", src) ;
    return -1 ;
  }

  if (*ptr)
    return 0 ;

  if (l < 0.0)
    return 0 ;

  if (fmod(l,1.0) >= 0.5)
    *ptr = (Ushort)(l + 1.0) ;
  else
    *ptr = (Ushort)l ;

  return 0 ;
}

int
ulong_a (const char *src, Ulong *ptr, int blank_ok)
{
  float l ;

  if (*src == '\0')
    return 0 ;

  l = 0.0 ;

  if (sscanf (src, "%f", &l) != 1)
  {
    printf ("ulong_a: %s\n", src) ;
    return (blank_ok) ? 0 : -1 ;
  }

  if (*ptr)
    return 0 ;

  if (l < 0.0)
    return 0 ;

  if (fmod(l,1.0) >= 0.5)
    *ptr = (Ulong)(l + 1.0) ;
  else
    *ptr = (Ulong)l ;

  return 0 ;
}

/*****************************************************************************
 */

TRH_TREE *
DAT_process_ppm (const char *line)
{
  char work [BUFFER_MAX_SIZE+2] ;
  int i ;

  /*
   * The ppmXXX.dat file has the following fixed field format
   *
   * [0]  - [6] PPM number
   * [19] - [22] Magnitude
   * [24] - [25] Spectrum
   * [27] - [38] RA J2000.0 (HH MM SS.SSS)
   * [41] - [52] DEC J2000.0 (+/- DD MM SS.SS)
   * [55] - [61] Annual Proper Motion RA (in seconds of time per year)
   * [63] - [68] Annual Proper Motion DEC (in arc seconds of angle per year)
   * [101]- [106] SAO catalogue number
   * [108]- [113] HD catalogue number
   * [126]- [130] Remarks. [127] contains a "D", is a double
   */

  zero_structure (&dummy[0]) ;
  strncpy (work, line, BUFFER_MAX_SIZE) ;
  work [BUFFER_MAX_SIZE] = work [BUFFER_MAX_SIZE+1] = '\0' ; /* force double termination */
  strip_space (&work[0], &work[6]) ; // PPM
  strip_space (&work[19], &work[22]) ; // Mag
  strip_space (&work[24], &work[25]) ; // spectrum
  strip_space (&work[27], &work[38]) ; // RA
  strip_space (&work[41], &work[52]) ; // DEC
  strip_space (&work[55], &work[61]) ; // APM RA
  strip_space (&work[63], &work[68]) ; // APM DEC
  strip_space (&work[101], &work[106]) ; // SAO
  strip_space (&work[108], &work[113]) ; // HD

  i = 0 ;
  sscanf (&work[0], "%d", &i) ;

  if (i == 0) /* disregard lines we can't read */
    return TRH_NULL ;

  sprintf (key_str, "PPM%d", i) ;
  dummy[0].PPM = i ;

  if (ra_h_space_min_space_sec (&work[27], &dummy[0].RA) ||
      dec_sign_DMS (&work[41], &dummy[0].DEC) ||
      uchar_a_dot_b (&work[19], &dummy[0].magnitude, 10, 1) ||
      char_apm (&work[55], &dummy[0].RA_apm, 1) ||
      char_apm (&work[63], &dummy[0].DEC_apm, 0) ||
      ulong_a (&work[108], &dummy[0].HD, 1) ||
      ulong_a (&work[101], &dummy[0].SAO, 1))
  {
    printf ("ppmXXX, ignoring line: \"%s\"\n", line) ;
    return TRH_NULL ;
  }

  if (dummy[0].magnitude > LIM_STAR_MAG)
    return TRH_NULL ;

  if (work[126] == 'D')
    dummy[0].type = DAT_classify_string ("double", DB_obj_type) ;

  strcat (spectrum, &work[24]) ;
  return &dummy[0] ;
}

/*****************************************************************************
 */

TRH_TREE *
DAT_process_ppm_supplement (const char *line)
{
  char work [BUFFER_MAX_SIZE+2] ;
  int i ;

  /*
   * The ppmb.dat file has the following fixed field format
   *
   * [0]  - [5] PPM number
   * [19] - [21] Magnitude
   * [23] - [34] RA J2000.0 (HH MM SS.SSS)
   * [36] - [47] DEC J2000.0 (+/- DD MM SS.SS)
   * [49] - [55] Annual Proper Motion RA (in seconds of time per year)
   * [57] - [62] Annual Proper Motion DEC (in arc seconds of angle per year)
   * [64] - [69] HD catalogue number
   * [73] - [75] Remarks. [74] contains a "D", is a double
   */

  zero_structure (&dummy[0]) ;
  strncpy (work, line, BUFFER_MAX_SIZE) ;
  work [BUFFER_MAX_SIZE] = work [BUFFER_MAX_SIZE+1] = '\0' ; /* force double termination */
  strip_space (&work[0], &work[5]) ; // PPM
  strip_space (&work[19], &work[21]) ; // Mag
  strip_space (&work[23], &work[34]) ; // RA
  strip_space (&work[36], &work[47]) ; // DEC
  strip_space (&work[49], &work[55]) ; // APM RA
  strip_space (&work[57], &work[62]) ; // APM DEC
  strip_space (&work[64], &work[69]) ; // HD

  i = 0 ;
  sscanf (&work[0], "%d", &i) ;

  if (i == 0) /* disregard lines we can't read */
    return TRH_NULL ;

  sprintf (key_str, "PPM%d", i) ;
  dummy[0].PPM = i ;

  if (ra_h_space_min_space_sec (&work[23], &dummy[0].RA) ||
      dec_sign_DMS (&work[36], &dummy[0].DEC) ||
      uchar_a_dot_b (&work[19], &dummy[0].magnitude, 10, 1) ||
      char_apm (&work[49], &dummy[0].RA_apm, 1) ||
      char_apm (&work[57], &dummy[0].DEC_apm, 0) ||
      ulong_a (&work[64], &dummy[0].HD, 1))
  {
    printf ("ppm_supplement, ignoring line: \"%s\"\n", line) ;
    return TRH_NULL ;
  }

  if (dummy[0].magnitude > LIM_STAR_MAG)
    return TRH_NULL ;

  if (work[74] == 'D')
    dummy[0].type = DAT_classify_string ("double", DB_obj_type) ;

  return &dummy[0] ;
}

/*****************************************************************************
 */

int
is_close (const int a, const int b)
{
  int diff ;

  diff = a - b ;

  if (diff < 0)
    diff = -diff ;

  return (diff < 10) ;
}

TRH_TREE *
DAT_process_al_star (const char *line, TRH_TREE **tree)
{
  char work [BUFFER_MAX_SIZE+2] ;
  const TRH_TREE *p ;

  /*
   * The alignment_star.txt file has the following fixed field format
   *
   * [0]-[13]  Common name
   * Other information ignored
   */

  zero_structure (&dummy[0]) ;
  strncpy (work, line, BUFFER_MAX_SIZE) ;
  work [BUFFER_MAX_SIZE] = work [BUFFER_MAX_SIZE+1] = '\0' ; // force double termination
  strip_space (&work[0], &work[13]) ; // Work[0] contains common name
  strip_space (&work[15], &work[35]) ; // Work[15] contains RA
  strip_space (&work[41], &work[66]) ; // Work[41] contains DEC
  strip_space (&work[70], &work[76]) ; // Work[70] contains magnitude

  if (ra_h_space_min_space_sec (&work[15], &dummy[0].RA) ||
      dec_sign_DMS (&work[41], &dummy[0].DEC) ||
      uchar_a_dot_b (&work[70], &dummy[0].magnitude, 10, 1))
  {
    printf ("alignment_star, ignoring line: \"%s\"\n", line) ;
    return TRH_NULL ;
  }

  p = TR_find_nth (*tree, 1) ;

  for (; p != TRH_NULL ; p=p->next)
  {
    if (is_close(p->RA, dummy[0].RA) && is_close(p->DEC, dummy[0].DEC))
    {
      if (strncasecmp(p->common_name, &work[0], strlen(&work[0])) != 0)
      {
	copy_fixed_elements (&dummy[0], p) ;
	strcpy (key_str, p->key) ;
	strcpy (common_name, &work[0]) ;
	strip_garbage (common_name) ;
	dummy[0].alignment = 0xff ;

	if (dummy[0].type == 0)
	  dummy[0].type = DAT_classify_string ("star", DB_obj_type) ;

	return &dummy[0] ;
      }

      p->alignment = 0xff ;

      if (dummy[0].type == 0)
	dummy[0].type = DAT_classify_string ("star", DB_obj_type) ;

      return TRH_NULL ;
    }
  }

  printf ("Couldn't find an object in the viscinity of <%s>\n", &work[0]) ;
  strcpy (key_str, &work[0]) ; // use the name as the key
  strcpy (common_name, &work[0]) ; // use the name as the key
  strip_garbage (key_str) ;
  dummy[0].type = DAT_classify_string ("star", DB_obj_type) ;
  dummy[0].alignment = 0xff ;
  return &dummy[0] ;
}

//----------------------------------------------------------------
void
append_string (const Ushort start, const Ushort end, char *str, const char *format, const Ushort size, char *dest)
{
  char c ;

  c = str[end+1] ;
  str[end+1] = '\0' ;
  STR_snaprintf (dest, size, format, &str[start]) ;
  str[end+1] = c ;
}

TRH_TREE *
DAT_process_nex_align (const char *line, TRH_TREE **tree)
{
  char work [BUFFER_MAX_SIZE+2] ;
  char line_buf [BUFFER_MAX_SIZE+2] ;
  char *pointers[MAX_POINTERS] ;
  const TRH_TREE *p ;

  /*
   * The nex_align.txt file has the following fixed field format
   *
   * [ 0]-[ 4] Nexstar catalogue number
   * [ 5]-[16] Common Name
   * [17]-[18] HH
   * [20]-[26] MM.MMMM
   * [30]-[37] +/- DEC.DEC
   * [39]-[44] SAO catalogue number
   * [46]-[49] Magnitude
   * [50]-[52] Spectrum
   * [54]-[56] Constellation
   */

  strcpy (line_buf, line) ;   /* create a dummy[0] work line which we'll later process */

  *work = '\0' ;
  /* [1] */
  append_string (5, 16, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; // common name
  append_string (17, 26, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; // HH MM.MMMM
  append_string (30, 37, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; // +/- DEC
  append_string (39, 44, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; // SAO number
  append_string (46, 49, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; // Mag
  append_string (50, 52, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; // spectrum
  append_string (54, 56, line_buf, "%s,,", BUFFER_MAX_SIZE, work) ; // constellation

  zero_structure (&dummy[0]) ;
  work [BUFFER_MAX_SIZE] = work [BUFFER_MAX_SIZE+1] = '\0' ; /* force double termination */
  process_delimiters (work, pointers, ',') ;

  strip_garbage (pointers[0]) ;
  strip_garbage (pointers[1]) ;
  strip_garbage (pointers[2]) ;
  strip_garbage (pointers[3]) ;
  strip_garbage (pointers[4]) ;
  strip_garbage (pointers[5]) ;

  if (ra_h_space_min_space_sec (pointers[0], &dummy[0].RA) ||
      dec_sign_DMS (pointers[1], &dummy[0].DEC))
  {
    printf ("nex_align.txt, ignoring line: \"%s\"\n", line) ;
    return TRH_NULL ;
  }

  /// BLAH BLAH


  zero_structure (&dummy[0]) ;
  strncpy (work, line, BUFFER_MAX_SIZE) ;
  work [BUFFER_MAX_SIZE] = work [BUFFER_MAX_SIZE+1] = '\0' ; // force double termination
  strip_space (&work[0], &work[13]) ; // Work[0] contains common name
  strip_space (&work[15], &work[35]) ; // Work[15] contains RA
  strip_space (&work[41], &work[66]) ; // Work[41] contains DEC
  strip_space (&work[70], &work[76]) ; // Work[70] contains magnitude

  if (ra_h_space_min_space_sec (&work[15], &dummy[0].RA) ||
      dec_sign_DMS (&work[41], &dummy[0].DEC) ||
      uchar_a_dot_b (&work[70], &dummy[0].magnitude, 10, 1))
  {
    printf ("alignment_star, ignoring line: \"%s\"\n", line) ;
    return TRH_NULL ;
  }

  p = TR_find_nth (*tree, 1) ;

  for (; p != TRH_NULL ; p=p->next)
  {
    if (is_close(p->RA, dummy[0].RA) && is_close(p->DEC, dummy[0].DEC))
    {
      if (strncasecmp(p->common_name, &work[0], strlen(&work[0])) != 0)
      {
	copy_fixed_elements (&dummy[0], p) ;
	strcpy (key_str, p->key) ;
	strcpy (common_name, &work[0]) ;
	strip_garbage (common_name) ;
	dummy[0].alignment = 0xff ;

	if (dummy[0].type == 0)
	  dummy[0].type = DAT_classify_string ("star", DB_obj_type) ;

	return &dummy[0] ;
      }

      p->alignment = 0xff ;

      if (dummy[0].type == 0)
	dummy[0].type = DAT_classify_string ("star", DB_obj_type) ;

      return TRH_NULL ;
    }
  }

  printf ("Couldn't find an object in the viscinity of <%s>\n", &work[0]) ;
  strcpy (key_str, &work[0]) ; // use the name as the key
  strcpy (common_name, &work[0]) ; // use the name as the key
  strip_garbage (key_str) ;
  dummy[0].type = DAT_classify_string ("star", DB_obj_type) ;
  dummy[0].alignment = 0xff ;
  return &dummy[0] ;
}


/*****************************************************************************
 */

TRH_TREE *
DAT_process_harvard_catalogue (const char *line)
{
  char work [BUFFER_MAX_SIZE+2] ;
  int i ;

  /*
   * The 5050_catalog.dat file has the following comma delimited format
   *
   * [0]-[3]     Harvard Revised Number       -> HR
   * [4]-[8]     Formal name (number)         -> formal name
   * [9]-[11]    Formal name (greek letter)   -> formal name
   * [12]        Formal name (numeric ext)    -> formal name
   * [13]-[15]   Formal name (constellation)  -> formal name
   *		  Note: number and letter fields also hold Messier designation!
   * [31]-[36]   Henry Draper catalogue number-> HD
   * [39]-[44]   SAO catalogue number         -> SAO
   * [53]        Double or multiple star code -> type
   * [107]-[116] HH MM SS.S                   -> RA J2000
   * [119]       dec sign                     -> DEC sign
   * [120]-[127] DD MM SS                     -> DEC J2000
   * [145]-[150] Magnitude (float)            -> mag
   * [191]-[209] Spectral type                -> spectrum
   * [211]       Spectral type code           -> spectrum
   * [212]-[219] RA ann prop motion (signed)  -> RA_apm
   * [220]-[227] DEC ann prop motion (signed) -> DEC_apm
   * [263]-[268] Binary Magnitude difference  -> mag_diff
   * [271]-[276] Separation of components     -> sep
   * [285]-[286] Number of components in mult -> num_components
   *
   * Other fields present but I've not decoded them as yet.
   */

  zero_structure (&dummy[0]) ;
  strncpy (work, line, BUFFER_MAX_SIZE) ;
  work[16] = '\0' ; // process Formal Name
  memmove (&work[0], &work[4], 13) ;
  strip_garbage (&work[0]) ;

  if ((work[0] == 'M') && isspace(work[1]))
    work[0] = '\0' ;

  memmove (&work[20], &line[0], 4) ; // process HR number
  work[24] = '\0' ; // HR termination
  work[37] = '\0' ; // HD termination
  work[45] = '\0' ; // SAO termination
  work[54] = '\0' ; // type termination
  work[117] = '\0' ; // RA termination
  work[128] = '\0' ; // DEC termination
  work[151] = '\0' ; // magnitude termination
  work[210] = '\0' ; // spectral type termination
  work[220] = '\0' ; // RA apm termination
  work[228] = '\0' ; // DEC apm termination
  work[269] = '\0' ; // mag diff termination
  work[277] = '\0' ; // separation termination
  work[287] = '\0' ; // number of components termination

  work [BUFFER_MAX_SIZE] = work [BUFFER_MAX_SIZE+1] = '\0' ; /* force double termination */

  //fprintf (stderr, "Formal:%s HR:%s HD:%s SAO:%s\n", &work[0], &work[20], &work[31], &work[39]) ;
  //fprintf (stderr, "type:%c RA:%s DEC:%s mag:%s\n", work[53], &work[107], &work[119], &work[145]) ;
  //fprintf (stderr, "spectrum:%s %c ra apm:%s dec apm:%s\n", &work[191], work[211], &work[212], &work[221]) ;
  //fprintf (stderr, "mag diff:%s separation:%s num_comp:%s\n", &work[263], &work[271], &work[285]) ;

  if (ra_h_space_min_space_sec (&work[107], &dummy[0].RA) ||
      dec_sign_DMS (&work[119], &dummy[0].DEC) ||
      ulong_a (&work[31], &dummy[0].HD, 1) ||
      ushort_a (&work[20], &dummy[0].HR) ||
      ulong_a (&work[39], &dummy[0].SAO, 1) ||
      char_apm (&work[212], &dummy[0].RA_apm, 0) ||
      char_apm (&work[221], &dummy[0].DEC_apm, 0) ||
      uchar_a_dot_b (&work[145], &dummy[0].magnitude, 10, 2) ||
      uchar_a_dot_b (&work[263], &dummy[0].mag_diff, 10, 0) ||
      ushort_a_dot_b (&work[271], &dummy[0].separation, 10, 0) ||
      uchar_a (&work[285], &dummy[0].num_components))
  {
    //    printf ("5050_catalog, ignoring line: \"%s\"\n", line) ;
    return TRH_NULL;
  }

  if (dummy[0].magnitude > LIM_STAR_MAG)
    return TRH_NULL ;

  sprintf (key_str, "HR%hu", dummy[0].HR) ;
  compress_spaces (formal_name, &work[0]) ;

  if ((*formal_name == '\0') && DAT_classify_string (&work[0], DB_constellations))
    compress_spaces (formal_name, &work[0]) ;

  i = strlen(formal_name) ;

  compress_spaces (spectrum, &work[191]) ;

  if (work[211] != ' ')
  {
    work[210] = ' ' ;
    work[212] = '\0' ;
    strcat (spectrum, &work[210]) ;
  }

  if (i >= 3)
    dummy[0].constellation = DAT_classify_string (&formal_name[i-3], DB_constellations) ;

  if (work[53] == 'A')
    dummy[0].type = DAT_classify_string ("astronometric", DB_obj_type) ;
  else if (work[53] != ' ')
    dummy[0].type = DAT_classify_string ("double", DB_obj_type) ;
  else
    dummy[0].type = DAT_classify_string ("star", DB_obj_type) ;

  return &dummy[0] ;
}

/*****************************************************************************
 */

TRH_TREE *
DAT_process_dsodel (const char *line)
{
  char work [BUFFER_MAX_SIZE+2] ;
  char *pointers[MAX_POINTERS] ;
  int j ;

  /*
   * The dsodel.txt file has the following semicolon delimited format
   *
   * [0]  Three letter constellation
   * [2]  Messier number
   * [3]  NGC number
   * [4]  NGC component
   * [5]  IC number
   * [6]  IC component
   * [7]  UGC number
   * [8]  UGC component
   * [9]  type attribute list, items delimited with +
   * [10] RA hours J2000
   * [11] RA minutes J2000
   * [12] RA seconds J2000
   * [13] DEC sign J2000
   * [14] DEC degrees J2000
   * [15] DEC minutes J2000
   * [16] DEC seconds J2000
   * [17] Major axis size (arc seconds)
   * [19] Minor axis size (not always present)
   * [20] Integrated magnitude
   * [22] Position angle (1=face on, 5=edge on)
   * [23] Subjective object rating
   * [24] Object name
   */

  zero_structure (&dummy[0]) ;
  strncpy (work, line, BUFFER_MAX_SIZE) ;
  work [BUFFER_MAX_SIZE] = work [BUFFER_MAX_SIZE+1] = '\0' ; /* force double termination */
  process_delimiters (work, pointers, ';') ;

  if ((*pointers[2] != '\0') &&(sscanf(pointers[2], "%d", &j) == 1))
  {
    sprintf (key_str, "M%d", j) ;
    dummy[0].messier = j ;
  }

  if ((*pointers[3] != '\0') && (sscanf(pointers[3], "%d", &j) == 1))
  {
    if (*key_str == '\0')
    {
      sprintf (key_str, "NGC%d", j) ;

      if (*pointers[4])
	STR_snaprintf (key_str, BUFFER_MAX_SIZE, "%c", *pointers[4]) ;
    }

    dummy[0].NGC = j ;
    dummy[0].NGC_comp = *pointers[4] ;
  }

  if ((*pointers[5] != '\0') && (sscanf(pointers[5], "%d", &j) == 1))
  {
    if (*key_str == '\0')
    {
      sprintf (key_str, "IC%d", j) ;

      if (*pointers[6])
	STR_snaprintf (key_str, BUFFER_MAX_SIZE, "%c", *pointers[6]) ;
    }

    dummy[0].IC = j ;
    dummy[0].IC_comp = *pointers[6] ;
  }

  if ((*pointers[7] != '\0') && (sscanf(pointers[7], "%d", &j) == 1))
  {
    if (*key_str == '\0')
    {
      sprintf (key_str, "UGC%d", j) ;

      if (*pointers[8])
	STR_snaprintf (key_str, BUFFER_MAX_SIZE, "%c", *pointers[8]) ;
    }

    dummy[0].UGC = j ;
    dummy[0].UGC_comp = *pointers[8] ;
  }

  *(pointers[11] - 1) = ' ' ;
  *(pointers[12] - 1) = ' ' ;
  *(pointers[14] - 1) = ' ' ;
  *(pointers[15] - 1) = ' ' ;
  *(pointers[16] - 1) = ' ' ;

  if (ra_h_space_min_space_sec (pointers[10], &dummy[0].RA) ||
      dec_sign_DMS (pointers[13], &dummy[0].DEC) ||
      uchar_a (pointers[2], &dummy[0].messier) ||
      ushort_a (pointers[3], &dummy[0].NGC) ||
      char_a (pointers[4], &dummy[0].NGC_comp) ||
      ushort_a (pointers[5], &dummy[0].IC) ||
      char_a (pointers[6], &dummy[0].IC_comp) ||
      ulong_a (pointers[7], &dummy[0].UGC, 1) ||
      char_a (pointers[8], &dummy[0].UGC_comp) ||
      uchar_a (pointers[17], &dummy[0].major) ||
      uchar_a (pointers[19], &dummy[0].minor) ||
      uchar_a (pointers[22], &dummy[0].angle) ||
      uchar_a_dot_b (pointers[20], &dummy[0].magnitude, 10, 1))
  {
    printf ("dsodel, ignoring line: \"%s\"\n", line) ;
    return TRH_NULL ;
  }

  dummy[0].type = DAT_classify_string (pointers[9], DB_obj_type) ;
  dummy[0].constellation = DAT_classify_string (pointers[0], DB_constellations) ;
  dummy[0].rating = DAT_classify_string (pointers[23], DB_rating_type) ;

  if (*pointers[24] != '\0')
  {
    strcpy (common_name, pointers[24]) ;
    capitalisation (common_name) ;
  }

  return &dummy[0] ;
}

/*****************************************************************************
 */

TRH_TREE *
DAT_process_rc3 (const char *line)
{
  char work [BUFFER_MAX_SIZE+2] ;
  char line_buf [BUFFER_MAX_SIZE+2] ;
  char *pointers[MAX_POINTERS] ;

  strcpy (line_buf, line) ;   /* create a dummy[0] work line which we'll later process */

  *work = '\0' ;
  /* [0] */
  append_string (0, 9, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; /* RA2000 HH MM SS.S */

  /* [1] */
  append_string (12, 20, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; /* signed DEC2000 (sign)DD MM SS */

  /* [2] [3] [4] */
  append_string (78, 91, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; /* name */
  append_string (93, 108, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; /* alt name */
  append_string (109, 124, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; /* other desig */

  /* [5] [6] [7] [8] */
  append_string (125, 132, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; /* PGC number */
  append_string (134, 137, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; /* type */
  append_string (259, 263, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; /* magnitude */
  append_string (271, 277, line_buf, "%s,", BUFFER_MAX_SIZE, work) ; /* magnitude */

  zero_structure (&dummy[0]) ;
  work [BUFFER_MAX_SIZE] = work [BUFFER_MAX_SIZE+1] = '\0' ; /* force double termination */
  process_delimiters (work, pointers, ',') ;

  strip_garbage (pointers[2]) ;
  strip_garbage (pointers[3]) ;
  strip_garbage (pointers[4]) ;
  strip_garbage (pointers[5]) ;
  strip_garbage (pointers[7]) ;
  strip_garbage (pointers[8]) ;

  if (ra_h_space_min_space_sec (pointers[0], &dummy[0].RA) ||
      dec_sign_DMS (pointers[1], &dummy[0].DEC))
  {
    printf ("rc3, ignoring line: \"%s\"\n", line) ;
    return TRH_NULL ;
  }

  /* take whichever comes first */
  uchar_a_dot_b (pointers[8], &dummy[0].magnitude, 10, 1) ;
  uchar_a_dot_b (pointers[7], &dummy[0].magnitude, 10, 1) ;

  sscanf (pointers[2], "IC%hu%c", &dummy[0].IC, &dummy[0].IC_comp) ;
  sscanf (pointers[2], "NGC%hu%c", &dummy[0].NGC, &dummy[0].NGC_comp) ;
  sscanf (pointers[3], "UGC%lu%c", &dummy[0].UGC, &dummy[0].UGC_comp) ;
  sscanf (pointers[5], "PGC%lu", &dummy[0].PGC) ;

  if (dummy[0].NGC != 0)
    sprintf (key_str, "NGC%hu%c", dummy[0].NGC, dummy[0].NGC_comp) ;
  else if (dummy[0].IC != 0)
    sprintf (key_str, "IC%hu%c", dummy[0].IC, dummy[0].IC_comp) ;
  else if (dummy[0].UGC != 0)
    sprintf (key_str, "UGC%ld%c", dummy[0].UGC, dummy[0].UGC_comp) ;
  else if (dummy[0].PGC != 0)
    sprintf (key_str, "PGC%ld", dummy[0].PGC) ;

  /* NEED TO DECODE TYPE when I work out what the codes mean */
  dummy[0].type = DAT_classify_string ("galaxy", DB_obj_type) ;

  return &dummy[0] ;
}

/*****************************************************************************
 */

TRH_TREE *
DAT_process_harvard_notes (const char *line)
{
  char work [BUFFER_MAX_SIZE+2] ;
  char *p, *q ;
  int i ;

  zero_structure (&dummy[0]) ;
  strncpy (work, line, BUFFER_MAX_SIZE) ;
  work [BUFFER_MAX_SIZE] = work [BUFFER_MAX_SIZE+1] = '\0' ; // force double termination
  work[4] = '\0' ;
  work[8] = '\0' ;
  work[15] = '\0' ;

  if (work[10] == 'C') /* disregard colour notes */
    return TRH_NULL ;

  if ((work[10] == 'D') && (work[11] == 'Y')) /* disregard dynamical parallex */
    return TRH_NULL ;

  if (work[10] == 'P') /* disregard polarisation notes */
    return TRH_NULL ;

  if ((work[10] == 'R') && (work[11] == 'V')) /* disregard rotational velocity */
    return TRH_NULL ;

  if (work[10] == 'S') /* disregard spectra notes */
    return TRH_NULL;

  if (work[10] == 'V') /* disregard variability notes */
    return TRH_NULL ;

  i = 0 ;
  sscanf (&work[0], "%d", &i) ;

  if (i == 0) /* disregard lines we can't read */
  {
    fprintf (stdout, "Can't read line: %s\n", line) ;
    return TRH_NULL ;
  }

  dummy[0].HR = i ;
  sprintf (key_str, "HR%d", i) ;

  if (SOR_search ((void *)HR, cmp_HR, num.hr, (void *)&dummy[0]) == SOR_FAIL)
    return TRH_NULL ;

  if (work[7] == '1')
  {
    if (work[10] == 'D')
    {
      STR_snaprintf (description, BUFFER_MAX_SIZE, "Double: %s", &work[16]) ;
      dummy[0].type = DAT_classify_string ("double", DB_obj_type) ;
    }

    else if (work[10] == 'G')
    { /* remove all occurences of the string "group" */
      while ((p = strstr(&work[16], " group")) != NULL)
      {
	for (q=p+6 ; *q ;)
	  *p++ = *q++ ;

	*p = '\0' ;
      }

      STR_snaprintf (description, BUFFER_MAX_SIZE, "Group: %s", &work[16]) ;
      strip_garbage (&work[16]) ;
      dummy[0].constellation = DAT_classify_string (&work[16], DB_constellations) ;

      if (dummy[0].constellation && (strlen(&work[16]) < 10))
	strcpy (formal_name, &work[16]) ;
    }

    else if (work[10] == 'M')
      STR_snaprintf (description, BUFFER_MAX_SIZE, "Misc: %s", &work[16]) ;

    else if (work[10] == 'N')
    {
      char *extra_info ;

      for (p=&work[16] ; *p ; p++) /* segregate into first & remaining sect */
	if ((*p == ';') || (*p == '.'))
	  break ;

      if (*p)
      {
	*p++ = '\0' ;

	while (*p && isspace(*p))
	  p++ ;
      }

      extra_info = p ; /* mark the remainder */
      capitalisation (&work[16]) ;
      p = strstr(&work[16], " in ") ;

      if (p != NULL) /* decode the constellation if marked */
      {
	*p = '\0' ; /* peg it off */
	p += 4 ;

	if (dummy[0].constellation == 0) /* identify the constellation */
	  dummy[0].constellation = DAT_classify_string (p, DB_constellations) ;
      }

      if (DAT_classify_string (&work[16], DB_constellations))
      {
	strcpy (formal_name, &work[16]) ;
	strip_garbage (formal_name) ;
      }
      else
      {
	strcpy (common_name, &work[16]) ;
	strip_garbage (common_name) ;
      }

      if (*extra_info)
	STR_snaprintf (description, BUFFER_MAX_SIZE, "Name: %s", extra_info) ;
    }

    else if (work[10] == 'R')
      STR_snaprintf (description, BUFFER_MAX_SIZE, "%s", &work[16]) ;
  }
  else
  {
    /* This is a continuation of the previous comment */

    STR_snaprintf (description, BUFFER_MAX_SIZE, " %s", &work[16]) ;
  }

  /*
   * strip any white space from the end of the new description
   */

  strip_garbage (formal_name) ;
  strip_garbage (description) ;

  return &dummy[0] ;
}

/*****************************************************************************
 */

TRH_TREE *
DAT_process_ngc2000 (const char *line)
{
  char work [BUFFER_MAX_SIZE+2] ;
  int i ;

  /*
   * The ngc2000.txt file has the following fixed field format
   *
   * [0]  NGC or IC number
   * [9]  Object type
   * [13] RA
   * [22] sign of declanation
   * [23] DEC
   * [32] constellation
   * [36] object size (in arcmin)
   * [43] magnitude
   * [49] description (!=remarkable, refer to dreyfus's notes)
   */

  zero_structure (&dummy[0]) ;
  strncpy (work, line, BUFFER_MAX_SIZE) ;
  work [BUFFER_MAX_SIZE] = work [BUFFER_MAX_SIZE+1] = '\0' ; /* force double termination */
  strip_space (&work[0], &work[6]) ; // NGC / IC
  strip_space (&work[8], &work[10]) ; // type
  strip_space (&work[12], &work[18]) ; // RA
  strip_space (&work[21], &work[26]) ; // DEC
  strip_space (&work[32], &work[34]) ; // Constellation
  strip_space (&work[40], &work[44]) ; // Size (arcmin)
  strip_space (&work[47], &work[50]) ; // Magnitude
  strip_space (&work[56], &work[104]) ; // Description

  i = 0 ;
  sscanf (&work[1], "%d", &i) ;

  if (i == 0) /* disregard lines we can't read */
    return TRH_NULL ;

  if (work[0] == 'I')
  {
    sprintf (key_str, "IC%d", i) ;
    dummy[0].IC = i ;
  }
  else
  {
    sprintf (key_str, "NGC%d", i) ;
    dummy[0].NGC = i ;
  }

  if (ra_h_space_min_dot_sec (&work[12], &dummy[0].RA) ||
      dec_sign_DMS (&work[21], &dummy[0].DEC) ||
      uchar_a (&work[40], &dummy[0].major) ||
      uchar_a_dot_b (&work[47], &dummy[0].magnitude, 10, 1))
  {
    printf ("ngc2000, ignoring line: \"%s\"\n", line) ;
    return TRH_NULL ;
  }

  dummy[0].type = DAT_classify_string (&work[8], DB_obj_type) ;
  dummy[0].constellation = DAT_classify_string (&work[32], DB_constellations) ;

  strip_garbage (&work[49]) ;
  dummy[0].rating = DAT_classify_string (&work[56], DB_rating_type) ;

  return &dummy[0] ;
}

/*****************************************************************************
 */

TRH_TREE *
DAT_process_ngc_names (const char *line)
{
  char work [BUFFER_MAX_SIZE+2] ;
  int i ;

  /*
   * The 7118_names.dat file has the following fixed field format
   *
   * [0]  - [34]  Common Name
   * [36] - [40]  NGC or IC name
   * [42] - [69]  Comment
   */

  zero_structure (&dummy[0]) ;
  strncpy (work, line, BUFFER_MAX_SIZE) ;
  work [BUFFER_MAX_SIZE] = work [BUFFER_MAX_SIZE+1] = '\0' ; /* force double termination */
  strip_space (&work[0], &work[34]) ; // Common name
  strip_space (&work[36], &work[40]) ; // NGC or IC
  strip_space (&work[42], &work[69]) ; // comment

  if ((work[0] == 'M') && (work[1] == ' '))
    return TRH_NULL ;

  i = 0 ;

  if (work[36] == 'I')
  {
    sscanf (&work[37], "%d", &i) ;
    sprintf (key_str, "IC%d", i) ;
    dummy[0].IC = i ;
  }
  else
  {
    sscanf (&work[36], "%d", &i) ;
    sprintf (key_str, "NGC%d", i) ;
    dummy[0].NGC = i ;
  }

  strcpy (common_name, &work[0]) ;
  return &dummy[0] ;
}

/*****************************************************************************
 */

void
DAT_check_null (TRH_TREE *tree)
{
  const TRH_TREE *p ;
  char Rmin, Rmax, Dmin, Dmax ;

  p = TR_find_nth (tree, 1) ;
  Rmin = Rmax = Dmin = Dmax = 0 ;

  for (; p != TRH_NULL ; p=p->next)
  {
    if ((p->RA == 0) && (p->DEC == 0))
    {
      DAT_display_node (p) ;
      printf ("Newly added node has RA AND DEC == 0!!!\n") ;
    }

    if (p->spectrum == (char *)0)
    {
      DAT_display_node (p) ;
      printf ("Spectrum is a null string\n") ;
    }

    if (p->separation >= 20000)
    {
      DAT_display_node (p) ;
      printf ("Separation is a bad value\n") ;
    }      

    if (p->RA_apm < Rmin)
      Rmin = p->RA_apm ;

    if (p->RA_apm > Rmax)
      Rmax = p->RA_apm ;

    if (p->DEC_apm < Dmin)
      Dmin = p->DEC_apm ;

    if (p->DEC_apm > Dmax)
      Dmax = p->DEC_apm ;
  }

  fprintf (stdout, "\nRA: %d to %d; DEC: %d to %d\n", Rmin, Rmax, Dmin, Dmax) ;
}


typedef TRH_TREE *(*FUNCTION)(const char *string) ;

void
process_file (const char *name, FUNCTION process_it, TRH_TREE **tree, int halt, int disp_err)
{
  FILE *f ;
  char string[BUFFER_MAX_SIZE+1] ;
  TRH_TREE *node ;
  int i ;

  if ((f = fopen (name, "r")) == (FILE *)0)
  {
    fprintf (stderr, "Can't open %s\n", name) ;
    return ;
  }

  printf ("\nAbout to process %s\n", name) ;

  for (i=0 ; RL_read_line(string, BUFFER_MAX_SIZE, f) ; i++)
  {
    //fprintf (stderr, "%s\n", string) ;

    if (*process_it == (FUNCTION)DAT_process_al_star)
      node = DAT_process_al_star (string, tree) ;
    else if (*process_it == (FUNCTION)DAT_process_nex_align)
      node = DAT_process_nex_align (string, tree) ;
    else
      node = (*process_it)(string) ;

    if (node == TRH_NULL)
      continue ;

    if (*node->key == '\0')
      fprintf (stdout, "Attempting to create a null node!\n") ;

    if (halt)
    {
      DAT_display_node (node) ;

      if (halt-- == 1)
	exit (1) ;
    }

    if (i >= 100)
    {
      printf ("#") ;
      fflush (stdout) ;
      i -= 100 ;
    }

    process_cross_reference (node, tree, disp_err, 0) ;
  }

  fclose (f) ;
  DAT_check_null (*tree) ;
  build_cross_reference (*tree, 1, 1) ;
}

int
main ()
{
  TRH_TREE *tree ;
  TRH_TREE *node ;

  printf ("LONG RUN TIME: ~4 Hours on PentII-400, ~9 Hours on PentI-200\n") ;
  printf ("Processing with limiting star mag=%.1f, limiting mag=%.1f\n",
	  LIMITING_STAR_MAG, LIMITING_MAGNITUDE) ;

  tree = node = TRH_NULL ;
  type_boundary = DAT_classify_string ("@boundary", DB_obj_type) ;

  process_file ("1206_ppmb.dat", DAT_process_ppm_supplement, &tree, 0, 0) ;
  process_file ("1146_ppmnorth.dat", DAT_process_ppm, &tree, 0, 0) ;
  process_file ("1193_ppmsouth.dat", DAT_process_ppm, &tree, 0, 0) ;

  printf ("\nThis next file takes a while (~90 #s)\n") ;
  process_file ("5050_catalog.dat", DAT_process_harvard_catalogue, &tree, 0, 0) ;

  printf ("\nThis next file also takes a while (~90 #s)\n") ;
  process_file ("5050_notes.dat", DAT_process_harvard_notes, &tree, 0, 0) ;

  printf ("\nThis next file takes a while (~220 #s)\n") ;
  process_file ("dsodel.txt", DAT_process_dsodel, &tree, 0, 0) ;

  printf ("\nThis next file takes a while (~230 #s)\n") ;
  process_file ("7155_rc3.dat", DAT_process_rc3, &tree, 0, 0) ;

  printf ("\nThis next file takes a while (~130 #s)\n") ;
  // SHOULD BE ONE OF THE LATTER FILES PROCESSED DUE TO LOW ACCURACY ON RA/DEC
  process_file ("7118_ngc2000.dat", DAT_process_ngc2000, &tree, 0, 0) ;

  process_file ("7118_names.dat", DAT_process_ngc_names, &tree, 0, 0) ;
  process_file ("alignment_stars.txt", (FUNCTION)DAT_process_al_star, &tree, 0, 0) ;
  //process_file ("nex_align.txt", (FUNCTION)DAT_process_nex_align, &tree, 0, 0) ;

  //  display_list (mes, 10) ;

  /*---------------- OUTPUT ----------------*/
  printf ("About to write compressed database\n") ;
  DAT_write_compressed_database (tree) ;

  exit (0) ;
}
