/*
 * tree.c
 *
 * This module contains routines to maintain binary trees. The routines can be
 * used with multiple tree types by altering the <info> union in <tree.h> to
 * incorporate all data structures required.
 *
 * To use the <tree.c> module, create your own tree definition struct to suit
 * the particular application. <tree.h> and <tree.c> should be copied from their
 * source directory every compile time.
 ******************************************************************************
 * The module supports automatic height balancing if required.
 * To activate automatic height balancing, define TRH_HBAL in treedef.h.
 *
 * Height balancing slows down the node addition and deletion operations
 * and speeds up the search operation. If the most common operation on the
 * tree will be searching, then height balancing should be enabled. If the
 * tree will be searched and added in equal ratios, then height balancing is
 * marginal for tree sizes of 50 nodes or less.
 ******************************************************************************
 * Written by Stefan Keller-Tuberg, March 1989, November 1994.
 ******************************************************************************
 * Copyright (C) 2002 Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ******************************************************************************
 */

#include	<string.h>
#include        <stdio.h>
#include	<stdlib.h>

#include        "types.h"
#include	"tree.h"

/*******************************************************************************
 * Local static variables.
 */

#ifdef  TRH_HBAL
static  void	TR_balance (TRH_TREE **, TRH_TREE *) ;
#endif

#ifdef  TRH_DEBUG
static  void	TR_local_display_tree (TRH_TREE *, char *, char *) ;
#endif

/*******************************************************************************
 * Assumptions made:
 *
 *  All memory used for the nodes of the trees manipulated by this software
 *  is assumed to have originated through calls to <malloc> or equivalent.
 *  This module uses calls to <free> to release nodes which are no longer
 *  required.
 *
 *  The block of memory allocated for a node (using <malloc>) is also
 *  assumed to contain the key AND data fields. From this assumption, it is
 *  OK to dispose of an entire memory block (using <free>) with a single
 *  call. Such nodes can be obtained by calling TR_create_node().
 */

/*******************************************************************************
 * TR_create_node
 *
 * A routine to fetch space for us so that a new node may be added to a tree.
 *
 * The paramater <data_size> should be the length of the user's area of the
 * structure. If no extra space is required, <data_size> should be 0.
 * Note that if the user's field will be a string, a byte should also be
 * reserved for the null ('\0') terminator! {strlen() + 1}
 */

TRH_TREE *
TR_create_node (const TRH_KEY_TYPE new_key, const Ushort data_size)
{
    Ushort	size ;
    Ushort	new_size ;
    TRH_TREE	*tmp ;

    size		= (Ushort)(sizeof(TRH_TREE) + TRH_key_len(new_key)) ;
    new_size		= data_size + size ;
    tmp			= (TRH_TREE *)malloc(new_size) ;

    if (tmp == (TRH_TREE *)0)
    {
      fprintf (stderr, "TR_create_node: Malloc can't allocate space\n") ;
      exit (1) ;
    }

    TRH_key_copy (tmp->key, new_key) ;
    tmp->size		= new_size ;
    tmp->user_offset	= size ;
    return (tmp) ;
}

/*******************************************************************************
 * TR_node_infocopy
 *
 * A routine to copy information strings into a newly created node
 */

#ifdef  TRH_STRING_DATA

void
TR_node_infocopy (TRH_TREE *node, const char *descr, const char *pri, const char *sec, const char *ter)
{
  char *p ;

  p = &node->key [strlen(node->key) + 1] ;
  node->DESCR_PTR = p ;
  strcpy (p, descr) ;
  p = &node->DESCR_PTR [strlen(node->DESCR_PTR) + 1] ;

  if (*pri != '\0')
  {
    node->PRIMARY_PTR = p ;
    strcpy (p, pri) ;
    p = &node->PRIMARY_PTR [strlen(node->PRIMARY_PTR) + 1] ;
  }
  else
    node->PRIMARY_PTR = p-1 ;

  if (*sec != '\0')
  {
    node->SECONDARY_PTR = p ;
    strcpy (p, sec) ;
    p = &node->SECONDARY_PTR [strlen(node->SECONDARY_PTR) + 1] ;
  }
  else
    node->SECONDARY_PTR = p-1 ;

  if (*ter != '\0')
  {
    node->TERTIARY_PTR = p ;
    strcpy (p, ter) ;
  }
  else
    node->TERTIARY_PTR = p-1 ;
}

#endif

/*******************************************************************************
 * TR_copy_user_data
 *
 * A routine to append an external user data structure to the end of a node.
 * This routine is meant to be used in conjunction with TR_create_node().
 */

void
TR_copy_user_data (TRH_TREE *node, const char *data)
{
    if (((int)node->size - (int)node->user_offset) <= 0)
    {
	fprintf (stderr,
#ifdef  TRH_STRING_KEY
		"Run time size / user offset error: node=<%s>\n",
#else
		"Run time size / user offset error: node=<%08lx>\n",
#endif
		node->key) ;
	exit (1) ;
    }

    memcpy (TRH_USER_PTR(node), data, node->size - node->user_offset) ;
}

/*******************************************************************************
 * TR_find_node
 *
 * A routine to scan the specified tree for a key and return a pointer
 * to the node when found.
 *
 * If the key is not located, a null pointer is returned.
 */

const TRH_TREE *
TR_find_node (const TRH_TREE *node, const TRH_KEY_TYPE key)
{
    int		c_val ;

    /*
     * Non recursive implementation.
     *
     * Zip through the tree searching for the key. The tree is ordered by
     * alphabetical order of the key.
     */

    for (; node != TRH_NULL ;)
    {
	c_val = TRH_key_compare(key, node->key) ;

	if (c_val == 0)
	    break ;

	node = (c_val < 0) ? node->left : node->right ;
    }

    return (node) ;
}

/*******************************************************************************
 * TR_find_closest
 *
 * A routine to scan the specified tree to find the next element equal to or
 * following the provided key.
 *
 * If the key is not located, a null pointer is returned.
 */

const TRH_TREE *
TR_find_closest (const TRH_TREE *node, const TRH_KEY_TYPE key)
{
  int		c_val ;
  const TRH_TREE *prev_node ;
  const TRH_TREE *next_node  ;

    /*
     * Non recursive implementation.
     *
     * Zip through the tree searching for the key. The tree is ordered by
     * alphabetical order of the key.
     */

    for (prev_node=node ; node!=TRH_NULL ; prev_node=node, node=next_node)
    {
	c_val = TRH_key_compare(key, node->key) ;

	if (c_val == 0)
	{
	    return (node) ;
	}

	next_node = (c_val < 0) ? node->left : node->right ;

	if (next_node == TRH_NULL)
	    return ((prev_node->left == node) ? prev_node : node) ;
    }

    return (TRH_NULL) ;
}

/*******************************************************************************
 * TR_find_nth
 *
 * A routine to scan the specified tree to find the nth element in the tree.
 * If there are fewer than n elements in the tree, a null pointer is returned.
 */

const TRH_TREE    *
TR_find_nth (const TRH_TREE *node, Ushort n)
{
    if (node == TRH_NULL)
    {
	fprintf (stderr, "TR_find_nth: Called with TRH_NULL pointer\n") ;
	exit (1) ;
    }

    for (; node->left != TRH_NULL ; node=node->left) ;

    for (; n>1 ; n--)
    {
	if (node->next == TRH_NULL)
	{
	    return (TRH_NULL) ;
	}

	node = node->next ;
    }

    return (node) ;
}

/*******************************************************************************
 * TR_balance
 *
 * A routine to rebalance a previously balanced tree which has had one
 * change made to it.
 *
 * It requires, as input, a pointer to the node which has been added, or the
 * lowest node which has been affected by the change. The procedure then scans
 * upwards in the tree hierachy correcting balance as it goes.
 *
 * The balancing algorithm is "Height Balancing". A height balanced tree is one
 * in which at every node, the heights of the two child nodes differ by at
 * most 1.
 *
 * TR_balance assumes that it has been called with interrupts disabled.
 */

#ifdef  TRH_HBAL

static void
TR_balance (TRH_TREE **root_p, TRH_TREE *node)
{
    TRH_TREE	*child, *child2 ;
    short	clh, crh, nlh, nrh ;

    /*
     * IMPORTANT ASSUMPTIONS:
     * 1) This routine has not been called with null parameters.
     *
     * 2) The tree was previously in balance and a single operation has
     *	  potentially put the tree out of balance.
     */

    for (;;)
    {
	child = node ;

	/*
	 * Firstly, reset the weight of the present node.
	 */

	nlh = nrh = clh = crh = 0 ;

	if (child->left != TRH_NULL)
	    clh = child->left->height ;

	if (child->right != TRH_NULL)
	    crh = child->right->height ;

	child->height = ((clh > crh) ? clh : crh) + 1 ;

	/*
	 * Check to see if we should exit the infinite loop
	 */

	if ((node = child->parent) == TRH_NULL)
	    break ;

	if (node->left != TRH_NULL)
	    nlh = node->left->height ;

	if (node->right != TRH_NULL)
	    nrh = node->right->height ;

	/*
	 * balance the present node
	 */

	if ((nlh - nrh) > 1)
	{
	    child = node->left ;
	    clh	  = crh = 0 ;

	    if (child->left != TRH_NULL)
		clh = child->left->height ;

	    if (child->right != TRH_NULL)
		crh = child->right->height ;

	    if (clh > crh)
	    {
		/*
		 * This node is out of balance because the left branch is
		 * heavier than the right by > 1.
		 * -> left branch MUST therefore be non null.
		 *
		 * The unbalance is suitable for correction by a single rotate
		 * because the child's left branch is heavier.
		 *
		 *  From:	     To:
		 *	   C		      B
		 *	* / \		     / \	   (*  = heavy branch)
		 *	 /   \		    /	\	   (Xn = anything)
		 *	B     X3	   A	 C
		 *   * / \		  /	/ \
		 *    /	  \		 /     /   \
		 *   A	   X2		X1    X2   X3
		 *  /
		 * X1
		 *
		 * node  ==> C
		 * child ==> B
		 *
		 * Place X2 (child->right) under C (node->left)
		 */

		if ((node->left = child->right) != TRH_NULL)
		    child->right->parent = node ;

		/*
		 * Place B (child) on top of C (node)
		 */

		if (node->parent == TRH_NULL)
		    *root_p = child ;
		else if (node->parent->left == node)
		    node->parent->left = child ;
		else
		    node->parent->right = child ;

		child->parent	= node->parent ;
		child->right	= node ;
		node->parent	= child ;
	    }
	    else
	    {
		/*
		 * This node is out of balance because the left branch is
		 * heavier than the right by > 1.
		 * -> left branch MUST therefore be non null.
		 *
		 * The unbalance is suitable for correction by a double rotate
		 * because the child's right branch is heavier.
		 *
		 *  From:	     To:
		 *	 C		  B
		 *    * / \		 / \	      (*  = heavy branch)
		 *     /   \		/   \	      (Xn = anything)
		 *    A	    X4	       A     C
		 *   / \ *	      / \   / \
		 *  /	\	     X1 X2 X3 X4
		 * X1	 B
		 *	/ \
		 *    X2   X3
		 *
		 * node	  ==> C
		 * child  ==> A
		 * child2 ==> B
		 */

		child2 = child->right ;

		if (node->parent == TRH_NULL)
		    *root_p = child2 ;
		else if (node->parent->left == node)
		    node->parent->left = child2 ;
		else
		    node->parent->right = child2 ;

		/*
		 * Place X2 (child2->left) into position under A (child)
		 */

		if ((child->right = child2->left) != TRH_NULL)
		    child2->left->parent = child ;

		/*
		 * Place X3 (child2->right) into position under C (node)
		 */

		if ((node->left = child2->right) != TRH_NULL)
		    child2->right->parent = node ;

		/*
		 * Place B (child2) into position over A and C
		 */

		child2->parent  = node->parent ;
		child2->left	= child ;
		child2->right	= node ;
		child->parent	= child2 ;
		node->parent	= child2 ;

		/*
		 * Adjust the height of A, it won't be touched again
		 */

		clh = crh = 0 ;

		if (child->left != TRH_NULL)
		    clh = child->left->height ;

		if (child->right != TRH_NULL)
		    crh = child->right->height ;

		child->height = ((clh > crh) ? clh : crh) + 1;
	    }
	}
	else if ((nrh - nlh) > 1)
	{
	    child = node->right ;
	    clh	  = crh = 0 ;

	    if (child->left != TRH_NULL)
		clh = child->left->height ;

	    if (child->right != TRH_NULL)
		crh = child->right->height ;

	    if (crh > clh)
	    {
		/*
		 * This node is out of balance because the right branch is
		 * heavier than the left by > 1.
		 * -> right branch MUST therefore be non null.
		 *
		 * The unbalance is suitable for correction by a single rotate
		 * because the child's right branch is heavier.
		 *
		 *  From:	     To:
		 *	   A		      B
		 *	  / \ *		     / \	   (*  = heavy branch)
		 *	 /   \		    /	\	   (Xn = anything)
		 *	X1    B		   A	 C
		 *	     / \ *	  / \	  \
		 *	    /	\	 /   \	   \
		 *	   X2	 C	X1   X2	   X3
		 *		  \
		 *		  X3
		 *
		 * node  ==> A
		 * child ==> B
		 *
		 * Place X2 (child->left) under A (node->right)
		 */

		if ((node->right = child->left) != TRH_NULL)
		    child->left->parent = node ;

		/*
		 * Place B (child) on top of A (node)
		 */

		if (node->parent == TRH_NULL)
		    *root_p = child ;
		else if (node->parent->left == node)
		    node->parent->left = child ;
		else
		    node->parent->right = child ;

		child->parent	= node->parent ;
		child->left	= node ;
		node->parent	= child ;
	    }
	    else
	    {
		/*
		 * This node is out of balance because the right branch is
		 * heavier than the left by > 1.
		 * -> right branch MUST therefore be non null.
		 *
		 * The unbalance is suitable for correction by a double rotate
		 * because the child's left branch is heavier.
		 *
		 *  From:	     To:
		 *	 A		  B
		 *	/ \ *		 / \	      (*  = heavy branch)
		 *     /   \		/   \	      (Xn = anything)
		 *    X1    C	       A     C
		 *	 * / \	      / \   / \
		 *	  /   \	     X1 X2 X3 X4
		 *	 B    X4
		 *	/ \
		 *    X2   X3
		 *
		 * node	  ==> A
		 * child  ==> C
		 * child2 ==> B
		 */

		child2 = child->left ;

		if (node->parent == TRH_NULL)
		    *root_p = child2 ;
		else if (node->parent->left == node)
		    node->parent->left = child2 ;
		else
		    node->parent->right = child2 ;

		/*
		 * Place X2 (child2->left) into position under A (node)
		 */

		if ((node->right = child2->left) != TRH_NULL)
		    child2->left->parent = node ;

		/*
		 * Place X3 (child2->right) into position under C (child)
		 */

		if ((child->left = child2->right) != TRH_NULL)
		    child2->right->parent = child ;

		/*
		 * Place B (child2) into position over A and C
		 */

		child2->parent  = node->parent ;
		child2->left	= node ;
		child2->right	= child ;
		child->parent	= child2 ;
		node->parent	= child2 ;

		/*
		 * Adjust the height of C, it won't be touched again
		 */

		clh = crh = 0 ;

		if (child->left != TRH_NULL)
		    clh = child->left->height ;

		if (child->right != TRH_NULL)
		    crh = child->right->height ;

		child->height = ((clh > crh) ? clh : crh) + 1;
	    }
	}
    }
}

#endif

/*******************************************************************************
 * TR_add_node
 *
 * A routine to add the specified node into the correct location in the
 * specified tree. If the node already exists in the tree, the old one is
 * replaced and discarded.
 *
 * The node to be added should be created by the caller.
 * <TR_create_node> (above) can be used to help with this operation.
 */

void
TR_add_node (TRH_TREE **root_p, TRH_TREE *newnode)
{
    int		c_val = 0 ;
    TRH_TREE	*ptr = *root_p ;
    TRH_TREE	*tmp = (TRH_TREE *)0 ;

    /*
     * Non recursive implementation.
     *
     * If the tree is empty, the new node becomes the root of the tree.
     *
     * Zip through the tree searching for the key. (The key of the new node)
     *
     * If the key is located, the node containing the key is replaced with
     * the new one and the old node disposed of.
     *
     * If the tree is non empty and the node is not located, the new node
     * is added at the correct place.
     */

    if (*root_p == TRH_NULL)
    {
	newnode->height = 1 ;
	newnode->left = newnode->right = newnode->parent =
	newnode->prev = newnode->next = TRH_NULL ;
	*root_p = newnode ;
    }
    else
    {
	while (ptr != TRH_NULL)
	{
	    c_val = TRH_key_compare (newnode->key, ptr->key) ;

	    if (c_val == 0)
		break ;

	    tmp = ptr ;
	    ptr = (c_val < 0) ? ptr->left : ptr->right ;
	}

	if (c_val == 0)
	{
	    /*
	     * The key has been located in the tree already. Replace it.
	     *
	     * The children and parent of the new node must be duplicated from
	     * the old node. If the old node had non null children, those
	     * children's parent pointers must be updated to point to the
	     * new node and not to the old node. The same goes for the old
	     * parent.
	     */

#ifdef  TRH_DEBUG
	    fprintf (stderr,
#ifdef  TRH_STRING_KEY
		"Identified duplicate node: node=<%s>\n",
#else
		"Identified duplicate node: node=<%08lx>\n",
#endif
		newnode->key) ;
#endif

	    newnode->height = ptr->height ;

	    if ((newnode->left = ptr->left) != TRH_NULL)
		newnode->left->parent = newnode ;

	    if ((newnode->right = ptr->right) != TRH_NULL)
		newnode->right->parent = newnode ;

	    if ((newnode->prev = ptr->prev) != TRH_NULL)
		newnode->prev->next = newnode ;

	    if ((newnode->next = ptr->next) != TRH_NULL)
		newnode->next->prev = newnode ;

	    if ((newnode->parent = ptr->parent) != TRH_NULL)
	    {
		if (newnode->parent->left == ptr)
			newnode->parent->left  = newnode ;
		else	newnode->parent->right = newnode ;
	    }
	    else
	    {
		/*
		 * If the parent of the node we are replacing is null, we must
		 * be the root node. The root pointer needs updating.
		 */

		*root_p = newnode ;
	    }

	    free (ptr) ;
	}
	else
	{
	    /*
	     * A node with a matching key has not been located. <tmp> points
	     * to the leaf node below which the new node will be added.
	     */

	    newnode->height = 1 ;
	    newnode->left = newnode->right = TRH_NULL ;
	    newnode->parent = tmp ;

	    if (c_val < 0)
	    {
		tmp->left = newnode ;
		newnode->next = tmp ;

		if ((newnode->prev = tmp->prev) != TRH_NULL)
		    newnode->prev->next = newnode ;

		tmp->prev = newnode ;
	    }
	    else
	    {
		tmp->right = newnode ;
		newnode->prev = tmp ;

		if ((newnode->next = tmp->next) != TRH_NULL)
		    newnode->next->prev = newnode ;

		tmp->next = newnode ;
	    }
	}

	/*
	 * Now that we've added the new node, we need to check the balance
	 * of the tree and correct it if necessary.
	 */

#ifdef  TRH_HBAL
	TR_balance (root_p, newnode) ;
#endif
    }
}

/*******************************************************************************
 * TR_addstr_node
 *
 * This routine is similar to TR_add_node except that the decision to
 * discard duplicate nodes is based on duplication of both the key and the
 * STRING_PTR parameter.
 */

#ifdef  TRH_STRING_DATA

void
TR_addstr_node (TRH_TREE **root_p, TRH_TREE *newnode)
{
    int		c_val ;
    TRH_TREE	*ptr = *root_p ;
    TRH_TREE	*tmp ;

    tmp = (TRH_TREE *)0 ;

    /*
     * Non recursive implementation.
     *
     * If the tree is empty, the new node becomes the root of the tree.
     *
     * Zip through the tree searching for the key. (The key of the new node)
     *
     * If the tree is non empty and the node is not located, the new node
     * is added at the correct place.
     */

    if (*root_p == TRH_NULL)
    {
	newnode->height = 1 ;
	newnode->left = newnode->right = newnode->parent =
	newnode->prev = newnode->next = TRH_NULL ;
	*root_p = newnode ;
    }
    else
    {
	while (ptr != TRH_NULL)
	{
	    c_val = TRH_key_compare (newnode->key, ptr->key) ;

	    if (c_val == 0)
		break ;

	    tmp = ptr ;
	    ptr = (c_val < 0) ? ptr->left : ptr->right ;
	}

	if (c_val == 0)
	{
	    /*
	     * The key has been located in the tree already. Search the
	     * STRING_PTR parameter to determine whether to keep or
	     * discard the new node.
	     */

	    c_val = strcmp (newnode->DESCR_PTR, ptr->DESCR_PTR) ;
	}

	if (c_val == 0)
	{
	    /*
	     * The key with accompanying STRING_PTR has been located in the
	     * tree already. Replace it.
	     *
	     * The children and parent of the new node must be duplicated from
	     * the old node. If the old node had non null children, those
	     * children's parent pointers must be updated to point to the
	     * new node and not to the old node. The same goes for the old
	     * parent.
	     */

#ifdef  TRH_DEBUG
	    fprintf (stderr,
		"Identified duplicate node: node=<%s>, data=<%s>\n",
		newnode->key, newnode->DESCR_PTR) ;
#endif

	    newnode->height = ptr->height ;

	    if ((newnode->left = ptr->left) != TRH_NULL)
		newnode->left->parent = newnode ;

	    if ((newnode->right = ptr->right) != TRH_NULL)
		newnode->right->parent = newnode ;

	    if ((newnode->prev = ptr->prev) != TRH_NULL)
		newnode->prev->next = newnode ;

	    if ((newnode->next = ptr->next) != TRH_NULL)
		newnode->next->prev = newnode ;

	    if ((newnode->parent = ptr->parent) != TRH_NULL)
	    {
		if (newnode->parent->left == ptr)
			newnode->parent->left  = newnode ;
		else	newnode->parent->right = newnode ;
	    }
	    else
	    {
		/*
		 * If the parent of the node we are replacing is null, we must
		 * be the root node. The root pointer needs updating.
		 */

		*root_p = newnode ;
	    }

	    free (ptr) ;
	}
	else
	{
	    /*
	     * A duplicate node has not been located. <tmp> points
	     * to the leaf node below which the new node will be added.
	     */

	    newnode->height	= 1 ;
	    newnode->left	= newnode->right = TRH_NULL ;
	    newnode->parent	= tmp ;

	    if (c_val < 0)
	    {
		tmp->left = newnode ;
		newnode->next = tmp ;

		if ((newnode->prev = tmp->prev) != TRH_NULL)
		    newnode->prev->next = newnode ;

		tmp->prev = newnode ;
	    }
	    else
	    {
		tmp->right = newnode ;
		newnode->prev = tmp ;

		if ((newnode->next = tmp->next) != TRH_NULL)
		    newnode->next->prev = newnode ;

		tmp->next = newnode ;
	    }

	    /*
	     * Now that we've added the new node, we need to check the balance
	     * of the tree and correct it if necessary.
	     */

#ifdef  TRH_HBAL
	    TR_balance (root_p, newnode) ;
#endif
	}
    }
}

#endif

/*******************************************************************************
 * TR_delete_node
 *
 * This routine is used to remove nodes from a tree. If the node is not
 * found, no action is taken. This routine will automatically reorder the tree
 * so that the binary tree structure is maintained after a deletion.
 */

int
TR_delete_node (TRH_TREE **root_p, const TRH_KEY_TYPE key)
{
  TRH_TREE	*node, *child_p, *lowest ;

  if ((node = (TRH_TREE *)TR_find_node (*root_p, key)) != TRH_NULL)
  {
    /*
     * <node> now points to the node which is to be deleted.
     *
     * The algorithm for deleting a node depends on how many children a node
     * has.
     */

    if (node->prev != TRH_NULL)
      node->prev->next = node->next ;

    if (node->next != TRH_NULL)
      node->next->prev = node->prev ;

    if (((node->left == TRH_NULL) || (node->right == TRH_NULL)) &&
	(node->left != node->right))
    {
      /*
       * The node to be deleted has one child only. Simply promote the
       * child to take the place of the deleted parent.
       *
       *	    A					    A
       *	   / \					   / \
       *	 X1   B	    <- eg. delete this node.	 X1   X2
       *	       \
       *		X2
       *
       *	    A					    A
       *	   / \					   / \
       *	 X1   B	    <- eg. delete this node.	 X1   X2
       *	     /
       *	   X2
       *
       * node    ==> B
       * child_p ==> X2
       */

      child_p = (node->left == TRH_NULL) ? node->right : node->left ;

      if ((child_p->parent = node->parent) == TRH_NULL)
      {
	/*
	 * We're deleting the root node (with 1 child).
	 */

	*root_p = child_p ;
      }
      else
      {
	/*
	 * we're deleting an interior node with 1 child.
	 */

	if (node->parent->left == node)
	{
	  node->parent->left  = child_p ;

#ifdef  TRH_HBAL
	  if (node->parent->right != TRH_NULL)
	    TR_balance (root_p, node->parent->right) ;
	  else
	    TR_balance (root_p, child_p) ;
#endif
	}
	else
	{
	  node->parent->right = child_p ;

#ifdef  TRH_HBAL
	  if (node->parent->left != TRH_NULL)
	    TR_balance (root_p, node->parent->left) ;
	  else
	    TR_balance (root_p, child_p) ;
#endif
	}
      }
    }
    else if (node->left != node->right)
    {
      /*
       * The node to be deleted has two children.
       */

      child_p = node->right ;

      while (child_p->left != TRH_NULL)
	child_p = child_p->left ;

      if (node->right == child_p)
      {
	/*
	 * the right child of the node being deleted is the leftmost
	 * in the sub-tree.
	 *
	 * This is a simple operation.....
	 *
	 *	  D				      D
	 *	 / \				     / \
	 *	A   X5				    A	X5
	 *	 \				     \
	 *	  B	    <- delete node B	      C
	 *	 / \				     / \
	 *     X1   C	    <- promote node C	   X1	X2
	 *	     \
	 *	      X2
	 *
	 * node	   ==> B
	 * child_p ==> C
	 */

	child_p->left	      = node->left ;
	child_p->left->parent = child_p ;
	lowest = child_p->left ;

	/*
	 * We still need to make node's former parent point to child_p
	 */
      }
      else
      {
	/*
	 * <child_p> points to the leftmost child, of the right sub-tree
	 * but it still may have a RIGHT child of its own.
	 *
	 * Remove the child as per deleting with 0 or 1 children.
	 * If <child_p> has any children at all, they MUST be in the
	 * RIGHT leg.
	 *
	 * Promote the leftmost child of the right branch of the node
	 * being deleted.
	 *
	 *	   D					 D
	 *	  / \					/ \
	 *	 A   X5				       A   X5
	 *	  \					\
	 *	   B	       <- delete node B		 C
	 *	  / \					/ \
	 *	X1   X3				      X1   X3
	 *	    /  \				  /  \
	 *	   C	X4     <- promote node C	X2    X4
	 *	    \
	 *	     X2
	 *
	 * node	   ==> B
	 * child_p ==> C
	 */

	if (child_p->right != TRH_NULL)
	{
	  lowest	   = child_p->right ;
	  lowest->parent = child_p->parent ;
	}
	else if (child_p->parent->right != TRH_NULL)
	  lowest = child_p->parent->right ;
	else
	  lowest = child_p->parent ;

	child_p->parent->left = child_p->right ;

	/*
	 * <child_p> has now been removed from the tree. It is free to
	 * replace <node>.
	 */

	child_p->left	       = node->left ;
	child_p->left->parent  = child_p ;
	child_p->right	       = node->right ;
	child_p->right->parent = child_p ;

	/*
	 * We still need to make node's former parent point to child_p
	 */
      }

      if ((child_p->parent = node->parent) == TRH_NULL)
	*root_p = child_p ;
      else if (node->parent->left == node)
	child_p->parent->left  = child_p ;
      else
	child_p->parent->right = child_p ;

      /*
       * Now that we've deleted the node, rebalance the tree
       */

#ifdef  TRH_HBAL
      TR_balance (root_p, lowest) ;
#endif
    }
    else
    {
      /*
       * The node to be deleted has no children. It is a leaf node. Simply
       * sever the tie between it and its parent and free the memory.
       *
       *		 C
       *		/ \
       *	       A   X2
       *	      / \
       *	    X1	 B	 <- eg. delete this node.
       */

      if (node->parent == TRH_NULL)
      {
	/*
	 * We're deleting the root node (no children).
	 */

	*root_p = TRH_NULL ;
      }
      else
      {
	/*
	 * we're deleting an exterior node (0 children).
	 */

	if (node->parent->left == node)
	{
	  node->parent->left  = TRH_NULL ;

#ifdef  TRH_HBAL
	  if (node->parent->right != TRH_NULL)
	    TR_balance (root_p, node->parent->right) ;
	  else
	    TR_balance (root_p, node->parent) ;
#endif
	}
	else
	{
	  node->parent->right = TRH_NULL ;

#ifdef  TRH_HBAL
	  if (node->parent->left != TRH_NULL)
	    TR_balance (root_p, node->parent->left) ;
	  else
	    TR_balance (root_p, node->parent) ;
#endif
	}
      }
    }

    /*
     * having delinked the node to be deleted from the tree,
     * return its memory allocation to the pool.
     */

    free (node) ;
    return 1 ;
  }
  else
  {
    /*
     * the node to be deleted has NOT been found!
     * Ignore it.
     */

    fprintf (stderr,
#ifdef  TRH_STRING_KEY
	     "Node not found: <%s>\n",
#else
	     "Node not found: <%#lx>\n",
#endif
	     key) ;
  }

  return 0 ;
}

/*******************************************************************************
 * TR_merge_tree
 *
 * This routine may be used to add all nodes from the second specified tree to
 * the first. A new copy of the nodes are made. The first tree is left intact.
 *
 * TRH_fix_ptr must be correctly defined to correct any internally directed
 * pointers. The node is otherwise copied byte for byte.
 *
 * Note the call to the macro TRH_fix_ptr(). This is defined by the user in
 * treedef.h and is indended to be used in the TR_merge_tree routine only.
 * Its purpose is to correct any pointers to user data after a node has been
 * literally copied.
 *
 *	eg.				SOURCE struct		unfixed COPY
 *
 *	TRH_TREE	*left		XXX			XXX
 *	TRH_TREE	*right		XXX			XXX
 *	TRH_TREE	*parent		XXX			XXX
 *	int		size		SSS			SSS
 *	TRH_TREE_INFO	STRING_PTR	-----			  -----
 *	char		key[]		key   |			key   |
 *	<<user data>>			  X----  <---------	  X   |
 *							  |	      |
 *							  -------------
 *
 * The fixed copy of the node will have the STRING_PTR correctly pointing to
 * the new copy of the user data. (Note: this is only for when the user data
 * contains a inwardly directed pointer)
 */

void
TR_merge_tree (TRH_TREE **into, TRH_TREE *root)
{
    TRH_TREE	*node ;
    TRH_TREE	*tmp ;

    /*
     * Find the left most node, which is the first ordered element. Then
     * scan along the linear linked list of leaves to merge the trees.
     */

    for (node=root ; node->left != TRH_NULL ; node=node->left) ;

    for (; node != TRH_NULL ; node=node->next)
    {
	tmp = (TRH_TREE *)malloc (node->size) ;
	memcpy ((char *)tmp, (char *)node, (unsigned)(node->size)) ;
	TRH_fix_ptr(tmp) ;
	TR_add_node (into, tmp) ;
    }
}

/*******************************************************************************
 * TR_free_tree
 *
 * This routine discards all nodes in the specified tree. Following
 * this operation, the <root> pointer is reset to point to a null tree.
 */

void
TR_free_tree (TRH_TREE **root_p)
{
    TRH_TREE	*node ;
    TRH_TREE	*tmp ;

    if (*root_p == TRH_NULL)
    {
	return ;
    }

    /*
     * Find the left most node, which is the first ordered element. Then
     * scan along the linear linked list of leaves to delete the nodes.
     */

    for (node=*root_p ; node->left != TRH_NULL ; node=node->left) ;

    while (node != TRH_NULL)
    {
	tmp = node->next ;
	free (node) ;
	node = tmp ;
    }

    *root_p = TRH_NULL ;
}

/*******************************************************************************
 * TR_local_display_tree and TR_display_tree
 *
 * These routines are used for debugging purposes only. They display a "pretty"
 * output representing the tree.
 *
 * It is assumed that TR_local_display_tree is called with interrupts disabled
 */

#ifdef  TRH_DEBUG

static void
TR_local_display_tree (TRH_TREE *node, char *str, char *trailer)
{
		char	*newstr = (char *)0 ;
    static	char	*s ;
		char	*d ;
    static	char	c ;

    /*
     * Display the contents of a tree on a CRT. Only keys are shown.
     *
     * Tree has the following format...
     *
     *		 -- h:node
     *		|
     *	     -- h:node
     *	    |	|
     *	    |	|    -- h:node
     *	    |	|   |
     *	    |	 -- h:node
     *	    |
     *  --- h:node
     *	    |
     *	     -- h:node
     *		|
     *		 -- h:node
     */

    if (node->right != TRH_NULL)
    {
	newstr = (char *)malloc (strlen(str) + 3) ;

	if (node->parent != TRH_NULL)
	{
	    if ((node->parent->left != node) && (node->parent->right != node))
		fprintf (stderr,
#ifdef  TRH_STRING_KEY
			"Node %d:%s, Node's parent doesn't point to it!%s\n",
#else
			"Node %d:%08lx, Node's parent doesn't point to it!%s\n",
#endif

			node->height, node->key, trailer) ;
	}
	else if (*str != '\0')
	{
	    fprintf (stderr,
#ifdef  TRH_STRING_KEY
			"Node %d:%s, Node has null parent but is not root%s\n",
#else
			"Node %d:%08lx, Node has null parent but is not root%s\n",
#endif
			node->height, node->key, trailer) ;
	}

	s = str ;
	d = newstr ;
	while ((*d++ = *s++) != 0) ;
	*d-- = '\0' ;
	*d++ = 'r' ;
	TR_local_display_tree (node->right, newstr, trailer) ;
	*d++ = 'x' ;
	*d = '\0' ;
	s = newstr ;
	c = *s++ ;

	while (*s)
	{
	    printf ((c == *s) ? "    " : "   |") ;
	    c = *s++ ;
	}

	printf ("%s\n", trailer) ;
    }

    s = str ;

    while ((c=*s++) != 0)
    {
	printf ("%s", ((c == *s) || (*s == '\0')) ? "    " : "   |") ;
    }

#ifdef  TRH_STRING_KEY
    printf ("-- %d:%s %s\n", node->height, node->key, trailer) ;
#else
    printf ("-- %d:%08lx%s\n", node->height, node->key, trailer) ;
#endif

    if (node->left != TRH_NULL)
    {
	if (newstr == (char *)0)
	    newstr = (char *)malloc (strlen(str) + 3) ;

	s = str ;
	d = newstr ;
	while ((*d++ = *s++) != 0) ;
	*--d = 'l' ;
	*++d = 'x' ;
	*++d = '\0' ;
	s = newstr ;
	c = *s++ ;

	while (*s)
	{
	    printf ((c == *s) ? "    " : "   |") ;
	    c = *s++ ;
	}

	printf ("%s\n", trailer) ;

	*--d = '\0' ;
	TR_local_display_tree (node->left, newstr, trailer) ;
    }

    if (newstr != (char *)0)
	free (newstr) ;
}

void
TR_display_tree (TRH_TREE *root, char *trailer)
{
    /*
     * Display the contents of a tree on a CRT. Only keys are shown.
     */

    if (root == TRH_NULL)
	printf ("Null tree%s\n", trailer) ;

    else TR_local_display_tree (root, "", trailer) ;
}

#endif
