/*
 * tree.h
 ******************************************************************************
 * Copyright (C) 2002 Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ******************************************************************************
 * This module contains routines to maintain binary trees. The routines can be
 * used with multiple tree types by altering the <TRH_TREE_INFO> union or
 * structure defined in the application specific tree definition structure.
 *
 * The <typedef> TRH_TREE_INFO can be declared by the user as they wish.
 * It may be a union, a structure or whatever.
 *
 * Dynamically alterable memory (eg. character strings) can be requested at
 * TR_create_node() time. It can be accessed as the 1st byte following the
 * '\0' termination character of the <<TRH_TREE -> key>> field.
 */

#ifndef		_TREE_H
#define		_TREE_H

/*
 * define TRH_HBAL if the height balancing procedures are required
 */

#define TRH_HBAL

/*
 * define TRH_STRING_KEY if the key will be non numeric. Comparisons about
 * order of nodes will be performed on a string rather than binary basis.
 */

#define	TRH_STRING_KEY

/*
 * define TRH_STRING_DATA if part of the user's additional data structure
 * will include an ascii string. This allows addition of nodes with the same
 * key but different contents.
 */

/*
#define	TRH_STRING_DATA
*/

/*
 * define TRH_DEBUG to turn on compilation of the code to dump trees in ascii
 */

#define TRH_DEBUG

/*******************************************************************************
 * The nodal data structure is defined here. Note that additional data
 * definitions (eg. a UNION to hold other info) should be added where noted
 *
 * left		These parameters store the pointers which make this data
 * right	structure a binary tree
 * parent
 *
 * prev		These parameters store the pointers which link the leaves
 * next		together into a linear linked list
 *
 * height	When using height balanced trees, this parameter stores the
 *		previously calculated height of the node. It is used to simplify
 *		the rebalancing operation after node addition or deletion.
 *
 * size		This is the number of bytes which have been <M_malloc>ed for the
 *		node. It includes the size of the key and user data.
 *
 * user_offset  This is used to calculate a pointer to the byte following the
 *		key. It is useful when the user declares additional bytes to
 *		be allocated over and above the size of the TRH_TREE header.
 *
 * <<additional data>>  if require, must be defined here.
 *		THE ADDITIONAL DATA SHOULD NOT BE VARIABLE IN LENGTH SO THAT THE
 *		RELATIVE LOCATION OF THE KEY REMAINS CONSTANT
 *
 * key		the <key> field MUST be the last field in the tree
 *		structure. This is because the <tree.c> software creates
 *		space for the key string at the END of the structure.
 *		The <[1]> is for the NULL terminator at the end of the key.
 *
 *		Note that the user can still delare free form space after the
 *		key for user specific info which cannot be practically stored in
 *		this node structure.
 */

struct btree
{
  struct btree	*left, *right, *parent ;
  struct btree	*prev, *next ;
  unsigned short	height ;
  unsigned short	size ;
  unsigned short	user_offset ;

  /*
   * Define other data HERE
   * Additional strings should be defined as POINTERS to constant strings
   * which FOLLOW the key
   */

  Uchar		messier ; /* Messier catalogue number */
  Uchar		constellation ;
  Uchar		season ;
  Uchar		type ;
  Uchar		magnitude ; /* in tenths (0=Mag 3, 1=mag 2.9, ...) */
  Uchar		mag_diff ;  /* in tenths */
  Uchar		num_components ;
  char		NGC_comp ;
  char		IC_comp ;
  char		UGC_comp ;
  Uchar		rating ; /* 1=best, 7=lowest (from various sources) */
  char		angle ; /*1=face on, 5=edge on */
  Uchar		major ; /* size along major axis */
  Uchar		minor ; /* size along minor axis (or zero) */
  Uchar		alignment ; /* whether an alignment star or not */

  char		RA_apm ; /* Annual proper motion in arcsec/yr */
  char		DEC_apm ; /* Annual proper motion in arcsec/yr */

  Ushort	NGC ; /* New General Catalogue number */
  Ushort        HR ;  /* Harvard Revised catalogue number */
  Ushort	IC ;  /* Index Catalogue number */
  Ulong		RA ;  /* H:M:S, 20 bit significance */
  long		DEC ; /* +/-90 deg, 20 bit significance */
  Ushort	separation ; /* in arc seconds */
  Ushort	radius ; /* in tenths of solar units */

  Ulong		PPM ; /* PPM Position and Proper Motion Catalogue number */
  Ulong		SAO ; /* SAO Smithsonian Astronomical Catalogue number */
  Ulong		PGC ; /* PGC Catalogue number */
  Ulong		UGC ; /* UGC Catalogue number */
  Ulong		HD ;  /* Henry Draper catalogue number */
  Ulong		LY ; /* distance in tenths of LY */

  char		*common_name ;
  char		*formal_name ;
  char		*description ;
  char		*spectrum ;

#ifdef  TRH_STRING_KEY
  char		key [1] ;
#else
  Ulong		key ;
#endif
} ;

typedef struct btree	TRH_TREE ;

#define TRH_NULL	((TRH_TREE *)0)

/*******************************************************************************
 *
 */

#ifdef  TRH_STRING_KEY
#define TRH_key_copy(A,B)	strcpy(A,B)
#define TRH_key_compare(A,B)	strcmp(A,B)
#define TRH_key_len(A)		strlen(A)
typedef char	*TRH_KEY_TYPE ;

#else

#define TRH_key_copy(A,B)	A=B
#define TRH_key_compare(A,B)	((A<B)?-1:((A>B)?1:0))
#define TRH_key_len(A)		0
typedef Ulong	TRH_KEY_TYPE ;

#endif

#ifdef  TRH_STRING_DATA
extern  void		TR_addstr_node  (TRH_TREE **, TRH_TREE *) ;
#endif

#ifdef  TRH_DEBUG
extern  void		TR_display_tree (TRH_TREE *, char *) ;
#endif

/*******************************************************************************
 * The following definitions relate to additional data defined at the noted
 * place within the nodal data definition
 *
 * TRH_USER_PTR returns a pointer to the byte following the last byte of the
 * key.
 *
 * TRH_fix_ptr should be defined so that any pointers (as above) are copied
 * correctly if the node is copied
 */

#define TRH_USER_PTR(NODE)	(&((char *)NODE) [NODE->user_offset])

#ifdef  TRH_STRING_DATA
 #define TRH_fix_ptr(PTR)	node->common_name=TRH_USER_PTR(PTR)
#else
 #define TRH_fix_ptr(PTR)
#endif

extern  TRH_TREE    *TR_create_node	(const TRH_KEY_TYPE, const Ushort) ;
extern  void	    TR_copy_user_data	(TRH_TREE *, const char *) ;
extern  TRH_TREE    *TR_find_node	(TRH_TREE *, const TRH_KEY_TYPE) ;
extern  TRH_TREE    *TR_find_closest	(TRH_TREE *, const TRH_KEY_TYPE) ;
extern  TRH_TREE    *TR_find_nth	(TRH_TREE *, Ushort) ;
extern  void	    TR_add_node		(TRH_TREE **, TRH_TREE *) ;
extern  int	    TR_delete_node	(TRH_TREE **, const TRH_KEY_TYPE) ;
extern  void	    TR_merge_tree	(TRH_TREE **, TRH_TREE *) ;
extern  void	    TR_free_tree	(TRH_TREE **) ;

#ifdef  TRH_STRING_DATA
 extern  void       TR_node_infocopy    (TRH_TREE *node, const char *, const char *, const char *, const char *) ;
#endif

/******************************************************************************/

#endif
