/*
 * aux_driver.c
 ******************************************************************************
 * Copyright (C) 2002  Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. http://www.gnu.org
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ******************************************************************************
 * Routines to driver the 4099 based auxiliary port
 */

#include <stdio.h>
#include <string.h>
#include <asm/MC68EZ328.h> /* for all I/O port locations */
#include <unistd.h> /* for sleep, exit */
#include "types.h"
#include "asm.h"
#include "aux_driver.h"

/******************************************************************************
 * AUX_write_aux_bit
 *
 * Routine to write to the 4099 auxiliary output registers.
 */

void
AUX_write_aux_bit (Ushort address, BOOL state, BOOL in_supervisor)
{
  Uchar add_val ;
  Uchar write_enable ;
  Ulong ssp = 0 ; //shut up the compiler complaint

  if (address >= AUX_NUM_BITS)
    return ;

  if (address < 0x08)
    write_enable = 0xfd ;
  else if (address < 0x10)
    write_enable = 0xfe ;
  else
    write_enable = 0xfb ;

  add_val = (address << 4) & 0x70 ;
    
  if (state) /* set the data bit to the correct value */
    add_val |= 0x80 ;

  /* set the data lines appropriately */
  if (!in_supervisor)
    ssp = supervisor_mode() ;

  PCDATA &= 0x0f ; /* clear top order bits */
  PCDATA |= add_val ; /* write the address and data control bit out */

  /* Now toggle the appropriate *WR line */
  PEDATA &= write_enable ;
  PEDATA |= 0x07 ;

  if (!in_supervisor)
    user_mode (ssp) ;
}

/******************************************************************************
 * xxSEL - sets whether the pin is used for dedicated or general purpose I/O
 *   0 = the dedicated function of the pin is selected
 *   1 = the general purpose I/O function is selected (via xxDATA port)
 *
 * xxDIR - sets the direction of a parallel port
 *   0 = input
 *   1 = output
 *
 * xxPUEN - for ports A, B, D, E, selects whether pull up resistors are enabled
 *   0 = pull up disabled
 *   1 = pull up enabled
 *
 * xxPDEN - for port C, selects whether pull down resistors are enabled
 *   0 = pull down disabled
 *   1 = pull down enabled
 *
 * NOTE: Pull up and down functions should only be enabled on INPUTs. Enabling
 * the pull up / down on outputs just uses more power.
 */

void
AUX_init_ports (void)
{
  Ulong ssp ;
  Ushort i ;

  ssp = supervisor_mode() ;

  PBSEL |= PB_INIT ; /* Configure bits 6 and 7 for general purpose I/O */
  PBDIR |= 0xc0 ; /* Configure bits 6 and 7 for output */
  PBPUEN &= ~PB_INIT ; /* disable the pull up function for bits 6 and 7 */
  PBDATA &= ~PB_INIT ;

  PCSEL = 0xff ; /* Configure bits 0-7 for general purpose I/O */
  PCDIR = 0xf0 ; /* Configure bits 4-7 for output, 0-3 for input */
  PCPDEN = 0x0f ; /* enable the pull down function for bits 0-3 only */
  PCDATA = 0x00 ;

  PDSEL = 0xff ; /* Choose GP I/O for bits 0 to 7 */
  PDDIR = PD_INIT ; /* Choose output for bit 7 */
  PDPUEN = 0xff ; /* Enable the pull up resistor on bits 0 to 7 */
  PDDATA = 0x00 ;

  PESEL |= 0x07 ; /* Choose GP I/O for bits 0 to 2 */
  PEDIR |= 0x07 ; /* Choose output for bits 0 to 2 */
  PEPUEN &= ~0x07 ; /* Disable the pull up resistor for bits 0 to 2 */
  PEDATA |= 0x07 ;

  /* Now initialise all of the 4099 driven bits */
  for (i=0 ; i < AUX_NUM_BITS ; i++)
    AUX_write_aux_bit (i, (i < 16) ? 0 : 1, True) ;

  user_mode (ssp) ;
}
