/*
 * database.h
 *
 * Defines data structures used in database.c
 *******************************************************************************
 */

#ifndef _ASTRO_DATA
#define _ASTRO_DATA

/******************************************************************************
 * Define the following two constants to compile in routines to write and
 * display a compressed database entry
 */

//#define DB_WRITE
//#define DB_DISPLAY

// These are normally set in "types.h"
/******************************************************************************
 * The following structure is used for compression of entries in a database.
 *
 * Create an array of these structures. The first entry of the array is the
 * default entry. The remaining entries are keys which will be compressed. The
 * last entry must have "keywords" set to (char *)0.
 *
 * The "keywords" field is a list of LOWER CASE ONLY keywords, each separated
 * by a comman. If white space is included, it becomes significant. Normaly,
 * this comma separated list should contain no white space.
 *
 * The "description" field is what the compressed entry will be described as.
 */

struct _STRING_CLASSIFICATION
{
  char  *keywords ;
  char  *description ;
} ;

typedef struct _STRING_CLASSIFICATION STRING_CLASS ;

/******************************************************************************
 * The following structure is the first to be written to the database file and
 * defines the number of entries which are contained in the database.
 */

struct _NUM_ENTRY_STRUCT
{
  Ulong messier ;
  Ulong ugc ;
  Ulong ngc ;
  Ulong hr ;
  Ulong ic ;
  Ulong sao ;
  Ulong pgc ;
  Ulong hd ;
  Ulong ppm ;
  Ulong align ;
} ;

typedef struct _NUM_ENTRY_STRUCT NUM_ENTRIES ;

/******************************************************************************
 * The following information relates to the compressed record format used in
 * the database file.
 *
 * Each record consists of several fixed fields and several optional fields.
 * In order to save space in the database, the presence or absence of an
 * optional field is stored in a bit map. If the appropriate bit is "1", then
 * the corresponding field is present in the following data structure,
 * otherwise, the specified field is absent and the space is saved.
 */

#define	BIT_MESSIER	0x80000000
#define	BIT_NGC		0x40000000
#define	BIT_NGC_COMP	0x20000000
#define	BIT_UGC		0x10000000

#define	BIT_UGC_COMP	0x08000000
#define	BIT_HR		0x04000000
#define	BIT_IC		0x02000000
#define	BIT_IC_COMP	0x01000000

#define	BIT_SAO		0x00800000
#define	BIT_PGC		0x00400000
#define	BIT_HD		0x00200000
#define BIT_PPM		0x00100000

#define	BIT_CONST	0x00080000
#define	BIT_LY		0x00040000
#define	BIT_RATING	0x00020000
#define	BIT_SEASON	0x00010000

#define	BIT_TYPE	0x00008000
#define	BIT_ANGLE	0x00004000
#define	BIT_MAG		0x00002000
#define	BIT_MAG_DIFF	0x00001000

#define	BIT_RADIUS	0x00000800
#define	BIT_SEP		0x00000400
#define	BIT_MINOR	0x00000200
#define	BIT_MAJOR	0x00000100

#define	BIT_NUM_COMP	0x00000080
#define	BIT_COMMON_NAME	0x00000040
#define	BIT_FORMAL_NAME	0x00000020
#define	BIT_DESCRIPTION	0x00000010

#define	BIT_SPECTRUM	0x00000008
#define BIT_RA_APM	0x00000004
#define BIT_DEC_APM	0x00000002
#define	BIT_ALIGNMENT	0x00000001

#define BUFFER_MAX_SIZE 3000

struct _OBJECT_INFO
{
  Ulong		ra ;
  long		dec ;
  Ulong		bit_fields ;

  Uchar		messier ;
  Ushort	ngc ;
  char		ngc_comp ;
  Ulong		ugc ;
  char		ugc_comp ;
  Ushort	hr ;
  Ushort	ic ;
  char		ic_comp ;
  Ulong		sao ;
  Ulong		pgc ;
  Ulong		hd ;
  Ulong		ppm ;

  Uchar		constellation ;
  Ulong		ly ;
  Uchar		rating ;
  Uchar		season ;
  Uchar		type ;
  Uchar		angle ;
  Uchar		mag ;
  Uchar		mag_diff ;
  Ushort	radius ;
  Ushort	sep ;
  Uchar		minor ;
  Uchar		major ;
  Uchar		num_comp ;

  char		ra_apm ;
  char		dec_apm ;

  char		*common_name ;
  char		*formal_name ;
  char		*description ;
  char		*spectrum ;
  char		buffer [BUFFER_MAX_SIZE] ;
} ;

typedef struct _OBJECT_INFO OBJECT_INFO ;

/******************************************************************************
 */

extern const STRING_CLASS DB_obj_type[] ;
extern const STRING_CLASS DB_constellations[] ;
extern const STRING_CLASS DB_seasons[] ;
extern const STRING_CLASS DB_rating_type[] ;

#ifdef DB_WRITE
extern void	DB_write_header (FILE *f, NUM_ENTRIES *num) ;
extern int	DB_write_object_info (const OBJECT_INFO *o, FILE *f) ;
#endif

extern Ushort	DB_read_number (FILE *f) ;
extern void	DB_read_header (FILE *f, NUM_ENTRIES *num) ;
extern int	DB_read_object_info (OBJECT_INFO *o, FILE *f, const int brief) ;

#ifdef DB_DISPLAY
extern void	DB_display_object_info (const OBJECT_INFO *o, FILE *f) ;
#endif

#endif
