/*
 * encoders.c
 ******************************************************************************
 * Copyright (C) 2002  Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. http://www.gnu.org
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ******************************************************************************
 * Read input from the shaft encoders and display it.
 */

#include <stdio.h>
#include <asm/MC68EZ328.h> /* for all I/O port locations */

#include "types.h"
#include "encoders.h"
#include "asm.h"
#include "aux_driver.h"
#include "process.h"

// #define DEBUG_EVENT

int ENC_X_coord, ENC_Y_coord ;

///////////////////////////////////////////////////////////////////////////////

void
ENC_write_byte (Ushort control_reg, Ushort y_reg, Uchar val)
{
  Ulong ssp ;

  ssp = supervisor_mode() ;

  /* Set port D to output */
  PDDIR = 0xff ; /* All 8 bits configured for output */
  PDDATA = val ;  /* write the value to the port */

  /* Set up the address bits appropriately */
  AUX_write_aux_bit (AUX_ENC_CD, control_reg, True) ;
  AUX_write_aux_bit (AUX_ENC_XY, y_reg, True) ;

  /* Clock out the data */
  AUX_write_aux_bit (AUX_ENC_CS, 0, True) ;
  AUX_write_aux_bit (AUX_ENC_WR, 0, True) ;
  ENC_kill_time(5) ;
  AUX_write_aux_bit (AUX_ENC_WR, 1, True) ;
  AUX_write_aux_bit (AUX_ENC_CS, 1, True) ;

  user_mode (ssp) ;
}

///////////////////////////////////////////////////////////////////////////////

void
ENC_set_counter (Ushort y_reg, Ulong val)
{
  ENC_write_byte (CONTROL_REG, y_reg, BP_RESET) ;
  ENC_write_byte (DATA_REG, y_reg,  val     & 0xff) ;
  ENC_write_byte (DATA_REG, y_reg, (val>>8) & 0xff) ;
  ENC_write_byte (DATA_REG, y_reg, (val>>16)& 0xff) ;
  ENC_write_byte (CONTROL_REG, y_reg, XFER_PRE_CNTR) ;
}

///////////////////////////////////////////////////////////////////////////////

Uchar
ENC_read_byte (Ushort control_reg, Ushort y_reg)
{
  Ulong ssp ;
  Uchar val ;

  ssp = supervisor_mode() ;

  /* Set port D to output */
  PDDIR = 0x00 ; /* All 8 bits configured for input */

  /* Set up the address bits appropriately */
  AUX_write_aux_bit (AUX_ENC_CD, control_reg, True) ;
  AUX_write_aux_bit (AUX_ENC_XY, y_reg, True) ;

  /* Clock out the data */
  AUX_write_aux_bit (AUX_ENC_CS, 0, True) ;
  AUX_write_aux_bit (AUX_ENC_RD, 0, True) ;
  ENC_kill_time(5) ;
  val = PDDATA ;  /* read the value from the port */
  AUX_write_aux_bit (AUX_ENC_RD, 1, True) ;
  AUX_write_aux_bit (AUX_ENC_CS, 1, True) ;

  user_mode (ssp) ;
  return val ;
}

///////////////////////////////////////////////////////////////////////////////

int
ENC_read_counter (Ushort y_reg)
{
  Ulong val ;
  Uchar flag ;

  ENC_write_byte (CONTROL_REG, y_reg, XFER_CNTR_OL) ; // CNTR --> output latch
  ENC_write_byte (CONTROL_REG, y_reg, BP_RESET) ; // Reset the byte pointer

  val =  (Ulong)(ENC_read_byte (DATA_REG, y_reg))       & 0x000000ff ;
  val |= (Ulong)(ENC_read_byte (DATA_REG, y_reg) << 8)  & 0x0000ff00 ;
  val |= (Ulong)(ENC_read_byte (DATA_REG, y_reg) << 16) & 0x00ff0000 ;

  if (val & 0x00800000) /* convert to a real signed number */
    val &= 0x007fffff ;
  else
  {
    val &= 0x007fffff ;
    val -= 0x00800000 ;
  }

  flag = ENC_read_byte (CONTROL_REG, y_reg) ;

  if (flag & 0x10)
  {
    PR_log_msg (ERR, "Excessive noise on channel %d (%02x)\n", y_reg, flag) ;
    ENC_write_byte (CONTROL_REG, y_reg, EFLAG_RESET) ;
  }

  if (flag & 0x03)
  {
    PR_log_msg (ERR, "Overflow error on channel %d (%02x)\n", y_reg, flag) ;
    ENC_write_byte (CONTROL_REG, y_reg, FLAGS_RESET) ;
  }

  return (int)val ;
}

///////////////////////////////////////////////////////////////////////////////

int
ENC_read_coordinates (void)
{
  int new_x, new_y ;
  int change ;

  change = 0 ;
  new_x = ENC_read_counter (X_REG) ;
  new_y = ENC_read_counter (Y_REG) ;

  if (new_x != ENC_X_coord)
  {
    ENC_X_coord = new_x ;
    change = 1 ;

    #ifdef DEBUG_EVENT
      PR_log_msg (NORM, "X: %d\n", new_x - ENC_X_coord) ;
    #endif    
  }

  if (new_y != ENC_Y_coord)
  {
    ENC_Y_coord = new_y ;
    change = 1 ;

    #ifdef DEBUG_EVENT
      PR_log_msg (NORM, "Y: %d\n", new_y - ENC_Y_coord) ;
    #endif
  }

  return change ;
}

///////////////////////////////////////////////////////////////////////////////

int
ENC_init_hardware (void)
{
  Ulong ssp ;
  int error ;

  error = 0 ;

  /* Firstly, set all control signals to the right value */
  ssp = supervisor_mode() ;

  AUX_write_aux_bit (AUX_ENC_WR, 1, True) ;
  AUX_write_aux_bit (AUX_ENC_CD, 1, True) ;
  AUX_write_aux_bit (AUX_ENC_XY, 1, True) ;
  AUX_write_aux_bit (AUX_ENC_RD, 1, True) ;
  AUX_write_aux_bit (AUX_ENC_CS, 1, True) ;

  /* Now, start the PWM output as a clock generator */
  /* PCLK = 16.67 MHz / (16*2)=521 kHz */
  PWMC = ((15 << PWMC_PRESCALER_SHIFT) & PWMC_PRESCALER_MASK) | PWMC_EN ;

  /* PWMO freq = PCLK / (3+2)= 104 kHz */
  PWMP = 3 ;

  /* I'm guessing here at how to set the samples */
  PWMS = 0x0202 ;

  user_mode (ssp) ;

#ifdef DEBUG_EVENT
  PR_log_msg (NORM, "XFLAG = %02x before init\n", ENC_read_byte(1,0)) ;
  PR_log_msg (NORM, "YFLAG = %02x before init\n", ENC_read_byte(1,1)) ;
#endif

  ENC_write_byte (CONTROL_REG, 0, INDEX_DISABLE | BOTH_CHANNELS) ;
  ENC_write_byte (CONTROL_REG, 0, FLAGS_RESET | BOTH_CHANNELS) ;
  ENC_write_byte (CONTROL_REG, 0, EFLAG_RESET | BOTH_CHANNELS) ;
  ENC_write_byte (CONTROL_REG, 0, BP_RESET | BOTH_CHANNELS) ;
  ENC_write_byte (DATA_REG, X_REG, CLOCK_DATA) ;
  ENC_write_byte (CONTROL_REG, 0, CLOCK_SETUP | BOTH_CHANNELS) ;
  ENC_write_byte (CONTROL_REG, 0, INPUT_SETUP | BOTH_CHANNELS) ;
  ENC_write_byte (CONTROL_REG, 0, QUAD_X4 | BOTH_CHANNELS) ;
  ENC_write_byte (CONTROL_REG, 0, CNTR_RESET | BOTH_CHANNELS) ;

  ENC_set_counter (X_REG, 0x800000) ;
  ENC_X_coord = ENC_read_counter (X_REG) ;

  ENC_set_counter (Y_REG, 0x800000) ;
  ENC_Y_coord = ENC_read_counter (Y_REG) ;

  if (ENC_X_coord != 0)
  {
    error++ ;

  #ifdef DEBUG_EVENT
    PR_log_msg (ERR, "X reg programmed to 0x800000: reads %x (should be 0)\n", ENC_X_coord) ;
  #endif
  }


  if (ENC_Y_coord != 0)
  {
    error++ ;

  #ifdef DEBUG_EVENT
    PR_log_msg (ERR, "Y reg programmed to 0x800000: reads %x (should be 0)\n", ENC_Y_coord) ;
  #endif
  }

#ifdef DEBUG_EVENT
  PR_log_msg (NORM, "XFLAG = %02x after init\n", ENC_read_byte(1,0)) ;
  PR_log_msg (NORM, "YFLAG = %02x after init\n", ENC_read_byte(1,1)) ;
#endif

  if (error)
    PR_log_msg (ERR, "X or Y register did not read back what was programmed\n") ;

  return error ;
}
