/*
 * lcd.c
 ******************************************************************************
 * Copyright (C) 2002  Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. http://www.gnu.org
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ******************************************************************************
 * A programme to drive a hitachi 44780 based LCD.
 */

#include <stdio.h>
#include <string.h>
#include <asm/MC68EZ328.h> /* for all I/O port locations */
#include <unistd.h> /* for sleep */
#include "types.h"
#include "asm.h"
#include "lcd.h"
#include "aux_driver.h"

/******************************************************************************
 */

const Uchar LCD_initialise_string[] =
{
  0x3f, 0x3f, 0x32, // force the LCD into 8 character mode

  0x22, // FUNCTION SET: select 4 bit operation
	// lower 4 bits are lost because LCD defaults to 8 bit mode

  0x21, // FUNCTION SET: Repeat so we now get chance to send low bits
  0x82, // 0x80 = 2 line display, ~0x40 = 5x7 font

  0x01, // CLEAR DISPLAY
  0x1f, // 0x10=cmd

  0x41, // Move cursor to start of programmable character #0
  0x01,

  0x01, // HOME CURSOR
  0x2f, // 0x20=cmd

  0x01, // DISPLAY ON/OFF
  0xc2, // 0x80=cmd, 0x40=disp on, ~0x20=cursor off, ~0x10=blink off

  0x01, // ENTRY MODE SET
  0x62, // 0x40=cmd, 0x20=inc cursor pos, ~0x10=don't shift display

  0x00
} ;

/******************************************************************************
 * LCD_write_lcd_control
 *
 * The parameter string is a '\0' terminated string of 4 bit commands to send
 * to the LCD device.
 *
 * A value of '\0' terminates the string and is not written to the LCD.
 *
 * For non zero values:
 *    the high order 4 bits of non zero values are written to the LCD.
 *
 *    the lower order 4 bits indicate the delay for this instruction...
 *
 *       1 indicates NO delay
 *       2 indicates >= 40 us delay
 *       F indicates >= 2 ms delay
 */

void
LCD_write_lcd_control (const Uchar *string)
{
  Ulong ssp ;
  const Uchar *p ;

  ssp = supervisor_mode() ;

  /* Set Register Select to "0" (control), E to "0" */
  RS_LOW() ;
  E_LOW() ;

  for (p=string ; *p ; p++)
  {
    /* set the data lines appropriately */
    PCDATA &= ~0xf0 ; /* clear top order bits */
    PCDATA |= *p & 0xf0 ; /* write new command value */

    /* Now toggle "E" line, falling edge latches the value */
    E_HIGH() ;
    E_LOW() ;
    LCD_kill_time (LCD_DELAY) ;

    /* Finally, delay */

    switch (*p & 0x0f)
    {
      case 0x02:
	LCD_kill_time (LCD_DELAY) ;
	break ;

      case 0x0f: // delay for 5ms
	usleep (5000) ; // usually, delay is >= one system tick (10ms)
	break ;
    }
  }

  user_mode (ssp) ;
}

/******************************************************************************
 * LCD_write_lcd_string
 */

#define LINE_LEN 1000

static Uchar line_init[5] ;
static char padded_line [LINE_LEN+LCD_LINE_LEN+1] ;

void
LCD_write_char (const Uchar c)
{
  /* write high order bits */
  PCDATA &= ~0xf0 ; /* clear top order bits */
  PCDATA |= c & 0xf0 ; /* write high order bits */

  /* Now toggle "E" line, falling edge latches the value */
  E_HIGH() ;
  E_LOW() ;
  LCD_kill_time (LCD_DELAY) ;

  /* write low order bits */
  PCDATA &= ~0xf0 ; /* clear top order bits */
  PCDATA |= (c & 0x0f) << 4 ; /* write low order bits */

  /* Now toggle "E" line, falling edge latches the value */
  E_HIGH() ;
  E_LOW() ;
  LCD_kill_time (LCD_DELAY) ; /* wait 40 microseconds between characters */
}

void
write_20 (const Uchar *control, const char *string)
{
  Ulong ssp ;
  const Uchar *p ;
  Ushort i ;

  LCD_write_lcd_control (control) ; // position the cursor
  ssp = supervisor_mode() ;

  // Set Register Select to "1" (control), E to "0"
  RS_HIGH() ;  // Set register select to 1
  E_LOW() ;

  // Write the message
  for (i=LCD_LINE_LEN, p=string ; i && *p ; i--, p++)
    LCD_write_char (*p) ;

  // For any other characters left on the line, output blanks
  for (; i ; i--)
    LCD_write_char (' ') ;

  user_mode (ssp) ;
}

void
LCD_write_lcd_string (const Uchar line, const char *string)
{
  Ushort length ;
  Ushort i ;

  /*
   * build a control command for putting cursor at start of the correct line
   */

  line_init[0] = 0x81 | ((line == 0) ? 0 : 0x40) ;
  line_init[1] = 0x02 ;
  line_init[2] = 0x01 ;			// MSB
  line_init[3] = 0xc2 ;			// LSB, cur off, blink off, ...
  line_init[4] = 0 ;

  length = strlen (string) ;

  if (length <= LCD_LINE_LEN)
    write_20 (line_init, string) ;
  else
  {
    for (i=0 ; i<(LCD_LINE_LEN-1) ; i++) // pad first ncharacters with space
      padded_line [i] = ' ' ;

    strncpy (&padded_line[LCD_LINE_LEN-1], string, LINE_LEN) ; // copy rest
    padded_line [LINE_LEN+LCD_LINE_LEN] = '\0' ;

    for (i=0 ; i < length ; i++)
    {
      write_20 (line_init, &padded_line[i]) ;
      usleep (175000) ;
    }
  }
}

/******************************************************************************
 * LCD_position_cursor
 */

void
LCD_position_cursor (const Uchar line, const Uchar location, const BOOL blink)
{
  Uchar address ;

  address = (0x40 * line) + location ;

  line_init[0] = 0x81 | (address & 0x70) ;		// MSB, no delay
  line_init[1] = 0x02 | ((address & 0x0f) << 4) ;	// LSB, 40us delay
  line_init[2] = 0x01 ;					// MSB, cur, blink, ...
  line_init[3] = (blink) ? 0xf2 : 0xc2 ;		// LSB, cur, blink, ...
  line_init[4] = 0 ;

  LCD_write_lcd_control (line_init) ;
}

/******************************************************************************
 * LCD_control_light
 */

void
LCD_control_light (Ushort state)
{
  AUX_write_aux_bit (AUX_LCD_LIGHT, state, False) ;
}

/******************************************************************************
 * LCD_initialise
 */

const Uchar PCG1[8] = {0, 0, 0, 0x1f, 0, 0, 0, 0} ;
const Uchar PCG2[8] = {0x10, 0x08, 0x08, 0x04, 0x04, 0x02, 0x02, 0x01} ;
const Uchar PCG3[8] = {0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04} ;
const Uchar PCG4[8] = {0x01, 0x02, 0x02, 0x04, 0x04, 0x08, 0x08, 0x10} ;

void
LCD_write_8 (const char *p)
{
  Ushort i ;

  for (i=8 ; i-- ;)
    LCD_write_char (*p++) ;
}

void
LCD_initialise (void)
{
  Ulong ssp ;

  LCD_write_lcd_control (LCD_initialise_string) ;

  // programme the programmable character generator
  line_init[0] = 0x41 ;
  line_init[1] = 0x82 ;
  line_init[2] = 0 ;
  LCD_write_lcd_control (line_init) ;

  ssp = supervisor_mode() ;
  RS_HIGH() ;
  E_LOW() ;
  LCD_write_8 (PCG1) ; // character 1
  LCD_write_8 (PCG2) ; // character 2
  LCD_write_8 (PCG3) ; // character 3
  LCD_write_8 (PCG4) ; // character 4
  user_mode (ssp) ;
}
