/*
 * menu.c
 ******************************************************************************
 * Copyright (C) 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ****************************************************************************(
 */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/time.h> // for setting the system time
#include <errno.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <ctype.h>

#include "types.h"
#include "menu.h"
#include "lcd.h"
#include "process.h"
#include "telescope.h"
#include "joystick.h"
#include "astro.h"
#include "database.h"
#include "access_data.h"

// All routines in this file assume that the system clock is set to GMT and
// NOT local time.
///////////////////////////////////////////////////////////////////////////////
#define OBJ_NAME_SIZE	14
#define NAME_SIZE	39
#define LINE_SIZE	100
#define FIND_LAST	1000000

MENU		*MEN_current ;
char		MEN_cursor ;
Uchar		MEN_state ;
char		MEN_obj_name [OBJ_NAME_SIZE+1] ;
OBJECT_INFO	MEN_star ;
double		MEN_alt, MEN_az ;
time_t		MEN_sample_time ;
ASH_POSITION_STRUCT MEN_pos ;
char		MEN_name [NAME_SIZE+1] ;
char		MEN_line [LINE_SIZE] ;

const char MEN_goto_req[] = "GoTo requested" ;
const char MEN_abort[] = "Aborted" ;
const char MEN_calc[] = "Calculating" ;
const char MEN_search[] = "Searching %c" ;

#ifdef UCLINUX
const char MEN_spinner[] = {0x01, 0x02, 0x03, 0x04} ;
#else
const char MEN_spinner[] = "/-\\|" ;
#endif

#define MEN_SET_ALT_AZ(ALTO, AZO, ALTI, AZI) \
{\
  ALTO = (ALTI) * 360.0 / 2.0 / M_PI ;\
  AZO  = (AZI)  * 360.0 / 2.0 / M_PI ;\
\
  if ((ALTO) > 90.0)\
    ALTO -= 360.0 ;\
\
  if ((AZO) < 0.0)\
    AZO += 360.0 ;\
}

#ifdef EMBEDED
#define MEN_SET_ALT_AZ_DELTA(ALT, AZ) \
{\
  double alt, az ;\
\
  MEN_SET_ALT_AZ(alt, az, ALT, AZ) ;\
  Global->alt_delta = MEN_alt - alt ;\
  Global->az_delta = MEN_az - az ;\
}
#else
#define MEN_SET_ALT_AZ_DELTA(ALT, AZ) \
{\
  double alt, az ;\
\
  MEN_SET_ALT_AZ(alt, az, ALT, AZ) ;\
}
#endif

#ifdef EMBEDED
#define MEN_GOTO() \
{\
  Global->altitude = ASH_ALT_to_int (MEN_alt) ;\
  Global->azimuth  = ASH_AZ_to_int (MEN_az) ;\
  TEL_lcd_msg (1, 1, MEN_goto_req) ;\
}
#else
#define MEN_GOTO() \
{\
  TEL_lcd_msg (1, 1, MEN_goto_req) ;\
}
#endif

///////////////////////////////////////////////////////////////////////////////
// The menu is an object oriented structure which is parsed by the routines in
// menu.c
//
// The ->edit routine is called when in the MEN_EDIT state
//	edit(0) called on ENTER
//	edit(100) called on ABORT
//	edit(1) called on UP
//	edit(-1) called on DOWN

MENU men1=
{
  NULL, NULL, &set_home, &tzone,
  "Time", MEN_disp_time, MEN_edit_time,
  0, 0, 0, NULL
} ;

MENU tzone=
{
  NULL, NULL, &men1, &location,
  "Timezone", MEN_disp_tz, MEN_edit_tz,
  0, 0, 0, NULL
} ;

MENU location=
{
  NULL, &place, &tzone, &go_to,
  "Location", NULL, NULL,
  0, 0, 1, NULL
} ;

MENU go_to=
{
  NULL, &messier, &location, &set_home,
  "GoTo", MEN_disp_tz, MEN_edit_tz,
  0, 0, 0, NULL
} ;

MENU set_home=
{
  NULL, NULL, &go_to, &men1,
  "HomePos", MEN_disp_home, MEN_edit_home,
  0, 0, 0, NULL
} ;

///////////////////////////////////////////////////////////////////////////////
// location options

MENU place=
{
  &location, NULL, &manual, &manual,
  "place", MEN_disp_file, MEN_edit_file,
  0, 0, 1, MEN_comit_lon_lat
} ;

MENU manual=
{
  &location, NULL, &place, &place,
  "manual", MEN_disp_int, MEN_edit_int,
  0, 0, 1, MEN_null_routine
} ;

///////////////////////////////////////////////////////////////////////////////
// Goto options

MENU messier=
{
  &go_to, NULL, &moon, &ngc,
  "Messier", MEN_disp_int, MEN_edit_int,
  1, 110, 1, MEN_lookup_database
} ;

MENU ngc=
{
  &go_to, NULL, &messier, &ugc,
  "NGC", MEN_disp_int, MEN_edit_int,
  1, 11000, 1, MEN_lookup_database
} ;

MENU ugc=
{
  &go_to, NULL, &ngc, &pgc,
  "UGC", MEN_disp_int, MEN_edit_int,
  1, 11000, 1, MEN_lookup_database
} ;

MENU pgc=
{
  &go_to, NULL, &ugc, &sao,
  "PGC", MEN_disp_int, MEN_edit_int,
  1, 11000, 1, MEN_lookup_database
} ;

MENU sao=
{
  &go_to, NULL, &pgc, &hr,
  "SAO", MEN_disp_int, MEN_edit_int,
  1, 250000, 1, MEN_lookup_database
} ;

MENU hr=
{
  &go_to, NULL, &sao, &hd,
  "HR", MEN_disp_int, MEN_edit_int,
  1, 250000, 1, MEN_lookup_database
} ;

MENU hd=
{
  &go_to, NULL, &hr, &ppm,
  "HD", MEN_disp_int, MEN_edit_int,
  1, 250000, 1, MEN_lookup_database
} ;

MENU ppm=
{
  &go_to, NULL, &hd, &ic,
  "PPM", MEN_disp_int, MEN_edit_int,
  1, 250000, 1, MEN_lookup_database
} ;

MENU ic=
{
  &go_to, NULL, &ppm, &align,
  "IC", MEN_disp_int, MEN_edit_int,
  1, 250000, 1, MEN_lookup_database
} ;

MENU align=
{
  &go_to, NULL, &ic, &planet,
  "AlgnStar", MEN_disp_align, MEN_edit_align,
  0, 0, -1, NULL // -1 triggers search on first display
} ;

MENU planet=
{
  &go_to, &mercury, &align, &moon,
  "Planet", NULL, NULL,
  0, 0, 0, NULL
} ;

MENU moon=
{
  &go_to, NULL, &planet, &messier,
  "Moon", MEN_disp_moon, MEN_edit_moon,
  0, 0, 0, NULL
} ;

MENU dummy_menu=
{
  &go_to, NULL, NULL, NULL,
  "", MEN_disp_dummy, MEN_edit_dummy,
  0, 0, 0, NULL
} ;

///////////////////////////////////////////////////////////////////////////////
// Goto Planet options

MENU mercury=
{
  &planet, NULL, &neptune, &venus,
  "Mercury", MEN_disp_plan, MEN_edit_plan,
  0, 0, MERCURY, NULL
} ;

MENU venus=
{
  &planet, NULL, &mercury, &mars,
  "Venus", MEN_disp_plan, MEN_edit_plan,
  0, 0, VENUS, NULL
} ;

MENU mars=
{
  &planet, NULL, &venus, &jupiter,
  "Mars", MEN_disp_plan, MEN_edit_plan,
  0, 0, MARS, NULL
} ;

MENU jupiter=
{
  &planet, NULL, &mars, &saturn,
  "Jupiter", MEN_disp_plan, MEN_edit_plan,
  0, 0, JUPITER, NULL
} ;

MENU saturn=
{
  &planet, NULL, &jupiter, &uranus,
  "Saturn", MEN_disp_plan, MEN_edit_plan,
  0, 0, SATURN, NULL
} ;

MENU uranus=
{
  &planet, NULL, &saturn, &neptune,
  "Uranus", MEN_disp_plan, MEN_edit_plan,
  0, 0, URANUS, NULL
} ;

MENU neptune=
{
  &planet, NULL, &uranus, &mercury,
  "Neptune", MEN_disp_plan, MEN_edit_plan,
  0, 0, NEPTUNE, NULL
} ;

///////////////////////////////////////////////////////////////////////////////
#define MIN_ELEVATION (20)	// degrees

void
MEN_disp_coord (void)
{
  char		*p ;
  Ushort	h, m ;
  Ushort	deg, min ;
  double	f, sec ;
  double	degrees ;

do_it_again:
  switch (MEN_cursor)
  {
    case 0:
      if ((MEN_az <= 22.5) || (MEN_az > 337.5))
	p = "S" ;
      else if (MEN_az <= 67.5)
	p = "SW" ;
      else if (MEN_az <= 112.5)
	p = "W" ;
      else if (MEN_az <= 157.5)
	p = "NW" ;
      else if (MEN_az <= 202.5)
	p = "N" ;
      else if (MEN_az <= 247.5)
	p = "NE" ;
      else if (MEN_az <= 292.5)
	p = "E" ;
      else
	p = "SE" ;

      if (MEN_alt < 0.0)
	TEL_lcd_msg (1, 0, "Below horizon (%s)", p) ;
      else if (MEN_alt < MIN_ELEVATION)
	TEL_lcd_msg (1, 0, "Alt only %d%c (%s)", (int)MEN_alt, DEG_CHAR, p) ;
      else
	TEL_lcd_msg (1, 0, "Alt=%d%c Az=%d%c (%s)",
		  (int)MEN_alt, DEG_CHAR, (int)MEN_az, DEG_CHAR, p) ;
      break ;

    case 1:
      f = MEN_pos.apa_ra ;

      if (f < 0.0)
	f += 2.0 * M_PI ;

      h = f / (2.0 * M_PI) * 24.0 ;
      m = (Ushort)(f / (2.0 * M_PI) * 24.0 * 60.0) % 60 ;
      sec = fmod ((f / (2.0 * M_PI) * 24.0 * 60.0 * 60.0), 60.0) ;

      if (sec > 59.5)
      {
	sec = 0.0 ;

	if (++m >= 60)
	{
	  m = 0 ;
	  h++ ;
	}
      }

      f = MEN_pos.apa_dec ;
      degrees = f / (2.0 * M_PI) * 360.0 ;

      if (degrees > 0.0)
	p = "" ;
      else
      {
	degrees = -degrees ;
	p = "-" ;
      }

      deg = degrees ;
      min = (degrees - deg) * 60 ;
      f = (((degrees - deg) * 60) - min) * 60.0 ;
 
      if (f > 59.5)
      {
	f = 0.0 ;

	if (++min >= 60)
	{
	  min = 0 ;
	  deg++ ;
	}
      }

      TEL_lcd_msg (1, 0, "%hu:%02hu:%02.0f, %s%hu%c%hu'%.0f\"",
		   h, m, sec, p, deg, DEG_CHAR, min, f) ;
      break ;

    case 2:
      if (MEN_star.bit_fields & BIT_FORMAL_NAME)
      {
	TEL_lcd_msg (1, 0, MEN_star.formal_name) ;
	break ;
      }

      MEN_cursor++ ;

    case 3:
      if (MEN_star.bit_fields & BIT_MAG)
      {
	TEL_lcd_msg (1, 0, "Magnitude: %01.1f", (MEN_star.mag - 30) / 10.0) ;
	break ;
      }

      MEN_cursor++ ;

    case 4:
      if (MEN_star.bit_fields & BIT_LY)
      {
	TEL_lcd_msg (1, 0, "Distance: %1.1f LY", MEN_star.ly /10.0) ;
	break ;
      }

      MEN_cursor++ ;

    case 5:
      if (MEN_star.bit_fields & BIT_TYPE)
      {
	TEL_lcd_msg (1, 0, "Type: %s", DB_obj_type[MEN_star.type].description) ;
	break ;
      }

      MEN_cursor++ ;

    case 6:
      if (MEN_star.bit_fields & BIT_NUM_COMP)
      {
	TEL_lcd_msg (1, 0, "Num components: %lu", (Ulong)MEN_star.num_comp) ;
	break ;
      }

      MEN_cursor++ ;

    case 7:
      if (MEN_star.bit_fields & BIT_MAG_DIFF)
      {
	TEL_lcd_msg (1, 0, "Mag difference: %1.1f", MEN_star.mag_diff / 10.0) ;
	break ;
      }

      MEN_cursor++ ;

    case 8:
      if (MEN_star.bit_fields & BIT_SEP)
      {
	TEL_lcd_msg (1, 0, "Sep: %lu arc sec", (Ulong)MEN_star.sep) ;
	break ;
      }

      MEN_cursor++ ;

    case 9:
      if (MEN_star.bit_fields & BIT_RADIUS)
      {
	TEL_lcd_msg (1, 0, "Radius: %lu SolarUnit", (Ulong)MEN_star.radius) ;
	break ;
      }

      MEN_cursor++ ;

    case 10:
      if (MEN_star.bit_fields & BIT_MAJOR)
      {
	snprintf (MEN_line, LINE_SIZE, "App size: %lu", (Ulong)MEN_star.major) ;

	if (MEN_star.bit_fields & BIT_MINOR)
	  snprintf (&MEN_line[strlen(MEN_line)], LINE_SIZE-strlen(MEN_line)-1, "x%lu", (Ulong)MEN_star.minor) ;

	snprintf (&MEN_line[strlen(MEN_line)], LINE_SIZE-strlen(MEN_line)-1, " arc sec") ;
	TEL_lcd_msg (1, 0, MEN_line) ;
	break ;
      }

      MEN_cursor++ ;

    case 11:
      if (MEN_star.bit_fields & BIT_CONST)
      {
	TEL_lcd_msg (1, 0, "Const: %s", DB_constellations[MEN_star.constellation].description) ;
	break ;
      }

      MEN_cursor++ ;

    case 12:
      if (MEN_star.bit_fields & BIT_RATING)
      {
	TEL_lcd_msg (1, 0, "Rating: %s", DB_rating_type[MEN_star.rating].description) ;
	break ;
      }

      MEN_cursor++ ;

    case 13:
      if (MEN_star.bit_fields & BIT_SEASON)
      {
	TEL_lcd_msg (1, 0, "Season: %s", DB_seasons[MEN_star.season].description) ;
	break ;
      }

      MEN_cursor++ ;

    case 14:
      if (MEN_star.bit_fields & BIT_ANGLE)
      {
	TEL_lcd_msg (1, 0, "Angle class: %lu", (Ulong)MEN_star.angle) ;
	break ;
      }

      MEN_cursor++ ;

    case 15:
      if (MEN_star.bit_fields & BIT_SPECTRUM)
      {
	TEL_lcd_msg (1, 0, "Spectrum: %s", MEN_star.spectrum) ;
	break ;
      }

      MEN_cursor++ ;

    case 16:
      if (MEN_star.bit_fields & BIT_DESCRIPTION)
      {
	TEL_lcd_msg (1, 0, MEN_star.description) ;
	break ;
      }

    default:
      if (MEN_cursor < 0)
	MEN_cursor = 1 ;
      else
	MEN_cursor = 0 ;

      goto do_it_again ;
  }
}
///////////////////////////////////////////////////////////////////////////////
void
MEN_disp_name (void)
{
  char		msg[30] ;
  struct tm	*raw ;
  time_t	tmp ;

  tmp = time (NULL) ; // get GMT
  tmp += tzone.value ; // convert time to localtime
  raw = gmtime (&tmp) ; // actually, get localtime

  snprintf (msg, LCD_LINE_LEN+1, "%-*.*s%02d:%02d",
	    LCD_LINE_LEN-5, LCD_LINE_LEN-4, MEN_obj_name,
	    raw->tm_hour, raw->tm_min) ;

  TEL_lcd_msg (0, 0, msg) ;
}
///////////////////////////////////////////////////////////////////////////////
#define NOW 0
#define FUTURE 1

void
MEN_get_time (ASH_POSITION_STRUCT *pos, const Uchar when)
{
  struct tm	*raw ;
  time_t	tmp ;

  if (when) // == future
    tmp = MEN_sample_time + TEL_DELTA_TIME ;
  else // == now
  {
    tmp = time (NULL) ; // get GMT if it's not been provided by the caller
    MEN_sample_time = tmp ;
  }

  raw = gmtime (&tmp) ;
  AS_initialise_struct (pos) ;
  AS_fill_time_struct (&pos->GMT,
			raw->tm_year+1900,
			raw->tm_mon,
			raw->tm_mday,
			raw->tm_hour,
			raw->tm_min,
			raw->tm_sec) ;
}
///////////////////////////////////////////////////////////////////////////////
void
MEN_process (void)
{
  char line [LCD_LINE_LEN+1] ; // n characters plus termination
  char nam [NAME_LEN+1] ;
  MENU *p ;
  int n, cur ;

  #ifdef PARANOID
    if (MEN_current == NULL)
    {
      PR_log_msg (ERR, "MEN_process: MEN_current set to null!") ;
      return ;
    }
  #endif

  cur = JOY_changes & JOY_current ;

  if (MEN_state == MEN_EDIT)
  {
    if (cur & JOY_ENTER_BIT)		// ENTER
    {
      MEN_current->edit (MEN_current, 0) ;

      if (MEN_current != &dummy_menu)
	MEN_state = MEN_LOOK ;
    }
    else if (cur & JOY_MODE_BIT)	// abort
    {
      MEN_current->edit (MEN_current, 100) ;
      MEN_state = MEN_LOOK ;
    }
    else if (cur & JOY_LEFT_BIT)	// Left
      MEN_cursor++ ;

    else if (cur & JOY_RIGHT_BIT)	// Right
      MEN_cursor-- ;

    else if (cur & JOY_UP_BIT)		// Up
      MEN_current->edit (MEN_current, 1) ;

    else if (cur & JOY_DOWN_BIT)	// Down
      MEN_current->edit (MEN_current, -1) ;
  }
  else // In MEN_LOOK state
  {
    if (cur & JOY_LEFT_BIT)		// Left
      MEN_current = MEN_current->left ;

    else if (cur & JOY_RIGHT_BIT)	// Right
      MEN_current = MEN_current->right ;

    else if (cur & JOY_UP_BIT)		// Up
    {
      if (MEN_current->up != NULL)
	MEN_current = MEN_current->up ;
    }
    else if (cur & JOY_DOWN_BIT)	// Down
    {
      if (MEN_current->down == NULL)
	MEN_state = MEN_EDIT ;
      else
	MEN_current = MEN_current->down ;
    }

    MEN_cursor = 0 ;
  }

  // Make these following keys appear to auto repeat
  JOY_current &= ~(JOY_UP_BIT | JOY_DOWN_BIT | JOY_LEFT_BIT | JOY_RIGHT_BIT) ;

  if (MEN_state == MEN_EDIT)
    MEN_current->disp (MEN_current, line) ;
  else
  {
    if (MEN_current->up == NULL)
    {
      strcpy (MEN_obj_name, "Main Menu") ;
    }
    else
    {
      // At an intermediate level
      memcpy (MEN_obj_name, MEN_current->up->name, NAME_LEN) ;
      MEN_obj_name [NAME_LEN] = '\0' ;
    }

    MEN_disp_name() ;

    // Display the surrounding menu choices on line 2
    memcpy (nam, MEN_current->left->name, NAME_LEN) ;
    nam [NAME_LEN] = '\0' ;
    n = snprintf (line, LCD_LINE_LEN+1, "%s", nam) ;

    if (MEN_current == MEN_current->left)
      cur = 0 ;
    else
      cur = n+1 ;

    for (p=MEN_current ; (n<LCD_LINE_LEN) && (p!=MEN_current->left) ; p=p->right)
    {
      char *c ;

      line[n++] = ' ' ;
      memcpy (nam, p->name, NAME_LEN) ;
      nam [NAME_LEN] = '\0' ;

      for (c=nam ; *c && (n < LCD_LINE_LEN) ;)
	line[n++] = *c++ ;

      line[n] = '\0' ;
      n = strlen (line) ;
    }

    TEL_lcd_msg (1, 0, line) ;
    TEL_lcd_cur (1, cur) ;
  }
}
///////////////////////////////////////////////////////////////////////////////
// For every menu object, there is a display routine and there's an edit
// routine.
//
// The display routine is responsible for printing the value of the object on
// the second line of the LCD display and for putting the cursor at the field
// which is currently being edited.
//
// The edit routine is responsible for changing the value of the field under
// the cursor either up or down and for committing the value when the user
// finishes editing.
//
// The edit routine is called with three possible delta parameters:
// -1 or +1: These values imply that the edited field needs to be changed
// 0: This value implies that the changed value should be committed.
///////////////////////////////////////////////////////////////////////////////
// MEN_disp_int displays an integer in a form which makes it easy to see what
// can be edited

int
MEN_null_routine (const int val)
{
  // a null routine - do nothing

  return 0 ; // Nothing done
}

int
MEN_lookup_database (const int val)
{
  DB_TYPE db_type ;
  OBJECT_INFO o ;

  if (MEN_current == &messier)
    db_type = DB_MES ;

  else if (MEN_current == &ugc)
    db_type = DB_UGC ;

  else if (MEN_current == &ngc)
    db_type = DB_NGC ;

  else if (MEN_current == &hr)
    db_type = DB_HR ;

  else if (MEN_current == &ic)
    db_type = DB_IC ;

  else if (MEN_current == &sao)
    db_type = DB_SAO ;

  else if (MEN_current == &pgc)
    db_type = DB_PGC ;

  else if (MEN_current == &hd)
    db_type = DB_HD ;

  else if (MEN_current == &ppm)
    db_type = DB_PPM ;

  else if (MEN_current == &align)
    db_type = DB_ALIGN ;

  else
  {
#ifdef PARANOID
    TEL_lcd_msg (1, 1, "Unknown switch!") ;
#endif
    return 0 ; // Nothing done
  }

  if (DB_lookup (db_type, val, &o, 0))
  {
    TEL_lcd_msg (1, 1, "Not in database!") ;
    return 0 ; // Nothing done
  }

  // We're going to change the current menu item to a special dummy value
  // so that the selected item can be correctly displayed and a goto
  // performed if appropriate.
  //
  // First, we need to create a dummy menu item based upon the selected data

  dummy_menu.left = MEN_current->left ;
  dummy_menu.right = MEN_current->right ;
  dummy_menu.value = MEN_current->value ;
  dummy_menu.other = MEN_current ;
  memmove (dummy_menu.name, MEN_current->name, sizeof(dummy_menu.name)) ;

  MEN_get_time (&MEN_pos, NOW) ;
  AS_time_functions (&MEN_pos) ;
  MEN_pos.obj_ra = AS_intern_ra_to_rad (o.ra) ;
  MEN_pos.obj_dec = AS_intern_dec_to_rad (o.dec) ;
  AS_nutation (&MEN_pos) ;
  AS_aberration (&MEN_pos) ;
  AS_allow_refraction (&MEN_pos) ;
  MEN_SET_ALT_AZ (MEN_alt, MEN_az, MEN_pos.altitude, MEN_pos.azimuth) ;

  if (o.bit_fields & BIT_COMMON_NAME)
  {
    strncpy (MEN_obj_name, o.common_name, OBJ_NAME_SIZE) ;
    MEN_obj_name[OBJ_NAME_SIZE] = '\0' ;
  }
  else
    snprintf (MEN_obj_name, OBJ_NAME_SIZE+1, "%s %d",
	      MEN_current->name, MEN_current->value) ;

  MEN_current = &dummy_menu ;
  return 1 ; // Something has been done=1
}

void
MEN_disp_dummy (MENU *m, char *buffer)
{
  MEN_disp_name () ;  // display line 1
  MEN_disp_coord () ; // display line 2
}

void
MEN_edit_dummy (MENU *m, char delta)
{
  if (delta > 1) // abort
  {
    TEL_lcd_msg (1, 1, MEN_abort) ;
    MEN_current = dummy_menu.other ;
    return ;
  }

  if (delta == 0) // SELECT!
  {
    ASH_POSITION_STRUCT pos ;

    MEN_GOTO() ;
    MEN_get_time (&pos, FUTURE) ; // calculate based on delta interval
    AS_time_functions (&pos) ;
    pos.obj_ra = AS_intern_ra_to_rad (MEN_star.ra) ;
    pos.obj_dec = AS_intern_dec_to_rad (MEN_star.dec) ;
    AS_nutation (&pos) ;
    AS_aberration (&pos) ;
    AS_allow_refraction (&pos) ;
    MEN_SET_ALT_AZ_DELTA (pos.altitude, pos.azimuth) ;
    MEN_current = dummy_menu.other ;
  }
}
///////////////////////////////////////////////////////////////////////////////
void
MEN_disp_int (MENU *m, char *buffer)
{
  char *p = buffer ;
  int val = m->value ;
  int i = m->min ;
  int j = m->max ;

  if (i < 0)
  {
    i = -i ;

    if (val < 0)
    {
      *p++ = '-' ;
      val = -val ;
    }
    else
      *p++ = '+' ;
  }

  // now work out the field width (in digits) into 'i'
  if (j < 0)
    j = -j ;

  if (i > j)
    j = i ;

  for (i=1 ; j>10 ; j/=10, i++) ;

  // fill the field with "blanks"
  for (; i ; i--)
    *p++ = '_' ;

  *p-- = '\0' ; // delimit the string with a null

  // Now display the number
  for (*p='0' ; val ; val /= 10)
    *p-- = (val % 10)+'0' ;

  // Now fix the cursor so that it's within range of the field
  i = strlen(buffer) - 1 ;

  if (MEN_cursor > i)
    MEN_cursor = 0 ;
  else if (MEN_cursor < 0)
    MEN_cursor = i ;

  memcpy (MEN_obj_name, m->name, NAME_LEN) ;
  MEN_obj_name [NAME_LEN] = '\0' ;
  MEN_disp_name () ;  // display line 1
  TEL_lcd_msg (1, 0, buffer) ;
  TEL_lcd_cur (1, i-MEN_cursor) ;
}

void
MEN_edit_int (MENU *m, char delta)
{
  int sign ;
  int abs ;
  int i = m->min ;
  int j = m->max ;
  int k ;
  int l = MEN_cursor ;

  if ((delta == 0) || (delta > 1)) // commit the change
  {
    if (MEN_current->value < MEN_current->min)
    {
      MEN_current->value = MEN_current->min ;
      MEN_cursor = 0 ;
    }
    else if (MEN_current->value > MEN_current->max)
    {
      MEN_current->value = MEN_current->max ;
      MEN_cursor = 0 ;
    }

    if (delta == 0) // perform the action
    {
      COM_DEF *ptr ;

      ptr = m->other ; // this is the call to the action routine

      if (ptr (m->value))
      {
	// Do something here if necessary
      }
    }

    return ;
  }

  // get sign and absolute value
  if (m->value < 0)
  {
    sign = -1 ;
    abs = -m->value ;
  }
  else
  {
    sign = 1 ;
    abs = m->value ;
  }

  // now work out the field width (in digits)
  if (i < 0)
    i = -i ;

  if (j < 0)
    j = -j ;

  if (i > j)
    j = i ;	// j now holds max (abs(min), abs(max))

  for (i=1, k=abs ; l && (j>10) ; i*=10, j/=10, k/=10, l--) ;

  if (l > 0)
    sign *= -1 ;	// Only the sign needs toggling
  else
  {
    // Need to bump the value of the appropriate column
    k %= 10 ;  // get k=column value, i=column weight

    if (((k==0) && (delta<0)) || ((k==9) && (delta>0)))
      delta *= -9 ;

    i *= delta ;
    abs += i ;
  }

  m->value = abs * sign ;
}
///////////////////////////////////////////////////////////////////////////////
// MEN_disp_time

#define MEN_MON 4
char MEN_mon[][MEN_MON] =
{"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"} ;

#define MEN_MON_MAXCUR 10
char MEN_moncur[MEN_MON_MAXCUR+1] = {17, 16, 14, 13, 11, 10, 8, 7, 5, 1, 0} ;

static struct tm MEN_time ;

void
MEN_disp_time (MENU *m, char *buffer)
{
  struct tm *raw ;
  time_t tmp ;

  if (m->value == 0)
  {
    tmp = time (NULL) ; // get GMT
    tmp += tzone.value ; // convert time to localtime
    raw = gmtime (&tmp) ; // actually, get localtime
    raw->tm_sec = 0 ;
    m->value = 1 ;
    memcpy (&MEN_time, raw, sizeof (struct tm)) ;
  }

  strcpy (MEN_obj_name, "Set time") ;
  MEN_disp_name () ;  // display line 1
  sprintf (buffer, "%2d/%s/%02d %02d:%02d:%02d",
	   MEN_time.tm_mday, &MEN_mon[MEN_time.tm_mon][0], MEN_time.tm_year%100,
	   MEN_time.tm_hour, MEN_time.tm_min, MEN_time.tm_sec) ;

  TEL_lcd_msg (1, 0, buffer) ;

  if (MEN_cursor > MEN_MON_MAXCUR)
    MEN_cursor = 0 ;
  else if (MEN_cursor < 0)
    MEN_cursor = MEN_MON_MAXCUR ;

  TEL_lcd_cur (1, MEN_moncur[(int)MEN_cursor]) ;
}

int
MEN_delta (const int current, const int min, const int max, const int delta)
{
  int tmp ;

  tmp = current + delta ;

  if (tmp > max)
    return min ;

  if (tmp < min)
    return max ;

  return tmp ;
}

void
MEN_edit_time (MENU *m, char delta)
{
  int tmp ;

  if (delta > 1) // abort
  {
    m->value = 0 ;
    return ;
  }

  if (delta == 0) // commit the change
  {
    struct timeval tv ;
    time_t t ;

    t = mktime (&MEN_time) ;

    // uClinux doesn't support timezones properly
    #ifdef UCLINUX
      t -= tzone.value ;
    #endif

    tv.tv_sec = t ;
    tv.tv_usec = 0 ;

    if (settimeofday (&tv, 0) != 0)
      PR_log_msg (ERR, "Set time failed: %s", strerror(errno)) ;

    m->value = 0 ;
    return ;
  }

  switch (MEN_cursor)
  {
    case 0: // seconds, units
      tmp = MEN_delta (MEN_time.tm_sec%10, 0, 9, delta) ;
      MEN_time.tm_sec = (MEN_time.tm_sec/10)*10 + tmp ;
      break ;

    case 1: // seconds, tens
      tmp = MEN_delta (MEN_time.tm_sec/10, 0, 5, delta) * 10 ;
      MEN_time.tm_sec = (MEN_time.tm_sec%10) + tmp ;
      break ;

    case 2: // minutes, units
      tmp = MEN_delta (MEN_time.tm_min%10, 0, 9, delta) ;
      MEN_time.tm_min = (MEN_time.tm_min/10)*10 + tmp ;
      break ;

    case 3: // minutes, tens
      tmp = MEN_delta (MEN_time.tm_min/10, 0, 5, delta) * 10 ;
      MEN_time.tm_min = (MEN_time.tm_min%10) + tmp ;
      break ;

    case 4: // hours, units
      tmp = MEN_delta (MEN_time.tm_hour%10, 0, 9, delta) ;
      MEN_time.tm_hour = (MEN_time.tm_hour/10)*10 + tmp ;
      break ;

    case 5: // hours, tens
      tmp = MEN_delta (MEN_time.tm_hour/10, 0, 2, delta) * 10 ;
      MEN_time.tm_hour = (MEN_time.tm_hour%10) + tmp ;
      break ;

    case 6: // year, units
      tmp = MEN_delta (MEN_time.tm_year%10, 0, 9, delta) ;
      MEN_time.tm_year = (MEN_time.tm_year/10)*10 + tmp ;
      break ;

    case 7: // year, tens
      tmp = MEN_delta ((MEN_time.tm_year%100)/10, 0, 2, delta) * 10 ;
      tmp += (MEN_time.tm_year/100)*100 ;
      MEN_time.tm_year = (MEN_time.tm_year%10) + tmp ;
      break ;

    case 8: // month
      MEN_time.tm_mon = MEN_delta (MEN_time.tm_mon, 0, 11, delta) ;
      break ;

    case 9: // day, units
      tmp = MEN_delta (MEN_time.tm_mday%10, 0, 9, delta) ;
      MEN_time.tm_mday = (MEN_time.tm_mday/10)*10 + tmp ;
      break ;

    case 10: // day, tens
      tmp = MEN_delta (MEN_time.tm_mday/10, 0, 2, delta) * 10 ;
      MEN_time.tm_mday = (MEN_time.tm_mday%10) + tmp ;
      break ;
  }

  if (MEN_cursor >= 6) // if modifying the date, account for daylight saving
  {
    MEN_disp_file (&place, NULL) ;
    MEN_edit_file (&place, 0) ;
  }
}
///////////////////////////////////////////////////////////////////////////////
void
MEN_disp_home (MENU *m, char *buffer)
{
  if (m->value == 0)
  {
    // Set the home position: it has not been set yet.
    m->value = 1 ;
    MEN_edit_home (m, 0) ;
  }
  else
  {
    // The home position has been previously set.
    strcpy (MEN_obj_name, "Confirm set") ;
    MEN_disp_name () ;  // display line 1
    TEL_lcd_msg (1, 0, "home position?") ;
  }
}

void
MEN_edit_home (MENU *m, char delta)
{
  if (delta > 1) // abort
  {
    TEL_lcd_msg (1, 1, MEN_abort) ;
    return ;
  }

  if (delta == 0) // SELECT!
  {
    // DO SOMETHING TO SET THE HOME POSITION
    TEL_one_sec_msg (INITIALISE) ;
    TEL_lcd_msg (1, 1, "Home Position set") ;
  }
}
///////////////////////////////////////////////////////////////////////////////
// MEN_disp_tz

#define MEN_TZ_MAXCUR 4
char MEN_tzcur[MEN_TZ_MAXCUR+1] = {17, 16, 14, 13, 12} ;

void
MEN_disp_tz (MENU *m, char *buffer)
{
  int min ;
  int hour ;
  char sign ;

  if (m->min == 0)  // if just entering, remember starting position
    m->min = m->value ;

  hour = m->value / 3600 ;
  min = (m->value / 60) % 60 ;
  sign = '+' ;

  if (hour < 0)
  {
    sign = '-' ;
    hour = -hour ;
  }

  if (min < 0)
    min = -min ;

  strcpy (MEN_obj_name, "Set timezone") ;
  MEN_disp_name () ;  // display line 1
  sprintf (buffer, "GMT offset: %c%02d:%02d", sign, hour, min) ;
  TEL_lcd_msg (1, 0, buffer) ;

  if (MEN_cursor > MEN_TZ_MAXCUR)
    MEN_cursor = 0 ;
  else if (MEN_cursor < 0)
    MEN_cursor = MEN_TZ_MAXCUR ;

  TEL_lcd_cur (1, MEN_tzcur[(int)MEN_cursor]) ;
}

void
MEN_edit_tz (MENU *m, char delta)
{
  int tmp ;
  int min ;
  int hour ;
  char sign ;

  if (delta > 1) // abort
  {
    m->value = m->min ; // restore the starting value set in disp_tz
    Global->time_offset = m->min ;
    m->min = 0 ;
    return ;
  }

  hour = m->value / 3600 ;
  min = (m->value / 60) % 60 ;
  sign = 1 ;

  if (hour < 0)
  {
    hour = -hour ;
    sign = -1 ;
  }

  if (min < 0)
    min = -min ;

  if (delta == 0) // commit the change
  {
    m->min = 0 ;
    return ;
  }

  switch (MEN_cursor)
  {
    case 0: // minutes, units
      tmp = MEN_delta (min%10, 0, 9, delta) ;
      min = (min/10)*10 + tmp ;
      break ;

    case 1: // minutes, tens
      tmp = MEN_delta (min/10, 0, 5, delta) * 10 ;
      min = (min%10) + tmp ;
      break ;

    case 2: // hours, units
      tmp = MEN_delta (hour%10, 0, 9, delta) ;
      hour = (hour/10)*10 + tmp ;
      break ;

    case 3: // hours, tens
      tmp = MEN_delta (hour/10, 0, 1, delta) * 10 ;
      hour = (hour%10) + tmp ;
      break ;

    case 4: // sign
      sign *= -1 ;
      break ;
  }

  m->value = sign * (hour*60 + min) * 60 ;
  Global->time_offset = m->value ;
}
///////////////////////////////////////////////////////////////////////////////
// MEN_disp_file

void
MEN_disp_file (MENU *m, char *buffer)
{
  FILE	*f ;
  int	line_count, line_get ;
  char	*ret, *dest, *p ;

  if (m->value == 0)
    line_get = FIND_LAST; // zero means "find last entry in file"
  else
    line_get = m->value ;

  // on entry, if m->value == 0, this indicates the last entry in file
  if (m->min == 0) // if just entering, remember starting position
    m->min = m->value ;

  f = (FILE *)0 ;

  for (line_count=0 ; line_count < line_get ; line_count++)
  {
    // open the file if this is first pass in the loop
    if (line_count == 0)
    {
      strncpy (MEN_name, FILE_PREFIX, NAME_SIZE) ;
      MEN_name[NAME_SIZE] = '\0' ;
      p = &MEN_name[strlen(MEN_name)] ;
      strncat (MEN_name, m->name, NAME_SIZE-strlen(MEN_name)) ;
      strncat (MEN_name, ".info", NAME_SIZE-strlen(MEN_name)) ;

      if (((f = fopen (MEN_name, "r")) == (FILE *)0) &&
	  ((f = fopen (p, "r")) == (FILE *)0))
      {
	strncat (MEN_name, "??", NAME_SIZE-strlen(MEN_name)) ;
	MEN_name[NAME_SIZE] = '\0' ;
	TEL_lcd_msg (1, 0, p) ;
	MEN_cursor = 0 ;
	return ;
      }
    }

    // read the next non-blank line of the file
    while (((ret=fgets (MEN_line, LINE_SIZE, f)) != NULL) &&
	   ((MEN_line[0] == '\n') || (MEN_line[0] == '*'))) ;

    // NULL could mean EOF or error
    if (ret == NULL)
    {
      fclose (f) ;
      f = (FILE *)0 ;

      if (line_count == 0)
      {
	TEL_lcd_msg (1, 0, "File empty error!") ;
	MEN_cursor = 0 ;
	return ;
      }

      if (m->value == 0) // set m->value to the number of entries in the file
      {
	line_get = m->value = line_count ;
	break ;
      }
      else
	line_get = m->value = 1 ;

      line_count = -1 ; // restart the loop and reopen the file
      continue ;
    }
  }

  // file is still open, name contains 1st param, line contains rest.
  if (f != (FILE *)0)
    fclose (f) ;

  for (ret=MEN_line ; *ret && isspace(*ret) ; ret++) ; // skip over white space
  for (dest=MEN_name ; *ret && (*ret != '|') ; *dest++=*ret++) ;
  *dest = '\0' ;

  if (*ret == '|')
    ret++ ;

  for (dest-- ; (dest > MEN_name) && isspace(*dest) ; *dest-- = '\0') ;
  for (; *ret && isspace(*ret) ; ret++) ; // skip over white space
  for (dest=MEN_line ; *ret && (*ret != '\n') ; *dest++=*ret++) ;
  *dest = '\0' ;

  if (MEN_current == m) // don't display when special circumstances
  {
    snprintf (MEN_obj_name, OBJ_NAME_SIZE, "Set %s", m->name) ;
    MEN_disp_name () ;  // display line 1
    TEL_lcd_msg (1, 0, MEN_name) ;
  }
}

int
MEN_comit_lon_lat (const int val)
{
  double lon, lat ;
  int offset, start, end, delta ;
  int num ;
  int min, hour ;
  char sign ;

  num = sscanf (MEN_line, "%lf %lf %d %d %d %d", &lon, &lat, &offset, &start, &end, &delta) ;

  if (num < 3)
  {
    TEL_lcd_msg (0, 1, "Lon/Lat spec error") ;
    return 0 ;
  }

  hour = offset / 3600 ;
  min = (offset / 60) % 60 ;
  sign = 1 ;

  if (hour < 0)
  {
    hour = -hour ;
    sign = -1 ;
  }

  if (min < 0)
    min = -min ;

  tzone.value = offset ; // first approximation excluding daylight savings
  AS_longitude = lon / 360.0 * 2.0 * M_PI ;
  AS_lattitude = lat / 360.0 * 2.0 * M_PI ;

  if ((num > 3) && (num < 6))
  {
    TEL_lcd_msg (0, 1, "Daylight spec error") ;
    return 0 ;
  }

  if (num == 6)	// daylight savings parameters were specified
  {
    struct tm	*raw ;
    time_t	tmp ;

    if ((start < 1) || (start > 12) || (end < 1) || (end > 12))
    {
      TEL_lcd_msg (0, 1, "Months spec error") ;
      return 0 ;
    }

    tmp = time (NULL) ; // get GMT
    tmp += tzone.value ; // convert time to localtime
    raw = gmtime (&tmp) ; // actually, get localtime
    raw->tm_mon++ ;

    if (((start < end) && (raw->tm_mon > start) && (raw->tm_mon < end))
	|| ((start > end) && ((raw->tm_mon > start) || (raw->tm_mon < end))))
    {
      // in daylight savings mode

      tzone.value += delta ;
    }
    // Now check for last sunday of the month
    else if (((raw->tm_mon == start) && (raw->tm_mday > (24+raw->tm_wday))) ||
	     ((raw->tm_mon == end) && (raw->tm_mday <= (24+raw->tm_wday))))
    {
      // Above works only for 31 day months! 30 and 29 day months will have
      // one day in error!

      tzone.value += delta ;
    }
  }

  return 1 ;
}

void
MEN_edit_file (MENU *m, char delta)
{
  COM_DEF *ptr ;

  if (delta == 0) // commit the change
  {
    ptr = m->other ;

    if (ptr (m->value))
    {
      m->min = 0 ;
      return ;
    }

    delta = 100 ;
  }

  if (delta > 1) // abort
  {
    m->value = m->min ; // restore the starting value set in disp_file
    m->min = 0 ;
    TEL_lcd_msg (1, 1, MEN_abort) ;
    return ;
  }

  if (delta == -1)
  {
    if (--m->value < 0)
      m->value = 0 ; // zero has a special meaning - means the last entry
  }

  if (delta == 1)
    m->value++ ;
}
///////////////////////////////////////////////////////////////////////////////
// MEN_disp_moon

void
MEN_disp_moon (MENU *m, char *buffer)
{
  if (m->min == 0)
  {
    m->min = 1 ;
    MEN_star.bit_fields = 0 ;
    strcpy (MEN_obj_name, m->name) ;
    MEN_disp_name () ;
    TEL_lcd_msg (1, 0, MEN_calc) ;
    MEN_get_time (&MEN_pos, NOW) ;
    AS_find_moon_apparent_location (&MEN_pos) ;
    MEN_SET_ALT_AZ (MEN_alt, MEN_az, MEN_pos.altitude, MEN_pos.azimuth) ;
    m->max = (int)MEN_alt ;
  }
  else
    MEN_disp_name () ;

  MEN_disp_coord () ;
}

void
MEN_edit_moon (MENU *m, char delta)
{
  ASH_POSITION_STRUCT pos ;

  if (delta > 1) // abort
  {
    TEL_lcd_msg (1, 1, MEN_abort) ;
    m->min = 0 ;
    return ;
  }

  if (delta == 0) // commit the change
  {
    MEN_GOTO() ;
    MEN_get_time (&pos, FUTURE) ; // calculate based on delta interval
    AS_find_moon_apparent_location (&pos) ;
    MEN_SET_ALT_AZ_DELTA (pos.altitude, pos.azimuth) ;
    m->min = 0 ;
  }
}
///////////////////////////////////////////////////////////////////////////////
// MEN_disp_plan

void
MEN_disp_plan (MENU *m, char *buffer)
{
  if (m->min == 0)
  {
    m->min = 1 ;
    MEN_star.bit_fields = 0 ;
    strcpy (MEN_obj_name, m->name) ;
    MEN_disp_name () ;
    TEL_lcd_msg (1, 0, MEN_calc) ;
    MEN_get_time (&MEN_pos, NOW) ;
    AS_find_planet_apparent_location (&MEN_pos, m->value) ;
    MEN_SET_ALT_AZ (MEN_alt, MEN_az, MEN_pos.altitude, MEN_pos.azimuth) ;
    m->max = (int)MEN_alt ;
  }
  else
    MEN_disp_name () ;

  MEN_disp_coord () ;
}

void
MEN_edit_plan (MENU *m, char delta)
{
  if (delta > 1) // abort
  {
    TEL_lcd_msg (1, 1, MEN_abort) ;
    m->min = 0 ;
    return ;
  }

  if (delta == 0) // commit the change
  {
    ASH_POSITION_STRUCT pos ;

    MEN_GOTO() ;
    MEN_get_time (&pos, FUTURE) ;
    AS_find_planet_apparent_location (&pos, m->value) ;
    MEN_SET_ALT_AZ_DELTA (pos.altitude, pos.azimuth) ;
    m->min = 0 ;
  }
}
///////////////////////////////////////////////////////////////////////////////
// MEN_disp_align
//
// In determining whether a star is obserable or not, some very rough
// approximations are made. As we scan the list, we don't care whether the
// the location is precise, we just need to know roughly where in the sky
// it is.

#define MIN_ALIGN_ELEVATION (35)				      // deg
#define MIN_ALIGN_ELEV_RAD (MIN_ALIGN_ELEVATION / 360.0 * M_PI * 2.0) // rads

void
MEN_find_next_align_star (MENU *m, const char delta, ASH_POSITION_STRUCT *pos, OBJECT_INFO *o)
{
  ASH_POSITION_STRUCT copy_pos ;
  char c ;

  // firstly, perform some common calculations which don't change later
  MEN_get_time (&copy_pos, NOW) ;
  AS_time_functions (&copy_pos) ;

  // Now, find the next item in the list which is visible at an appropriate
  // altitude

  for (;;)
  {
    m->value += delta ;

    if (m->value < 0)
      m->value = DB_ref[DB_ALIGN].num - 1 ;

    if (m->value >= DB_ref[DB_ALIGN].num)
      m->value = 0 ;

    c = MEN_spinner[m->value & 0x03] ;
    TEL_lcd_msg (1, 0, MEN_search, c) ;

    if (DB_lookup (DB_ALIGN, m->value, o, 1)) // brief lookup
    {
#ifdef PARANOID
      TEL_lcd_msg (1, 1, "Not in database!") ;
#endif
      continue ;
    }

    memmove (pos, &copy_pos, sizeof(ASH_POSITION_STRUCT)) ;
    pos->obj_dec = AS_intern_dec_to_rad (o->dec) ;

    // Don't bother with calc if star not visible from this lattitude
    if (pos->obj_dec > (AS_lattitude + (M_PI/2) - MIN_ALIGN_ELEV_RAD))
      continue ;

    if (pos->obj_dec < (AS_lattitude - (M_PI/2) + MIN_ALIGN_ELEV_RAD))
      continue ;

    pos->obj_ra = AS_intern_ra_to_rad (o->ra) ;
    AS_nutation (pos) ;
    AS_aberration (pos) ;
    AS_allow_refraction (pos) ;
    MEN_SET_ALT_AZ (MEN_alt, MEN_az, pos->altitude, pos->azimuth) ;

    if (MEN_alt >= MIN_ALIGN_ELEVATION)
      break ;
  }
}

void
MEN_disp_align (MENU *m, char *buffer)
{
  if (m->value < 0) // initialise
    MEN_find_next_align_star (m, 1, &MEN_pos, &MEN_star) ;

  if (DB_lookup (DB_ALIGN, m->value, &MEN_star, 0))
  {
#ifdef PARANOID
    TEL_lcd_msg (1, 1, "Not in database!") ;
#endif
    return ; // Nothing done
  }

  MEN_get_time (&MEN_pos, NOW) ;
  AS_time_functions (&MEN_pos) ;
  MEN_pos.obj_ra = AS_intern_ra_to_rad (MEN_star.ra) ;
  MEN_pos.obj_dec = AS_intern_dec_to_rad (MEN_star.dec) ;
  AS_nutation (&MEN_pos) ;
  AS_aberration (&MEN_pos) ;
  AS_allow_refraction (&MEN_pos) ;
  MEN_SET_ALT_AZ (MEN_alt, MEN_az, MEN_pos.altitude, MEN_pos.azimuth) ;

  if (MEN_star.bit_fields & BIT_COMMON_NAME)
  {
    strncpy (MEN_obj_name, MEN_star.common_name, OBJ_NAME_SIZE) ;
    MEN_obj_name[OBJ_NAME_SIZE] = '\0' ;
  }
#ifdef PARANOID
  else
    strcpy (MEN_obj_name, "No common name") ;
#endif
  MEN_disp_name() ; // display line 1
  MEN_disp_coord() ; // display line 2
}

void
MEN_edit_align (MENU *m, char delta)
{
  ASH_POSITION_STRUCT pos ;

  if ((delta == 1) || (delta == -1))
  {
    // Find the next alignment star
    MEN_find_next_align_star (m, delta, &pos, &MEN_star) ;
  }

  if (delta > 1) // abort
  {
    TEL_lcd_msg (1, 1, MEN_abort) ;
    m->min = 0 ;
    return ;
  }

  if (delta == 0) // SELECT!
  {
    MEN_GOTO() ;
    MEN_get_time (&pos, FUTURE) ; // calculate based on delta interval
    AS_time_functions (&pos) ;
    pos.obj_ra = AS_intern_ra_to_rad (MEN_star.ra) ;
    pos.obj_dec = AS_intern_dec_to_rad (MEN_star.dec) ;
    AS_nutation (&pos) ;
    AS_aberration (&pos) ;
    AS_allow_refraction (&pos) ;
    MEN_SET_ALT_AZ_DELTA (pos.altitude, pos.azimuth) ;
    m->min = 0 ;
  }
}
