/*
 * motors.c
 ******************************************************************************
 * Copyright (C) 2002  Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. http://www.gnu.org
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ******************************************************************************
 * Routines to drive a bipolar stepper motor.
 */

#include <stdio.h>
#include <string.h>
#include <asm/MC68EZ328.h> /* for all I/O port locations */
#include <sched.h>
#include <sys/types.h>
#include <unistd.h>
#include "types.h"
#include "asm.h"
#include "motors.h"
#include "aux_driver.h"
#include "process.h"
#include "telescope.h"
#include "joystick.h"

/******************************************************************************
 * The following array, speed_struct, contains CBA bit values for setting the
 * free running mode of the stepper motor controllers.
 *
 * The values in the array are a bit mask.
 *	0x20==> 1=automatic run mode, 0=manual step mode
 *	0x10==> 1=half step, 0=full step
 *	0x04=*C,
 *	0x02=*B,
 *	0x01=*A
 *
 * WHEN WRITING TO THE 4099 PORT, 0 SENDS a +ve VOLTAGE TO THE STEPPER CHIP
 */

SPEED_STRUCT MOT_speed [MOT_MAX_SPEED] =
{
// bits, half steps per minute
  {0x17,	0}, // stopped  =0
  {0x37,     1840}, //  4.6 RPM =1
  {0x34,     2920}, //  7.3 RPM =2
  {0x33,     3680}, //  9.2 RPM =3
  {0x32,     4800}, // 12  RPM  =4
  {0x24,     5840}, // 14.6 RPM =5
  {0x21,     7280}, // 18.2 RPM =6
  {0x22,     9600}, // 24   RPM =7
  {0x30,    14560}, // 36.4 RPM =8
  {0x20,    28240}  // 70.6 RPM =9
} ;

struct MOTOR_MSG_STRUCT
{
  GENERIC_MSG	command ;
  MOT_DIR	dir ;
} ;

typedef struct	MOTOR_MSG_STRUCT	MOTOR_MSG ;

/******************************************************************************
 */

void
MOT_msg (const Ulong steps_per_min, const MOT_DIR dir, const char alt_az)
{
  MOTOR_MSG	msg ;
  int		queue ;

  msg.command.mtype = STEPPER_SPEED ;
  msg.command.sub_command = steps_per_min ;
  msg.command.reply_to = This_Proc->msgq ;
  msg.dir = dir ;

  queue = (alt_az == 'l') ? Global->alt_msgq : Global->az_msgq ;

  #ifdef MOT_DEBUG
    PR_log_msg (NORM, "Sending motor msg to %s: rate=%d, dir=%d", PR_proc_from_msgq (queue), msg.command.sub_command, (int)msg.dir) ;
  #endif

  PR_send_mesg (queue, (LINUX_MSG_BUF *)&msg, sizeof(msg)) ;
}

/******************************************************************************
 * MOT_set_run_speed (speed, alt_az, direction)
 *
 * speed is from 0 to MAX_SPEED
 * alt_az may be either AUX_ALT or AUX_AZ
 * direction
 */

void
MOT_set_run_speed (Ushort steps_per_min, Uchar alt_az, MOT_DIR direction)
{
  Ulong  ssp ;
  Uchar  bits ;
  Ushort i ;

  for (i=0 ; i < MOT_MAX_SPEED ; i++)
  {
    if (MOT_speed[i].rate >= steps_per_min)
      break ;
  }

  #ifdef MOT_DEBUG
    PR_log_msg (NORM, "Set_run_speed: index=%hu, steps_per_min=%hu, alt_az=%d, dir=%d", i, steps_per_min, alt_az, direction) ;
  #endif

  #ifdef PARANOID
    if (i >= MOT_MAX_SPEED)
    {
      PR_log_msg (ERR, "Tried to set run speed (%hu) greater than max!\n", steps_per_min) ;
      return ;
    }

    if ((alt_az != AUX_ALT) && (alt_az != AUX_AZ))
    {
      PR_log_msg (ERR, "Set_run_speed: Specified bad alt_az (%d)!\n", (int)alt_az) ;
      return ;
    }
  #endif

  /*
   * Set the half stepping and speed control bits according to speed. Speed=0
   * is treated specially.
   */

  bits = MOT_speed [i].bits ;
  ssp = supervisor_mode() ;
  AUX_write_aux_bit (alt_az+MOT_DIRECTION, (BOOL)(direction==MOT_FORWARD), True) ;
  AUX_write_aux_bit (alt_az+MOT_HALF_STEP_MODE, (BOOL)(bits & 0x10), True) ;
  AUX_write_aux_bit (alt_az+MOT_STEP_A, (BOOL)(bits & 0x01), True) ;
  AUX_write_aux_bit (alt_az+MOT_B, (BOOL)(bits & 0x02), True) ;
  AUX_write_aux_bit (alt_az+MOT_C, (BOOL)(bits & 0x04), True) ;
  AUX_write_aux_bit (alt_az+MOT_RUN_MODE, (BOOL)(bits & 0x20), True) ;
  user_mode (ssp) ;
}

/******************************************************************************
 */

void
MOT_advance_one_step (Uchar alt_az, MOT_DIR direction, BOOL half_step)
{
  Ulong ssp ;

  #ifdef MOT_DEBUG
    PR_log_msg (NORM, "Adv_one_step: alt_az=%hu, dir=%hu, halfstep=%hu", (Ushort)alt_az, (Ushort)direction, (Ushort)half_step) ;
  #endif

  #ifdef PARANOID
    if ((alt_az != AUX_ALT) && (alt_az != AUX_AZ))
    {
      PR_log_msg (ERR, "Adv_one_step: Specified bad alt_az (%d)!\n", (int)alt_az) ;
      return ;
    }
  #endif

  ssp = supervisor_mode() ;
  AUX_write_aux_bit (alt_az+MOT_DIRECTION, (BOOL)(direction==MOT_FORWARD), True) ;
  AUX_write_aux_bit (alt_az+MOT_HALF_STEP_MODE, half_step, True) ;
  AUX_write_aux_bit (alt_az+MOT_RUN_MODE, False, True) ;

  /* Now perform a step. The step happens on a falling edge */
  AUX_write_aux_bit (alt_az+MOT_STEP_A, False, True) ;
  MOT_kill_time (40) ;
  AUX_write_aux_bit (alt_az+MOT_STEP_A, True, True) ; 
  user_mode (ssp) ;
}

/******************************************************************************
 */

#define MDIR(X) (((X) >= 0) ? MOT_FORWARD : MOT_REVERSE)

void
MOT_process (const Ushort arcsec_per_sec)
{
  if (JOY_AZ)
    MOT_msg ((Ulong)(arcsec_per_sec*60.0*STEPS_PER_REV/AZ_ARCSEC_PER_REV), MDIR(JOY_AZ), 'z') ;
  else
    MOT_msg (0, MOT_FORWARD, 'z') ;

  if (JOY_ALT)
    MOT_msg ((Ulong)(arcsec_per_sec*60.0*STEPS_PER_REV/AZ_ARCSEC_PER_REV), MDIR(JOY_ALT), 'l') ;
  else
    MOT_msg (0, MOT_FORWARD, 'l') ;
}
/******************************************************************************
 */

int
MOT_motor_process (const char *param)
{
  Uchar		alt_az ;
  MOTOR_MSG	msg ;
  Ushort	step_ticks ;
  MOT_DIR	dir = MOT_FORWARD ;
  Ushort	i ;

  #ifdef BASIC_DEBUG
    ANNOUNCE_PROCESS() ;
    PR_log_msg (FILE_ONLY, "Called with parameter \"%s\"", param) ;
  #endif

  if (*param == 'l')
  {
    alt_az = AUX_ALT ;
    Global->alt_msgq = This_Proc->msgq ;
  }
  else if (*param == 'z')
  {
    alt_az = AUX_AZ ;
    Global->az_msgq = This_Proc->msgq ;
  }
  else
    return 10 ;

  PR_set_process_priority (SCHED_RR, 2) ;
  MOT_set_run_speed (0, alt_az, MOT_FORWARD) ;

  for (;;)
  {
    PR_read_mesg (This_Proc->msgq, (LINUX_MSG_BUF *)(&msg), sizeof(msg), 0) ;

    switch (msg.command.mtype)
    {
      case TIMEOUT_NOTIFICATION:
	if (step_ticks != 0)
	{
	  PR_send_timeout_req_msg (This_Proc->msgq, step_ticks, &step_ticks) ;

	  #ifdef MOT_DEBUG
	    PR_log_msg (NORM, "Timeout: Stepping in direction=%d", dir) ; //dir = 9,10
	  #endif

	  MOT_advance_one_step (alt_az, dir, True) ; // "true" specs half step
	}
	break ;

      case STEPPER_SPEED:
	if (msg.command.sub_command == 0)
	{
	  #ifdef MOT_DEBUG
	    PR_log_msg (NORM, "Stepper_speed: Speed=0") ;
	  #endif

	  step_ticks = 0 ;
	  MOT_set_run_speed (0, alt_az, MOT_FORWARD) ;
	}
	else if (msg.command.sub_command < MOT_speed[1].rate)
	{
	  // manually step if below minimum auto run rate of the motor controller
	  step_ticks = (60 * TICK_FREQ) / msg.command.sub_command ;

	  if (step_ticks == 0)
	    step_ticks++ ;

	  dir = msg.dir ;
	  PR_send_timeout_req_msg (This_Proc->msgq, step_ticks, &step_ticks) ;

	  #ifdef MOT_DEBUG
	    PR_log_msg (NORM, "Stepping manually: Step interval=%d ticks", step_ticks) ;
	  #endif
	}
	else
	{
	  for (i=2 ; i<MOT_MAX_SPEED ; i++)
	    if (msg.command.sub_command < MOT_speed[i].rate)
	      break ;

	  if (i >= MOT_MAX_SPEED)
	    i = MOT_MAX_SPEED-1 ;

	  #ifdef MOT_DEBUG
	    PR_log_msg (NORM, "Stepping auto: Run speed=%d (index=%d)", MOT_speed[i].rate, i) ;
	  #endif

	  MOT_set_run_speed (MOT_speed[i].rate, alt_az, msg.dir) ;
	  step_ticks = 0 ;
	}
	break ;

      #ifdef PARANOID
        default:
	  PR_log_msg (ERR, "Bad msg recd: %s, Subcmd: %#0x, From %s", PR_decode_message_type (msg.command.mtype), msg.command.sub_command, PR_proc_from_msgq (msg.command.reply_to)) ;
	  break ;
      #endif
    }
  }

  return 98 ; // Ungracefully exit
}
