/*
 * Copyright (C) 2001  Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. http://www.gnu.org
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ******************************************************************************
 * process.h
 *
 * The purpose of this code is to automate the process of implementing a multi
 * process system. It is intended to work under uClinux and full blown Linux.
 * The two operating systems have different capabiltiies because of the absence
 * of support for a MMU in uClinux. This code tries to hide this from the
 * application code.
 */

#ifndef  __PROCESS_H
#define  __PROCESS_H

/******************************************************************************
 * DIFFERENT SYSTEM MESSAGES ARE DEFINED HERE.
 *
 * Each type of message has a different set of message values and structure.
 */

enum MESSAGE_TYPES
{
  LOG_MESSAGE=1,	// 1	messages to the logging process
  CLOSE_LOG,		// 2	stop logging
  TICK_EVENT,		// 3	the timekeeper is getting a tick
  TIMEOUT_REQUEST,	// 4	timeout messages to the timekeeper
  TIMEOUT_NOTIFICATION,	// 5	timeout messages from the timekeeper
  SLEEP_REQUEST,	// 6	sleep messages to the timekeeper
  SLEEP_NOTIFICATION,	// 7	sleep messages from the timekeeper
  MONITOR,		// 8	trigger logging of a status msg

	/* define your messages in a list AFTER here */
  NUM_PROC_MSGS
} ;

typedef enum MESSAGE_TYPES PROC_MSG_TYPE ;

/******************************************************************************
 */

#include "process_setup.h" /* defines some of the constants required here */

/******************************************************************************
 * These message routines support any type of arbitrary messages. However, the
 * first few bytes of each message structure are defined and must be adhered
 * to, strictly.
 */

typedef struct msgbuf      LINUX_MSG_BUF ;

struct GENERIC_MSG_STRUCT
{
  long       mtype ;             /* a unique id */
  int        sub_command ;       /* additional message classification */
  int        reply_to ;          /* the queue id to which to send replies */

  /* DEFINE YOUR OWN FIELDS HERE - or define new structures based on this one */
} ;

typedef struct GENERIC_MSG_STRUCT  GENERIC_MSG ;

/* The following NULL structure is used by PR_clear_mesg. Be careful to ensure
 * that messages are always shorter than this!
 */

#define MAX_FIELD_SIZE  300

struct GENERIC_NULL_MSG_STRUCT
{
  GENERIC_MSG header ;
  char        dummy_field [MAX_FIELD_SIZE] ;
} ;

typedef struct GENERIC_NULL_MSG_STRUCT GENERIC_NULL_MSG ;

#define MSG_OVERHEAD  (sizeof(long))  /* the mtype field is overhead to LINUX */

/******************************************************************************
 * The LOG_PRIORITY type is a list of the different possible priorities that a
 * log message could have.
 */

#ifdef HAVE_LOG_PROCESS

enum LOG_PRIORITY_TYPES
{
  NORM,           /* normal priority - log to file if file, otherwise screen */
  SCREEN,         /* Log to screen even if a file is defined */
  ERR,            /* this is an error, log to screen also */
  FILE_ONLY,      /* log message to file but not screen */
  LOG_ABORT       /* display this message and bail */
} ;

typedef enum LOG_PRIORITY_TYPES LOG_PRIORITY ;

#define LOG_MSG_SIZE	(100)	/* max characters in a log message */

struct LOG_MSG_STRUCT
{
  long         mtype ;             /* a unique id */
  LOG_PRIORITY pri ;
  char         msg [LOG_MSG_SIZE+1] ;
} ;

typedef struct LOG_MSG_STRUCT LOG_MSG ;

#endif

/******************************************************************************
 * The following constants and structures relate to the timer process.
 */

#ifdef HAVE_TIMEKEEPER

#define	FRAC_TIME(A)	(A*100/TICK_FREQ)

struct TIMER_STRUCT
{
  Ushort              interval ;/* number of ticks remaining */
  Ushort              type ;	/* number of ticks remaining */
  int                 ref_id ;	/* magic number specific to the originator */
  int                 reply_to ;/* message queue to notify of timeout */
  struct TIMER_STRUCT *next ;	/* linked list of timers */
} ;

struct SYS_TIME_STRUCT
{
  Ushort             ticks ;
  Ulong              seconds ;
} ;

typedef struct TIMER_STRUCT TIMER ;
typedef struct SYS_TIME_STRUCT SYS_TIME ;

struct TIMER_MSG_STRUCT
{
  GENERIC_MSG	command ;
  Ushort	interval ; /* measured in TICKS */
} ;

typedef struct TIMER_MSG_STRUCT   TIMER_MSG ;

#endif

/******************************************************************************
 * These definitions must be recognised as command line parameters by the
 * executable.
 */

#define SPAWN	"*spawn"
#define DATA	"*data"
#define PARAM	"*param"
#define CMD_SIZE 50		/* CMD_SIZE includes null termination */
#define LINE_TERM "\n"		/* NULL if using real logger, else \n */
#define LOGGER "Logger" 	/* MUST use this name for the log process */
#define TIMEKEEPER "Timekeeper" /* MUST use this name for the timekeeper process */

///////////////////////////////////////////////////////////////////////////////
#ifndef HAVE_FORK

extern char	Child [CMD_SIZE] ;
extern int	Data ;
extern char	Param [CMD_SIZE] ;
#endif

///////////////////////////////////////////////////////////////////////////////
#ifdef HAVE_LOG_PROCESS
#define __XA (1)
#else
#define __XA (0)
#endif

///////////////////////////////////////////////////////////////////////////////
#ifdef HAVE_TIMEKEEPER
#define __XB (1)
#else
#define __XB (0)
#endif

#define NUM_SYSTEM_PROCESSES (__XA + __XB)

/******************************************************************************
 * PROCESS DEFINITION ARRAY
 *
 * The following information is used to map process information to process id
 * Only the children are known and stored here. There are two parts in the
 * structure: a fixed and variable part.
 *
 * FIXED PART (never modified by programme)
 *   name - this is the name of the process - for use in debugging
 *   proc - this is the pointer to the function implementing the process
 *   param - this parameter will be passed in the function invocation
 *   shm_size - this specifies the number of bytes of shared memory to allocate
 *   io - zero or one child processes will inherit the open file descriptors for
 *        stdin, stdout, stderr. All others will receive /dev/null descriptors
 *
 * NOTE: There cannot be any spaces in the "name" field.
 *
 * VARIABLE PART: (if fields are non zero, they're never modified)
 *   pid - when the process is forked, the system pid is stored here
 *   msgq - when a message queue is created, the system queue id is stored here
 *   sem_id - when a semaphore is created, this will be its identifier
 *   shm_id - when shared memory is created, this will be its identifier
 *   shm - this is a pointer to the shared memory itself
 *
 * Setting msgq or sem_id to "-1" causes them not to be initialised
 * Setting msgq or sem_id to any other value causes them to be initialised
 *
 * Shared memory will be initialised if shm_size is set to a non zero value
 *
 * THIS ARRAY MUST BE DEFINED IN A MANNER WHICH IS GLOBALLY ACCESSIBLE BY
 * ALL PROCESSES. IT IS THE USER'S RESPONSIBILITY TO DECLARE THE MEMORY AS
 * SHARED MEMORY.
 */

typedef int PROC_DEF(const char *) ;

struct PROC_INFO_STRUCT
{
  /* The fixed part of this structure is first */
  /* These fields are defined but never changed */
  char		name[PROCESS_NAME_SIZE+1] ;
  PROC_DEF	*proc ;
  const char	*param ;
  int		shm_size ;
  short 	io ;

  /* The variable part of this structure follows */
  int		pid ;
  int		msgq ;	/* These fields may be set but could be changed */
  int		sem_id ;
  int		shm_id ;
  void		*shm ;
} ;

typedef struct PROC_INFO_STRUCT PROC_INFO ;

/******************************************************************************
 * SHARED GLOBAL STORAGE
 *
 * The following structure contains shared global storage to be available to
 * all processes in the system.
 *
 * This is only the structure definition. It is the responsibility of the
 * user's code to declare the shared memory and then fill it with valid
 * information.
 */

struct SHARED_GLOBAL_STORAGE_STRUCT
{
  /* FIXED ENTRIES */

  PROC_INFO	*Processes ;	/* This is the array */
  Ushort	Num_Proc ;	/* This is the number of entries in the array */

#ifdef HAVE_LOG_PROCESS
  int		Log_Proc_id ;	/* message Q of log proc, -1 if none */
#endif

#ifdef HAVE_TIMEKEEPER
  int		Timekeeper_ID ;	/* message Q of timekeeper, -1 if none */
  SYS_TIME	Up_Time ;	/* Up time of programme in ticks and seconds */
#endif

  char		index ;		/* A reference for creating IPC structures */

  /* USER DEFINED (application specific) ENTRIES FOLLOW */

  USER_DEFINED_GLOBALS
} ;

typedef struct SHARED_GLOBAL_STORAGE_STRUCT SHARED_GLOBAL_STORAGE ;

/******************************************************************************
 * Global variables
 */

#ifndef HAVE_FORK
  #define DECLARE_PROCESS_OPTIONS \
	{SPAWN, "Spawn a child", STRING, &Child, 0, "", 0, CMD_SIZE, 1}, \
	{DATA, "Pointer to common data", INT, &Data, 0, NULL, 0x80000000, 0x7fffffff, 1}, \
	{PARAM, "Process parameter", STRING, &Param, 0, "", 0, CMD_SIZE, 1},
#else
  #define DECLARE_PROCESS_OPTIONS
#endif

extern PROC_INFO *This_Proc ;
extern SHARED_GLOBAL_STORAGE *Global ;

#define ANNOUNCE_PROCESS()	PR_log_msg(FILE_ONLY, "STARTING. PID=%d, Queue_Id=%d", getpid(), This_Proc->msgq)

/******************************************************************************
 * The following public services are provided by the process.c module
 */

extern void PR_set_process_info (const char *name, const int pid, const int msgq1, const int msgq2) ;
extern char *PR_proc_from_msgq (const int msgq) ;
extern char *PR_proc_from_pid (const int pid) ;
extern short PR_proc_from_name (const char *name) ;

extern void PR_init_shared_memory (void) ;
extern int  PR_process (char *exec_name, const int argc, char **argv) ;
extern int  PR_launch_process (char *proc, const int argc, char **argv) ;
extern int  PR_create_system_structures (const char *logfilename) ;
extern void PR_register_new_process (const char *name, PROC_DEF *proc, const char *param, const int shm_size, const short io, const int msgq, const int sem_id) ;
extern void PR_disp_process (void) ;

extern void PR_set_process_priority (const int policy, const int priority) ;

extern char *PR_decode_message_type (const long mtype) ;
extern void PR_clear_mesg (const int msgq_id) ;
extern int  PR_create_mesg (const char token) ;
extern void PR_read_mesg (const int queue, LINUX_MSG_BUF *buf, const int max_msg_len, const long msg_type) ;
extern void PR_send_mesg (const int queue, LINUX_MSG_BUF *buf, const int msg_len) ;
extern void PR_rm_mesg (const int queue) ;

extern int  PR_create_shm (const char token, const int size, void **shm_addr) ;
extern void PR_rm_shm (const int shm) ;

#ifdef HAVE_TIMEKEEPER
extern int  PR_timekeeper_process (const char *param) ;
extern void PR_send_timeout_req_msg (const int rep_to, const Ushort interval, const void *id) ;
extern void PR_sleep (const Ushort interval) ;
#endif

extern void PR_create_sem (const char token, int *sem) ;
extern void PR_rm_sem (int *sem) ;
extern void PR_unlock_sem (int *sem) ;
extern void PR_lock_sem (int *sem) ;

#ifdef HAVE_LOG_PROCESS
extern void PR_log_msg (const LOG_PRIORITY priority, const char *fmt, ...) ;
extern int  PR_log_process (const char *filename) ;

#else
#define PR_log_msg fprintf
#define ERR	  stderr
#define SCREEN	  stdout
#define FILE_ONLY stdout
#define NORM	  stdout
#endif

#endif
