/*
 * setcirc.h
 ******************************************************************************
 * Header file for digital setting circles code
 *
 * In preparing this code, I drew heavily from algorithms first published in
 * the February 1989 issue of "Sky and Telescope" in an article by T.Taki
 * concerning the pointing of Dobsonian Telescopes. The software appearing in
 * this article may be obtained from the following link:
 * http://skyandtelescope.com/resources/software/programs/taki.bas
 *
 * I also drew from a translation of the T.Taki code into the C language by
 * Mel Bartels. His translation may be found on his web site within a zip
 * file of the source code. You may find it at:
 * http://www.bbastrodesigns.com/cot/cot.html and
 * http://zebu.uoregon.edu/~mbartels/dnld/source_code.zip (convert.c)
 *
 * I have chosen to release my adaptation of these algorithms under the GNU
 * public license.
 ******************************************************************************
 * Copyright (C) 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************/
// Assumptions:
//	The microprocessor running this application has an accurate real time
//	clock. Accuracy needs to be in terms of frequency rather than in terms
//	of having the right time set.
//
//	Preferably, the real time clock can be read to a fraction of a second.
//	It's OK if it can only be read to a second accuracy however this will
//	limit the accuracy of the setting circle to around 15".
//
//	The shaft encoders for the ALT and AZ axes should be as accurate as
//	possible - certainly accurate to an arc minute or hopefully better.
//
//	The routine to poll the shaft encoder value should convert the reading
//	to radians. This means that the size of the increment needs to be
//	accurately calculated for each telescope.
//
// To use:
//	(a) fill Z1, Z2 and Z3 with the telescope error values. 0.0=no error.
//		The three values are Kalman error coefficients. You can find
//		out how to calculate them by reading "Telescope Control" by
//		Trueblood and Genet. An alternate algorithm is provided in
//		Mel Bartel's source code zip file.
//		It's also OK to use the routines in the no-error mode.
//
//	(b) align on either two or three stars, filling STAR structs for each.
//		When the shaft encoders are polled to determine the telescope
//		coordinates, the system tick counter must also be polled and
//		the value stored in the STAR structure.
//
//	(c) call DSC_generate_transform_matrix (&star1, &star2, &star3)
//		For two star alignment, set star3 to (STAR *)0
//
//	(d) To determine RA/DEC from a known ALT/AZ/sys_ticks reading,
//		partially fill a STAR structure and then call
//		DSC_Alt_Az_from_Ra_Dec (&star_structure)
//
//	    To determine ALT/AZ from a known RA/DEC/sys_ticks reading,
//		partially fill a STAR structure and then call
//		DSC_Ra_Dec_from_Alt_Az (&star_structure)
//
//		The unfilled fields of the STAR structure will be filled out
//		upon return.
//
// Note:
//	It is feasible to perform a two star alignment and then, later in the
//	night, call DSC_generate_transform_matrix with a third star. In this
//	case, there is no need to reinitialise any matrices. Simply call
//	DSC_generate_transform_matrix with any three valid STAR structures.

#ifndef __SETCIRC_H
#define __SETCIRC_H

///////////////////////////////////////////////////////////////////////////////
// Constants

#define DEGREES_PER_RAD	  (360.0 / 2.0 / M_PI)
#define SID_TO_SOLAR_CONSTANT (1.00273790935) // multiply sidereal time by this to get solar time
#define TICK_FREQ 4 // frequency of the system ticker in Hz. This must be very accurate.
#define SECONDS_IN_DAY (60 * 60 * 24)

///////////////////////////////////////////////////////////////////////////////
// Data structures
//
// The STAR structure holds actual and measured angles in RADIANS
//	sys_ticks holds the elapsed since the programme started running

struct _star
{
  double	ra ;		// ra and dec of actual star
  double	dec ;
  double	alt ;		// alt and az reading from the shaft encoders
  double 	az ;
  Ulong		sys_ticks ;	// time at which the alt/az readings were taken
} ;

typedef struct _star STAR ;


typedef double MATRIX_1x3 [3] ;
typedef double MATRIX_3x3 [3][3] ;


///////////////////////////////////////////////////////////////////////////////
// Templates

extern MATRIX_3x3 Telescope ;
extern MATRIX_3x3 Actual ;
extern MATRIX_3x3 Transform_to_Ra_Dec ;
extern MATRIX_3x3 Transform_to_Alt_Az ;

extern double Z1, Z2, Z3 ;

extern void	DSC_fill_3x3_matrix (MATRIX_3x3 *matrix, const double value) ;
extern void	DSC_clear_matrices (void) ;
extern void 	DSC_init_array_column (MATRIX_3x3 *dest, const Ushort col, MATRIX_1x3 *src) ;
extern void	DSC_print_3x3_matrix (MATRIX_3x3 *matrix, const char *txt) ;
extern void	DSC_print_1x3_matrix (MATRIX_1x3 *matrix, const char *txt) ;
extern void	DSC_show_star (const STAR *star, const char *txt) ;

extern double	DSC_determinant (MATRIX_3x3 *A) ;
extern void	DSC_copy_3x3_matrix (MATRIX_3x3 *to, MATRIX_3x3 *from) ;
extern void	DSC_multiply_3x3_with_3x3 (MATRIX_3x3 *result, MATRIX_3x3 *A, MATRIX_3x3 *B) ;
extern void	DSC_multiply_3x3_with_1x3 (MATRIX_1x3 *result, MATRIX_3x3 *A, MATRIX_1x3 *B) ;
extern void	DSC_transpose_3x3 (MATRIX_3x3 *A) ;

extern void	DSC_spherical_to_rect (MATRIX_1x3 *result, const STAR *star) ;
extern double	DSC_calc_sidereal_time (const STAR *star) ;
extern void	DSC_init_star (const STAR *star, const Ushort col) ;
extern void	DSC_kludge_third_star (MATRIX_3x3 *matrix) ;

extern void	DSC_generate_transform_matrices (const STAR *star1, const STAR *star2, const STAR *star3) ;
extern void	DSC_Ra_Dec_from_Alt_Az (STAR *star) ;
extern void	DSC_Alt_Az_from_Ra_Dec (STAR *star) ;

#endif
