/*
 * telescope.c
 ******************************************************************************
 * Copyright (C) 2002  Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. http://www.gnu.org
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sched.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <stdarg.h>
#include <math.h>

#include "types.h"	/* for definitions of common types */
#include "parse.h"	/* routines to parse command line parameters */
#include "process.h"	/* routines to support multi processing */
#include "readline.h"	/* text I/O */
#include "stringfn.h"	/* manipulation of strings */
#include "aux_driver.h"
#include "joystick.h"
#include "lcd.h"
#include "encoders.h"
#include "motors.h"
#include "telescope.h"
#include "menu.h"
#include "astro.h"
#include "database.h"
#include "access_data.h"

///////////////////////////////////////////////////////////////////////////////
// Here are other definitons which don't fall into the above categories.

char	Logfilename [FILE_NAME_SIZE] ;

#define LCD_MSG_SIZE	150	// max chars in LCD message

struct LCD_MSG_STRUCT
{
  long	mtype ;
  Uchar	line ;
  Uchar delay ;
  char	msg [LCD_MSG_SIZE+1] ;
} ;

typedef struct LCD_MSG_STRUCT LCD_MSG ;

///////////////////////////////////////////////////////////////////////////////
// The following routines make message passing a little easier

void
TEL_lcd_msg (const Uchar line, const Uchar delay, const char *fmt, ...)
{
  int		n ;
  va_list	args ;
  LCD_MSG	msg ;

  va_start(args, fmt) ;
  vsnprintf (&msg.msg[0], LOG_MSG_SIZE-2, fmt, args) ;
  va_end(args) ;
  msg.mtype = LCD_MESSAGE ;
  msg.line = line ;
  msg.delay = delay ;
  n = strlen (msg.msg) ;

  #ifdef BASIC_DEBUG
    PR_log_msg (SCREEN, "TEL_lcd_msg: \"%s\"", msg.msg) ;
  #endif

  PR_send_mesg (Global->lcd_msgq, (LINUX_MSG_BUF *)&msg, MSG_OVERHEAD+sizeof(msg.line)+sizeof(msg.delay)+n+1) ;
}

void
TEL_lcd_cur (const Uchar line, const Uchar cursor)
{
  LCD_MSG	msg ;

  msg.mtype = LCD_CURSOR ;
  msg.line = line ;
  msg.msg[0] = cursor ;

  #ifdef BASIC_DEBUG
    PR_log_msg (FILE_ONLY, "TEL_lcd_cur: line=%d, cur=%d", msg.line, msg.msg[0]) ;
  #endif

  PR_send_mesg (Global->lcd_msgq, (LINUX_MSG_BUF *)&msg, MSG_OVERHEAD+sizeof(msg.line)+sizeof(msg.delay)+1) ;
}

/******************************************************************************
 * The LCD is a two line 20 character per line display. You write ascii text
 * to it.
 *
 * Any process can write to the LCD using a printf-like command called TEL_lcd_msg.
 * This routine sends a Linux message containing the formatted text to a
 * dedicated LCD controller process.
 *
 * The LCD control process displays one message at at time. The driver takes
 * care of scrolling text if you try to display something that's greater
 * than 20 characters.
 *
 * It's the nature of Linux messages that they form a queue to wait to be
 * processed. The LCD driver can be as fast or slow as it likes and any text
 * messages will just patiently queue to wait their turn to be displayed.
 */

int
lcd_process (const char *param)
{
  LCD_MSG msg ;

  #ifdef BASIC_DEBUG
    ANNOUNCE_PROCESS() ;
    PR_log_msg (FILE_ONLY, "Called with parameter \"%s\"", param) ;
  #endif

  PR_set_process_priority (SCHED_RR, 1) ; // Priority above normal
  Global->lcd_msgq = This_Proc->msgq ;
  LCD_initialise () ;
  LCD_control_light (True) ;
  TEL_lcd_msg (0, 0, "") ;
  TEL_lcd_msg (1, 1, "                   Welcome to Stefan's Telescope! ") ;

  for (;;)
  {
    PR_read_mesg (This_Proc->msgq, (LINUX_MSG_BUF *)(&msg), sizeof(msg), 0) ;

    if (msg.mtype == LCD_MESSAGE)
    {
      #ifdef BASIC_DEBUG
        PR_log_msg (FILE_ONLY, "Writing message: \"%s\"", msg.msg) ;
      #endif

      LCD_write_lcd_string (msg.line, msg.msg) ;

      if (msg.delay)
	PR_sleep (TICK_FREQ) ;
    }
    else if (msg.mtype == LCD_CURSOR)
    {
      #ifdef BASIC_DEBUG
        PR_log_msg (FILE_ONLY, "Setting cursor: line=%d, state=%d", msg.line, msg.msg[0]) ;
      #endif

      LCD_position_cursor (msg.line, msg.msg[0], True) ;
    }
  }

  return 100 ; // Not reached
}
///////////////////////////////////////////////////////////////////////////////
void
TEL_one_sec_msg (const long mes_type)
{
  GENERIC_MSG	msg ;

  msg.mtype = mes_type ;
  msg.sub_command = 0 ;
  msg.reply_to = This_Proc->msgq ;

  #ifdef BASIC_DEBUG
    PR_log_msg (NORM, "Sending one_sec msg: type=%ld", msg.mtype) ;
  #endif

  PR_send_mesg (Global->one_sec_msgq, (LINUX_MSG_BUF *)&msg, sizeof(msg)) ;
}

int
one_second_process (const char *param)
{
  GENERIC_MSG	msg ;

  #ifdef BASIC_DEBUG
    ANNOUNCE_PROCESS() ;
    PR_log_msg (FILE_ONLY, "Called with parameter \"%s\"", param) ;
  #endif

  PR_set_process_priority (SCHED_RR, 1) ;
  Global->one_sec_msgq = This_Proc->msgq ;

  if (ENC_init_hardware())
    return 9 ; // exit ungracefully

  // Start the one second timer
  PR_send_timeout_req_msg (This_Proc->msgq, TICK_FREQ, &msg) ;

  for (;;)
  {
    PR_read_mesg (This_Proc->msgq, (LINUX_MSG_BUF *)(&msg), sizeof(msg), 0) ;

    switch (msg.mtype)
    {
      case TIMEOUT_NOTIFICATION:
	PR_send_timeout_req_msg (This_Proc->msgq, TICK_FREQ, &msg) ;

	if (ENC_read_coordinates())
	{
	  // We saw a change in the XY encoders

	  #ifdef BASIC_DEBUG
	    TEL_lcd_msg (1, 0, "X=%d, Y=%d", ENC_X_coord, ENC_Y_coord) ;
	  #endif
	}
	break ;

      case INITIALISE:
	if (ENC_init_hardware())
	  return 8 ; // exit ungracefully

	break ;
    }
  }

  return 99 ; // Not Reached
}
/******************************************************************************
 * The joystick has four directions (up, down, left, right), two buttons (big
 * and little) and a slide switch.
 *
 * The little button controls panning speed.
 * The big button is a panic switch to turn off the motors. I'll find another
 *     function with this when I get some confidence in the algorithms.
 * The slide switch controls the LCD back light.
 */

#define SCAN_PERIOD 4 // Scan joystick each quarter of a second

struct SPEED_STRUCT
{
  char	 name[8] ;
  Ushort setting ;
} ;

typedef struct SPEED_STRUCT SPEED ;

#define NUM_SPEEDS 5

SPEED speed_array [NUM_SPEEDS] =
{ // Speeds are specified in 1/3600ths of a degree per second
  {"Slowest",	60},	// 1 arcminute per second
  {"Slow",	240},	// 4 arcminutes per second
  {"Medium",	600},	// 10 arcminutes per second
  {"Fast",	3600},	// 1 degree per second
  {"Fastest",	14400}	// 4 degrees per second
} ;

int
joy_process (const char *param)
{
  Uchar		mode ;
  time_t	tmp ;
  int		hour, minute ;
  struct tm	*raw ;
  Ushort	i ;
  Uchar		speed ;

  #ifdef BASIC_DEBUG
    ANNOUNCE_PROCESS() ;
    PR_log_msg (FILE_ONLY, "Called with parameter \"%s\"", param) ;
  #endif

  PR_set_process_priority (SCHED_RR, 1) ;
  mode = TEL_SLEW ;
  raw = 0 ;
  minute = hour = -1 ;
  speed = 0 ;

  TEL_lcd_msg (0, 0, "Reading database") ;
  DB_initialise() ; // populate the database tables
  MEN_current = &place ; // set LON / LAT to the default location
  TEL_lcd_msg (0, 0, "Setting default loc") ;
  MEN_disp_file (MEN_current, NULL) ;
  MEN_edit_file (MEN_current, 0) ;
  TEL_lcd_msg (1, 1, MEN_name) ; // display the name of the default location

  MEN_current = &men1 ; // initialise menu to the top
  MEN_cursor = 0 ;

  // uClinux doesn't support timezones properly
  #ifdef UCLINUX
    Global->time_offset = tzone.value ;
  #endif

  Global->altitude = Global->azimuth = 0 ;
  Global->alt_delta = Global->az_delta = 0 ;

  for (i=1 ;;) // initialise i to trigger at the first iteration
  {
    PR_sleep (TICK_FREQ / SCAN_PERIOD) ;

    if (--i == 0)
    {
      i = SCAN_PERIOD ;
      tmp = time (NULL) ;

      // uClinux doesn't support timezones properly
      #ifdef UCLINUX
	tmp += Global->time_offset ;
      #endif

      raw = localtime (&tmp) ;
    }

    if (JOY_scan_inputs() || (minute != raw->tm_min))
    {
      LCD_control_light (JOY_current & JOY_LIGHT_BIT) ;

      if (mode == TEL_CMD)
      {
	#ifdef BASIC_DEBUG
	  PR_log_msg (NORM, "TEL_CMD: change=%02x, current=%02x", JOY_changes, JOY_current) ;
	#endif

	minute = raw->tm_min ;
	hour = raw->tm_hour ;

	// Check to see if we should change to SLEW mode
	if (JOY_changes & JOY_current & JOY_MODE_BIT)
	{
	  mode = TEL_SLEW ;
	  minute = -1 ; // force a display of the time at next itteration
	}

	MEN_process() ;
      }
      else // in slew mode
      {
	#ifdef BASIC_DEBUG
	  PR_log_msg (NORM, "TEL_SLEW: change=%02x, current=%02x", JOY_changes, JOY_current) ;
	#endif

	if (minute != raw->tm_min)
	{
	  TEL_lcd_msg (0, 0, "%*.*c%02d:%02d",
	    LCD_LINE_LEN-5, LCD_LINE_LEN-4, ' ', raw->tm_hour, raw->tm_min) ;
	  TEL_lcd_msg (1, 0, "") ;
	  minute = raw->tm_min ;
	  hour = raw->tm_hour ;
	}

	// Check to see if we should change to CMD mode
	if (JOY_changes & JOY_current & JOY_MODE_BIT)
	{
	  mode = TEL_CMD ;
	  JOY_changes &= ~(JOY_UP_BIT | JOY_DOWN_BIT | JOY_LEFT_BIT | JOY_RIGHT_BIT) ;
	  MEN_process () ; // call with direction bits off to display screen
	}
	else if (JOY_changes & JOY_current & JOY_ENTER_BIT)
	{
	  if (++speed >= NUM_SPEEDS)
	    speed = 0 ;

	  TEL_lcd_msg (1, 0, speed_array [speed].name) ;
	}
	else if (JOY_changes & (JOY_UP_BIT | JOY_DOWN_BIT | JOY_LEFT_BIT | JOY_RIGHT_BIT))
	{
	  // Something about the state of the directions has changed
	  JOY_get_direction () ;

	  #ifdef BASIC_DEBUG
	    PR_log_msg (NORM, "AZ:%d, ALT:%d\n", JOY_AZ, JOY_ALT) ;
	  #endif

	  MOT_process (speed_array[speed].setting) ;
	}
      }
    }
  }

  return 100 ; // Not Reached
}
///////////////////////////////////////////////////////////////////////////////
OPTION opt[] =
{
  /* abbr, desc, type, var, defint, defstring, min, max/len, hidden */

  DECLARE_PROCESS_OPTIONS
  {"l", "Save log messages in a file rather than stdout", FILENAME, &Logfilename, 0, "logfile.txt", 0, FILE_NAME_SIZE, 0},
  DECLARE_PARSE_OPTIONS
} ;
///////////////////////////////////////////////////////////////////////////////
int
main (int argc, char **argv)
{
  Error = PA_parse (&argc, &argv) ;
  PROCESS_PARSE_OPTIONS(argv[0]) ;
  PROCESS_FINAL_OPTIONS (Error, argv[0], "control Stefan's Telescope") ;

  AUX_init_ports() ;

  if (PR_create_system_structures (Logfilename))
  {
    //			    (*name,		&proc(),	&param, shm_size, io, msgq, sem_id)
    PR_register_new_process ("LCD_proc",	lcd_process,	    "", 0, 0, 1, 0) ;
    PR_register_new_process ("1_sec_proc",	one_second_process, "", 0, 0, 1, 0) ;
    PR_register_new_process ("ALT_Motor",	MOT_motor_process,  "l",0, 0, 1, 0) ;
    PR_register_new_process ("AZ_Motor",	MOT_motor_process,  "z",0, 0, 1, 0) ;
    PR_register_new_process ("Joy_proc",	joy_process,	    "", 0, 0, 1, 0) ;
  }

  if (Error == 0)
    Error = PR_launch_process (argv[0], argc, argv) ;

  AUX_init_ports() ; // Do our last paranoid clean up
  MOT_set_run_speed (0, AUX_ALT, MOT_FORWARD) ;
  MOT_set_run_speed (0, AUX_AZ, MOT_FORWARD) ;
  exit (Error) ;
}
