/*
 * astro.c 
 ******************************************************************************
 * Astronomical calculations
 *
 * This code is based entirely upon algorithms published in the book
 * "Astronomical Algorithms" by Jean Meeus. 2nd Edition. Willmann-Bell.
 * It can be ordered from the publishers and can be located at
 * http://www.willbell.com/math/MC1.HTM
 *
 * You will need to obtain this reference in order to understand the code
 * and to understand how it should be used.
 *
 * The publisher has not made the algorithms available in electronic format.
 * As it was not a trivial task to copy the algorithms into C and then
 * meticulously find all the typos in the numerous constants, I thougth that
 * it might be of use to some others to make my translations available.
 *
 * I have chosen to release my adaptation of these algorithms under the GNU
 * public license.
 ******************************************************************************
 * Copyright (C) 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ******************************************************************************
 */

#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <float.h>
#include <string.h>

#include "types.h"
#include "astro.h"

//#define ASTRO_DEBUG

/*****************************************************************************
 * The following constants must be set by an external module. They will hold
 * the longitude and lattitude of the observer's location.
 */

double AS_longitude, AS_lattitude ; // MEASURED IN RADIANS

/*****************************************************************************
 * Representing angles using minimal storage space
 *
 * In order to build a reasonable database of stellar objects, the data needs
 * to be arranged in such a way as to minimise the storage space. If you have
 * a database containing 30k objects, every byte you save, will save 30k of
 * memory. This can amount to considerable savings in an otherwise small
 * microcontroller system.
 *
 * The proposed "compressed" storage system presented here allows RA and DEC
 * to be stored in 20 bits each. This leads to an maximum error of 0.08
 * seconds of RA (1/1,048,576 of a revolution) and 0.6 arcseconds of DEC
 * (1/2,097,152 of a revolution).
 *
 * The 20 bit representation of RA is an unsigned value. When multiplied by
 * 2*PI/1048576, it yields radians.
 *
 * The 20 bit representation of DEC is a signed value. When multiplied by
 * PI/2/524288, it also yields radians.
 */

Ulong
AS_rad_to_intern_ra (const double x)
{
  double result ;

  result = x * 1048576.0 / (2.0 * M_PI) ;

  if ((result - floor(result)) >= 0.5)
    return (Ulong)(result + 1) ;
  else
    return (Ulong)result ;
}

long
AS_rad_to_intern_dec (const double x)
{
  double result, abs_val ;
  long sign ;

  result = x * 524288.0 / (M_PI / 2.0) ;

  abs_val = fabs (result) ;
  sign = (result < 0.0) ? -1 : 1 ;

  if ((abs_val - floor(abs_val)) >= 0.5)
    return sign * (long)(abs_val + 1) ;
  else
    return sign * (long)abs_val ;
}

double
AS_intern_ra_to_rad (const Ulong x)
{
  return (double)(x) * 2.0 * M_PI / 1048576.0 ;
}

double
AS_intern_dec_to_rad (const long x)
{
  return (double)(x) * M_PI / 2.0 / 524288.0 ;
}

char *
AS_display_intern_RA (char *ptr, const Ulong ra, const int accurate)
{
  Ushort h, m ;
  double seconds ;

  seconds = (double)ra / 1048576.0 * ASH_DAY_LENGTH ;
  h = seconds / 3600 ;
  m = (seconds - h*3600) / 60 ;
  seconds -= h*3600 + m*60 ;

  if (seconds >= 59.95)
  {
    seconds = 0.0 ;

    if (++m == 60)
    {
      m = 0 ;
      h++ ;
    }
  }

  if (accurate)
    sprintf (ptr, "%02huh %02hum %1.1fs", h, m, seconds) ;
  else
  {
    if (seconds >= 30.0)
    {
      if (m == 59)
	{m=0; h++;}
      else
	m++ ;
    }

    sprintf (ptr, "%02huh %02hum", h, m) ;
  }

  return ptr ;
}

char *
AS_display_intern_DEC (char *ptr, long dec, const int accurate)
{
  char sign ;
  Ushort deg, min ;
  double seconds ;

  if (dec < 0)
  {
    sign = '-' ;
    dec = -dec ;
  }
  else
    sign = '+' ;

  seconds = (double)dec / 524288.0 * 90.0 * 3600.0 ;
  deg = seconds / 3600 ;
  min = (seconds - deg*3600) / 60 ;
  seconds -= deg*3600 + min*60 ;

  if (seconds >= 59.5)
  {
    seconds = 0.0 ;

    if (++min == 60)
    {
      min = 0 ;
      deg++ ;
    }
  }

  if (accurate)
    sprintf (ptr, "%c%hu%c %hu' %1.0f\"", sign, deg, DEG_CHAR, min, seconds) ;
  else
  {
    if (seconds >= 30.0)
    {
      if (min == 59)
	{min=0; deg++;}
      else
	min++ ;
    }

    sprintf (ptr, "%c%hu%c %hu'\"", sign, deg, DEG_CHAR, min) ;
  }

  return ptr ;
}

/*****************************************************************************
 * The following routines manipulate angles.
 *
 * AS_normalise_XXX takes an angle which might be many revolutions and
 * converts it to its equivalent angle between 0 and 360 degrees.
 */

double
AS_normalise_degrees (double angle)
{
  double ret ;

  ret = fmod (angle, 360.0) ;

  if (ret < 0.0)
    ret += 360.0 ;

  return ret ;
}

double
AS_normalise_rads (double angle)
{
  double ret ;

  ret = fmod (angle, 2.0 * M_PI) ;

  if (ret < 0.0)
    ret += 2.0 * M_PI ;

  return ret ;
}

double
AS_deg_to_rad (double x)
{
  return AS_normalise_degrees (x) * 2.0 * M_PI / 360.0 ;
}

double
AS_signed_deg_to_rad (double x)
{
  double norm ;

  norm = fmod (x, 360.0) ;

  if (norm < 0.0)
    norm += 360.0 ;

  if (norm > 180.0)
    norm -= 360.0 ;

  return norm * 2.0 * M_PI / 360.0 ;
}

double
AS_rad_to_deg (double x)
{
  return x / (2.0 * M_PI) * 360.0 ;
}

double
AS_signed_rad_to_deg (double x)
{
  double norm ;

  norm = fmod (x, 2.0 * M_PI) ;

  if (norm < 0.0)
    norm += 2.0 + M_PI ;

  if (norm > M_PI)
    norm -= 2.0 * M_PI ;

  return norm / (2.0 * M_PI) * 360.0 ;
}

/*****************************************************************************
 * Other routines
 */

void
AS_fill_time_struct (AS_TIME_STRUCT *t, const Ushort year, const Ushort month, const Ushort date, const Ushort hour, const Ushort minute, const Ushort seconds)
{
  t->year = year ;
  t->month = month ; // 0=JAN, 11=DEC
  t->date = date ;
  t->hour = hour ;
  t->minute = minute ;
  t->seconds = seconds ;
}

void
AS_initialise_struct (ASH_POSITION_STRUCT *pos)
{
  memset (&pos->GMT, '\0', sizeof(AS_TIME_STRUCT)) ;

  pos->Grnwch_SidTime =
    pos->obj_ra =
    pos->obj_dec =
    pos->apa_ra =
    pos->apa_dec =
    pos->apa_distance =
    pos->delta_ra =
    pos->delta_dec =
    pos->altitude =
    pos->azimuth =
    pos->delta_psi =
    pos->delta_epsilon =
    pos->epsilon0 =
    pos->lon_sun =
    pos->e =
    pos->pi =
    pos->JD =
    pos->JDE =
    pos->T =
    pos->T2 =
    pos->T3 =
    pos->T4 =
    0.0 ;
}

char *
AS_rad_to_DMS (char *ptr, double rad)
{
  int deg, min ;
  double degrees, sec ;
  char sign ;

  degrees = rad / (2.0 * M_PI) * 360.0 ;

  if (degrees > 0.0)
    sign = ' ' ;
  else
  {
    degrees = -degrees ;
    sign = '-' ;
  }

  deg = degrees ;
  min = (degrees - deg) * 60 ;
  sec = (((degrees - deg) * 60) - min) * 60.0 ;

  if (sec >= 59.995)
  {
    sec = 0.0 ;
    min++ ;

    if (min >= 60)
    {
      min = 0 ;
      deg++ ;
    }
  }

  sprintf (ptr, "%c%d%c%d'%.3f\"", sign, deg, DEG_CHAR, min, sec) ;
  return ptr ;
}

char *
AS_rad_to_HMS (char *ptr, double rad)
{
  int h, m ;
  double sec ;

  if (rad < -0.0000000001) // funny rounding problem with 0.0
    rad += 2.0 * M_PI ;

  h = rad / (2.0 * M_PI) * 24.0 ;
  m = (int)(rad / (2.0 * M_PI) * 24.0 * 60.0) % 60 ;
  sec = fmod ((rad / (2.0 * M_PI) * 24.0 * 60.0 * 60.0), 60.0) ;
  sprintf (ptr, "%dh %dm %.3fs", h, m, sec) ;
  return ptr ;
}

char *
AS_display_RA (char *ptr, const double ra)
{
  Ushort h, m ;
  double seconds ;

  seconds = ra / (2.0 * M_PI) * ASH_DAY_LENGTH ;
  h = seconds / 3600 ;
  m = (seconds - h*3600) / 60 ;
  seconds = (seconds - h*3600 - m*60) ;

  sprintf (ptr, "%02hu:%02hu:%1.4f, %8.6f%c, %8.6f rad", h, m, seconds, AS_rad_to_deg(ra), DEG_CHAR, ra) ;
  return ptr ;
}

char *
AS_display_date (char *ptr, const AS_TIME_STRUCT *ts)
{
  sprintf (ptr, "%hu/%hu/%hu %02hu:%02hu:%02hu", ts->date, ts->month+1, ts->year, ts->hour, ts->minute, ts->seconds) ;
  return ptr ;
}

/*****************************************************************************
 * The following routines manipulate Julian Days.
 * ALL TIMES ASSUMED IN UNIVERSAL TIME FORMAT
 *
 * Month: 0=Jan, 1=Feb, ....
 * Day of month: 1=1st, 2=2nd, ....
 *
 * The routines are valid for years > 1582 (ie. Gregorian Calendar only)
 */

// AS_julian_day: refer to pp 60-61
double
AS_julian_day (const AS_TIME_STRUCT *tim)
{
  Ushort year, month ;
  double time ;
  double a, b, t1, t2 ;
  double jd ;

  year = tim->year ;
  month = tim->month + 1 ;
  time = (((tim->hour * (60 * 60)) + (tim->minute * 60) + (tim->seconds))
		  / (double)ASH_DAY_LENGTH) ;

  time += tim->date ;

  if (month < 3) /* Jan or Feb? */
  {
    year -= 1 ;
    month += 12 ;
  }

  a = floor (year / 100.0) ;
  b = 2.0 - a + floor (a / 4.0) ;

  t1 = floor (365.25 * (year + 4716)) ;
  t2 = floor (30.6001 * (month + 1)) ;

  jd = t1 + t2 + time + b - 1524.5 ;

  #ifdef ASTRO_DEBUG
  {
    char buf[100] ;

    printf ("%s --> JD %12.10f\n", AS_display_date (buf, tim), jd) ;
  }
  #endif

  return jd ;
}

// AS_date_from_julian: refer to p63
void
AS_date_from_julian (const double julian_time, AS_TIME_STRUCT *tim)
{
  double Z, F, A, alph ;
  double B, C, D, E ;

  F = modf (julian_time+0.5, &Z) ;

  if (Z < 2299161)
    A = Z ;
  else
  {
    alph = floor ((Z - 1867216.25) / 36524.25) ;
    A = Z + 1.0 + alph - floor (alph / 4) ;
  }

  B = A + 1524.0 ;
  C = floor ((B - 122.1) / 365.25) ;
  D = floor (365.25 * C) ;
  E = floor ((B - D) / 30.6001) ;

  alph = B - D - floor (30.6001 * E) + F ;
  tim->date = alph ;
  tim->month = (E < 14.0) ? E-2.0 : E-14.0 ;
  tim->year = (tim->month < 2) ? C-4715 : C-4716 ;

  tim->hour = F * 24.0 ;
  tim->minute = (F * 24.0 * 60.0 + 0.01 ) - tim->hour * 60.0 ;
  tim->seconds = (F * (60.0 * 60.0 * 24.0) + 0.01) - tim->hour * 60.0 * 60.0 - tim->minute * 60.0 ;

  #ifdef ASTRO_DEBUG
  {
    char buf[100] ;

    printf ("JD: %.2f = %s\n", julian_time, AS_display_date(buf, tim)) ;
  }
  #endif
}

/*****************************************************************************
 * The following routines manipulates time.
 *
 * in pos->GMT....
 *	Month: 0=Jan, 1=Feb, ....
 *	Day of month: 1=1st, 2=2nd, ....
 *
 * The routines are valid for years > 2000
 */

// AS_greenwich_sidereal_time: refer to p87
void
AS_greenwich_sidereal_time (ASH_POSITION_STRUCT *pos)
{
  Ulong		time_in_seconds ;
  double	T, sidtime ;
  AS_TIME_STRUCT tmp_struct ;

  time_in_seconds = pos->GMT.hour * (60 * 60) +
		pos->GMT.minute * 60 +
		pos->GMT.seconds ;

  /* First calculate mean sidereal time at midnight */
  memcpy (&tmp_struct, &pos->GMT, sizeof(pos->GMT)) ;
  tmp_struct.hour = 0 ;
  tmp_struct.minute = 0 ;
  tmp_struct.seconds = 0 ;

  T = (AS_julian_day(&tmp_struct) - 2451545.0) / 36525.0 ;

  // Calc mean sid time (in seconds) (modified eqn 12.2/12.3)
  sidtime = 24110.54841 +
	    (8640184.812866 + (0.09310392 - 0.000006199948*T) * T) * T  ;

  /* Now add the offset for the time of day */
  sidtime += SID_TO_SOLAR_CONSTANT * time_in_seconds ;

  /* But we must now add a multiple of DAY_LENGTH to normalise the result */
  pos->Grnwch_SidTime = fmod (sidtime, (double)ASH_DAY_LENGTH) ;

  if (pos->Grnwch_SidTime < 0.0)
    pos->Grnwch_SidTime += (double)ASH_DAY_LENGTH ;

  pos->Grnwch_SidTime = (pos->Grnwch_SidTime / (double)ASH_DAY_LENGTH) * 2.0 * M_PI ;

  #ifdef ASTRO_DEBUG
  {
    char buf[100] ;

    printf ("Centuries from J2000.0 (T): %14.12f\n", T) ;
    printf ("Greenwich Sidereal Time: %s\n", AS_display_RA(buf, pos->Grnwch_SidTime)) ;
  }
  #endif
}

// AS_JD_JDE refer to p78 & p87
static void
AS_JD_JDE (ASH_POSITION_STRUCT *pos)
{
  double	t, dt ;

  pos->JD = AS_julian_day(&pos->GMT) ;

  // refer to p78

  if (pos->GMT.year >= 2000)
  {
    // t = (pos->GMT.year - 2000.0) / 100.0 ;
    t = (pos->JD - 2451545.0) / 36525.0 ;

    dt = 102.0 + 102.0*t + 25.3*t*t - 0.37 * (2100.0 - pos->GMT.year) ;
  }
  else
  {
    // kludge this back to 1980. Refer to table 10.A on p79

    dt = 50.5 + (pos->GMT.year - 1980) * (63.0 - 50.5) / 20.0 ;
  }

  pos->JDE = pos->JD + (dt / (24*60*60)) ;

  // refer to p87
  pos->T = (pos->JDE - 2451545.0) / 36525.0 ;
  pos->T2 = pos->T * pos->T ;
  pos->T3 = pos->T2 * pos->T ;
  pos->T4 = pos->T3 * pos->T ;

  #ifdef ASTRO_DEBUG
    printf ("JD=%12.10f, JDE=%12.10f, dt=%12.10f\n", pos->JD, pos->JDE, dt) ;
    printf ("T=%12.10f, T2=%12.10f, T3=%12.10f, T4=%12.10f\n",
	    pos->T, pos->T2, pos->T3, pos->T4) ;
  #endif
}

// AS_time_functions refer to p78 & p87
void
AS_time_functions (ASH_POSITION_STRUCT *pos)
{
  AS_greenwich_sidereal_time (pos) ;
  AS_JD_JDE (pos) ;
}

/*****************************************************************************
 * The following routine accounts for precession
 *
 * ON ENTRY: pos contains pre-calculated Julian Ephemeris Century referred to
 *		J2000.0
 *	     pos->obj_x contains location of object
 *
 * EXIT:     pos->apa_x = apparent location of object (accounting precession)
 */

// AS_precession: refer to p134
void
AS_precession (ASH_POSITION_STRUCT *pos)
{
  double a1, a2, a3 ;
  double A, B, C ;

  a1 = AS_deg_to_rad ((2306.218*pos->T + 0.30188*pos->T2 + 0.017998*pos->T3) / 3600.0) ;
  a2 = AS_deg_to_rad ((2306.218*pos->T + 1.09468*pos->T2 + 0.018203*pos->T3) / 3600.0) ;
  a3 = AS_deg_to_rad ((2004.3109*pos->T - 0.42665*pos->T2 -0.041833*pos->T3) / 3600.0) ;

  A = cos(pos->obj_dec) *
	sin(pos->obj_ra + a1) ;

  B = cos(a3) * cos(pos->obj_dec) * cos(pos->obj_ra + a1) -
	  sin(a3) * sin(pos->obj_dec) ;

  C = sin(a3) * cos(pos->obj_dec) * cos(pos->obj_ra + a1) +
	  cos(a3) * sin(pos->obj_dec) ;

  pos->apa_ra = atan2(A, B) + a2 ;

  if (pos->apa_ra < 0.0)
    pos->apa_ra += 2.0 * M_PI ;

  pos->apa_dec = asin(C) ;

  #ifdef ASTRO_DEBUG
    printf ("Precession: a1=%f%c, a2=%f%c, a3=%f%c\n",
	    AS_rad_to_deg(a1), DEG_CHAR,
	    AS_rad_to_deg(a2), DEG_CHAR,
	    AS_rad_to_deg(a3), DEG_CHAR) ;
    printf ("Precession: obj_ra=%f%c, obj_dec=%f%c\n",
	    AS_rad_to_deg(pos->obj_ra), DEG_CHAR,
	    AS_rad_to_deg(pos->obj_dec), DEG_CHAR) ;
    printf ("Precession: apa_ra=%f%c, apa_dec=%f%c\n",
	    AS_rad_to_deg(pos->apa_ra), DEG_CHAR,
	    AS_rad_to_deg(pos->apa_dec), DEG_CHAR) ;
  #endif
}

/*****************************************************************************
 * The following routine accounts for nutation and obliquity of the ecliptic
 *
 * Nutation is caused by the motion of Lunar orbital nodes. It has a period of
 * 18.6 years and cause the celestial poles to follow small ellipses with a
 * semimajor axis of 9.202 seconds. This is called the Constant of Nutation.
 *
 * ON ENTRY: pos contains pre-calculated Julian Ephemeris Century referred to
 *		J2000.0
 */

// AS_nutation: refer to pp143-147, p151
void
AS_nutation (ASH_POSITION_STRUCT *pos)
{
  double D, M, m, F, O, x ;
  double delta_psi, delta_epsilon ;
  const struct _table22A *p ;
  double da, dd ;

  D = AS_deg_to_rad (297.85036 + 445267.111480*pos->T - 0.0019142*pos->T2 + pos->T3/189474.0) ;
  M = AS_deg_to_rad (357.52772 +  35999.050340*pos->T - 0.0001603*pos->T2 - pos->T3/300000.0) ;
  m = AS_deg_to_rad (134.96298 + 477198.867398*pos->T + 0.0086972*pos->T2 + pos->T3/56250.0 ) ;
  F = AS_deg_to_rad ( 93.27191 + 483202.017538*pos->T - 0.0036825*pos->T2 + pos->T3/327270.0) ;
  O = AS_deg_to_rad (125.04452 -   1934.136261*pos->T + 0.0020708*pos->T2 + pos->T3/450000.0) ;

  delta_psi = delta_epsilon = 0.0 ;

  for (p=&Nutation[0] ; p->D != 0x7f ; p++)
  {
    x = 0 ;

    if (p->D)
      x = p->D * D ;

    if (p->M)
      x += p->M * M ;

    if (p->m)
      x += p->m * m ;

    if (p->F)
      x += p->F * F ;

    if (p->O)
      x += p->O * O ;

    delta_psi += (p->psi_const + p->psi_T*pos->T/10.0) / 10000.0 * sin(x) ;
    delta_epsilon += (p->eps_const + p->eps_T*pos->T/10.0) / 10000.0 * cos(x) ;
  }

  // Refer to p147
  pos->delta_psi = delta_psi / 3600.0 / 360.0 * 2.0 * M_PI ;
  pos->delta_epsilon = delta_epsilon / 3600.0 / 360.0 * 2.0 * M_PI ;
  pos->epsilon0 = AS_deg_to_rad ((84381.448 - 46.815*pos->T - 0.00059*pos->T2 + 0.001813*pos->T3) / 3600.0) ;
  pos->e = pos->epsilon0 + pos->delta_epsilon ;

  // refer to p151
  /* NOTE:
   * THE FOLLOWING IS INVALID IF STAR IS CLOSE TO CELESTIAL POLE
   */

  da = (cos(pos->e) + sin(pos->e)*sin(pos->obj_ra)*tan(pos->obj_dec)) * pos->delta_psi
	- (cos(pos->obj_ra)*tan(pos->obj_dec)) * pos->delta_epsilon ;

  dd = (sin(pos->e) * cos(pos->obj_ra)) * pos->delta_psi
	+ sin(pos->obj_ra) * pos->delta_epsilon ;

  pos->delta_ra += da  ;
  pos->delta_dec += dd ;

  #ifdef ASTRO_DEBUG
  {
    char buf[100], buf1[100] ;

    printf ("Nutation: T=%12.10f, T2=%12.10f, T3=%12.10f\n", pos->T, pos->T2, pos->T3) ;

    printf ("Nutation obj RA: %s, %.8f deg\n", 
	   AS_rad_to_HMS (buf, pos->obj_ra),
	   AS_rad_to_deg(pos->obj_ra)) ;

    printf ("Nutation obj DEC: %s, %.8f deg\n", 
	   AS_rad_to_DMS (buf, pos->obj_dec),
	   AS_rad_to_deg(pos->obj_dec)) ;

    printf ("Nutation: D=%f, M=%f, m=%f, F=%f, O=%f\n",
	  AS_rad_to_deg(D),
	  AS_rad_to_deg(M),
	  AS_rad_to_deg(m),
	  AS_rad_to_deg(F),
	  AS_rad_to_deg(O)) ;

    printf ("Nutation: delta_psi: %s, %f deg, %f rads\n", AS_rad_to_DMS (buf, pos->delta_psi), AS_rad_to_deg(pos->delta_psi), pos->delta_psi) ;
    printf ("Nutation: delta_eps: %s, %f deg, %f rads\n", AS_rad_to_DMS (buf, pos->delta_epsilon), AS_rad_to_deg(pos->delta_epsilon), pos->delta_epsilon) ;
    printf ("Nutation: epsilon0: %s, %f deg, %f rads\n", AS_rad_to_DMS (buf, pos->epsilon0), AS_rad_to_deg(pos->epsilon0), pos->epsilon0) ;
    printf ("Nutation: epsilon: %s, %f deg, %f rads\n", AS_rad_to_DMS (buf, pos->e), AS_rad_to_deg(pos->e), pos->e) ;

    printf("Nutation delta_ra: %s, %s, %.8f rad\n", 
	   AS_rad_to_HMS (buf, da),
	   AS_rad_to_DMS (buf1, da),
	   da) ;

    printf("Nutation delta_dec: %s, %.8f rad\n", 
	   AS_rad_to_DMS (buf, dd),
	   dd) ;
  }
  #endif
}

/*****************************************************************************
 * The following routine calculates the longitude of the sun
 *
 * ON ENTRY: pos contains pre-calculated Julian Ephemeris Century referred to
 *		J2000.0
 */

// refer to p163
void
AS_lon_sun (ASH_POSITION_STRUCT *pos)
{
  double L0, Ma, Cs ;

  Ma = AS_deg_to_rad(357.52911 + 35999.05029*pos->T + 0.0001537*pos->T2) ;
  L0 = 280.46646 + 36000.76983*pos->T + 0.0003032*pos->T2 ;

  Cs = (1.914602 - 0.004817*pos->T - 0.000014*pos->T2) * sin(Ma)
	+ (0.019993 - 0.000101*pos->T) * sin(2.0 * Ma)
	+ 0.000289 * sin(3.0 * Ma) ;

  pos->lon_sun = AS_deg_to_rad (L0 + Cs) ;

  #ifdef ASTRO_DEBUG
  {
    char buf[100] ;

    printf("Lon_sun: %s, %f deg, %f rads\n", AS_rad_to_DMS (buf, pos->lon_sun), AS_rad_to_deg(pos->lon_sun), pos->lon_sun) ;
  }
  #endif
}

/*****************************************************************************
 * The following routine calculates the effect of aberration.
 *
 * Aberration is due to the Earth's movement through space. It makes celestial
 * objects appear to be located in a different position than if the Earth was
 * at rest. Stars appear to be displaced around an ellipse of approximately
 * 20.49552 seconds.
 */

#define KAPPA (20.49552 / 3600.0 / 360.0 * 2.0 * M_PI) // constant of aberration (in RADIANS)

// refer to p151-2
void
AS_aberration (ASH_POSITION_STRUCT *pos)
{
  double da, dd ;

  AS_lon_sun (pos) ;

  pos->e = 0.016708634 - 0.000042037*pos->T - 0.0000001267*pos->T2 ;
  pos->pi = AS_deg_to_rad(102.93735 + 1.71946*pos->T + 0.00046*pos->T2) ;

  da = -(KAPPA)*((cos(pos->obj_ra)*cos(pos->lon_sun)*cos(pos->epsilon0) + sin(pos->obj_ra)*sin(pos->lon_sun)) / cos(pos->obj_dec))
	+ pos->e*(KAPPA)*((cos(pos->obj_ra)*cos(pos->pi)*cos(pos->epsilon0) + sin(pos->obj_ra)*sin(pos->pi)) / cos(pos->obj_dec)) ;

  dd = -(KAPPA)*(cos(pos->lon_sun)*cos(pos->epsilon0)*(tan(pos->epsilon0)*cos(pos->obj_dec) - sin(pos->obj_ra)*sin(pos->obj_dec)) + cos(pos->obj_ra)*sin(pos->obj_dec)*sin(pos->lon_sun))
    + pos->e*(KAPPA)*(cos(pos->pi)*cos(pos->epsilon0)*(tan(pos->epsilon0)*cos(pos->obj_dec) - sin(pos->obj_ra)*sin(pos->obj_dec)) + cos(pos->obj_ra)*sin(pos->obj_dec)*sin(pos->pi)) ;

  pos->delta_ra += da  ;
  pos->delta_dec += dd ;

  #ifdef ASTRO_DEBUG
  {
    char buf[100], buf1[100] ;

    printf("Aberration obj RA: %s, %.8f deg\n", 
	   AS_rad_to_HMS (buf, pos->obj_ra),
	   AS_rad_to_deg(pos->obj_ra)) ;

    printf("Aberration obj DEC: %s, %.8f deg\n", 
	   AS_rad_to_DMS (buf, pos->obj_dec),
	   AS_rad_to_deg(pos->obj_dec)) ;

    printf("Aberration: e=%.8f, pi=%.8f deg\n", pos->e, AS_rad_to_deg(pos->pi)) ;

    printf("Aberration delta_ra: %s, %s, %.8f rad\n", 
	   AS_rad_to_HMS (buf, da),
	   AS_rad_to_DMS (buf1, da),
	   AS_rad_to_deg(da)) ;

    printf("Aberration delta_dec: %s, %.8f rad\n", 
	   AS_rad_to_DMS (buf, dd),
	   AS_rad_to_deg(dd)) ;
  }
  #endif
}

/*****************************************************************************
 * This routine makes an allowance for atmospheric refaction on the apparent
 * location of an object.
 *
 * Refraction causes objects to appear to be higher in the sky than they really
 * are. Refraction is most apparent for objects near the horizon where it
 * appears to "lift" them by approximately 35 minutes (more than a moon
 * diameter) but tends towards negligible at elevations higher than 45 degrees.
 *
 * Upon input:
 *	pos->obj_ra/dec == position of object
 *	pos->delta_ra/dec == change due to precession and nutation
 *
 * Upon exit:
 *	pos->obj_ra/dec = position of object
 *	pos->apa_ra/dec = position of object INCLUDING precession and nutation
 *	pos->delta_ra/dec = change in apparent due to REFRACTION
 */

/* refer to p106 */

void
AS_allow_refraction (ASH_POSITION_STRUCT *pos)
{
  double deg, argument, refraction, SidTime, H, h ;
  double sin_lat, cos_lat ;

  pos->apa_ra = pos->obj_ra + pos->delta_ra ;
  pos->apa_dec = pos->obj_dec + pos->delta_dec ;
  pos->delta_ra = pos->delta_dec = 0.0 ;

  sin_lat = sin(AS_lattitude) ;
  cos_lat = cos(AS_lattitude) ;

  // p95, example 13.b
  SidTime = pos->Grnwch_SidTime + pos->delta_psi * cos(pos->e) ;

  // p92. H=local hour angle
  H = SidTime - AS_longitude - pos->apa_ra ;

  // p93 equations 13.5 and 13.6. h=local altitude
  pos->azimuth = atan2 (sin(H), cos(H)*sin_lat - tan(pos->apa_dec)*cos_lat) ;

  h = asin (sin_lat*sin(pos->apa_dec)
	    + cos_lat*cos(pos->apa_dec)*cos(H)) ;

  // p106. eq 16.4
  if (h >= -(35.0/60.0)/360.0*2.0*M_PI) // refract anything above -35'
  {
    deg = AS_rad_to_deg (h) ;
    argument = AS_deg_to_rad (deg + 10.3/(deg + 5.11)) ;
    refraction = AS_deg_to_rad (1.02 / tan (argument) / 60.0) ;
  }
  else
    refraction = 0.0 ;

  pos->altitude = h + refraction ;

  if (pos->altitude >= M_PI) // show negatives when below the horizon
    pos->altitude -= 2.0 * M_PI ;

  // calculate delta based on eq 13.5 and 13.6, p93
  pos->delta_ra = (SidTime - AS_longitude - atan2 (sin(pos->azimuth), cos(pos->azimuth)*sin_lat + tan(pos->altitude)*cos_lat)) - pos->apa_ra ;
  pos->delta_dec = asin (sin_lat*sin(pos->altitude) - cos_lat*cos(pos->altitude)*cos(pos->azimuth)) - pos->apa_dec ;

  if (pos->delta_ra > M_PI)
    pos->delta_ra -= 2 * M_PI ;

  if (pos->delta_ra < -M_PI)
    pos->delta_ra += 2 * M_PI ;

  if (pos->delta_dec > M_PI)
    pos->delta_dec -= 2 * M_PI ;

  if (pos->delta_dec < -M_PI)
    pos->delta_dec += 2 * M_PI ;

  #ifdef ASTRO_DEBUG
  {
    char buf[100] ;

    if (H < 0.0)
      H += 2 * M_PI ;

    printf ("Refraction Greenwich Sidereal Time: %s\n", AS_display_RA(buf, SidTime)) ;

    printf ("Refraction Actual RA: %s, %.8f deg, %f rad\n",
	    AS_rad_to_HMS (buf, pos->obj_ra),
	    AS_rad_to_deg (pos->obj_ra),
	    pos->obj_ra) ;

    printf ("Refraction Actual DEC: %s, %.8f deg, %f rad\n",
	    AS_rad_to_DMS (buf, pos->obj_dec),
	    AS_rad_to_deg (pos->obj_dec),
	    pos->obj_dec) ;

    printf ("Refraction Apparent RA: %s, %.8f deg, %f rad\n",
	    AS_rad_to_HMS (buf, pos->apa_ra),
	    AS_rad_to_deg (pos->apa_ra),
	    pos->apa_ra) ;

    printf ("Refraction Apparent DEC: %s, %.8f deg, %f rad\n",
	    AS_rad_to_DMS (buf, pos->apa_dec),
	    AS_rad_to_deg (pos->apa_dec),
	    pos->apa_dec) ;

    printf ("Refraction Hour Angle: %s, %.8f deg, %f rad\n",
	    AS_rad_to_HMS (buf, (H)),
	    AS_rad_to_deg (H),
	    H) ;

    printf ("Refraction Raw ALT: %s, %.8f deg, %f rad\n",
	    AS_rad_to_DMS (buf, (h)),
	    AS_rad_to_deg (h),
	    h) ;

    printf ("Refraction Amount: %s, %.8f deg, %f rad\n",
	    AS_rad_to_DMS (buf, refraction),
	    AS_rad_to_deg (refraction),
	    refraction) ;

    printf ("Refraction ALT+refraction: %s, %.8f deg, %f rad\n",
	    AS_rad_to_DMS (buf, (pos->altitude)),
	    AS_rad_to_deg (pos->altitude),
	    pos->altitude) ;

    printf ("Refraction AZ: %s, %.8f deg, %f rad\n",
	    AS_rad_to_HMS (buf, (pos->azimuth)),
	    AS_rad_to_deg (pos->azimuth),
	    pos->azimuth) ;

    printf ("Refraction Refracted RA: %s, %.8f deg, %f rad\n",
	    AS_rad_to_HMS (buf, pos->apa_ra+pos->delta_ra),
	    AS_rad_to_deg (pos->apa_ra+pos->delta_ra),
	    pos->apa_ra+pos->delta_ra) ;

    printf ("Refraction Refracted DEC: %s, %.8f deg, %f rad\n",
	    AS_rad_to_DMS (buf, pos->apa_dec+pos->delta_dec),
	    AS_rad_to_deg (pos->apa_dec+pos->delta_dec),
	    pos->apa_dec+pos->delta_dec) ;

    printf ("Refraction Delta RA: %s, %.8f deg, %f rad\n",
	    AS_rad_to_HMS (buf, pos->delta_ra),
	    AS_rad_to_deg (pos->delta_ra),
	    pos->delta_ra) ;

    printf ("Refraction Delta DEC: %s, %.8f deg, %f rad\n",
	    AS_rad_to_DMS (buf, pos->delta_dec),
	    AS_rad_to_deg (pos->delta_dec),
	    pos->delta_dec) ;
  }
  #endif
}

/*****************************************************************************
 * The following routine is an example of a calculation of the aparant
 * position of an object.
 *
 * ON ENTRY
 *	pos initialised and the locale fields have been set
 *	pos contains the time and date to use in calculating the position
 */

void
AS_find_apparent_location (ASH_POSITION_STRUCT *pos)
{
  pos->delta_dec = pos->delta_ra = 0.0 ;
  AS_time_functions (pos) ;
  AS_nutation (pos) ;
  AS_aberration (pos) ;
  AS_allow_refraction (pos) ;
}

/*****************************************************************************
 * The following routine calculates the position of the moon
 *
 * On entry, the pos struct should contain only the time of interest
 */

// refer to p338, p342
void
AS_find_moon_apparent_location (ASH_POSITION_STRUCT *pos)
{
  double D, M, m, F, l, A1, A2, A3 ;
  double E, E2, x ;
  double sigma_l, sigma_r, sigma_b ;
  double lambda, beta ;
  const struct Table47A *p ;
  const struct Table47B *q ;

  AS_time_functions (pos) ; // calculate T

  l = AS_normalise_degrees (218.3164477 + 481267.88123421* pos->T - 0.0015786 * pos->T2 + pos->T3/538841.0  - pos->T4/65194000.0) ;
  D = AS_normalise_degrees (297.8501921 + 445267.1114034 * pos->T - 0.0018819 * pos->T2 + pos->T3/545868.0  - pos->T4/113065000.0) ;
  M = AS_normalise_degrees (357.5291092 +  35999.0502909 * pos->T - 0.0001536 * pos->T2 + pos->T3/24490000.0) ;
  m = AS_normalise_degrees (134.9633964 + 477198.8675055 * pos->T + 0.0087414 * pos->T2 + pos->T3/69699.0   - pos->T4/14712000.0) ;
  F = AS_normalise_degrees ( 93.272095  + 483202.0175233 * pos->T - 0.0036539 * pos->T2 - pos->T3/3526000.0 + pos->T4/863310000.0) ;

  A1 = AS_normalise_degrees (119.75 +    131.849 * pos->T) ;
  A2 = AS_normalise_degrees ( 53.09 + 479264.290 * pos->T) ;
  A3 = AS_normalise_degrees (313.45 + 481266.484 * pos->T) ;

  E = 1.0 - 0.002516*pos->T - 0.0000074*pos->T2 ;
  E2 = E * E ;
  sigma_l = sigma_r = sigma_b = 0.0 ;

  // Firstly, calculate the longitude and distance of the moon
  for (p=&Lon_And_Distance[0] ; p->D != 0x7f ; p++)
  {
    x = 0.0 ;

    if (p->D)
      x = p->D * D ;

    if (p->M)
      x += p->M * M ;

    if (p->m)
      x += p->m * m ;

    if (p->F)
      x += p->F * F ;

    x = AS_deg_to_rad(x) ;

    if (p->M == 0)
    {
      sigma_l += p->sigma_l * sin(x) ;
      sigma_r += p->sigma_r * cos(x) ;
    }
    else if ((p->M == 1) || (p->M == -1))
    {
      sigma_l += p->sigma_l * E * sin(x) ;
      sigma_r += p->sigma_r * E * cos(x) ;
    }
    else
    {
      sigma_l += p->sigma_l * E2 * sin(x) ;
      sigma_r += p->sigma_r * E2 * cos(x) ;
    }
  }

  // Now, calculate the latitude of the moon
  for (q=&Lat[0] ; q->D != 0x7f ; q++)
  {
    x = 0.0 ;

    if (q->D)
      x = q->D * D ;

    if (q->M)
      x += q->M * M ;

    if (q->m)
      x += q->m * m ;

    if (q->F)
      x += q->F * F ;

    x = sin(AS_deg_to_rad(x)) ;

    if (q->M == 0)
      sigma_b += q->sigma_b * x ;

    else if ((q->M == 1) || (q->M == -1))
      sigma_b += q->sigma_b * E * x ;

    else
      sigma_b += q->sigma_b * E2 * x ;
  }

  // add the final terms

  sigma_l += 3958.0 * sin(AS_deg_to_rad(A1))
	   + 1962.0 * sin(AS_deg_to_rad(l-F))
	   +  318.0 * sin(AS_deg_to_rad(A2)) ;

  sigma_b += -2235.0 * sin(AS_deg_to_rad(l))
	   +   382.0 * sin(AS_deg_to_rad(A3))
	   +   175.0 * (sin(AS_deg_to_rad(A1-F)) + sin(AS_deg_to_rad(A1+F)))
	   +   127.0 * sin(AS_deg_to_rad(l-m))
	   -   115.0 * sin(AS_deg_to_rad(l+m)) ;

  // Now calculate the geocentric coordinates of the moon
  lambda = AS_deg_to_rad (l + sigma_l / 1000000.0) ;
  beta   = AS_signed_deg_to_rad (sigma_b / 1000000.0) ;

  // Allow Nutation
  pos->delta_dec = pos->delta_ra = 0.0 ;
  AS_nutation (pos) ;

  // Convert coordinates to aparent RA and DEC
  lambda += pos->delta_psi ;

  // Now calculate apparent RA and dec using eq 13.3 and 13.4 (p93)
  pos->obj_ra  = atan2 (sin(lambda)*cos(pos->e) - tan(beta)*sin(pos->e), cos(lambda)) ;
  pos->obj_dec = asin  (sin(beta)*cos(pos->e) + cos(beta)*sin(pos->e)*sin(lambda)) ;
  pos->apa_distance = (385000.56 + sigma_r / 1000.0) / 149597870.0 ;

  // Allow refraction
  pos->delta_dec = pos->delta_ra = 0.0 ;
  AS_allow_refraction (pos) ;

  #ifdef ASTRO_DEBUG
  {
    char buf[100] ;

    printf ("Moon: T=%12.10f, T2=%12.10f, T3=%12.10f T4=%12.10f\n", pos->T, pos->T2, pos->T3, pos->T4) ;
    printf ("Moon: l=%f, D=%f, M=%f, m=%f, F=%f\n", l, D, M, m, F) ;
    printf ("Moon: A1=%f, A2=%f, A3=%f\n", A1, A2, A3) ;
    printf ("Moon: E=%f, E2=%f\n", E, E2) ;
    printf ("Moon: sigma_l=%f, sigma_b=%f, sigma_r=%f\n", sigma_l, sigma_b, sigma_r) ;

    printf ("Moon apparent lambda: %s, %.8f deg\n", 
	   AS_rad_to_DMS (buf, lambda),
	   AS_rad_to_deg (lambda)) ;

    printf ("Moon epsilon: %s, %.8f deg\n", 
	   AS_rad_to_DMS (buf, pos->e),
	   AS_rad_to_deg (pos->e)) ;

    printf ("Moon beta: %s, %.8f deg\n", 
	   AS_rad_to_DMS (buf, beta),
	   AS_rad_to_deg (beta)) ;

    printf ("Moon Actual RA: %s, %.8f deg, %f rad\n",
	    AS_rad_to_HMS (buf, pos->obj_ra),
	    AS_rad_to_deg (pos->obj_ra),
	    pos->obj_ra) ;

    printf ("Moon Actual DEC: %s, %.8f deg, %f rad\n",
	    AS_rad_to_DMS (buf, pos->obj_dec),
	    AS_rad_to_deg (pos->obj_dec),
	    pos->obj_dec) ;

    printf ("Moon Apparent RA: %s, %.8f deg, %f rad\n",
	    AS_rad_to_HMS (buf, pos->apa_ra),
	    AS_rad_to_deg (pos->apa_ra),
	    pos->apa_ra) ;

    printf ("Moon Apparent DEC: %s, %.8f deg, %f rad\n",
	    AS_rad_to_DMS (buf, pos->apa_dec),
	    AS_rad_to_deg (pos->apa_dec),
	    pos->apa_dec) ;

    printf ("Moon Apparent Distance: %.10f A.U., %.1f km\n",
	    pos->apa_distance,
	    pos->apa_distance * 149597870.0) ;
  }
  #endif
}

/*****************************************************************************
 * The following routine calculates the position of the planets
 *
 * On entry, the pos struct should contain only the time of interest
 */

// refer to p218
static double
AS_calculate_VSOP_series (const char *name, const VSOP_PTR *p, const double t)
{
  double result, element ;
  VSOP_PTR q ;
  Ushort i, j ;
  char series = 0 ;

  result = 0.0 ;

  for (i=0 ; i<6 ; i++,p++)
  {
    if (*p == 0)
      break ;

    for (q=*p, element=0.0 ; q->A >= 0.0 ; q++)
      element += q->A * cos(q->B + q->C*t) ;

    series = -(char)q->A ;

    #ifdef ASTRO_DEBUG
      printf ("VSOP %s: %c%d = %.10f\n", name, series, i, element) ;
    #endif

    for (j=i ; j ; j--)
      element *= t ;

    result += element ;
  }

  result = AS_normalise_rads (result) ;

  if ((series == 'B') && (result > M_PI))
    result -= 2.0 * M_PI ;

  #ifdef ASTRO_DEBUG
  {
    char buf[100] ;

    if (series != 'R')
      printf ("VSOP %s %c: %.10f rad, %s, %f deg\n",
	   name, series, result,
	   AS_rad_to_DMS (buf, result),
	   AS_rad_to_deg(result)) ;
    else
      printf ("VSOP %s %c: %.10f A.U.\n",
	   name, series, result) ;
  }
  #endif

  return result ;
}

// refer to Chapters 32 and 33
void
AS_find_planet_apparent_location (ASH_POSITION_STRUCT *pos, const Uchar planet)
{
  double L, B, R, L0, B0, R0, Le ;
  double x, y, z ;
  double t, T ;
  double lambda, beta ;
  double delta_lambda, delta_beta ;

  #ifdef ASTRO_DEBUG
    if (planet >= NEPTUNE)
    {
      fprintf (stderr, "Called AS_find_planet_apparent_location with illegal planet number\n") ;
      return ;
    }

    if (planet == EARTH)
    {
      fprintf (stderr, "Called AS_find_planet_apparent_location with Earth as parameter!\n") ;
      return ;
    }
  #endif

  // The calculation is performed in two parts. Firstly, we calculate the position
  // of the planet and of Earth at time pos->T. However, due to the speed of light,
  // the location of the planet as the light left it will be different. Thus we need
  // to find the location of the planet at time pos->T - (time for light to reach E)
  // A second calculation takes place once we know how distant the planet is.

  AS_time_functions (pos) ; // calculate T
  t = pos->T / 10.0 ;

  #ifdef ASTRO_DEBUG
    printf ("----------Pass 1----------\n") ;
    printf ("Planet %s: t=%.10f\n", Planets[planet].name, t) ;
  #endif

  L = AS_calculate_VSOP_series (Planets[planet].name, Planets[planet].L, t) ;
  B = AS_calculate_VSOP_series (Planets[planet].name, Planets[planet].B, t) ;
  R = AS_calculate_VSOP_series (Planets[planet].name, Planets[planet].R, t) ;

  Le = L0 = AS_calculate_VSOP_series (Planets[EARTH].name, Planets[EARTH].L, t) ;
  B0 = AS_calculate_VSOP_series (Planets[EARTH].name, Planets[EARTH].B, t) ;
  R0 = AS_calculate_VSOP_series (Planets[EARTH].name, Planets[EARTH].R, t) ;

  // Now convert L and B to the FK5 system p219
  x = AS_deg_to_rad(1.397)*pos->T + AS_deg_to_rad(0.00031)*pos->T2 ;

  L += AS_deg_to_rad(-0.09033/3600.0) + AS_deg_to_rad(0.03916/3600.0)*tan(B)*(cos(L-x) + sin(L-x)) ;
  B += AS_deg_to_rad(0.03916/3600.0) * (cos(L-x) - sin(L-x)) ;

  L0 += AS_deg_to_rad(-0.09033/3600.0) + AS_deg_to_rad(0.03916/3600.0)*tan(B0)*(cos(L0-x) + sin(L0-x)) ;
  B0 += AS_deg_to_rad(0.03916/3600.0) * (cos(L0-x) - sin(L0-x)) ;

  // Convert coordinates to aparent RA and DEC p223
  x = R * cos(B) * cos(L) - R0 * cos(B0) * cos(L0) ;
  y = R * cos(B) * sin(L) - R0 * cos(B0) * sin(L0) ;
  z = R * sin(B)	  - R0 * sin(B0) ;

  pos->apa_distance = sqrt (x*x + y*y + z*z) ;

  #ifdef ASTRO_DEBUG
    printf ("Planet %s: x=%f, y=%f, z=%f\n", Planets[planet].name, x, y, z) ;
    printf ("Planet %s Apparent distance: %.12f A.U.\n",
	    Planets[planet].name,
	    pos->apa_distance) ;
  #endif

  // Secondly, account for the speed of light and recalculate
  T = ((pos->T * 36525.0) - pos->apa_distance*0.0057755183) / 36525.0 ;
  t = T / 10.0 ;

  #ifdef ASTRO_DEBUG
    printf ("----------Pass 2----------\n") ;
    printf ("Planet %s: t=%.10f\n", Planets[planet].name, t) ;
  #endif

  L = AS_calculate_VSOP_series (Planets[planet].name, Planets[planet].L, t) ;
  B = AS_calculate_VSOP_series (Planets[planet].name, Planets[planet].B, t) ;
  R = AS_calculate_VSOP_series (Planets[planet].name, Planets[planet].R, t) ;

  // Now convert L and B to the FK5 system p219
  x = AS_deg_to_rad(1.397)*T + AS_deg_to_rad(0.00031)*T*T ;

  L += AS_deg_to_rad(-0.09033/3600.0) + AS_deg_to_rad(0.03916/3600.0)*tan(B)*(cos(L-x) + sin(L-x)) ;
  B += AS_deg_to_rad(0.03916/3600.0) * (cos(L-x) - sin(L-x)) ;

  L0 += AS_deg_to_rad(-0.09033/3600.0) + AS_deg_to_rad(0.03916/3600.0)*tan(B0)*(cos(L0-x) + sin(L0-x)) ;
  B0 += AS_deg_to_rad(0.03916/3600.0) * (cos(L0-x) - sin(L0-x)) ;

  // Convert coordinates to aparent RA and DEC p223
  x = R * cos(B) * cos(L) - R0 * cos(B0) * cos(L0) ;
  y = R * cos(B) * sin(L) - R0 * cos(B0) * sin(L0) ;
  z = R * sin(B)          - R0 * sin(B0) ;

  lambda = atan2 (y, x) ;
  beta = atan2 (z, sqrt(x*x + y*y)) ;
  pos->apa_distance = sqrt (x*x + y*y + z*z) ;

  if (lambda < 0.0)
    lambda += M_PI * 2.0 ;

  #ifdef ASTRO_DEBUG
  {
    char buf[100] ;

    printf ("Planet %s: L=%s\n", Planets[planet].name, AS_rad_to_DMS(buf, L)) ;
    printf ("Planet %s: B=%s\n", Planets[planet].name, AS_rad_to_DMS(buf, B)) ;
    printf ("Planet %s: R=%.8f A.U.\n", Planets[planet].name, R) ;

    printf ("Planet %s: x=%f, y=%f, z=%f\n", Planets[planet].name, x, y, z) ;

    printf ("Planet %s Apparent RA: %s, %.8f deg, %f rad\n",
	    Planets[planet].name,
	    AS_rad_to_HMS (buf, lambda),
	    AS_rad_to_deg (lambda),
	    lambda) ;

    printf ("Planet %s Apparent DEC: %s, %.8f deg, %f rad\n",
	    Planets[planet].name,
	    AS_rad_to_DMS (buf, beta),
	    AS_rad_to_deg (beta),
	    beta) ;

    printf ("Planet %s Apparent distance: %.12f A.U.\n",
	    Planets[planet].name,
	    pos->apa_distance) ;
  }
  #endif

  // Now correct coordinates for aberration. p151 with input as per p226
  pos->e = 0.016708634 - 0.000042037*pos->T - 0.0000001267*pos->T2 ;
  pos->pi = AS_deg_to_rad(102.93735 + 1.71946*pos->T + 0.00046*pos->T2) ;
  z = Le + M_PI ;

  delta_lambda  = (KAPPA)*(-cos(z-lambda) + pos->e*cos(pos->pi-lambda)) / cos(beta) ;
  delta_beta = -(KAPPA)*sin(beta) * (sin(z-lambda) - pos->e*sin(pos->pi-lambda)) ;

  if (delta_lambda > M_PI)
    delta_lambda -= M_PI * 2.0 ;

  if (delta_beta > M_PI)
    delta_beta -= M_PI * 2.0 ;

  #ifdef ASTRO_DEBUG
  {
    char buf[100], buf1[100] ;

    printf("Planet aberration: e=%.8f, pi=%.8f deg\n", pos->e, AS_rad_to_deg(pos->pi)) ;
    printf("Planet lon_sun: %.8f rad, %.8f deg\n", z, AS_rad_to_deg(z)) ;

    printf("Planet aberration delta_ra: %s, %s, %.8f rad, %.8f deg\n", 
	   AS_rad_to_HMS (buf, delta_lambda),
	   AS_rad_to_DMS (buf1, delta_lambda),
	   delta_lambda, AS_rad_to_deg(delta_lambda)) ;

    printf("Planet aberration delta_dec: %s, %.8f rad, %.8f deg\n", 
	   AS_rad_to_DMS (buf, delta_beta),
	   delta_beta, AS_rad_to_deg(delta_beta)) ;

    printf("Planet aberration obj RA: %s, %.8f deg\n", 
	   AS_rad_to_HMS (buf, lambda+delta_lambda),
	   AS_rad_to_deg(lambda+delta_lambda)) ;

    printf("Planet aberration obj DEC: %s, %.8f deg\n", 
	   AS_rad_to_DMS (buf, beta+delta_beta),
	   AS_rad_to_deg(beta+delta_beta)) ;

  }
  #endif

  // Now correct the aparent values for the FK5 system
  x = AS_deg_to_rad(1.397)*pos->T + AS_deg_to_rad(0.00031)*pos->T2 ;

  y = AS_deg_to_rad(-0.09033/3600.0) + AS_deg_to_rad(0.03916/3600.0)*tan(beta+delta_beta)*(cos(lambda+delta_lambda-x) + sin(lambda+delta_lambda-x)) ;
  z = AS_deg_to_rad(0.03916/3600.0) * (cos(lambda+delta_lambda-x) - sin(lambda+delta_lambda-x)) ;

  if (y > M_PI)
    y -= M_PI * 2.0 ;

  if (z > M_PI)
    z -= M_PI * 2.0 ;

  delta_lambda += y ;
  delta_beta += z ;

  #ifdef ASTRO_DEBUG
  {
    char buf[100], buf1[100] ;

    printf("Planet FK5 corrections to ra: %s, %s, %.8f rad\n", 
	   AS_rad_to_HMS (buf, y),
	   AS_rad_to_DMS (buf1, y),
	   x) ;

    printf("Planet FK5 corrections to dec: %s, %.8f rad\n", 
	   AS_rad_to_DMS (buf, z),
	   z) ;

    printf("Planet FK5 corrected apa RA: %s, %.8f deg\n", 
	   AS_rad_to_HMS (buf, lambda+delta_lambda),
	   AS_rad_to_deg(lambda+delta_lambda)) ;

    printf("Planet FK5 corrected apa DEC: %s, %.8f deg\n", 
	   AS_rad_to_DMS (buf, beta+delta_beta),
	   AS_rad_to_deg(beta+delta_beta)) ;
  }
  #endif

  // Allow nutation
  pos->obj_ra = lambda ;
  pos->obj_dec = beta ;
  AS_nutation (pos) ;
  lambda = pos->obj_ra ;
  beta = pos->obj_dec ;

  // Transformation from ecliptical to equatorial coordinates - p93
  pos->obj_ra = atan2 (sin(lambda+delta_lambda)*cos(pos->e) - tan(beta+delta_beta)*sin(pos->e), cos(lambda+delta_lambda)) ;
  pos->obj_dec = asin (sin(beta+delta_beta)*cos(pos->e) + cos(beta+delta_beta)*sin(pos->e)*sin(lambda+delta_lambda)) ;

  if (pos->obj_ra < 0.0)
    pos->obj_ra += M_PI * 2.0 ;

  // Allow refraction
  AS_allow_refraction (pos) ;

  #ifdef ASTRO_DEBUG
  {
    char buf[100], buf1[100] ;

    printf("Planet ra corrected for nutation: %s, %s, %f deg, %.8f rad\n", 
	   AS_rad_to_HMS (buf, pos->apa_ra + pos->delta_ra),
	   AS_rad_to_DMS (buf1, pos->apa_ra + pos->delta_ra),
	   AS_rad_to_deg (pos->apa_ra + pos->delta_ra),
	   pos->apa_ra + pos->delta_ra) ;

    printf("Planet Actual RA: %s, %.8f deg\n", 
	   AS_rad_to_HMS (buf, pos->apa_ra+pos->delta_ra),
	   AS_rad_to_deg(pos->apa_ra+pos->delta_ra)) ;

    printf("Planet Actual DEC: %s, %.8f deg\n", 
	   AS_rad_to_DMS (buf, pos->apa_dec+pos->delta_dec),
	   AS_rad_to_deg(pos->apa_dec+pos->delta_dec)) ;

    printf("Planet Apparent RA: %s, %.8f deg\n", 
	   AS_rad_to_HMS (buf, pos->apa_ra),
	   AS_rad_to_deg(pos->apa_ra+pos->delta_ra)) ;

    printf("Planet Apparent DEC: %s, %.8f deg\n", 
	   AS_rad_to_DMS (buf, pos->apa_dec),
	   AS_rad_to_deg(pos->apa_dec+pos->delta_dec)) ;
  }
  #endif
}
