/*
 ******************************************************************************
 * sort.c
 ******************************************************************************
 * Copyright (C) 2002 Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ******************************************************************************
 */

#include <stdlib.h>
#include <stdio.h>

#include "types.h"
#include "sort.h"

/******************************************************************************
 * Modify the following comparison funciton to compare two entries in your
 * array
 */

int
SOR_cmpint(void *a, void *b)
{
  if ((int)a > (int)b)
    return 1 ;

  else if ((int)a < (int)b)
    return -1;

  else
    return 0;
}
 
/******************************************************************************
 * SOR_qsort
 *
 * Sorts a user supplied array of data.
 *
 * I obtained this routine from http://www.concentric.net/~Ttwang/sort/sort.htm
 * (Thomas Wang). It appears below virtually unmodified from the original
 * source code.
 */

void
SOR_qsort (void *source_array[], COMPARE compare, Ulong first, Ulong last)
{
  Ulong stack_pointer = 0;
  int first_stack[32];
  int last_stack[32];

  for (;;)
  {
    if (last - first <= INSERTION_SORT_LIMIT)
    {
      /* for small sort, use insertion sort */
      Ulong indx ;
      void *prev_val ;
      void *cur_val ;

      prev_val = source_array[first] ;

      for (indx = first + 1; indx <= last; ++indx)
      {
        cur_val = source_array[indx];

        if ((*compare)(prev_val, cur_val) > 0)
        {
          /* out of order: array[indx-1] > array[indx] */
          Ulong indx2;
          source_array[indx] = prev_val; /* move up the larger item first */

          /* find the insertion point for the smaller item */
          for (indx2 = indx - 1; indx2 > first; )
          {
            void *temp_val = source_array[indx2 - 1];

            if ((*compare)(temp_val, cur_val) > 0)
            {
              source_array[indx2--] = temp_val;
              /* still out of order, move up 1 slot to make room */
            }
            else
              break;
          }
          source_array[indx2] = cur_val; /* insert the smaller item right here */
        }
        else
        {
          /* in order, advance to next element */
          prev_val = cur_val;
        }
      }
    }
    else
    {
      void *pivot;
 
      /* try quick sort */
      {
        void *temp ;
        Ulong med = (first + last) >> 1 ;

        /* Choose pivot from first, last, and median position. */
        /* Sort the three elements. */

        temp = source_array[first] ;

        if ((*compare)(temp, source_array[last]) > 0)
	{
          source_array[first] = source_array[last] ;
	  source_array[last] = temp ;
	}

        temp = source_array[med];

        if ((*compare)(source_array[first], temp) > 0)
	{
          source_array[med] = source_array[first] ;
	  source_array[first] = temp ;
	}

        temp = source_array[last];

        if ((*compare)(source_array[med], temp) > 0)
	{
          source_array[last] = source_array[med] ;
	  source_array[med] = temp ;
	}

        pivot = source_array[med] ;
      }
      {
        Ulong up ;
        {
	  Ulong down ;

          /* First and last element will be loop stopper. */
	  /* Split array into two partitions. */

	  down = first ;
	  up = last ;

	  for (;;)
	  {
	    do
	    {
	      ++down;
	    } while ((*compare)(pivot, source_array[down]) > 0) ;
 
	    do
	    {
	      --up;
	    } while ((*compare)(source_array[up], pivot) > 0) ;
 
	    if (up > down)
	    {
	      void *temp ;
	      /* interchange L[down] and L[up] */
	      temp = source_array[down]; source_array[down]= source_array[up]; source_array[up] = temp;
	    }
	    else
	      break;
	  }
	}
	{
	  Ulong len1 ; /* length of first segment */
	  Ulong len2 ; /* length of second segment */

	  len1 = up - first + 1 ;
	  len2 = last - up ;

	  /* stack the partition that is larger */

	  if (len1 >= len2)
	  {
	    first_stack[stack_pointer] = first ;
	    last_stack[stack_pointer++] = up ;
 
	    first = up + 1 ;
	    /*  tail recursion elimination of
	     *  Qsort(source_array,compare,up + 1,last)
	     */
	  }
	  else
	  {
	    first_stack[stack_pointer] = up + 1 ;
	    last_stack[stack_pointer++] = last ;

	    last = up ;
	    /* tail recursion elimination of
	     * Qsort(source_array,compare,first,up)
	     */
	  }
	}
        continue;
      }
      /* end of quick sort */
    }

    if (stack_pointer > 0)
    {
      /* Sort segment from stack. */
      first = first_stack[--stack_pointer];
      last = last_stack[stack_pointer];
    }
    else
      break;
  } /* end for */
}

/******************************************************************************
 * SOR_search
 *
 * Searches a sorted array for a particular element. The index of the matching
 * element is returned.
 *
 * The algorithm is the Binary Search.
 *
 * Note that the LENGTH parameter is one larger than the largest element index
 */

Ulong
SOR_search (void *source_array[], COMPARE compare, Ulong length, void *key)
{
  Ulong lo, hi ;
  int result ;

  if (length == 0)
    return SOR_FAIL ;

  for (lo=0, hi=length-1 ; lo != hi ;)
  {
    result = (*compare)(source_array [(lo+hi)/2], key) ;

    if (result < 0)
    {
      /* The current half way point is BEFORE the key */
 
      if (lo == (hi-1)) /* check to prevent a rounding to lo loop */
	lo = hi ;
      else
	lo = (lo + hi) / 2 ;
    }
    else if (result > 0)
    {
      /* The current half way point is AFTER the key */
 
      hi = (lo + hi) / 2 ; /* This automatically rounds down */
    }
    else
      return (lo + hi) / 2 ;
  }

  /* We can only be here because lo == hi */

  result = (*compare)(source_array[lo], key) ;

  if (result != 0)
    return SOR_FAIL ;

  return lo ;
}
