/*
 * parse.c
 ******************************************************************************
 * Copyright (C) 2002  Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. http://www.gnu.org
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>           /* for calls to getenv, ... */
#include <stdarg.h>           /* for calls to vsnprintf etc */
#include <string.h>           /* for strncpy, memcpy, memset, ... */
#include <ctype.h>            /* for toupper, ... */
#include <unistd.h>           /* for getcwd, chdir, getuid, geteuid ... */
#include <stdarg.h>           /* for calls to vsnprintf, ... */
#include <fcntl.h>            /* static defs for open, creat, stat */
#include <sys/stat.h>
#include <sys/types.h>

#include "types.h"
#include "stringfn.h"
#include "parse.h"

int	Copyleft ;
int	Verbose ;
int	Error ;
int	Help ;

/******************************************************************************
 * PA_parse - parses the command line parameters
 */

#define BAD_ERROR 100

int
PA_parse (int *argc, char ***argv)
{
  char	*p, *q, *r ;
  int	error ;
  OPTION *option ;
  char	**new_argv ;
  int	count, i ;

  /*
   * set defaults
   */

#ifdef PARSE_DEBUG
  printf ("PA_parse called with the following options:\n") ;
  PA_dump_options (*argv) ;
#endif

  error = 0 ;

  for (option = &opt[0] ; *(option->op_str) ; option++)
  {
    switch (option->type)
    {
      case BOOLEAN:
	*(int *)(option->var) = 0 ;
	break ;

      case INT:
	*(int *)(option->var) = option->default_int ;
	break ;

      case STRING:
      case FILENAME:
	if ((option->default_string != NULL) && (option->default_int != 0))
	  strncpy (option->var, option->default_string, option->max) ;
	else
	  *(char *)(option->var) = '\0' ;
	break ;
    }
  }

  new_argv = malloc ((*argc+2) * sizeof (char **)) ; /* leave space for NULL */
  i = 0 ;

  if (new_argv == NULL)
  {
    fprintf (stderr, "PA_parse: Cannot malloc\n") ;
    return 100 ;
  }

  new_argv[i++] = *argv[0] ; /* set the programme name */

  /*
   * parse the command line options
   */

  count = *argc ;

  while ((error < BAD_ERROR) && (--count > 0))
  {
    if (*(p=*++(*argv)) == '-')
    {
      for (p++ ; (error < BAD_ERROR) && *p ;) /* scan along cluster of parameters */
      {
	for (option=&opt[0] ; *p && *(option->op_str) ; option++) /* check legal options */
	{
	  for (q=p, r=option->op_str ; *r && (*q == *r) ; q++, r++) ; /* string comparison */

	  if (*r == '\0') /* matched the string */
	  {
	    p = q ; 

	    switch (option->type)
	    {
	      case BOOLEAN:
		*(int *)(option->var) = 1 ;

#               ifdef PARSE_DEBUG
		  printf ("Parsed boolean (-%s)\n", option->op_str) ;
#               endif

	        break ;

	      case INT:
		if (*p == '\0')
	        {
		  if (--count > 0)
		    p = *++(*argv) ;

		}

		if (!isdigit(*p))
		{
		  *(int *)(option->var) = option->default_int ;
		  break ;
		}

		if ((sscanf(p, "%d", (int *)(option->var)) != 1) || (*(int *)(option->var) < option->min) || (*(int *)(option->var) > option->max))
		{
		  fprintf (stderr, "'-%s' parameter (%d) is outside the range %d to %d\n\n", option->op_str, *(int *)(option->var), option->min, option->max) ;
		  *(int *)(option->var) = option->default_int ;
		  error++ ;
		}

		p = "" ;

#               ifdef PARSE_DEBUG
		  printf ("Parsed integer (-%s). Value=%d\n", option->op_str, *(int *)(option->var)) ;
#               endif

	        break ;

	      case STRING:
	      case FILENAME:
		if (*p == '\0')
		{
		  if (--count > 0)
		    p = *++(*argv) ;
		  else if (option->default_string == NULL)
		  {
		    fprintf (stderr, "Parameter missing from '-%s' switch\n\n", option->op_str) ;
		    error += BAD_ERROR ;
		    break ;
		  }
		}

		if (((*p == '\0') || (*p == '-')) && (option->default_string != NULL))
		{
		  strncpy (option->var, option->default_string, option->max) ;

		  if (*p == '-')
		    p++ ;
		}
		else
		{
		  strncpy (option->var, p, option->max-1) ;
		  p = "" ;
		}

#               ifdef PARSE_DEBUG
		  printf ("Parsed string/filename (-%s). Value=\"%s\"\n", option->op_str, (char *)(option->var)) ;
#               endif

	        break ;
	    } /* switch */

	    break ; /* don't scan any more options - we've already found one */
	  } /* *r == 0 */
	} /* scan list of command string switches */

	if (*(option->op_str) == '\0')
	{
	  fprintf (stderr, "Unrecognised command line option: %c\n\n", *p) ;
	  error = BAD_ERROR ;
	  break ;
	}
      }
    } /* scan current argv */
    else
      new_argv[i++] = p ;
  }

  new_argv[i] = NULL ;
  i = 0 ;

  do
    (*argv)[i] = new_argv[i] ;
  while (new_argv[i++]) ;

  *argc = i ;
  return error ;
}

/******************************************************************************
 * PA_usage
 */

void
PA_usage (char *progname, char *use, char *extra)
{
  OPTION *option ;

  printf ("%s - A programme to %s%s\n", progname, use, extra) ;
  printf ("Written by Stefan Keller-Tuberg, (C) 2002%s\n", extra) ;
  printf ("\nUsage: %s {-", progname) ;

  for (option = &opt[0] ; *(option->op_str) ; option++)
    if (option->hidden == 0)
      printf ("%s", option->op_str) ;

  printf ("}%s\n", extra) ;

  for (option = &opt[0] ; *(option->op_str) ; option++)
    if (option->hidden == 0) switch (option->type)
    {
      case BOOLEAN:
	printf ("-%s\t%s%s\n", option->op_str, option->description, extra) ;
	break ;

      case INT:
	printf ("-%s num\t%s (%d <= num <= %d)%s\n", option->op_str, option->description, option->min, option->max, extra) ;
	break ;

      case STRING:
	printf ("-%s \"xx\"\t%s", option->op_str, option->description) ;

	if ((option->default_string != NULL) && (*(option->default_string) >= ' '))
	  printf (" (default=\"%s\")%s\n", option->default_string, extra) ;
	else
	  printf ("%s\n", extra) ;
	break ;

      case FILENAME:
	printf ("-%s file\t%s", option->op_str, option->description) ;

	if ((option->default_string != NULL) && (*(option->default_string) >= ' '))
	  printf (" (default=\"%s\")%s\n", option->default_string, extra) ;
	else
	  printf ("%s\n", extra) ;
	break ;
    }
}

/******************************************************************************
 * PA_output_config
 */

void
PA_output_config (void)
{
  OPTION *option ;

  printf ("Configured parameters:\n") ;

  for (option = &opt[0] ; *(option->op_str) ; option++) switch (option->type)
  {
    case BOOLEAN:
      printf ("  %s: %s\n", option->description, (*(int *)(option->var) == 1) ? "On" : "Off") ;
      break ;

    case INT:
      printf ("  %s: %d\n", option->description, *(int *)(option->var)) ;
      break ;

    case STRING:
    case FILENAME:
      if ((option->var != NULL) && (*(char *)(option->var) >= ' '))
	printf ("  %s: '%s'\n", option->description, (char *)(option->var)) ;
      break ;
  }

  printf ("\n") ;
}

/******************************************************************************
 * PA_dump_options
 */

void
PA_dump_options (char **argv)
{
  int i ;

  for (i=0 ; i <= 100 ; i++)
  {
    if (argv[i] == NULL)
      return ;

    printf ("argv[%d] == \"%s\"\n", i, argv[i]) ;
  }
}

/******************************************************************************
 * PA_copyleft
 */

void
PA_copyleft (char *name)
{
  STR_extract_filename (name) ;
  printf ("%s - Copyright and written by Stefan Keller-Tuberg. 2002\n\n", Filename) ;
  printf ("This program is free software; you can redistribute it and/or modify\n") ;
  printf ("it under the terms of the GNU General Public License Version 2 as published\n") ;
  printf ("by the Free Software Foundation.\n\n") ;
  printf ("This program is distributed in the hope that it will be useful,\n") ;
  printf ("but WITHOUT ANY WARRANTY; without even the implied warranty of\n") ;
  printf ("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n") ;
  printf ("GNU General Public License for more details.\n\n") ;
  printf ("\thttp://www.gnu.org/copyleft/gpl.html\n\n") ;
  printf ("You should have received a copy of the GNU General Public License\n") ;
  printf ("along with this program; if not, write to the Free Software\n") ;
  printf ("Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\n") ;
}
