/*
 * process.c
 ******************************************************************************
 * Copyright (C) 2002  Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. http://www.gnu.org
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ******************************************************************************
 * The purpose of this code is to automate the process of implementing a multi
 * process system. It is intended to work under uClinux and full blown Linux.
 * The two operating systems have different capabiltiies because of the absence
 * of support for a MMU in uClinux. This code tries to hide this from the
 * application code.
 */

#include <stdio.h>
#include <unistd.h>	/* for calls to ioperm, sleep, usleep, dup2, getcwd, ... */
#include <stdlib.h>
#include <errno.h>	/* for definitions of sys call error numbers */
#include <stdarg.h>	/* for calls to vsnprintf, va_start, ... */
#include <string.h>	/* for strncpy, memcpy, memset, ... */
#include <signal.h>	/* for kill, ... */
#include <sys/types.h>	/* for calls to wait, kill, ... */
#include <sys/ipc.h>	/* for ftok, semctl, ... */
#include <sys/msg.h>	/* for msgget, ... */
#include <sys/stat.h>	/* for calls to open, ... */
#include <sys/wait.h>	/* for calls to wait */
#include <fcntl.h>	/* for calls to open, ... */
#include <time.h>

#include "types.h"
#include "process.h"

#include <sys/sem.h>
#include <sys/shm.h>

#ifdef HAVE_TIMEKEEPER
#include <sys/time.h>
#endif

#include <sched.h>
#include "parse.h"
#include "stringfn.h"

PROC_INFO		*This_Proc ;
SHARED_GLOBAL_STORAGE	*Global ;

#ifndef HAVE_FORK
 char	Child [CMD_SIZE] ;
 int	Data ;
 char	Param [CMD_SIZE] ;
#endif

/******************************************************************************
 */

char *
PR_proc_from_msgq (const int msgq)
{
  Ushort i ;
  static char err_str[25] ;

  for (i=0 ; i < Global->Num_Proc ; i++)
    if (Global->Processes[i].msgq == msgq)
      break ;

  if (i >= Global->Num_Proc)
  {
    sprintf (err_str, "Unknown (msgq=%d)", msgq) ;
    return err_str ;
  }

  return Global->Processes[i].name ;
}

/******************************************************************************
 */

char *
PR_proc_from_pid (const int pid)
{
  Ushort i ;
  static char err_str[25] ;

  for (i=0 ; i < Global->Num_Proc ; i++)
    if (Global->Processes[i].pid == pid)
      break ;

  if (i >= Global->Num_Proc)
  {
    sprintf (err_str, "Unknown (pid=%d)", pid) ;
    return err_str ;
  }

  return Global->Processes[i].name ;
}

/******************************************************************************
 */

short
PR_proc_from_name (const char *name)
{
  Ushort i ;

  for (i=0 ; i < Global->Num_Proc ; i++)
    if (strcmp (Global->Processes[i].name, name) == 0)
      break ;

  if (i >= Global->Num_Proc)
    return -1 ;

  return i ;
}

/*******************************************************************************
 * PR_log_msg
 *
 * This routine is called from all over the place. It sends a message to the
 * logging process. For more information, see the logging process header below
 */

#ifdef HAVE_LOG_PROCESS

void
PR_log_msg (const LOG_PRIORITY priority, const char *fmt, ...)
{
  int		n ;
  va_list	args ;
  LOG_MSG	msg ;

  va_start(args, fmt) ;
  sprintf (msg.msg, "%-*.*s:%3lu.%02hu: ", PROCESS_NAME_SIZE, PROCESS_NAME_SIZE, This_Proc->name, Global->Up_Time.seconds, (Ushort)(FRAC_TIME(Global->Up_Time.ticks))) ;

  if (priority == ERR)
    strcat (msg.msg, "*") ;

  n = strlen (msg.msg) ;

  vsnprintf (&msg.msg[n], LOG_MSG_SIZE-2-n, fmt, args) ; /* changed from -1 due to sig11*/
  n = strlen (msg.msg) ;
  msg.msg[n++] = '\n' ;
  msg.msg[n] = '\0' ;
  va_end(args) ;
  msg.mtype = LOG_MESSAGE ;
  msg.pri = priority ;

  if (Global->Log_Proc_id != -1)
  {
    while (msgsnd (Global->Log_Proc_id, (LINUX_MSG_BUF *)&msg, n+sizeof(msg.pri)+1, 0) == -1)
    {
      if (errno != EINTR)
      {
	perror ("Unexpected error when writing to log msg queue") ;
	fprintf (stderr, "Failed message: %s", (char *)&msg.msg[0]) ;
	fprintf (stderr, "%s: Msg_buf log msg error location: %p\n", This_Proc->name, &msg) ;
	_exit (2) ;
      }

      usleep (DELAY_PERIOD) ;
    }
  }
  else if ((msg.pri == ERR) || (msg.pri == LOG_ABORT))
    fprintf (stderr, "%s", (char *)&msg.msg[0]) ;
  else
    printf ("%s", (char *)&msg.msg[0]) ;
}

#endif

/******************************************************************************
 * log_process
 *
 * This process is a single point of message logging.
 *
 * This process receives messages and logs them to a file. This may seem a
 * dumb way to do it - but it seemed to be the only way I could reliably get
 * many different processes to simulataneously be writing to a single file
 * and havinf the outputs appear in order and without loss. I know that this
 * could have been done by opening the file in O_SYNC mode (which like stderr,
 * blocks writers until the bytes are actually written to the hardware) but
 * this would not have been good in a real time application where processes
 * need to run quickly.
 *
 * There is probably a better way to do this under Linux of which I am not
 * aware.
 */

#ifdef HAVE_LOG_PROCESS

int
PR_log_process (const char *filename)
{
  FILE               *logfile ;
  int                len ;
  LOG_MSG            msg ;

  #ifdef BASIC_DEBUG
    PR_log_msg (FILE_ONLY, "STARTING. PID=%d, Queue_Id=%d", getpid(), Global->Log_Proc_id) ;
  #endif

  nice (1) ; /* specify the priority of this process: +ve numbers decrease priority */

  if ((filename != (char *)0) && (*filename != '\0'))
  {
    PR_log_msg (SCREEN, "LOG messages being written to \"%s\"", filename) ;

    if ((logfile = fopen (filename, "w")) == NULL)
    {
      PR_log_msg (ERR, "* Can't open \"%s\" for writing\n", filename) ;
      logfile = stdout ;
    }
  }
  else
    logfile = stdout ;

  for (;;)
  {
    while ((len = msgrcv (Global->Log_Proc_id, (LINUX_MSG_BUF *)&msg, sizeof(msg)-MSG_OVERHEAD, 0, 0)) == -1)
    {
      if (errno != EINTR)
      {
	sprintf (msg.msg, "Msg_buf wait error: %s, Id=%d, Loc=%p, Msg_len=%d, Exp_Len=%d\n", strerror(errno), Global->Log_Proc_id, &msg, len, (int)sizeof(msg)) ;
	msg.mtype = LOG_MESSAGE ;
	msg.pri = LOG_ABORT ;
	break ;
      }
    }

    if (msg.mtype == LOG_MESSAGE)
    {
      /*
       * Write the message to the log file and flush it to disk immediately.
       * This will cause this process to block for a short time but will
       * also ensure that if the programme crashes or is abnormally killed,
       * that as many messages as possible will have been written to disk
       *
       * If writing to disk rather than STDOUT, the logfile should always
       * conclude with the close message - this indicates that no messages
       * have been lost because the file was closed before the output
       * buffers were flushed.
       */

      if ((msg.pri == FILE_ONLY) && (logfile == stdout))
	continue ;

      if ((msg.pri == ERR) || (msg.pri == LOG_ABORT))
      {
	fprintf (stderr, "\r%s", msg.msg) ;

	if ((logfile == stdout) && (msg.pri != LOG_ABORT))
	  continue ;
      }

      if ((msg.pri == SCREEN) && (logfile != stdout))
      {
	fputs (msg.msg, stdout) ;
	fflush (stdout) ;
      }

      fputs (msg.msg, logfile) ;
      fflush (logfile) ;

      if (msg.pri == LOG_ABORT)
	goto log_abort ;
    }
    else if (msg.mtype == CLOSE_LOG)
    {
log_abort:
      if (logfile != stdout)
      {
	fputs ("\nLogfile closed normally\n", logfile) ;
	fclose (logfile) ;
      }

      _exit (0) ;
    }
    else
      PR_log_msg (LOG_ABORT, "Illegal message type received %d", msg.mtype) ;
  }
}

#endif

/******************************************************************************
 * This is called immediately prior to PR_process. It is called from
 * within LAUNCH_PROCESSES which is defined in process.h
 */

void
PR_init_shared_memory (void)
{
  Ushort i ;

  Global->index = 1 ;

  for (i=0 ; (i < Global->Num_Proc) ; i++)
  {
    if (Global->Processes[i].msgq)
    {
      Global->Processes[i].msgq = PR_create_mesg (Global->index++) ;

      if (strcmp (LOGGER, Global->Processes[i].name) == 0)
	Global->Log_Proc_id = Global->Processes[i].msgq ;

      else if (strcmp (TIMEKEEPER, Global->Processes[i].name) == 0)
	Global->Timekeeper_ID = Global->Processes[i].msgq ;

      #ifdef BASIC_DEBUG
	PR_log_msg (FILE_ONLY, "%s: Msgq=%d", Global->Processes[i].name, Global->Processes[i].msgq) ;
      #endif
    }

    if (Global->Processes[i].shm_size)
    {
      Global->Processes[i].shm_id = PR_create_shm (Global->index++, Global->Processes[i].shm_size, &Global->Processes[i].shm) ;

      #ifdef BASIC_DEBUG
	PR_log_msg (FILE_ONLY, "%s: Shm_id=%d, ptr=%p, size=%d", Global->Processes[i].name, Global->Processes[i].shm_id, Global->Processes[i].shm, Global->Processes[i].shm_size) ;
      #endif
    }

    if (Global->Processes[i].sem_id)
    {
      PR_create_sem (Global->index++, &Global->Processes[i].sem_id) ;

      #ifdef BASIC_DEBUG
	PR_log_msg (FILE_ONLY, "%s: Sem_id=%d", Global->Processes[i].name, Global->Processes[i].sem_id) ;
      #endif
    }
  }
}

/******************************************************************************
 */

int
PR_process (char *exec_name, const int argc, char **argv)
{
  Ushort i, logger ;
  int	ret_pid ;
  int	ret_status ;
  int	error ;
  int	fd ;

#ifndef HAVE_FORK
  char	name [CMD_SIZE] ;
  char	spawn [CMD_SIZE] ;
  char	data [CMD_SIZE] ;
  char	param [CMD_SIZE] ;
  char	**new_argv ;
#endif

  error = 0 ;

  /*
   * Create the required processes
   */

  fflush (stderr) ;

  #ifndef HAVE_FORK
    if (*exec_name != '/')
    {
      getcwd (name, CMD_SIZE) ;
      STR_extract_filename (exec_name) ;
      STR_snaprintf (name, CMD_SIZE, "/%s", FileExtension) ;
    }
    else
    {
      strncpy (name, exec_name, CMD_SIZE-1) ;
      name [CMD_SIZE-1] = '\0' ;
    }
  #endif

  for (i=0 ; (i < Global->Num_Proc) && (error == 0) ; i++)
  {
    #ifdef HAVE_FORK
      ret_pid = fork() ;

    #else
      new_argv = malloc ((argc+5) * sizeof (char **)) ;

      if (new_argv == NULL)
	_exit (0xf2) ;

      sprintf (spawn, "-" SPAWN "%s", Global->Processes[i].name) ;
      sprintf (data,  "-" DATA  "%d", (int)Global) ;
      sprintf (param, "-" PARAM "%s", Global->Processes[i].param) ;
      new_argv[0] = exec_name ;
      new_argv[1] = spawn ;
      new_argv[2] = data ;
      new_argv[3] = param ;

      /* copy from argv[1] to the null termination in argv[argc] */
      memcpy (new_argv + 4, argv+1, argc * sizeof (char **)) ;

      ret_pid = vfork() ;
    #endif

    if (ret_pid == 0) /* I am a child */
    {
      if (Global->Processes[i].io == 0)
      {
	close (0) ; /* close the things we're not supposed to see */
	close (1) ;
	close (2) ;

	/* make sure the stdin, stdout and stderr get directed to /dev/null */
	fd = open ("/dev/null", O_RDWR); /* open stdin */

        if (fd >= 0)
	{
	  if (fd != 0)
	  {
	    dup2(fd, 0);
	    close(fd);
	  }
	  dup2(0, 1); /* create stdout */
	  dup2(0, 2); /* create stderr */
        }
      }

      umask (0) ;

      #ifdef HAVE_FORK
	This_Proc = &Global->Processes[i] ;
	_exit ((*Global->Processes[i].proc)(Global->Processes[i].param)) ;
      #else
	execvp (name, new_argv) ;
	_exit (0xf0) ;
      #endif
    }
    else if (ret_pid == -1) /* it didn't work */
    {
      error += 100 ;
      PR_log_msg (ERR, "Fork failed: %s, Failed_Proc: %s" LINE_TERM, strerror(errno), Global->Processes[i].name) ;
    }

    /* I am the parent process if I got this far */
    Global->Processes[i].pid = ret_pid ;
  }

  #ifdef BASIC_DEBUG
    PR_log_msg (NORM, "Parent waiting for first child to quit") ;
  #endif

  ret_pid = wait (&ret_status) ;

  // DISPLAY THE DIAGNOSTICS IF THE CHILD EXITS WITH RETURN CODE > 1
  // OR IF THE CHILD TERMINATES BECAUSE OF UNCAUGHT SIGNAL
  //
  // Don't display the diagnostics (ie silent exit) with ret status 1

  if ((WIFEXITED(ret_status) && ((WEXITSTATUS(ret_status)) > 1)))
  {
    PR_log_msg (ERR, "A child quit with an abnormal return code: %s", PR_proc_from_pid (ret_pid)) ;
    PR_log_msg (ERR, "It indicated status %d", WEXITSTATUS(ret_status)) ;
  }
  else if (WIFSIGNALED(ret_status))
  {
    PR_log_msg (ERR, "A child quit because of an uncaught signal: %s", PR_proc_from_pid (ret_pid)) ;
    PR_log_msg (ERR, "The signal number was: %d (%#02x)", ret_status, ret_status) ;
    PR_log_msg (ERR, "(execute 'man 7 signal' for details of this type)") ;
  }

  /*
   * Close up, kill processes and return resources.
   */

  for (logger=Global->Num_Proc, i=0 ; i < Global->Num_Proc ; i++)
  {
    if (strcmp (LOGGER, Global->Processes[i].name) == 0) /* don't kill logger yet */
      logger = i ;
    else if (Global->Processes[i].pid != 0)
    {
      kill (Global->Processes[i].pid, SIGTERM) ;
      Global->Processes[i].pid = 0 ;
    }
  }

  if ((logger < Global->Num_Proc) && (Global->Processes[logger].pid != 0))
  {
    LOG_MSG	msg ;

    msg.mtype = CLOSE_LOG ;
    PR_send_mesg (Global->Processes[logger].msgq, (LINUX_MSG_BUF *)&msg, sizeof(msg.mtype)) ;
    sleep (2) ; /* now kill it if it's still there, after it's flushed */
    kill (Global->Processes[logger].pid, SIGTERM) ;
    Global->Processes[logger].pid = 0 ;
  }

  for (i=0 ; i < Global->Num_Proc ; i++)
  {
    if (Global->Processes[i].msgq)
    {
      PR_rm_mesg (Global->Processes[i].msgq) ;
      Global->Processes[i].msgq = 1 ;
    }

    if (Global->Processes[i].sem_id)
    {
      PR_rm_sem (&Global->Processes[i].sem_id) ;
      Global->Processes[i].sem_id = 1 ;
    }

    if (Global->Processes[i].shm_size)
    {
      PR_rm_shm (Global->Processes[i].shm_id) ;
      Global->Processes[i].shm_id = 0 ;
      Global->Processes[i].shm = (void *)0 ;
    }
  }

  return error ;
}

/******************************************************************************
 */

int
PR_launch_process (char *proc, const int argc, char **argv)
{
  #ifdef HAVE_FORK
    PR_init_shared_memory() ;
    return PR_process (proc, argc, argv) ;
  #else

    int i ;

    if (*Child != '\0')
    {
      for (i=0 ; i < Global->Num_Proc ; i++)
      {
	if (strcmp (Global->Processes[i].name, Child) == 0)
	{
	  This_Proc = &Global->Processes[i] ;
	  return (*Global->Processes[i].proc)(Param) ; // don't expect return
	}
      }

      /*
       * Something wrong - didn't find the right process
       */

      PR_log_msg (ERR, "Can't locate and launch (%s)" LINE_TERM, Child) ;
      return 0xe0 ;
    }
    else
    {
      PR_init_shared_memory() ;
      return PR_process (proc, argc, argv) ;
    }

  #endif
}

/******************************************************************************
 * PR_create_system_structures
 *
 * Returns true if this is main() and false if this is a child process.
 */

int
PR_create_system_structures (const char *logfilename)
{
#ifndef HAVE_FORK
  if (*Child == '\0')
  {
#endif

    if ((PR_create_shm ('!', sizeof(SHARED_GLOBAL_STORAGE), (void **)&Global) == 0) ||
	(PR_create_shm ('!'+1, sizeof(PROC_INFO[MAX_NUM_PROC]), (void **)&Global->Processes) == 0))
    {
      perror ("Create_system_struct failed to create shared memory") ;
      exit (1) ;
    }
    else
    {
      This_Proc = &Global->Processes[MAX_NUM_PROC] ;
      strncpy (This_Proc->name, "Main", PROCESS_NAME_SIZE) ;
      Global->index = '!'+2 ;
      Global->Log_Proc_id = -1 ;
      Global->Timekeeper_ID = -1 ;

      #ifdef HAVE_LOG_PROCESS
	//			(*name,	&proc(),	 &param, shm_size, io, msgq, sem_id)
	PR_register_new_process (LOGGER, PR_log_process, logfilename, 0, 1, 1, 0) ;
      #endif

      #ifdef HAVE_TIMEKEEPER
	PR_register_new_process (TIMEKEEPER, PR_timekeeper_process, "", 0, 0, 1, 0) ;
      #endif
    }

    return 1 ;

#ifndef HAVE_FORK
  }

  if (Data != 0)
    Global = (SHARED_GLOBAL_STORAGE *)Data ;
  else
  {
    fprintf (stderr, "Data location not specified in Child Process!\n") ;
    return 100 ;
  }

  return 0 ;
#endif
}

/******************************************************************************
 * PR_register_new_process
 *
 * Parameters:
 *	name:	 A pointer to the name of this process
 *
 *	proc:	 The address of the routine which holds this process
 *
 *	param:	 The address of a character string to pass as a parameter
 *
 *	shm_size:The number of bytes to reserve for this proc's shared mem
 *
 *	io:	 Non zero if this process will directly use stderr, stdout.
 *		 One and only one process should have this set to non zero.
 *		 Normally, this is the log process and other processes will
 *		 call PR_log_msg to log output.
 *
 *	msgq:	 Non zero if a message queue needs to be created.
 *
 *	sem_id:	 Non zero if a semaphore needs to be created.
 */

void
PR_register_new_process (const char *name, PROC_DEF *proc, const char *param, const int shm_size, const short io, const int msgq, const int sem_id)
{
  PROC_INFO *ptr ;

  #ifdef BASIC_DEBUG
    #define __test(A) (A)?"Y":"N"
    PR_log_msg (FILE_ONLY, "Creating: %s(%s). Shm=%d, I/O=%s, msgq=%s, sem=%s", name, param, shm_size, __test(io), __test(msgq), __test(sem_id)) ;
  #endif

  if (Global->Num_Proc >= (MAX_NUM_PROC - 1))
  {
    PR_log_msg (ERR, "Tried to register more processes than table size (=%hu)", MAX_NUM_PROC) ;
    return ;
  }

  ptr = &Global->Processes[Global->Num_Proc] ;
  Global->Num_Proc++ ;

  sprintf (ptr->name, "%.*s", PROCESS_NAME_SIZE, name) ;
  ptr->proc = proc ;
  ptr->param = param ;
  ptr->shm_size = shm_size ;
  ptr->io = io ;
  ptr->pid = 0 ;
  ptr->msgq = msgq ;
  ptr->sem_id = sem_id ;
  ptr->shm_id = 0 ;
  ptr->shm = (void *)0 ;
}

void
PR_disp_process (void)
{
  PROC_INFO	*ptr ;
  int		i ;

  PR_log_msg (NORM, "%d processes defined", Global->Num_Proc) ;

  for (i=0 ; i < Global->Num_Proc ; i++)
  {
    ptr = &Global->Processes[i] ;
    PR_log_msg (NORM, "%s(%s): pid=%d, msgq=%d", ptr->name, ptr->param, ptr->pid, ptr->msgq) ;
  }
}

/******************************************************************************
 */

void
PR_set_process_priority (const int policy, const int priority)
{
// only code this routine if not running uClinux
#ifndef EMBEDED

  struct sched_param param ;

  param.sched_priority = sched_get_priority_min(policy) + priority ;

  if (sched_setscheduler (0, policy, &param) == -1)
  {
    PR_log_msg (ERR, "Error setting scheduler. Are you root or suid root?") ;
    _exit (0) ;
  }

  if (sched_setparam (0, &param) == -1)
  {
    PR_log_msg (ERR, "Error setting process priority to %d", param.sched_priority) ;
    _exit (0) ;
  }

  #ifdef BASIC_DEBUG
    PR_log_msg (FILE_ONLY, "Process priority=%d, scheduler=%d", param.sched_priority, policy) ;
  #endif

#endif
}

/******************************************************************************
 * PR_clear_mesg
 *
 * These are housekeeping routines to make the tasks of managing Linux's
 * interprocess communications a little easier.
 *
 * The PR_read_mesg routine is blocking and will cause a process to go to
 * sleep until a message arrives. Same for lock_sem.
 *
 * The PR_send_mesg routine is also potentially blocking if there is no room left
 * in the kernel's message buffer area for another message to be written.
 */

char *
PR_decode_message_type (const long mtype)
{
  static char err_string[15] ;

  switch (mtype)
  {
    case 0:			return "Any Msg (=0)" ;
    case LOG_MESSAGE:		return "LOG_MSG" ;
    case CLOSE_LOG:		return "CLOSE_LOG" ;
    case TICK_EVENT:		return "TICK_EVENT" ;
    case TIMEOUT_REQUEST:	return "TIMEOUT_REQ" ;
    case TIMEOUT_NOTIFICATION:	return "TIMEOUT_NOTIFICATION" ;
    case SLEEP_REQUEST:		return "SLEEP_REQ" ;
    case SLEEP_NOTIFICATION:	return "SLEEP_NOTIFICATION" ;
    case MONITOR:		return "MONITOR" ;
    default:
      sprintf (err_string, "%ld", mtype) ;
      return err_string ;
  }
}

void
PR_clear_mesg (const int msgq_id)
{
  GENERIC_NULL_MSG  null_buffer ;

  /*
   * We call this routine to empty out all messages from a specified message
   * queue. On return, the queue is empty.
   */

  #ifdef MESSAGE_TRACE
    PR_log_msg (FILE_ONLY, "Clearing msg queue id: %d", msgq_id) ;
  #endif

  while (msgrcv (msgq_id, &null_buffer, sizeof(null_buffer)-MSG_OVERHEAD, 0, IPC_NOWAIT|MSG_NOERROR) != -1) ;

  if (errno != ENOMSG)
  {
    perror ("unexpected error when flushing message queue") ;
    fprintf (stderr, "Process: %s, queue id: %d\n", This_Proc->name, msgq_id) ;
    exit (4) ;
  }
}

int
PR_create_mesg (const char token)
{
  key_t         key ;
  int           msgq_id ;

  /*
   * In order to open a message queue, we first need a magic number with which
   * the operating system can identify the queue. We make this key using the
   * ftok call which convolutes the major and minor inode numbers of the
   * current durectory "." with another token which is specific to this
   * software.
   *
   * When we open the message queue, we also check to see if it is empty.
   * If not empty, we assume that it is still hanging around from a previous
   * (crashed) instance of this programme and we need to clear out all the
   * old messages before this programme can work properly.
   */

  key = ftok (".", token) ;

  if ((msgq_id = msgget (key, IPC_CREAT | 0600)) == -1)
  {
    perror ("Can't create message queue") ;
    fprintf (stderr, "Process: %s, key: %d\n", This_Proc->name, key) ;
    exit (3) ;
  }

  #ifdef MESSAGE_TRACE
    PR_log_msg (FILE_ONLY, "Created message queue token '%c', id %d", token, msgq_id) ;
  #endif

  PR_clear_mesg (msgq_id) ;
  return (msgq_id) ;
}

void
PR_read_mesg (const int queue, LINUX_MSG_BUF *buf, const int max_msg_len, const long msg_type)
{
  int len ;

  #ifdef MESSAGE_TRACE
    PR_log_msg (FILE_ONLY, "read_mesg: Waiting on %s", PR_decode_message_type (msg_type)) ;
  #endif

  while ((len = msgrcv (queue, buf, max_msg_len-MSG_OVERHEAD, msg_type, 0)) == -1)
  {
    if (errno != EINTR)
    {
      PR_log_msg (ERR, "read_mesg err: %s, Queue_Id=%d, Buf_Loc=%p, Len=%d, Exp_Len=%d", strerror(errno), queue, buf, len, max_msg_len) ;

      #ifdef PARANOID
        if (errno == E2BIG)
	{
	  #define SIZE_A 1000
	  unsigned char array[SIZE_A] ;
	  GENERIC_MSG   *p ;

	  p = (GENERIC_MSG *)array ;

	  if ((len = msgrcv (queue, array, SIZE_A-MSG_OVERHEAD, 0, MSG_NOERROR)) != -1)
	    PR_log_msg (ERR, "Msg_Size=%d, Type=%s, Sub_Cmd=%d, Rep_To=%d", len, PR_decode_message_type (p->mtype), p->sub_command, p->reply_to) ;
	}
      #endif

      _exit (0) ;
    }

    #ifdef MESSAGE_TRACE
      PR_log_msg (ERR, "Looping in msg_recv: %s", strerror(errno)) ;
    #endif

    usleep (DELAY_PERIOD) ; /* delay a while before trying again */
  }

  #ifdef MESSAGE_TRACE
    if (len > sizeof(long))
      PR_log_msg (FILE_ONLY, "%d byte message read from Msg_Queue_Id=%d, Type=%s", len, queue, PR_decode_message_type (((GENERIC_MSG *)buf)->mtype)) ;
  #endif
}

void
PR_send_mesg (const int queue, LINUX_MSG_BUF *buf, const int msg_len)
{
  #ifdef MESSAGE_TRACE
    if (msg_len > sizeof(long))
      PR_log_msg (FILE_ONLY, "%d byte message sent to %s, Type=%s", msg_len, PR_proc_from_msgq(queue), PR_decode_message_type (((GENERIC_MSG *)buf)->mtype)) ;
  #endif

  while (msgsnd (queue, buf, msg_len-MSG_OVERHEAD, 0) == -1)
  {
    if (errno != EINTR)
    {
      PR_log_msg (ERR, "send_msg err: %s, Id: %s, Msg_Buf_Loc: %p", strerror(errno), PR_proc_from_msgq(queue), buf) ;
      _exit (0) ;
    }

    usleep (DELAY_PERIOD) ; /* delay a while before trying again */
  }
}

void
PR_rm_mesg (const int queue)
{
  if (msgctl (queue, IPC_RMID, 0) != 0)
  {
    perror ("Couldn't remove a message queue") ;
    fprintf (stderr, "Process: %s, Queue_Id=%d", This_Proc->name, queue) ;
    exit (13) ;
  }

  #ifdef MESSAGE_TRACE
    PR_log_msg (FILE_ONLY, "Removed msg Queue_Id=%d", queue) ;
  #endif
}

/******************************************************************************
 * SEMAPHORES
 *
 * Semaphores are created in the locked state. The owner of the semaphore
 * should unlock the semaphore when everything is OK to start using it.
 */

void
PR_create_sem (const char token, int *sem)
{
  key_t         key ;
  int           arg ;

  key = ftok (".", token) ;

  if ((*sem = semget (key, 1, IPC_CREAT | 0600)) == -1)
  {
    perror ("Can't create semaphore") ;
    fprintf (stderr, "Process: %s, key: %d\n", This_Proc->name, key) ;
    exit (5) ;
  }

  #ifdef SEM_TRACE
    PR_log_msg (FILE_ONLY, "Created semaphore token '%c', id %d", token, *sem) ;
  #endif

  /*
   * The semaphore is created. But it could have been hanging
   * around from a previous instance of this programme.
   *
   * When a semaphore is created, its count should be initialised to 0. A
   * count of zero means that all processes which call unlock will be suspended
   * until another process sets the count above 0.
   *
   * Because the semaphore may not be newly created (such as if there was a
   * previous instance of the semaphore haning around from a previous crash)
   * then the count may not be 0 upon return. The only reliable fix for this
   * I have found is as follows.
   */

  arg = 0 ;

  if (semctl (*sem, 0, SETVAL, arg) == -1)
  {
    perror ("Can't initialise semaphore to 0") ;
    fprintf (stderr, "Process: %s, key: %d, id: %d\n", This_Proc->name, key, *sem) ;
    exit (6) ;
  }
}

void
PR_rm_sem (int *sem)
{
  if (semctl (*sem, 0, IPC_RMID, 0) != 0)
  {
    perror ("Couldn't remove a semaphore") ;
    fprintf (stderr, "Process %s, Sem_Id=%d\n", This_Proc->name, *sem) ;
    exit (14) ;
  }

  #ifdef SEM_TRACE
    PR_log_msg (FILE_ONLY, "Removed Sem_Id=%d", *sem) ;
  #endif

  *sem = 1 ;
}

void
PR_unlock_sem (int *sem)
{
  struct sembuf unlock ;

  unlock.sem_num = 0 ;
  unlock.sem_op  = 1 ;
  unlock.sem_flg = 0 ;

  if (semop(*sem, &unlock, 1) == -1)
  {
    PR_log_msg (ERR, "Sem unlock error: %s, Sem_Id=%d", strerror(errno), sem) ;
    _exit (0) ;
  }

  #ifdef SEM_TRACE
    PR_log_msg (FILE_ONLY, "Unlocked Sem_Id=%d", sem) ;
  #endif
}

void
PR_lock_sem (int *sem)
{
  struct sembuf lock ;
  int           ret ;

  lock.sem_num =  0 ;
  lock.sem_op  = -1 ;
  lock.sem_flg =  0 ;

  /*
   * Need to implement the following loop as a while loop with explicit delay
   * - because we have many different processes in the system with different
   * priorities. If a lower priority process gets the semaphore and a higher
   * priority process starts waiting, then the lower priority process will
   * never get a chance to give up the semaphore!!
   */

  while ((ret = semop(*sem, &lock, 1)) != 0)
  {
    if (errno != EAGAIN)
    {
      PR_log_msg (ERR, "Sem lock error: %s, Sem_Id=%d", strerror(errno), *sem) ;
      _exit (0) ;
    }

    #ifdef SEM_TRACE
      PR_log_msg (FILE_ONLY, "Looping for Sem_Id=%d", *sem) ;
    #endif

    usleep (DELAY_PERIOD) ; /* wait a while before trying again */
  }

  #ifdef SEM_TRACE
    PR_log_msg (FILE_ONLY, "Locked Sem_Id=%d", *sem) ;
  #endif
}

/******************************************************************************
 * SHARED MEMORY
 *
 * If the approriate field for PR_process is set to non zero, a block of shared
 * memory is defined for that process.
 */

int
PR_create_shm (const char token, const int size, void **shm_addr)
{
  int	shm_id = -1 ;

#ifdef HAVE_SHM
  key_t	key ;

  key = ftok (".", token) ;

  if ((shm_id = shmget (key, size, IPC_CREAT | 0600)) == -1)
  {
    perror ("Couldn't create shm") ;
    return 0 ;
  }

  if ((*shm_addr = (void *)shmat(shm_id, NULL, 0)) == (void *)-1)
  {
    perror ("Couldn't attach to shm") ;
    shmctl (shm_id, IPC_RMID, NULL) ;
    return 0 ;
  }
#else
  *shm_addr = malloc (size) ;
#endif

  memset (*shm_addr, 0, size) ; /* zero the block of memory */
  return shm_id ;
}

void
PR_rm_shm (const int shm)
{
#ifdef HAVE_SHM
  if (shmctl (shm, IPC_RMID, NULL) != 0)
  {
    perror ("Couldn't remove a shared memory segment") ;
    fprintf (stderr, "Process %s, shared Mem_Id=%d\n", This_Proc->name, shm) ;
    exit (3) ;
  }

#else
  free ((void *)shm) ;
#endif

  #ifdef SHM_TRACE
    PR_log_msg (FILE_ONLY, "Removed shared Mem_Id=%d", shm) ;
  #endif
}

/******************************************************************************
 * PR_timekeeper_process
 *
 * This routine is the basis for the timekeeper process.
 *
 * There are probably many ways I could have implemented timers in this
 * project. I have chosen to write a centralised timer management process (this
 * process) which is entirely message based.
 *
 * Other processes wishing to start a timer send a message to this process.
 * The message will contain a magic number identifying the specific timer which
 * is to be started. (The magic number is usually the address of a process
 * significant local variable).
 *
 * This process checks to see if a timer with that magic ID already exists
 * and if so, restarts it - using the specified timeout. If not, a new timer is
 * created and added to the queue.
 *
 * If a process wishes to stop a timer, it sends a timer restart message with
 * a new timeout of 0. The timer process then deletes the timeout from the
 * queue (silently).
 *
 * This process uses a periodic alarm at the rate of TICK_FREQ Hz.
 * This sets the minimum granularity of the timers.
 */

#ifdef HAVE_TIMEKEEPER

static void
PR_time_sig_handler (const int sig)
{
  long   timer_msg_type ;  /* This is a VERY SHORT message - 1 field */

  #ifdef SIG_HANDLER_DEBUG
    PR_log_msg (FILE_ONLY, "Time_sig_handler: Timer expired") ;
  #endif

  timer_msg_type = TICK_EVENT ;
  PR_send_mesg (Global->Timekeeper_ID, (LINUX_MSG_BUF *)(&timer_msg_type), sizeof(timer_msg_type)) ;
}

int
PR_timekeeper_process (const char *param)
{
  TIMER             timer_array [NUM_TIMERS] ;
  TIMER             *shortest_timer ;
  TIMER             *free_timer ;
  TIMER             *end_free_timer_list ;
  struct itimerval  timer ;
  TIMER_MSG         timer_msg ;
  Ushort            i ;
  TIMER             *tmp, *tmp1, *tmp2 ;

  #ifdef BASIC_DEBUG
    ANNOUNCE_PROCESS() ;
  #endif

  #ifdef TIMER_DEBUG
    PR_log_msg (SCREEN, "Timer debug messages being written to log file") ;
  #endif

  #ifdef SIG_HANDLER_DEBUG
    PR_log_msg (SCREEN, "Time_sig_handler debug messages being written to log file") ;
  #endif

  /*
   * Initialise the timer array and pointers
   */

  Global->Up_Time.seconds = 0 ; /* uptime = 0 */
  Global->Up_Time.ticks = 0 ;

  for (i=0 ; i < NUM_TIMERS ; i++)
  {
    timer_array [i].interval = 0 ;
    timer_array [i].ref_id = 0 ;
    timer_array [i].type = 0 ;
    timer_array [i].reply_to = 0 ;
    timer_array [i].next = &timer_array[i+1] ;
  }

  timer_array [NUM_TIMERS-1].next = NULL ;
  shortest_timer = (TIMER *)0 ;
  free_timer = &timer_array[0] ;
  end_free_timer_list = &timer_array[NUM_TIMERS-1] ;

  /*
   * Define a periodic timer to serve as the heartbeat which triggers the
   * processing of timers. By setting it_interval to a non zero value,
   * the timer becomes a periodic timeout - each time the timer expires, it
   * generates a SIGALM and at the same time, restarts the timer.
   */

  signal (SIGALRM, &PR_time_sig_handler) ;
  timer.it_interval.tv_sec = 0 ;
  timer.it_interval.tv_usec = 1000000 / TICK_FREQ ; /* microseconds */
  timer.it_value.tv_sec = 0 ;
  timer.it_value.tv_usec = 1000000 / TICK_FREQ ;    /* microseconds */
  setitimer (ITIMER_REAL, &timer, NULL) ;

  #ifdef TIMER_DEBUG
    PR_log_msg (FILE_ONLY, "Initialised. Tick Frequency: %d Hz", TICK_FREQ) ;
  #endif

  for (;;)
  {
    /*
     * This loop is where the timekeeper process receives requests
     * from other processes.
     *
     * Also, a tick message is received from the alarm handler each
     * time an interval expires. Ticks are processed in order and
     * could be queued after timeout request messages.
     */

    PR_read_mesg (Global->Timekeeper_ID, (LINUX_MSG_BUF *)(&timer_msg), sizeof(timer_msg), 0) ;

    switch (timer_msg.command.mtype)
    {
      case TICK_EVENT:
	/*
	 * The periodic timer has expired
	 */

	if (++(Global->Up_Time.ticks) >= TICK_FREQ)
	{
	  Global->Up_Time.ticks = 0 ;
	  Global->Up_Time.seconds++ ;
	}

        if (shortest_timer != NULL)
	{
	  shortest_timer->interval-- ;

	  /*
	   * If the head timer just decrmtd to zero, we need to process it.
	   */

	  while (shortest_timer->interval == 0)
	  {
	    tmp = shortest_timer ;
	    timer_msg.command.mtype = (tmp->type == TIMEOUT_REQUEST) ? TIMEOUT_NOTIFICATION : SLEEP_NOTIFICATION ;
	    timer_msg.command.sub_command = tmp->ref_id ;

            #ifdef TIMER_DEBUG
	      if (timer_msg.command.mtype == TIMEOUT_NOTIFICATION)
	        PR_log_msg (FILE_ONLY, "Timeout sent to %s, Id=%p", PR_proc_from_msgq (tmp->reply_to), (void *)timer_msg.command.sub_command) ;
	      else
	        PR_log_msg (FILE_ONLY, "Sleep notification sent to %s", PR_proc_from_msgq (tmp->reply_to)) ;
            #endif

	    PR_send_mesg (tmp->reply_to, (LINUX_MSG_BUF *)(&timer_msg), sizeof(GENERIC_MSG)) ;

	    shortest_timer = tmp->next ;
	    tmp->ref_id = 0 ;
	    tmp->type = 0 ;
	    tmp->reply_to = 0 ;
	    tmp->next = NULL ;

	    if (end_free_timer_list != NULL)
	    {
	      end_free_timer_list->next = tmp ;
	      end_free_timer_list = tmp ;
	    }
	    else /* The free timer list is empty */
	      end_free_timer_list = free_timer = tmp ;

            if (shortest_timer == NULL)
	      break ;
	  }
	}

	break ;

      case TIMEOUT_REQUEST:
      case SLEEP_REQUEST:
        /*
	 * A valid request has been received. First, scan the list of active
	 * timers to see if a timer with the same id already exists. If so,
	 * restart the timer by removing it from the queue and reinserting
	 * it. Otherwise, just start a new timer.
	 */

        for (tmp1=NULL,tmp=shortest_timer ; tmp!=NULL ; tmp1=tmp,tmp=tmp->next)
	{
	  if ((timer_msg.command.sub_command == tmp->ref_id) && (timer_msg.command.reply_to == tmp->reply_to))
	  {
	    /*
	     * An existing timer has been identified. Remove it from the list
	     * and reset it. Be careful to fix the list up though.
	     */

            #ifdef TIMER_DEBUG
	      PR_log_msg (FILE_ONLY, "Timer restart request received from %s, duration %d", PR_proc_from_msgq (tmp->reply_to), timer_msg.interval) ;
            #endif

	    if (tmp1 == NULL)
	    {
	      /*
	       * Existing timer. Remove it from head of the list
	       */

	      shortest_timer = tmp->next ;
	    }
	    else
	    {
	      /*
	       * Existing timer. Remove it from within the timer list
	       */

	      tmp1->next = tmp->next ;
	    }

	    if (tmp->next != NULL)
	    {
	      /*
	       * Another timer follows the existing timer which was just
	       * removed. Add the remaining interval of the deleted timer to
	       * the following timer.
	       */

	      tmp->next->interval += tmp->interval ;
	    }

	    /*
	     * Reset the parameters on the deleted timer to the spec'd values
	     */

	    tmp->interval = timer_msg.interval ;
	    tmp->reply_to = timer_msg.command.reply_to ;
	    break ;
	  }
	}

	if (timer_msg.interval == 0)
	{
	  /*
	   * The request was to delete the timer. Add the entry to the
	   * free list and get out of here
	   */

          #ifdef TIMER_DEBUG
	    PR_log_msg (FILE_ONLY, "Request was to delete timer") ;
          #endif

	  if (tmp != NULL)
	  {
	    tmp->ref_id = 0 ;
	    tmp->type = 0 ;
	    tmp->reply_to = 0 ;
	    tmp->next = NULL ;

	    if (end_free_timer_list != NULL)
	    {
	      end_free_timer_list->next = tmp ;
	      end_free_timer_list = tmp ;
	    }
	    else /* The free timer list is empty */
	      end_free_timer_list = free_timer = tmp ;
	  }

	  break ;
	}

	if (tmp == NULL)
	{
	  /*
	   * A new timer needs to be started. Find a free timer entry from the
	   * head of the free list and fill out the fields.
	   */

          #ifdef TIMER_DEBUG
	    if (timer_msg.command.mtype == TIMEOUT_REQUEST)
	      PR_log_msg (FILE_ONLY, "New timer from %s, Dur=%d, Id=%p", PR_proc_from_msgq (timer_msg.command.reply_to), timer_msg.interval, (void *)timer_msg.command.sub_command) ;
	    else
	      PR_log_msg (FILE_ONLY, "Sleep request from %s, Dur=%d", PR_proc_from_msgq (timer_msg.command.reply_to), timer_msg.interval) ;
          #endif

	  if ((tmp = free_timer) == NULL)
	  {
	    PR_log_msg (ERR, "Tried to start a new timer but no free timer entries") ;
	    return (0) ;
	  }

	  free_timer = tmp->next ;

	  if (free_timer == NULL)
	    end_free_timer_list = NULL ;

	  tmp->ref_id = timer_msg.command.sub_command ;
	  tmp->type = timer_msg.command.mtype ;
	  tmp->reply_to = timer_msg.command.reply_to ;
	  tmp->interval = timer_msg.interval ;
	}

	/*
	 * Add the new timer (pointed to by tmp) into the timer list
	 */

	if (shortest_timer == NULL)
	{
	  /*
	   * add the new timer to an empty list
	   */

	  shortest_timer = tmp ;
	  tmp->next = NULL ;
	}
	else
	{
	  /*
	   * add the new timer into the list but make sure to add it in at the
	   * right place.
	   */

	  for (tmp2=NULL, tmp1=shortest_timer ; tmp1 != NULL ; tmp2=tmp1, tmp1=tmp1->next)
	  {
	    if (tmp->interval < tmp1->interval)
	    {
	      /*
	       * add the new timer into the middle of the timer list
	       */

	      tmp1->interval -= tmp->interval ;

	      if (tmp2 == NULL)
	      {
		/*
		 * actually, the new timer should be the first entry in the
		 * timer list and all other entries should shift along.
		 */

		shortest_timer = tmp ;
	      }
	      else
		tmp2->next = tmp ;

	      tmp->next = tmp1 ;
	      break ;
	    }
	    else
	      tmp->interval -= tmp1->interval ;
	  }

	  if (tmp1 == NULL)
	  {
	    /*
	     * add the new timer to the end of the list
	     */

	    if (tmp2 == NULL)
	      shortest_timer = tmp ;
	    else
	      tmp2->next = tmp ;

	    tmp->next = NULL ;
	  }
	}

	break ;

      case MONITOR:
        for (tmp=shortest_timer, i=0 ; tmp != NULL ; tmp=tmp->next)
	  PR_log_msg (FILE_ONLY, "Timer %2d: owner=%s, interval=%d, ref_id=%d, type=%s", \
		      i++, PR_proc_from_msgq (tmp->reply_to), \
		      tmp->interval, tmp->ref_id, \
		      PR_decode_message_type (tmp->type)) ;

	break ;

      default:
	#ifdef PARANOID
	  PR_log_msg (ERR, "Bad msg recd: %s, Subcmd: %#0x, From %s", PR_decode_message_type (timer_msg.command.mtype), timer_msg.command.sub_command, PR_proc_from_msgq (timer_msg.command.reply_to)) ;
	  return (0) ;
	#endif
    }
  }

  return (254) ;
}

void
PR_send_timeout_req_msg (const int rep_to, const Ushort interval, const void *timer_id)
{
  TIMER_MSG  msg ;

  msg.command.mtype = TIMEOUT_REQUEST ;
  msg.command.sub_command = (int)timer_id ;
  msg.command.reply_to = rep_to ;
  msg.interval = interval;

  #ifdef TIMER_DEBUG
    PR_log_msg (FILE_ONLY, "PR_send_timeout: Sending Req Dur=%d, Id=%p", interval, timer_id) ;
  #endif

  if (Global->Timekeeper_ID != -1)
    PR_send_mesg (Global->Timekeeper_ID, (LINUX_MSG_BUF *)(&msg), sizeof(msg)) ;
  else
    PR_log_msg (ERR, "PR_send_timeout: No timekeeper!") ;
}

void
PR_sleep (const Ushort interval)
{
  TIMER_MSG  msg ;

  msg.command.mtype = SLEEP_REQUEST ;
  msg.command.sub_command = 0 ;
  msg.command.reply_to = This_Proc->msgq ;
  msg.interval = interval;

  #ifdef TIMER_DEBUG
    PR_log_msg (FILE_ONLY, "sleep_req_msg being sent, Dur=%d", interval) ;
  #endif

  if (Global->Timekeeper_ID != -1)
  {
    PR_send_mesg (Global->Timekeeper_ID, (LINUX_MSG_BUF *)(&msg), sizeof(msg)) ;
    PR_read_mesg (This_Proc->msgq, (LINUX_MSG_BUF *)(&msg), sizeof(msg), SLEEP_NOTIFICATION) ;

    #ifdef TIMER_DEBUG
      PR_log_msg (FILE_ONLY, "Sleep notification received") ;
    #endif
  }
  else
    PR_log_msg (ERR, "PR_sleep: No timekeeper!") ;
}

#endif
