/*
 * stringfn.c
 ******************************************************************************
 * Copyright (C) 2002 Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ******************************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>           /* for calls to vsnprintf etc */
#include <string.h>           /* for strncpy, memcpy, memset, ... */
#include <ctype.h>            /* for toupper, ... */
#include <unistd.h>           /* for getcwd, ... */
#include <stdarg.h>           /* for calls to vsnprintf, ... */

#include "types.h"
#include "stringfn.h"

/******************************************************************************
 * This area defines a bunch of things relating to files.
 *
 * The extract_filename routine has some definitions relating to MSDOS. If you
 * want to use this routine under DOS, these will make it easier for you.
 * The train software however has not been written to run in the DOS / WINDOWS
 * environment.
 */

char	Directory [FILE_NAME_SIZE] ;
char    *Filename, *Extension ;
char	Filename_String  [FILE_NAME_SIZE] ;
char	Extension_String [FILE_NAME_SIZE] ;
char	FileExtension [FILE_NAME_SIZE] ;

/*******************************************************************************
 * STR_extract_filename
 *
 * takes a system file specification and sets the global variables <Directory>,
 * <Filename> and <Extension>.
 *
 * eg.... /top/second/third/file.ext
 *
 * Directory == "/top/second/third/"
 * Filename == "file"
 * Extension == ".ext"
 * FileExtension == "file.ext"
 */

void
STR_extract_filename (const char *fspec)
{
  unsigned long	i, last_file, last_ext ;

  strncpy (Directory, fspec, FILE_NAME_SIZE-1) ;
  strncpy (Filename_String, fspec, FILE_NAME_SIZE-1) ;
  strncpy (Extension_String, fspec, FILE_NAME_SIZE-1) ;

  last_file = last_ext = 0 ;

  for (i=0 ; fspec[i] != '\0' ; i++)
  {
    if (fspec[i] == DIR_DELIMITER)
      last_file = i+1 ;

    if ((fspec[i] == EXT_DELIMITER) && (i >= 1) && (fspec[i-1] != DIR_DELIMITER))
      last_ext = i ;
  }

  if ((last_ext < last_file) || (last_ext == 0))
  {
    last_ext = 0 ;
    Extension_String [0] = '\0' ;
  }
  else
    Filename_String [last_ext] = '\0' ;

  Directory [last_file] = '\0' ;
  Filename = &Filename_String [last_file] ;
  Extension = &Extension_String [last_ext] ;
  sprintf (FileExtension, "%s%s", Filename, Extension) ;
}

/*******************************************************************************
 * STR_remove_special_characters()
 *
 * This routine prepares a file reference for inclusion in a web page. Special
 * characters are converted to their %xx equivalent
 */

void
STR_remove_special_characters (char *dest, const char *string, const int length, const int plus_space)
{
  const char *i ;
  char *o ;
  int  n ;

  n = 0 ;
  i = string ;
  o = dest ;

  for (; *i && (n < length) ;)
  {
    if ((o == dest) && (*i <= ' ')) /* skip leading white space */
    {
      i++ ;
      continue ;
    }

    if ((*i <= ' ') && (*(i+1) <= ' ')) /* skip multiple white space */
    {
      i++ ;
      continue ;
    }

    if ((*i <= ' ') && plus_space)
    {
      i++ ;
      *o++ = '+' ;
      n++ ;
      continue ;
    }

    if ((*i > ' ') && !((*i == '%') ||
			(*i == '&') ||
			(*i == '#') ||
			(*i == '\'') ||
			(*i == '\"') ||
			(*i == '?') ||
			(*i == '<') ||
			(*i == '>') ||
			(*i == '\\')))
    {
      *o++ = *i++ ;
      n++ ;
      continue ;
    }

    snprintf (&dest[n], FILE_NAME_SIZE-2-n, "%%%02X", *i++) ;
    o = &dest[strlen(dest)] ;
    n += 3 ;
  }

  *o = '\0' ;
}

void
STR_restore_special_characters (char *dest, const char *string)
{
  char *d ;
  const char *s ;
  int i, j ;

  for (d=dest, s=string ; *s != '\0' ; s++)
  {
    if (*s == '%')
    {
      if (sscanf (s+1, "%02x%n", &i, &j) == 1)
      {
	s += j ;

	if (i == 0x0d)
	  continue ;

	if (i == 0x0a)
	  i = ' ' ;

	*d++ = (char)i ;
      }
      else
	*d++ = *s ;

      continue ;
    }

    if (*s == '+')
    {
      *d++ = ' ' ;
      continue ;
    }

    *d++ = *s ;
  }

  *d = '\0' ;
}

/******************************************************************************
 * STR_snaprintf
 *
 * same as snprintf but append to the current string
 */

int
STR_snaprintf (char *string, const int maxlen, const char *fmt, ...)
{
  int     n ;
  va_list args ;

  va_start (args, fmt) ;
  n = strlen (string) ;

  if (n >= maxlen)
    return 0 ;

  return vsnprintf (&string [n], maxlen-n-2, fmt, args) ;
}
