/*
 * test_process.c
 */

#include <unistd.h>		// for calls to ioperm, sleep, usleep
#include <stdlib.h>
#include <string.h>		// for strncpy, memcpy, memset, ...
#include <ctype.h>		// for toupper, ...
#include <sched.h>		// for schedular types
#include <time.h>
#include <stdio.h>

/* Uncomment the following if your compiler complains about missing information */
/* #include <asm/io.h>           for inb, outb */


#include "types.h"	/* for definitions of common types */
#include "parse.h"	/* routines to parse command line parameters */
#include "process.h"	/* routines to support multi processing */
#include "readline.h"	/* text I/O */
#include "stringfn.h"	/* manipulation of strings */

/*******************************************************************************
 * Here are other definitons which don't fall into the above categories.
 */

char	Logfilename [FILE_NAME_SIZE] ;

/******************************************************************************
 */

#define NUM_TEST_ITERATIONS (5)

int
timertest_process (const char *param)
{
  int i ;
  TIMER_MSG msg ;

  #ifdef BASIC_DEBUG
    ANNOUNCE_PROCESS() ;
    PR_log_msg (NORM, "Called with parameter \"%s\"", param) ;
  #endif

  PR_set_process_priority (SCHED_RR, 1) ;
  PR_log_msg (SCREEN, "This message is always written to screen, even with '-l' option") ;
  PR_log_msg (FILE_ONLY, "This message only written to a file and if the '-l' option was specified") ;
  PR_log_msg (NORM, "This message is written to screen or file depending on '-l'") ;
  PR_log_msg (ERR, "This message is written to stderr") ;

  for (i=1 ; i <= NUM_TEST_ITERATIONS ; i++)
  { 
    PR_log_msg (SCREEN, "Iteration %d", i) ;

    // NOTE: the last parameter in the next call would normally be a pointer
    // to a local variable. Things will get hairy if you use integers in more
    // than one process!
    PR_send_timeout_req_msg (This_Proc->msgq, TICK_FREQ*(NUM_TEST_ITERATIONS+i), (const void *)i) ;
    PR_sleep (8) ;
  }

  for (;;)
  {
    PR_read_mesg (This_Proc->msgq, (LINUX_MSG_BUF *)(&msg), sizeof(msg), 0) ;

    switch (msg.command.mtype)
    {
      case TIMEOUT_NOTIFICATION:
	PR_log_msg (SCREEN, "Received timeout from iteration %d", msg.command.sub_command) ;
	break ;

      default:
	PR_log_msg (ERR, "Bad msg recd: %s, Subcmd: %#0x, From %s", PR_decode_message_type (msg.command.mtype), msg.command.sub_command, PR_proc_from_msgq (msg.command.reply_to)) ;
	continue ;
    }

    if (msg.command.sub_command == NUM_TEST_ITERATIONS)
      break ;
  }

  PR_log_msg (SCREEN, "Test complete") ;
  return 100 ; // ret status 1 is a silent exit, > 1 displays the details.
}

/******************************************************************************
 */

#define SLEEP_GRANULARITY 15
#define SEMTEST_NAME	"SemTest1"

int
semtest_process (const char *param)
{
  short i ;
  int *sem_id ;
  int *shm_ptr ;

  #ifdef BASIC_DEBUG
    ANNOUNCE_PROCESS() ;
    PR_log_msg (NORM, "Called with parameter \"%s\"", param) ;
  #endif

  PR_set_process_priority (SCHED_RR, 0) ;

  // All the semtest processes will share the semaphore created for semtest1
  // It is initially created in the locked state. Semtest 1 will need to
  // open the semaphore up in order for anyone to use it.

  if ((i=PR_proc_from_name (param)) < 0)
  {
    PR_log_msg (ERR, "Couldn't find process id for name %s", param) ;
    return (1) ;
  }

  sem_id = &Global->Processes[i].sem_id ;
  shm_ptr = (int *)Global->Processes[i].shm ;
  srand ((unsigned int)&This_Proc->sem_id + (unsigned int)time(NULL)) ;

  if (strcmp (This_Proc->name, SEMTEST_NAME) == 0)
  {
    PR_unlock_sem (sem_id) ;
    *shm_ptr = 0 ;
  }

  for (i=0 ; i < 5 ; i++)
  {
    // sleep for a random time
    PR_sleep (((rand() % SLEEP_GRANULARITY) * TICK_FREQ / 3) + 1) ;

    PR_lock_sem (sem_id) ;
    (*shm_ptr)++ ;
    PR_log_msg (SCREEN, "Got the semaphore. Shm_Count: %d", *shm_ptr) ;
    PR_sleep (1) ;
    PR_log_msg (FILE_ONLY, "Releasing the semaphore") ;
    PR_unlock_sem (sem_id) ;
  }

  PR_log_msg (SCREEN, "Test complete") ;
  return 99 ; // ret status 1 is a silent exit, > 1 displays the details.
}

/******************************************************************************
 */

OPTION opt[] =
{
  // abbr, desc,					type,    var,	defint, defstring, min, max/len, hidden

  DECLARE_PROCESS_OPTIONS
  {"l", "Save log messages in a file rather than stdout", FILENAME, &Logfilename, 0, "logfile.txt", 0, FILE_NAME_SIZE, 0},
  DECLARE_PARSE_OPTIONS
} ;

/******************************************************************************
 */

int
main (int argc, char **argv)
{
  Error = PA_parse (&argc, &argv) ;
  PROCESS_PARSE_OPTIONS(argv[0]) ;
  PROCESS_FINAL_OPTIONS (Error, argv[0], "test process.c") ;

  if (PR_create_system_structures (Logfilename))
  {
    //			    (*name,	   &proc(),		&param, shm_size, io, msgq, sem_id)
    PR_register_new_process ("TimerTest",  timertest_process,	"param1",     0, 0, 1, 0) ;
    PR_register_new_process (SEMTEST_NAME, semtest_process,	SEMTEST_NAME, 4, 0, 1, 1) ;
    PR_register_new_process ("SemTest2",   semtest_process,	SEMTEST_NAME, 0, 0, 1, 0) ;
    PR_register_new_process ("SemTest3",   semtest_process,	SEMTEST_NAME, 0, 0, 1, 0) ;
  }

  if (Error == 0)
    Error = PR_launch_process (argv[0], argc, argv) ;

  exit (0) ;
}
