/******************************************************************************
 * setcirc.c
 *
 * Digital Setting Circles code
 *
 * In preparing this code, I drew heavily from algorithms first published in
 * the February 1989 issue of "Sky and Telescope" in an article by T.Taki
 * concerning the pointing of Dobsonian Telescopes. The software appearing in
 * this article may be obtained from the following link:
 * http://skyandtelescope.com/resources/software/programs/taki.bas
 *
 * I also drew from a translation of the T.Taki code into the C language by
 * Mel Bartels. His translation may be found on his web site within a zip
 * file of the source code. You may find it at:
 * http://www.bbastrodesigns.com/cot/cot.html and
 * http://zebu.uoregon.edu/~mbartels/dnld/source_code.zip (convert.c)
 *
 * I have chosen to release my adaptation of these algorithms under the GNU
 * public license.
 ******************************************************************************
 * Copyright (C) 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************/
// Assumptions:
//	The microprocessor running this application has an accurate real time
//	clock. Accuracy needs to be in terms of frequency rather than in terms
//	of having the right time set.
//
//	Preferably, the real time clock can be read to a fraction of a second.
//	It's OK if it can only be read to a second accuracy however this will
//	limit the accuracy of the setting circle to around 15".
//
//	The shaft encoders for the ALT and AZ axes should be as accurate as
//	possible - certainly accurate to an arc minute or hopefully better.
//
//	The routine to poll the shaft encoder value should convert the reading
//	to radians. This means that the size of the increment needs to be
//	accurately calculated for each telescope.
//
// To use:
//	(a) fill Z1, Z2 and Z3 with the telescope error values. 0.0=no error.
//		The three values are Kalman error coefficients. You can find
//		out how to calculate them by reading "Telescope Control" by
//		Trueblood and Genet. An alternate algorithm is provided in
//		Mel Bartel's source code zip file.
//		It's also OK to use the routines in the no-error mode.
//
//	(b) align on either two or three stars, filling STAR structs for each.
//		When the shaft encoders are polled to determine the telescope
//		coordinates, the system tick counter must also be polled and
//		the value stored in the STAR structure.
//
//	(c) call DSC_generate_transform_matrix (&star1, &star2, &star3)
//		For two star alignment, set star3 to (STAR *)0
//
//	(d) To determine RA/DEC from a known ALT/AZ/sys_ticks reading,
//		partially fill a STAR structure and then call
//		DSC_Alt_Az_from_Ra_Dec (&star_structure)
//
//	    To determine ALT/AZ from a known RA/DEC/sys_ticks reading,
//		partially fill a STAR structure and then call
//		DSC_Ra_Dec_from_Alt_Az (&star_structure)
//
//		The unfilled fields of the STAR structure will be filled out
//		upon return.
//
// Note:
//	It is feasible to perform a two star alignment and then, later in the
//	night, call DSC_generate_transform_matrix with a third star. In this
//	case, there is no need to reinitialise any matrices. Simply call
//	DSC_generate_transform_matrix with any three valid STAR structures.

#include <stdio.h>
#include <float.h> /* for limits on number sizes */
#include <math.h>  /* for all the trig functions */

#include "types.h"
#include "setcirc.h"

///////////////////////////////////////////////////////////////////////////////

MATRIX_3x3 Telescope ;
MATRIX_3x3 Actual ;
MATRIX_3x3 Transform_to_Ra_Dec ;
MATRIX_3x3 Transform_to_Alt_Az ;

double Z1, Z2, Z3 ;

#define DEGREES_PER_RAD	  (360.0 / 2.0 / M_PI)
#define SID_TO_SOLAR_CONSTANT (1.00273790935) // multiply sidereal time by this to get solar time

///////////////////////////////////////////////////////////////////////////////
// initialisation routines

void
DSC_fill_3x3_matrix (MATRIX_3x3 *matrix, const double value)
{
  Ushort i, j ;

  for (i=0 ; i<3 ; i++)
    for (j=0 ; j<3 ; j++)
      (*matrix)[i][j] = value ;
}

void
DSC_clear_matrices (void)
{
  DSC_fill_3x3_matrix (&Telescope, 0.0) ;
  DSC_fill_3x3_matrix (&Actual, 0.0) ;
  DSC_fill_3x3_matrix (&Transform_to_Ra_Dec, 0.0) ;
  DSC_fill_3x3_matrix (&Transform_to_Alt_Az, 0.0) ;
}

void
DSC_init_array_column (MATRIX_3x3 *dest, const Ushort col, MATRIX_1x3 *src)
{
  Ushort i ;

  for (i=0 ; i < 3 ; i++)
    (*dest)[i][col] = (*src)[i] ;
}

void
DSC_print_3x3_matrix (MATRIX_3x3 *matrix, const char *txt)
{
  Ushort i, j ;

  printf ("%s\n", txt) ;

  for (i=0 ; i<3 ; i++)
  {
    printf ("|") ;

    for (j=0 ; j<3 ; j++)
      printf ("%9.5f", (*matrix)[i][j]) ;

    printf (" |\n") ;
  }    

  printf ("\n") ;
}

void
DSC_print_1x3_matrix (MATRIX_1x3 *matrix, const char *txt)
{
  Ushort i ;

  printf ("%s\n", txt) ;

  for (i=0 ; i<3 ; i++)
    printf ("|%9.5f |\n", (*matrix)[i]) ;

  printf ("\n") ;
}

void
DSC_show_star (const STAR *star, const char *txt)
{
  printf ("%s\n", txt) ;
  printf( "Ra: %8.4f deg, %8.4f rad\n", star->ra * DEGREES_PER_RAD, star->ra) ;
  printf( "Dec: %8.4f deg, %8.4f rad\n", star->dec * DEGREES_PER_RAD, star->dec) ;
  printf( "Alt: %8.4f deg, %8.4f rad\n", star->alt * DEGREES_PER_RAD, star->alt) ;
  printf( "Az: %8.4f deg, %8.4f rad\n", star->az * DEGREES_PER_RAD, star->az) ;
  printf( "SidT: %8.4f deg, %8.4f rad\n\n", DSC_calc_sidereal_time (star) * DEGREES_PER_RAD, DSC_calc_sidereal_time (star)) ;
}

///////////////////////////////////////////////////////////////////////////////
// matrix manipulation routines

double
DSC_determinant (MATRIX_3x3 *A)
{
  return (*A)[0][0] * ((*A)[1][1]*(*A)[2][2] - (*A)[1][2]*(*A)[2][1])
       - (*A)[0][1] * ((*A)[1][0]*(*A)[2][2] - (*A)[1][2]*(*A)[2][0])
       + (*A)[0][2] * ((*A)[1][0]*(*A)[2][1] - (*A)[1][1]*(*A)[2][0]) ;
}

void
DSC_copy_3x3_matrix (MATRIX_3x3 *to, MATRIX_3x3 *from)
{
  Ushort i, j ;

  for (i=0 ; i < 3 ; i++)
    for (j=0 ; j < 3 ; j++)
      (*to)[i][j] = (*from)[i][j] ;
}

void
DSC_multiply_3x3_with_3x3 (MATRIX_3x3 *result, MATRIX_3x3 *A, MATRIX_3x3 *B)
{
  // calculate Result = A * B
  // Scan across the rows of A and down the columns of B summing the products

  Ushort i, j, k ;

  DSC_fill_3x3_matrix (result, 0.0) ;

  for (i=0 ; i<3 ; i++)
    for (j=0 ; j<3 ; j++)
      for (k=0 ; k<3 ; k++)
	(*result)[i][j] += (*A)[i][k] * (*B)[k][j] ;
}

void
DSC_multiply_3x3_with_1x3 (MATRIX_1x3 *result, MATRIX_3x3 *A, MATRIX_1x3 *B)
{
  // calculate Result = A * B
  // Scan across the rows of A and down the column of B summing the products

  Ushort i, j ;

  for (i=0 ; i < 3 ; i++)
      (*result)[i] = 0.0 ;

  for (i=0 ; i < 3 ; i++)
    for (j=0 ; j < 3 ; j++)
      (*result)[i] += (*A)[i][j] * (*B)[j] ;
}

void
DSC_transpose_3x3 (MATRIX_3x3 *A)
{
  // Transpose the rows and columns of the matrix

  Ushort i, j ;
  double tmp ;

  for (i=0 ; i<3 ; i++)
    for (j=0 ; j<3 ; j++)
    {
      tmp = (*A)[i][j] ;
      (*A)[i][j] = (*A)[j][i] ;
      (*A)[j][i] = tmp ;
    }
}

///////////////////////////////////////////////////////////////////////////////
// other common conversion routines

void
DSC_spherical_to_rect (MATRIX_1x3 *result, const STAR *star)
{
  double x, y ;

  x = (2.0 * M_PI) - star->az ;
  y = star->alt + Z3 ;

  if ((Z1 != 0.0) || (Z2 != 0.0))
  {
    (*result)[0] = cos(x)*cos(y)*cos(Z2) - sin(x)*cos(Z1)*sin(Z2) + sin(x)*sin(y)*sin(Z1)*cos(Z2) ;
    (*result)[1] = sin(x)*cos(y)*cos(Z2) + cos(x)*cos(Z1)*sin(Z2) - cos(x)*sin(y)*sin(Z1)*cos(Z2) ;
    (*result)[2] = sin(y)*cos(Z1)*cos(Z2) + sin(Z1)*sin(Z2) ;
  }
  else
  {
    (*result)[0] = cos(x) * cos(y) ;
    (*result)[1] = sin(x) * cos(y) ;
    (*result)[2] = sin(y) ;
  }
}

double
DSC_calc_sidereal_time (const STAR *star)
{
  return (double)(star->sys_ticks) *
		((2.0 * M_PI) /
		(SID_TO_SOLAR_CONSTANT * (double)(TICK_FREQ) * (double)(SECONDS_IN_DAY))) ;
}

void
DSC_init_star (const STAR *star, const Ushort col)
{
  MATRIX_1x3 vector ;
  double sid_time ;

  sid_time = DSC_calc_sidereal_time (star) ;
  vector[0] = cos(star->dec)*cos(star->ra - sid_time) ;
  vector[1] = cos(star->dec)*sin(star->ra - sid_time) ;
  vector[2] = sin(star->dec) ;

  DSC_init_array_column (&Actual, col, &vector) ;
  DSC_spherical_to_rect (&vector, star) ;
  DSC_init_array_column (&Telescope, col, &vector) ;
}

void
DSC_kludge_third_star (MATRIX_3x3 *matrix)
{
  double a, b, c ;
  double length ;
  MATRIX_1x3 vector ;

  a = (*matrix)[1][0]*(*matrix)[2][1] - (*matrix)[2][0]*(*matrix)[1][1] ;
  b = (*matrix)[2][0]*(*matrix)[0][1] - (*matrix)[0][0]*(*matrix)[2][1] ;
  c = (*matrix)[0][0]*(*matrix)[1][1] - (*matrix)[1][0]*(*matrix)[0][1] ;

  length = sqrt (a*a + b*b + c*c) ;

  if (length == 0.0)
  {
    vector[0] = vector[1] = vector[2] = DBL_MAX ;
  }
  else
  {
    vector[0] = a / length ;
    vector[1] = b / length ;
    vector[2] = c / length ;
  }

  DSC_init_array_column (matrix, 2, &vector) ;
}

///////////////////////////////////////////////////////////////////////////////

void
DSC_generate_transform_matrices (const STAR *star1, const STAR *star2, const STAR *star3)
{
  double determinant ;
  Ushort i, j, k ;
  MATRIX_3x3 tmp1, tmp2 ;

  // initialise the Actual and Telescope matrices

  DSC_clear_matrices () ;
  DSC_init_star (star1, 0) ;
  DSC_init_star (star2, 1) ;

  if (star3 == (STAR *)0)
  {
    DSC_kludge_third_star (&Actual) ;
    DSC_kludge_third_star (&Telescope) ;
  }
  else
    DSC_init_star (star3, 2) ;

  // Now calculate the transform matrices

  determinant = DSC_determinant (&Actual) ;

  if (determinant == 0.0)
    DSC_fill_3x3_matrix (&tmp2, DBL_MAX) ;
  else
  {
    for (i=0 ; i<3 ; i++)
    {
      DSC_copy_3x3_matrix (&tmp1, &Actual) ;

      for (j=0 ; j<3 ; j++)
      {
	for (k=0 ; k<3 ; k++)
	  tmp1[k][i] = (k==j) ? 1.0 : 0.0 ;

	tmp2[i][j] = DSC_determinant(&tmp1) / determinant ;
      }
    }
  }

  DSC_multiply_3x3_with_3x3 (&Transform_to_Alt_Az, &Telescope, &tmp2) ;
  DSC_transpose_3x3 (&Transform_to_Alt_Az) ;

  determinant = DSC_determinant (&Transform_to_Alt_Az) ;

  if (determinant == 0.0)
    DSC_fill_3x3_matrix (&Transform_to_Ra_Dec, DBL_MAX) ;
  else
  {
    for (i=0 ; i<3 ; i++)
    {
      DSC_copy_3x3_matrix (&tmp1, &Transform_to_Alt_Az) ;

      for (j=0 ; j<3 ; j++)
      {
	for (k=0 ; k<3 ; k++)
	  tmp1[k][i] = (k==j) ? 1.0 : 0.0 ;

	Transform_to_Ra_Dec[i][j] = DSC_determinant(&tmp1) / determinant ;
      }
    }
  }
}

///////////////////////////////////////////////////////////////////////////////

void
DSC_Ra_Dec_from_Alt_Az (STAR *star)
{
  // The RA/DEC is calculated by converting the ALT/AZ to rectangular coordinates
  // and multiplying by the transform matrix. The resulting rectangular transform
  // must be converted back to spherical to determine the RA/DEC. Incremental
  // sidereal time must then be added to the RA.

  MATRIX_1x3 vector, result ;

  DSC_spherical_to_rect (&vector, star) ;
  DSC_multiply_3x3_with_1x3 (&result, &Transform_to_Ra_Dec, &vector) ;

  star->dec = atan2 (result[2], sqrt(result[0]*result[0] + result[1]*result[1])) ;

  star->ra = atan2 (result[1], result[0]) ;
  star->ra += DSC_calc_sidereal_time (star) ;

  if (star->ra < 0) // convert from -PI..PI to 0..2*PI
    star->ra += M_PI + M_PI ;
}

void
DSC_Alt_Az_from_Ra_Dec (STAR *star)
{
  MATRIX_1x3 vector, result ;
  double normalised_ra ;
  double az, alt ;

  normalised_ra = star->ra - DSC_calc_sidereal_time (star) ;

  vector[0] = cos(star->dec) * cos(normalised_ra) ;
  vector[1] = cos(star->dec) * sin(normalised_ra) ;
  vector[2] = sin(star->dec) ;

  DSC_multiply_3x3_with_1x3 (&result, &Transform_to_Alt_Az, &vector) ;

  alt = atan2 (result[2], sqrt(result[0]*result[0] + result[1]*result[1])) ;
  az = atan2 (result[1], result[0]) ;

  // account for mount errors

  if ((Z1 != 0.0) || (Z2 != 0.0))
  {
    result[0] = cos(az)*cos(alt) + sin(az)*sin(Z2)*cos(Z1)/cos(Z2) - sin(az)*sin(alt)*sin(Z1) ;
    result[1] = sin(az)*cos(alt) - cos(az)*sin(Z2)*cos(Z1)/cos(Z2) + cos(az)*sin(alt)*sin(Z1) ;
    result[2] = (sin(alt) - sin(Z1)*sin(Z2)) / (cos(Z1)*cos(Z2)) ;
  }
  else
  {
    result[0] = cos(az)*cos(alt) ;
    result[1] = sin(az)*cos(alt) ;
    result[2] = sin(alt) ;
  }

  star->alt = atan2 (result[2], sqrt(result[0]*result[0] + result[1]*result[1])) - Z3 ;
  star->az = -atan2 (result[1], result[0]) ;

  if (star->az < 0) // convert from -PI..PI to 0..2*PI
    star->az += M_PI + M_PI ;
}
