/******************************************************************************
 * test_setcirc.c
 *
 * Digital Setting Circles test code
 ******************************************************************************
 * In preparing this code, I drew heavily from algorithms first published in
 * the February 1989 issue of "Sky and Telescope" in an article by T.Taki
 * concerning the pointing of Dobsonian Telescopes. The software appearing in
 * this article may be obtained from the following link:
 * http://skyandtelescope.com/resources/software/programs/taki.bas
 *
 * I also drew from a translation of the T.Taki code into the C language by
 * Mel Bartels. His translation may be found on his web site within a zip
 * file of the source code. You may find it at:
 * http://www.bbastrodesigns.com/cot/cot.html and
 * http://zebu.uoregon.edu/~mbartels/dnld/source_code.zip (convert.c)
 *
 * I have chosen to release my adaptation of these algorithms under the GNU
 * public license.
 ******************************************************************************
 * Copyright (C) 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************/
// Assumptions:
//	The microprocessor running this application has an accurate real time
//	clock. Accuracy needs to be in terms of frequency rather than in terms
//	of having the right time set.
//
//	Preferably, the real time clock can be read to a fraction of a second.
//	It's OK if it can only be read to a second accuracy however this will
//	limit the accuracy of the setting circle to around 15".
//
//	The shaft encoders for the ALT and AZ axes should be as accurate as
//	possible - certainly accurate to an arc minute or hopefully better.
//
//	The routine to poll the shaft encoder value should convert the reading
//	to radians. This means that the size of the increment needs to be
//	accurately calculated for each telescope.
//
// To use:
//	(a) fill Z1, Z2 and Z3 with the telescope error values. 0.0=no error.
//		The three values are Kalman error coefficients. You can find
//		out how to calculate them by reading "Telescope Control" by
//		Trueblood and Genet. An alternate algorithm is provided in
//		Mel Bartel's source code zip file.
//		It's also OK to use the routines in the no-error mode.
//
//	(b) align on either two or three stars, filling STAR structs for each.
//		When the shaft encoders are polled to determine the telescope
//		coordinates, the system tick counter must also be polled and
//		the value stored in the STAR structure.
//
//	(c) call DSC_generate_transform_matrix (&star1, &star2, &star3)
//		For two star alignment, set star3 to (STAR *)0
//
//	(d) To determine RA/DEC from a known ALT/AZ/sys_ticks reading,
//		partially fill a STAR structure and then call
//		DSC_Alt_Az_from_Ra_Dec (&star_structure)
//
//	    To determine ALT/AZ from a known RA/DEC/sys_ticks reading,
//		partially fill a STAR structure and then call
//		DSC_Ra_Dec_from_Alt_Az (&star_structure)
//
//		The unfilled fields of the STAR structure will be filled out
//		upon return.
//
// Note:
//	It is feasible to perform a two star alignment and then, later in the
//	night, call DSC_generate_transform_matrix with a third star. In this
//	case, there is no need to reinitialise any matrices. Simply call
//	DSC_generate_transform_matrix with any three valid STAR structures.

#include <stdio.h>
#include <float.h> /* for limits on number sizes */
#include <math.h>  /* for all the trig functions */

#include "types.h"
#include "setcirc.h"

int
main()
{
  STAR star1, star2, star3 ;

  Z1 = -0.04 / DEGREES_PER_RAD ; // These are some imaginary error constants that
  Z2 = 0.4 / DEGREES_PER_RAD ;   // work with the examples below. Normally, you'd
  Z3 = -1.63 / DEGREES_PER_RAD ; // calculate them for each telescope. OR you could
				 // simply set them to zero and ignore mount errors

  star1.ra = 79.172 / DEGREES_PER_RAD ;
  star1.dec = 45.998 / DEGREES_PER_RAD ;
  star1.alt = 39.9 / DEGREES_PER_RAD ;
  star1.az = (360-39.9) / DEGREES_PER_RAD ;
  star1.sys_ticks = 9.826831508 * SID_TO_SOLAR_CONSTANT * TICK_FREQ * SECONDS_IN_DAY / 360.0 ;
  DSC_show_star (&star1, "star 1 at init") ;

  star2.ra = 37.96 / DEGREES_PER_RAD ;
  star2.dec = 89.264 / DEGREES_PER_RAD ;
  star2.alt = 36.2 / DEGREES_PER_RAD ;
  star2.az = (360-94.6) / DEGREES_PER_RAD ;
  star2.sys_ticks = 10.10258443 * SID_TO_SOLAR_CONSTANT * TICK_FREQ * SECONDS_IN_DAY / 360.0 ;
  DSC_show_star (&star2, "star 2 at init") ;

  DSC_generate_transform_matrices(&star1, &star2, (STAR *)0) ;

  star3.ra = 326.05 / DEGREES_PER_RAD ;
  star3.dec = 9.88 / DEGREES_PER_RAD ;
  star3.alt = 0 / DEGREES_PER_RAD ;
  star3.az = (360-0) / DEGREES_PER_RAD ;
  star3.sys_ticks = 11.78217043 * SID_TO_SOLAR_CONSTANT * TICK_FREQ * SECONDS_IN_DAY / 360.0 ;
  DSC_Alt_Az_from_Ra_Dec (&star3) ;
  printf ("Should be Alt: 42.16, Az: %7.3f\n", 360-202.54) ;
  DSC_show_star (&star3, "star #1 after alt_az calc") ;
  DSC_Ra_Dec_from_Alt_Az (&star3) ;
  DSC_show_star (&star3, "star #1 after ra_dec calc") ;


  star3.ra = 71.53 / DEGREES_PER_RAD ;
  star3.dec = 17.07 / DEGREES_PER_RAD ;
  star3.alt = 0 / DEGREES_PER_RAD ;
  star3.az = (360-0) / DEGREES_PER_RAD ;
  star3.sys_ticks = 15.54243759 * SID_TO_SOLAR_CONSTANT * TICK_FREQ * SECONDS_IN_DAY / 360.0 ;
  DSC_Alt_Az_from_Ra_Dec (&star3) ;
  printf ("Should be Alt: 40.31, Az: %7.3f\n", 360-359.98) ;
  DSC_show_star (&star3, "star #2 after alt_az calc") ;
  DSC_Ra_Dec_from_Alt_Az (&star3) ;
  DSC_show_star (&star3, "star #2 after ra_dec calc") ;


  star3.ra = 0 / DEGREES_PER_RAD ;
  star3.dec = 0 / DEGREES_PER_RAD ;
  star3.alt = 35.5 / DEGREES_PER_RAD ;
  star3.az = (360-24.1) / DEGREES_PER_RAD ;
  star3.sys_ticks = 18.02421392 * SID_TO_SOLAR_CONSTANT * TICK_FREQ * SECONDS_IN_DAY / 360.0 ;
  DSC_Ra_Dec_from_Alt_Az (&star3) ;
  printf ("Should be Ra: 87.99, Dec: 32.51\n") ;
  DSC_show_star (&star3, "star #3 after ra_dec calc") ;
  DSC_Alt_Az_from_Ra_Dec (&star3) ;
  DSC_show_star (&star3, "star #3 after alt_az calc") ;

  return 0 ;
}
