/*
 * astro_data.c
 ******************************************************************************
 * This file contains data structures that hold algorithmic data which will be
 * used in various calculations.
 ******************************************************************************
 * Copyright (C) 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ******************************************************************************
 */

#include "types.h"
#include "astro.h"

// The following table holds prefixes used in calculation of the obliquity
//  D,  M,  m,  F,  O,    del psi,        del e
// refer to pp145-146
const struct _table22A Nutation[] =
{
  { 0,  0,  0,  0,  1, -171996, -1742, 92025,  89},
  {-2,  0,  0,  2,  2,  -13187,   -16,  5736, -31},
  { 0,  0,  0,  2,  2,   -2274,    -2,   977,  -5},
  { 0,  0,  0,  0,  2,    2062,     2,  -895,   5},
  { 0,  1,  0,  0,  0,    1426,   -34,    54,  -1},
  { 0,  0,  1,  0,  0,     712,     1,    -7,   0},
  {-2,  1,  0,  2,  2,    -517,    12,   224,  -6},
  { 0,  0,  0,  2,  1,    -386,    -4,   200,   0},
  { 0,  0,  1,  2,  2,    -301,     0,   129,  -1},
  {-2, -1,  0,  2,  2,     217,    -5,   -95,   3},
  {-2,  0,  1,  0,  0,    -158,     0,     0,   0},
  {-2,  0,  0,  2,  1,     129,     1,   -70,   0},
  { 0,  0, -1,  2,  2,     123,     0,   -53,   0},
  { 2,  0,  0,  0,  0,      63,     0,     0,   0},
  { 0,  0,  1,  0,  1,      63,     1,   -33,   0},
  { 2,  0, -1,  2,  2,     -59,     0,    26,   0},
  { 0,  0, -1,  0,  1,     -58,    -1,    32,   0},
  { 0,  0,  1,  2,  1,     -51,     0,    27,   0},
  {-2,  0,  2,  0,  0,      48,     0,     0,   0},
  { 0,  0, -2,  2,  1,      46,     0,   -24,   0},
  { 2,  0,  0,  2,  2,     -38,     0,    16,   0},
  { 0,  0,  2,  2,  2,     -31,     0,    13,   0},
  { 0,  0,  2,  0,  0,      29,     0,     0,   0},
  {-2,  0,  1,  2,  2,      29,     0,   -12,   0},
  { 0,  0,  0,  2,  0,      26,     0,     0,   0},
  {-2,  0,  0,  2,  0,     -22,     0,     0,   0},
  { 0,  0, -1,  2,  1,      21,     0,   -10,   0},
  { 0,  2,  0,  0,  0,      17,    -1,     0,   0},
  { 2,  0, -1,  0,  1,      16,     0,    -8,   0},
  {-2,  2,  0,  2,  2,     -16,     1,     7,   0},
  { 0,  1,  0,  0,  1,     -15,     0,     9,   0},
  {-2,  0,  1,  0,  1,     -13,     0,     7,   0},
  { 0, -1,  0,  0,  1,     -12,     0,     6,   0},
  { 0,  0,  2, -2,  0,      11,     0,     0,   0},
  { 2,  0, -1,  2,  1,     -10,     0,     5,   0},
  { 2,  0,  1,  2,  2,      -8,     0,     3,   0},
  { 0,  1,  0,  2,  2,       7,     0,    -3,   0},
  {-2,  1,  1,  0,  0,      -7,     0,     0,   0},
  { 0, -1,  0,  2,  2,      -7,     0,     3,   0},
  { 2,  0,  0,  2,  1,      -7,     0,     3,   0},
  { 2,  0,  1,  0,  0,       6,     0,     0,   0},
  {-2,  0,  2,  2,  2,       6,     0,    -3,   0},
  {-2,  0,  1,  2,  1,       6,     0,    -3,   0},
  { 2,  0, -2,  0,  1,      -6,     0,     3,   0},
  { 2,  0,  0,  0,  1,      -6,     0,     3,   0},
  { 0, -1,  1,  0,  0,       5,     0,     0,   0},
  {-2, -1,  0,  2,  1,      -5,     0,     3,   0},
  {-2,  0,  0,  0,  1,      -5,     0,     3,   0},
  { 0,  0,  2,  2,  1,      -5,     0,     3,   0},
  {-2,  0,  2,  0,  1,       4,     0,     0,   0},
  {-2,  1,  0,  2,  1,       4,     0,     0,   0},
  { 0,  0,  1, -2,  0,       4,     0,     0,   0},
  {-1,  0,  1,  0,  0,      -4,     0,     0,   0},
  {-2,  1,  0,  0,  0,      -4,     0,     0,   0},
  { 1,  0,  0,  0,  0,      -4,     0,     0,   0},
  { 0,  0,  1,  2,  0,       3,     0,     0,   0},
  { 0,  0, -2,  2,  2,      -3,     0,     0,   0},
  {-1, -1,  1,  0,  0,      -3,     0,     0,   0},
  { 0,  1,  1,  0,  0,      -3,     0,     0,   0},
  { 0, -1,  1,  2,  2,      -3,     0,     0,   0},
  { 2, -1, -1,  2,  2,      -3,     0,     0,   0},
  { 0,  0,  3,  2,  2,      -3,     0,     0,   0},
  { 2, -1,  0,  2,  2,      -3,     0,     0,   0},
  {0x7f}
} ;

// The following tables hold prefixes used to calculate the moon's position
// D,  M,  m,  F, sigma_l,  sigma_r
// refer to pp339-340
const struct Table47A Lon_And_Distance[] =
{
  {0,  0,  1,  0, 6288774, -20905355},
  {2,  0, -1,  0, 1274027,  -3699111},
  {2,  0,  0,  0,  658314,  -2955968},
  {0,  0,  2,  0,  213618,   -569925},
  {0,  1,  0,  0, -185116,     48888},
  {0,  0,  0,  2, -114332,     -3149},
  {2,  0, -2,  0,   58793,    246158},
  {2, -1, -1,  0,   57066,   -152138},
  {2,  0,  1,  0,   53322,   -170733},
  {2, -1,  0,  0,   45758,   -204586},
  {0,  1, -1,  0,  -40923,   -129620},
  {1,  0,  0,  0,  -34720,    108743},
  {0,  1,  1,  0,  -30383,    104755},
  {2,  0,  0, -2,   15327,     10321},
  {0,  0,  1,  2,  -12528,         0},
  {0,  0,  1, -2,   10980,     79661},
  {4,  0, -1,  0,   10675,    -34782},
  {0,  0,  3,  0,   10034,    -23210},
  {4,  0, -2,  0,    8548,    -21636},
  {2,  1, -1,  0,   -7888,     24208},
  {2,  1,  0,  0,   -6766,     30824},
  {1,  0, -1,  0,   -5163,     -8379},
  {1,  1,  0,  0,    4987,    -16675},
  {2, -1,  1,  0,    4036,    -12831},
  {2,  0,  2,  0,    3994,    -10445},
  {4,  0,  0,  0,    3861,    -11650},
  {2,  0, -3,  0,    3665,     14403},
  {0,  1, -2,  0,   -2689,     -7003},
  {2,  0, -1,  2,   -2602,         0},
  {2, -1, -2,  0,    2390,     10056},
  {1,  0,  1,  0,   -2348,      6322},
  {2, -2,  0,  0,    2236,     -9884},
  {0,  1,  2,  0,   -2120,      5751},
  {0,  2,  0,  0,   -2069,         0},
  {2, -2, -1,  0,    2048,     -4950},
  {2,  0,  1, -2,   -1773,      4130},
  {2,  0,  0,  2,   -1595,         0},
  {4, -1, -1,  0,    1215,     -3958},
  {0,  0,  2,  2,   -1110,         0},
  {3,  0, -1,  0,    -892,      3258},
  {2,  1,  1,  0,    -810,      2616},
  {4, -1, -2,  0,     759,     -1897},
  {0,  2, -1,  0,    -713,     -2117},
  {2,  2, -1,  0,    -700,      2354},
  {2,  1, -2,  0,     691,         0},
  {2, -1,  0, -2,     596,         0},
  {4,  0,  1,  0,     549,     -1423},
  {0,  0,  4,  0,     537,     -1117},
  {4, -1,  0,  0,     520,     -1571},
  {1,  0, -2,  0,    -487,     -1739},
  {2,  1,  0, -2,    -399,         0},
  {0,  0,  2, -2,    -381,     -4421},
  {1,  1,  1,  0,     351,         0},
  {3,  0, -2,  0,    -340,         0},
  {4,  0, -3,  0,     330,         0},
  {2, -1,  2,  0,     327,         0},
  {0,  2,  1,  0,    -323,      1165},
  {1,  1, -1,  0,     299,         0},
  {2,  0,  3,  0,     294,         0},
  {2,  0, -1, -2,       0,      8752},
  {0x7f}
} ;
    
// D,  M,  m,  F, sigma_b
// refer to pp341
const struct Table47B Lat[] =
{
  {0,  0,  0,  1, 5128122},
  {0,  0,  1,  1,  280602},
  {0,  0,  1, -1,  277693},
  {2,  0,  0, -1,  173237},
  {2,  0, -1,  1,   55413},
  {2,  0, -1, -1,   46271},
  {2,  0,  0,  1,   32573},
  {0,  0,  2,  1,   17198},
  {2,  0,  1, -1,    9266},
  {0,  0,  2, -1,    8822},
  {2, -1,  0, -1,    8216},
  {2,  0, -2, -1,    4324},
  {2,  0,  1,  1,    4200},
  {2,  1,  0, -1,   -3359},
  {2, -1, -1,  1,    2463},
  {2, -1,  0,  1,    2211},
  {2, -1, -1, -1,    2065},
  {0,  1, -1, -1,   -1870},
  {4,  0, -1, -1,    1828},
  {0,  1,  0,  1,   -1794},
  {0,  0,  0,  3,   -1749},
  {0,  1, -1,  1,   -1565},
  {1,  0,  0,  1,   -1491},
  {0,  1,  1,  1,   -1475},
  {0,  1,  1, -1,   -1410},
  {0,  1,  0, -1,   -1344},
  {1,  0,  0, -1,   -1335},
  {0,  0,  3,  1,    1107},
  {4,  0,  0, -1,    1021},
  {4,  0, -1,  1,     833},
  {0,  0,  1, -3,     777},
  {4,  0, -2,  1,     671},
  {2,  0,  0, -3,     607},
  {2,  0,  2, -1,     596},
  {2, -1,  1, -1,     491},
  {2,  0, -2,  1,    -451},
  {0,  0,  3, -1,     439},
  {2,  0,  2,  1,     422},
  {2,  0, -3, -1,     421},
  {2,  1, -1,  1,    -366},
  {2,  1,  0,  1,    -351},
  {4,  0,  0,  1,     331},
  {2, -1,  1,  1,     315},
  {2, -2,  0, -1,     302},
  {0,  0,  1,  3,    -283},
  {2,  1,  1, -1,    -229},
  {1,  1,  0, -1,     223},
  {1,  1,  0,  1,     223},
  {0,  1, -2, -1,    -220},
  {2,  1, -1, -1,    -220},
  {1,  0,  1,  1,    -185},
  {2, -1, -2, -1,     181},
  {0,  1,  2,  1,    -177},
  {4,  0, -2, -1,     176},
  {4, -1, -1, -1,     166},
  {1,  0,  1, -1,    -164},
  {4,  0,  1, -1,     132},
  {1,  0, -1, -1,    -119},
  {4, -1,  0, -1,     115},
  {2, -2,  0,  1,     107},
  {0x7f}
} ;

/******************************************************************************
 * The following (long list) of structures hold the constants which are used
 * for calculation of the planetary positions. Refer to Chapter 32 and appendix
 * III.
 */

const struct VSOP_87 Mercury_L0[] =
{
  {4.40250710144, 0.00000000000, 0.00000000000},
  {0.40989414976, 1.48302034194, 26087.90314157420},
  {0.05046294199, 4.47785489540, 52175.80628314840},
  {0.00855346843, 1.16520322351, 78263.70942472259},
  {0.00165590362, 4.11969163181, 104351.61256629678},
  {0.00034561897, 0.77930765817, 130439.51570787099},
  {0.00007583476, 3.71348400510, 156527.41884944518},
  {0.00003559740, 1.51202669419, 1109.37855209340},
  {0.00001726012, 0.35832239908, 182615.32199101939},
  {0.00001803463, 4.10333178410, 5661.33204915220}, //10
  {0.00001364682, 4.59918318745, 27197.28169366760},
  {0.00001589923, 2.99510417815, 25028.52121138500},
  {0.00001017332, 0.88031439040, 31749.23519072640},
  {0.00000714182, 1.54144865265, 24978.52458948080},
  {0.00000643759, 5.30266110787, 21535.94964451540},
  {0.00000404200, 3.28228847025, 208703.22513259359},
  {0.00000352441, 5.24156297101, 20426.57109242200},
  {0.00000343313, 5.76531885335, 955.59974160860},
  {0.00000339214, 5.86327765000, 25558.21217647960},
  {0.00000451137, 6.04989275289, 51116.42435295920}, //20
  {0.00000325335, 1.33674334780, 53285.18483524180},
  {0.00000259587, 0.98732428184, 4551.95349705880},
  {0.00000345212, 2.79211901539, 15874.61759536320},
  {0.00000272947, 2.49451163975, 529.69096509460},
  {0.00000234830, 0.26672118900, 11322.66409830440},
  {0.00000238793, 0.11343953378, 1059.38193018920},
  {0.00000264336, 3.91705094013, 57837.13833230060},
  {0.00000216645, 0.65987207348, 13521.75144159140},
  {0.00000183359, 2.62878670784, 27043.50288318280},
  {0.00000175965, 4.53636829858, 51066.42773105500}, //30
  {0.00000181629, 2.43413502466, 25661.30495069820},
  {0.00000208995, 2.09178234008, 47623.85278608960},
  {0.00000172643, 2.45200164173, 24498.83024629040},
  {0.00000142316, 3.36003948842, 37410.56723987860},
  {0.00000137942, 0.29098447849, 10213.28554621100},
  {0.00000118233, 2.78149786369, 77204.32749453338},
  {0.00000096860, 6.20398202740, 234791.12827416777},
  {0.00000125219, 3.72079804425, 39609.65458316560},
  {-76.0}
} ;

const struct VSOP_87 Mercury_L1[] =
{
  {26088.14706222746, 0.00000000000, 0.00000000000},
  {0.01126007832, 6.21703970996, 26087.90314157420},
  {0.00303471395, 3.05565472363, 52175.80628314840},
  {0.00080538452, 6.10454743366, 78263.70942472259},
  {0.00021245035, 2.83531934452, 104351.61256629678},
  {0.00005592094, 5.82675673328, 130439.51570787099},
  {0.00001472233, 2.51845458395, 156527.41884944518},
  {0.00000352244, 3.05238094403, 1109.37855209340},
  {0.00000388318, 5.48039225891, 182615.32199101939},
  {0.00000093540, 6.11791163931, 27197.28169366760}, //10
  {0.00000090579, 0.00045481669, 24978.52458948080},
  {0.00000102743, 2.14879173777, 208703.22513259359},
  {0.00000051941, 5.62107554052, 5661.33204915220},
  {0.00000044370, 4.57348500464, 25028.52121138500},
  {0.00000028070, 3.04195430989, 51066.42773105500},
  {0.00000022003, 0.86475371243, 955.59974160860},
  {0.00000027295, 5.09210138837, 234791.12827416777},
  {-76.0}
} ;

const struct VSOP_87 Mercury_L2[] =
{
  {0.00053049845, 0.00000000000, 0.00000000000},
  {0.00016903658, 4.69072300649, 26087.90314157420},
  {0.00007396711, 1.34735624669, 52175.80628314840},
  {0.00003018297, 4.45643539705, 78263.70942472259},
  {0.00001107419, 1.26226537554, 104351.61256629678},
  {0.00000378173, 4.31998055900, 130439.51570787099},
  {0.00000122998, 1.06868541052, 156527.41884944518},
  {0.00000038663, 4.08011610182, 182615.32199101939},
  {0.00000014898, 4.63343085810, 1109.37855209340},
  {0.00000011861, 0.79187646439, 208703.22513259359}, //10
  {-76.0}
} ;

const struct VSOP_87 Mercury_L3[] =
{
  {0.00000188077, 0.03466830117, 52175.80628314840},
  {0.00000142152, 3.12505452600, 26087.90314157420},
  {0.00000096877, 3.00378171915, 78263.70942472259},
  {0.00000043669, 6.01867965826, 104351.61256629678},
  {0.00000035395, 0.00000000000, 0.00000000000},
  {0.00000018045, 2.77538373991, 130439.51570787099},
  {0.00000006971, 5.81808665742, 156527.41884944518},
  {0.00000002556, 2.57014364454, 182615.32199101939},
  {-76.0}
} ;

const struct VSOP_87 Mercury_L4[] =
{
  {0.00000114078, 3.14159265359, 0.00000000000},
  {0.00000003247, 2.02848007619, 26087.90314157420},
  {0.00000001914, 1.41731803758, 78263.70942472259},
  {0.00000001727, 4.50137643801, 52175.80628314840},
  {0.00000001237, 4.49970181057, 104351.61256629678},
  {0.00000000645, 1.26591776986, 130439.51570787099},
  {-76.0}
} ;

const struct VSOP_87 Mercury_L5[] =
{
  {0.00000000877, 3.14159265359, 0.00000000000},
  {-76.0}
} ;

const struct VSOP_87 Mercury_B0[] =
{
  {0.11737528962, 1.98357498767, 26087.90314157420},
  {0.02388076996, 5.03738959685, 52175.80628314840},
  {0.01222839532, 3.14159265359, 0.00000000000},
  {0.00543251810, 1.79644363963, 78263.70942472259},
  {0.00129778770, 4.83232503961, 104351.61256629678},
  {0.00031866927, 1.58088495667, 130439.51570787099},
  {0.00007963301, 4.60972126348, 156527.41884944518},
  {0.00002014189, 1.35324164694, 182615.32199101939},
  {0.00000513953, 4.37835409309, 208703.22513259359},
  {0.00000207674, 4.91772564073, 27197.28169366760}, //10
  {0.00000208584, 2.02020294153, 24978.52458948080},
  {0.00000132013, 1.11908492283, 234791.12827416777},
  {0.00000100454, 5.65684734206, 20426.57109242200},
  {0.00000121395, 1.81271752059, 53285.18483524180},
  {-66.0}
} ;

const struct VSOP_87 Mercury_B1[] =
{
  {0.00429151362, 3.50169780393, 26087.90314157420},
  {0.00146233668, 3.14159265359, 0.00000000000},
  {0.00022675295, 0.01515366880, 52175.80628314840},
  {0.00010894981, 0.48540174006, 78263.70942472259},
  {0.00006353462, 3.42943919982, 104351.61256629678},
  {0.00002495743, 0.16051210665, 130439.51570787099},
  {0.00000859585, 3.18452433647, 156527.41884944518},
  {0.00000277503, 6.21020774184, 182615.32199101939},
  {0.00000086233, 2.95244391822, 208703.22513259359},
  {0.00000026133, 5.97708962692, 234791.12827416777}, //10
  {0.00000027696, 0.29068938889, 27197.28169366760},
  {-66.0}
} ;

const struct VSOP_87 Mercury_B2[] =
{
  {0.00011830934, 4.79065585784, 26087.90314157420},
  {0.00001913516, 0.00000000000, 0.00000000000},
  {0.00001044801, 1.21216540536, 52175.80628314840},
  {0.00000266213, 4.43418336532, 78263.70942472259},
  {0.00000170280, 1.62255638714, 104351.61256629678},
  {0.00000096300, 4.80023692017, 130439.51570787099},
  {0.00000044692, 1.60758267772, 156527.41884944518},
  {0.00000018316, 4.66904655377, 182615.32199101939},
  {0.00000006927, 1.43404888930, 208703.22513259359},
  {-66.0}
} ;

const struct VSOP_87 Mercury_B3[] =
{
  {0.00000235423, 0.35387524604, 26087.90314157420},
  {0.00000160537, 0.00000000000, 0.00000000000},
  {0.00000018904, 4.36275460261, 52175.80628314840},
  {0.00000006376, 2.50715381439, 78263.70942472259},
  {0.00000004580, 6.14257817571, 104351.61256629678},
  {0.00000003061, 3.12497552681, 130439.51570787099},
  {0.00000001732, 6.26642412058, 156527.41884944518},
  {-66.0}
} ;

const struct VSOP_87 Mercury_B4[] =
{
  {0.00000004276, 1.74579932115, 26087.90314157420},
  {0.00000001023, 3.14159265359, 0.00000000000},
  {-66.0}
} ;

const struct VSOP_87 Mercury_R0[] =
{
  {0.39528271652, 0.00000000000, 0.00000000000},
  {0.07834131817, 6.19233722599, 26087.90314157420},
  {0.00795525557, 2.95989690096, 52175.80628314840},
  {0.00121281763, 6.01064153805, 78263.70942472259},
  {0.00021921969, 2.77820093975, 104351.61256629678},
  {0.00004354065, 5.82894543257, 130439.51570787099},
  {0.00000918228, 2.59650562598, 156527.41884944518},
  {0.00000260033, 3.02817753482, 27197.28169366760},
  {0.00000289955, 1.42441936951, 25028.52121138500},
  {0.00000201855, 5.64725040350, 182615.32199101939}, //10
  {0.00000201499, 5.59227724202, 31749.23519072640},
  {0.00000141980, 6.25264202645, 24978.52458948080},
  {0.00000100144, 3.73435608689, 21535.94964451540},
  {-82.0}
} ;

const struct VSOP_87 Mercury_R1[] =
{
  {0.00217347739, 4.65617158663, 26087.90314157420},
  {0.00044141826, 1.42385543975, 52175.80628314840},
  {0.00010094479, 4.47466326316, 78263.70942472259},
  {0.00002432804, 1.24226083435, 104351.61256629678},
  {0.00001624367, 0.00000000000, 0.00000000000},
  {0.00000603996, 4.29303116561, 130439.51570787099},
  {0.00000152851, 1.06060779810, 156527.41884944518},
  {0.00000039202, 4.11136751416, 182615.32199101939},
  {-82.0}
} ;

const struct VSOP_87 Mercury_R2[] =
{
  {0.00003117867, 3.08231840296, 26087.90314157420},
  {0.00001245396, 6.15183317423, 52175.80628314840},
  {0.00000424822, 2.92583352960, 78263.70942472259},
  {0.00000136130, 5.97983925842, 104351.61256629678},
  {0.00000042175, 2.74936980629, 130439.51570787099},
  {0.00000021759, 3.14159265359, 0.00000000000},
  {0.00000012793, 5.80143162209, 156527.41884944518},
  {-82.0}
} ;

const struct VSOP_87 Mercury_R3[] =
{
  {0.00000032676, 1.67971635359, 26087.90314157420},
  {0.00000024166, 4.63403168997, 52175.80628314840},
  {0.00000012133, 1.38983781545, 78263.70942472259},
  {0.00000005140, 4.43915386930, 104351.61256629678},
  {0.00000001981, 1.20733880274, 130439.51570787099},
  {0.00000001460, 3.14159265359, 0.00000000000},
  {-82.0}
} ;

const struct VSOP_87 Venus_L0[] =
{
  {3.17614666774, 0.00000000000, 0.00000000000},
  {0.01353968419, 5.59313319619, 10213.28554621100},
  {0.00089891645, 5.30650048468, 20426.57109242200},
  {0.00005477201, 4.41630652531, 7860.41939243920},
  {0.00003455732, 2.69964470778, 11790.62908865880},
  {0.00002372061, 2.99377539568, 3930.20969621960},
  {0.00001317108, 5.18668219093, 26.29831979980},
  {0.00001664069, 4.25018935030, 1577.34354244780},
  {0.00001438322, 4.15745043958, 9683.59458111640},
  {0.00001200521, 6.15357115319, 30639.85663863300}, //10
  {0.00000761380, 1.95014702120, 529.69096509460},
  {0.00000707676, 1.06466707214, 775.52261132400},
  {0.00000584836, 3.99839884762, 191.44826611160},
  {0.00000769314, 0.81629615911, 9437.76293488700},
  {0.00000499915, 4.12340210074, 15720.83878487840},
  {0.00000326221, 4.59056473097, 10404.73381232260},
  {0.00000429498, 3.58642859752, 19367.18916223280},
  {0.00000326967, 5.67736583705, 5507.55323866740},
  {0.00000231937, 3.16251057072, 9153.90361602180},
  {0.00000179695, 4.65337915578, 1109.37855209340}, //20
  {0.00000128263, 4.22604493736, 20.77539549240},
  {0.00000155464, 5.57043888948, 19651.04848109800},
  {0.00000127907, 0.96209822685, 5661.33204915220},
  {0.00000105547, 1.53721191253, 801.82093112380},
  {-76.0}
} ;

const struct VSOP_87 Venus_L1[] =
{
  {10213.52943052898, 0.00000000000, 0.00000000000},
  {0.00095707712, 2.46424448979, 10213.28554621100},
  {0.00014444977, 0.51624564679, 20426.57109242200},
  {0.00000213374, 1.79547929368, 30639.85663863300},
  {0.00000151669, 6.10635282369, 1577.34354244780},
  {0.00000173904, 2.65535879443, 26.29831979980},
  {0.00000082233, 5.70234133730, 191.44826611160},
  {0.00000069734, 2.68136034979, 9437.76293488700},
  {0.00000052408, 3.60013087656, 775.52261132400},
  {0.00000038318, 1.03379038025, 529.69096509460}, //10
  {0.00000029633, 1.25056322354, 5507.55323866740},
  {0.00000025056, 6.10664792855, 10404.73381232260},
  {-76.0}
} ;

const struct VSOP_87 Venus_L2[] =
{
  {0.00054127076, 0.00000000000, 0.00000000000},
  {0.00003891460, 0.34514360047, 10213.28554621100},
  {0.00001337880, 2.02011286082, 20426.57109242200},
  {0.00000023836, 2.04592119012, 26.29831979980},
  {0.00000019331, 3.53527371458, 30639.85663863300},
  {0.00000009984, 3.97130221102, 775.52261132400},
  {0.00000007046, 1.51962593409, 1577.34354244780},
  {0.00000006014, 0.99926757893, 191.44826611160},
  {-76.0}
} ;

const struct VSOP_87 Venus_L3[] =
{
  {0.00000135742, 4.80389020993, 10213.28554621100},
  {0.00000077846, 3.66876371591, 20426.57109242200},
  {0.00000026023, 0.00000000000, 0.00000000000},
  {-76.0}
} ;

const struct VSOP_87 Venus_L4[] =
{
  {0.00000114016, 3.14159265359, 0.00000000000},
  {0.00000003209, 5.20514170164, 20426.57109242200},
  {0.00000001714, 2.51099591706, 10213.28554621100},
  {-76.0}
} ;

const struct VSOP_87 Venus_L5[] =
{
  {0.00000000874, 3.14159265359, 0.00000000000},
  {-76.0}
} ;

const struct VSOP_87 Venus_B0[] =
{
  {0.05923638472, 0.26702775813, 10213.28554621100},
  {0.00040107978, 1.14737178106, 20426.57109242200},
  {0.00032814918, 3.14159265359, 0.00000000000},
  {0.00001011392, 1.08946123021, 30639.85663863300},
  {0.00000149458, 6.25390296069, 18073.70493865020},
  {0.00000137788, 0.86020146523, 1577.34354244780},
  {0.00000129973, 3.67152483651, 9437.76293488700},
  {0.00000119507, 3.70468812804, 2352.86615377180},
  {0.00000107971, 4.53903677647, 22003.91463486980},
  {-66.0}
} ;

const struct VSOP_87 Venus_B1[] =
{
  {0.00513347602, 1.80364310797, 10213.28554621100},
  {0.00004380100, 3.38615711591, 20426.57109242200},
  {0.00000196586, 2.53001197486, 30639.85663863300},
  {0.00000199162, 0.00000000000, 0.00000000000},
  {-66.0}
} ;

const struct VSOP_87 Venus_B2[] =
{
  {0.00022377665, 3.38509143877, 10213.28554621100},
  {0.00000281739, 0.00000000000, 0.00000000000},
  {0.00000173164, 5.25563766915, 20426.57109242200},
  {0.00000026945, 3.87040891568, 30639.85663863300},
  {0.00000001174, 0.09768632072, 10186.98722641120},
  {0.00000000685, 3.19139067811, 11790.62908865880},
  {0.00000000788, 4.36515965295, 10239.58386601080},
  {0.00000000592, 5.22270440328, 40853.14218484400},
  {-66.0}
} ;

const struct VSOP_87 Venus_B3[] =
{
  {0.00000646671, 4.99166565277, 10213.28554621100},
  {0.00000019952, 3.14159265359, 0.00000000000},
  {0.00000005540, 0.77376923951, 20426.57109242200},
  {0.00000002526, 5.44493763020, 30639.85663863300},
  {-66.0}
} ;

const struct VSOP_87 Venus_B4[] =
{
  {0.00000014102, 0.31537190181, 10213.28554621100},
  {-66.0}
} ;

const struct VSOP_87 Venus_R0[] =
{
  {0.72334820905, 0.00000000000, 0.00000000000},
  {0.00489824185, 4.02151832268, 10213.28554621100},
  {0.00001658058, 4.90206728012, 20426.57109242200},
  {0.00001632093, 2.84548851892, 7860.41939243920},
  {0.00001378048, 1.12846590600, 11790.62908865880},
  {0.00000498399, 2.58682187717, 9683.59458111640},
  {0.00000373958, 1.42314837063, 3930.20969621960},
  {0.00000263616, 5.52938185920, 9437.76293488700},
  {0.00000237455, 2.55135903978, 15720.83878487840},
  {0.00000221983, 2.01346776772, 19367.18916223280}, //10
  {0.00000119467, 3.01975365264, 10404.73381232260},
  {0.00000125896, 2.72769833559, 1577.34354244780},
  {-82.0}
} ;

const struct VSOP_87 Venus_R1[] =
{
  {0.00034551039, 0.89198710598, 10213.28554621100},
  {0.00000234203, 1.77224942714, 20426.57109242200},
  {0.00000233998, 3.14159265359, 0.00000000000},
  {-82.0}
} ;

const struct VSOP_87 Venus_R2[] =
{
  {0.00001406587, 5.06366395190, 10213.28554621100},
  {0.00000015529, 5.47321687981, 20426.57109242200},
  {0.00000013059, 0.00000000000, 0.00000000000},
  {-82.0}
} ;

const struct VSOP_87 Venus_R3[] =
{
  {0.00000049582, 3.22263554520, 10213.28554621100},
  {-82.0}
} ;

const struct VSOP_87 Venus_R4[] =
{
  {0.00000000573, 0.92229697820, 10213.28554621100},
  {-82.0}
} ;

const struct VSOP_87 Earth_L0[] =
{
  {1.75347045673, 0.00000000000, 0.00000000000},
  {0.03341656456, 4.66925680417, 6283.07584999140},
  {0.00034894275, 4.62610241759, 12566.15169998280},
  {0.00003417571, 2.82886579606, 3.52311834900},
  {0.00003497056, 2.74411800971, 5753.38488489680},
  {0.00003135896, 3.62767041758, 77713.77146812050},
  {0.00002676218, 4.41808351397, 7860.41939243920},
  {0.00002342687, 6.13516237631, 3930.20969621960},
  {0.00001273166, 2.03709655772, 529.69096509460},
  {0.00001324292, 0.74246356352, 11506.76976979360}, //10
  {0.00000901855, 2.04505443513, 26.29831979980},
  {0.00001199167, 1.10962944315, 1577.34354244780},
  {0.00000857223, 3.50849156957, 398.14900340820},
  {0.00000779786, 1.17882652114, 5223.69391980220},
  {0.00000990250, 5.23268129594, 5884.92684658320},
  {0.00000753141, 2.53339053818, 5507.55323866740},
  {0.00000505264, 4.58292563052, 18849.22754997420},
  {0.00000492379, 4.20506639861, 775.52261132400},
  {0.00000356655, 2.91954116867, 0.06731030280},
  {0.00000284125, 1.89869034186, 796.29800681640}, //20
  {0.00000242810, 0.34481140906, 5486.77784317500},
  {0.00000317087, 5.84901952218, 11790.62908865880},
  {0.00000271039, 0.31488607649, 10977.07880469900},
  {0.00000206160, 4.80646606059, 2544.31441988340},
  {0.00000205385, 1.86947813692, 5573.14280143310},
  {0.00000202261, 2.45767795458, 6069.77675455340},
  {0.00000126184, 1.08302630210, 20.77539549240},
  {0.00000155516, 0.83306073807, 213.29909543800},
  {0.00000115132, 0.64544911683, 0.98032106820},
  {0.00000102851, 0.63599846727, 4694.00295470760}, //30
  {0.00000101724, 4.26679821365, 7.11354700080},
  {0.00000099206, 6.20992940258, 2146.16541647520},
  {0.00000132212, 3.41118275555, 2942.46342329160},
  {0.00000097607, 0.68101272270, 155.42039943420},
  {0.00000085128, 1.29870743025, 6275.96230299060},
  {0.00000074651, 1.75508916159, 5088.62883976680},
  {0.00000101895, 0.97569221824, 15720.83878487840},
  {0.00000084711, 3.67080093025, 71430.69561812909},
  {0.00000073547, 4.67926565481, 801.82093112380},
  {0.00000073874, 3.50319443167, 3154.68708489560}, //40
  {0.00000078756, 3.03698313141, 12036.46073488820},
  {0.00000079637, 1.80791330700, 17260.15465469040},
  {0.00000085803, 5.98322631256, 161000.68573767410},
  {0.00000056963, 2.78430398043, 6286.59896834040},
  {0.00000061148, 1.81839811024, 7084.89678111520},
  {0.00000069627, 0.83297596966, 9437.76293488700},
  {0.00000056116, 4.38694880779, 14143.49524243060},
  {0.00000062449, 3.97763880587, 8827.39026987480},
  {0.00000051145, 0.28306864501, 5856.47765911540},
  {0.00000055577, 3.47006009062, 6279.55273164240}, //50
  {0.00000041036, 5.36817351402, 8429.24126646660},
  {0.00000051605, 1.33282746983, 1748.01641306700},
  {0.00000051992, 0.18914945834, 12139.55350910680},
  {0.00000049000, 0.48735065033, 1194.44701022460},
  {0.00000039200, 6.16832995016, 10447.38783960440},
  {0.00000035566, 1.77597314691, 6812.76681508600},
  {0.00000036770, 6.04133859347, 10213.28554621100},
  {0.00000036596, 2.56955238628, 1059.38193018920},
  {0.00000033291, 0.59309499459, 17789.84561978500},
  {0.00000035954, 1.70876111898, 2352.86615377180}, //60
  {0.00000040938, 2.39850881707, 19651.04848109800},
  {0.00000030047, 2.73975123935, 1349.86740965880},
  {0.00000030412, 0.44294464135, 83996.84731811189},
  {0.00000023663, 0.48473567763, 8031.09226305840},
  {0.00000023574, 2.06527720049, 3340.61242669980},
  {-76.0}
} ;

const struct VSOP_87 Earth_L1[] =
{
  {6283.31966747491, 0.00000000000, 0.00000000000},
  {0.00206058863, 2.67823455584, 6283.07584999140},
  {0.00004303430, 2.63512650414, 12566.15169998280},
  {0.00000425264, 1.59046980729, 3.52311834900},
  {0.00000108977, 2.96618001993, 1577.34354244780},
  {0.00000093478, 2.59212835365, 18849.22754997420},
  {0.00000119261, 5.79557487799, 26.29831979980},
  {0.00000072122, 1.13846158196, 529.69096509460},
  {0.00000067768, 1.87472304791, 398.14900340820},
  {0.00000067327, 4.40918235168, 5507.55323866740}, //10
  {0.00000059027, 2.88797038460, 5223.69391980220},
  {0.00000055976, 2.17471680261, 155.42039943420},
  {0.00000045407, 0.39803079805, 796.29800681640},
  {0.00000036369, 0.46624739835, 775.52261132400},
  {0.00000028958, 2.64707383882, 7.11354700080},
  {0.00000019097, 1.84628332577, 5486.77784317500},
  {0.00000020844, 5.34138275149, 0.98032106820},
  {0.00000018508, 4.96855124577, 213.29909543800},
  {0.00000016233, 0.03216483047, 2544.31441988340},
  {0.00000017293, 2.99116864949, 6275.96230299060}, //20
  {0.00000015832, 1.43049285325, 2146.16541647520},
  {0.00000014615, 1.20532366323, 10977.07880469900},
  {0.00000011877, 3.25804815607, 5088.62883976680},
  {0.00000011514, 2.07502418155, 4694.00295470760},
  {0.00000009721, 4.23925472239, 1349.86740965880},
  {0.00000009969, 1.30262991097, 6286.59896834040},
  {0.00000009452, 2.69957062864, 242.72860397400},
  {0.00000012461, 2.83432285512, 1748.01641306700},
  {0.00000011808, 5.27379790480, 1194.44701022460},
  {0.00000008577, 5.64475868067, 951.71840625060}, //30
  {0.00000010641, 0.76614199202, 553.56940284240},
  {0.00000007576, 5.30062664886, 2352.86615377180},
  {0.00000005834, 1.76649917904, 1059.38193018920},
  {0.00000006385, 2.65033984967, 9437.76293488700},
  {-76.0}
} ;

const struct VSOP_87 Earth_L2[] =
{
  {0.00052918870, 0.00000000000, 0.00000000000},
  {0.00008719837, 1.07209665242, 6283.07584999140},
  {0.00000309125, 0.86728818832, 12566.15169998280},
  {0.00000027339, 0.05297871691, 3.52311834900},
  {0.00000016334, 5.18826691036, 26.29831979980},
  {0.00000015752, 3.68457889430, 155.42039943420},
  {0.00000009541, 0.75742297675, 18849.22754997420},
  {0.00000008937, 2.05705419118, 77713.77146812050},
  {0.00000006952, 0.82673305410, 775.52261132400},
  {0.00000005064, 4.66284525271, 1577.34354244780}, //10
  {0.00000004061, 1.03057162962, 7.11354700080},
  {0.00000003463, 5.14074632811, 796.29800681640},
  {0.00000003169, 6.05291851171, 5507.55323866740},
  {0.00000003020, 1.19246506441, 242.72860397400},
  {0.00000002886, 6.11652627155, 529.69096509460},
  {0.00000003810, 3.44050803490, 5573.14280143310},
  {0.00000002714, 0.30637881025, 398.14900340820},
  {0.00000002371, 4.38118838167, 5223.69391980220},
  {0.00000002538, 2.27992810679, 553.56940284240},
  {0.00000002079, 3.75435330484, 0.98032106820}, //20
  {-76.0}
} ;

const struct VSOP_87 Earth_L3[] =
{
  {0.00000289226, 5.84384198723, 6283.07584999140},
  {0.00000034955, 0.00000000000, 0.00000000000},
  {0.00000016819, 5.48766912348, 12566.15169998280},
  {0.00000002962, 5.19577265202, 155.42039943420},
  {0.00000001288, 4.72200252235, 3.52311834900},
  {0.00000000635, 5.96925937141, 242.72860397400},
  {0.00000000714, 5.30045809128, 18849.22754997420},
  {-76.0}
} ;

const struct VSOP_87 Earth_L4[] =
{
  {0.00000114084, 3.14159265359, 0.00000000000},
  {0.00000007717, 4.13446589358, 6283.07584999140},
  {0.00000000765, 3.83803776214, 12566.15169998280},
  {-76.0}
} ;

const struct VSOP_87 Earth_L5[] =
{
  {0.00000000878, 3.14159265359, 0.00000000000},
  {-76.0}
} ;

const struct VSOP_87 Earth_B0[] =
{
  {0.00000279620, 3.19870156017, 84334.66158130829},
  {0.00000101643, 5.42248619256, 5507.55323866740},
  {0.00000080445, 3.88013204458, 5223.69391980220},
  {0.00000043806, 3.70444689758, 2352.86615377180},
  {0.00000031933, 4.00026369781, 1577.34354244780},
  {-66.0}
} ;

const struct VSOP_87 Earth_B1[] =
{
  {0.00000009030, 3.89729061890, 5507.55323866740},
  {0.00000006177, 1.73038850355, 5223.69391980220},
  {-66.0}
} ;

const struct VSOP_87 Earth_R0[] =
{
  {1.00013988799, 0.00000000000, 0.00000000000},
  {0.01670699626, 3.09846350771, 6283.07584999140},
  {0.00013956023, 3.05524609620, 12566.15169998280},
  {0.00003083720, 5.19846674381, 77713.77146812050},
  {0.00001628461, 1.17387749012, 5753.38488489680},
  {0.00001575568, 2.84685245825, 7860.41939243920},
  {0.00000924799, 5.45292234084, 11506.76976979360},
  {0.00000542444, 4.56409149777, 3930.20969621960},
  {0.00000472110, 3.66100022149, 5884.92684658320},
  {0.00000328780, 5.89983646482, 5223.69391980220}, //10
  {0.00000345983, 0.96368617687, 5507.55323866740},
  {0.00000306784, 0.29867139512, 5573.14280143310},
  {0.00000174844, 3.01193636534, 18849.22754997420},
  {0.00000243189, 4.27349536153, 11790.62908865880},
  {0.00000211829, 5.84714540314, 1577.34354244780},
  {0.00000185752, 5.02194447178, 10977.07880469900},
  {0.00000109835, 5.05510636285, 5486.77784317500},
  {0.00000098316, 0.88681311277, 6069.77675455340},
  {0.00000086499, 5.68959778254, 15720.83878487840},
  {0.00000085825, 1.27083733351, 161000.68573767410}, //20
  {0.00000062916, 0.92177108832, 529.69096509460},
  {0.00000057056, 2.01374292014, 83996.84731811189},
  {0.00000064903, 0.27250613787, 17260.15465469040},
  {0.00000049384, 3.24501240359, 2544.31441988340},
  {0.00000055736, 5.24159798933, 71430.69561812909},
  {0.00000042515, 6.01110242003, 6275.96230299060},
  {0.00000046963, 2.57805070386, 775.52261132400},
  {0.00000038968, 5.36071738169, 4694.00295470760},
  {0.00000044661, 5.53715807302, 9437.76293488700},
  {0.00000035660, 1.67468058995, 12036.46073488820}, //30
  {0.00000031921, 0.18368229781, 5088.62883976680},
  {0.00000031846, 1.77775642085, 398.14900340820},
  {0.00000033193, 0.24370300098, 7084.89678111520},
  {0.00000038245, 2.39255343974, 8827.39026987480},
  {0.00000028464, 1.21344868176, 6286.59896834040},
  {0.00000037490, 0.82952922332, 19651.04848109800},
  {0.00000036957, 4.90107591914, 12139.55350910680},
  {0.00000034537, 1.84270693282, 2942.46342329160},
  {0.00000026275, 4.58896850401, 10447.38783960440},
  {0.00000024596, 3.78660875483, 8429.24126646660}, //40
  {-82.0}
} ;

const struct VSOP_87 Earth_R1[] =
{
  {0.00103018608, 1.10748969588, 6283.07584999140},
  {0.00001721238, 1.06442301418, 12566.15169998280},
  {0.00000702215, 3.14159265359, 0.00000000000},
  {0.00000032346, 1.02169059149, 18849.22754997420},
  {0.00000030799, 2.84353804832, 5507.55323866740},
  {0.00000024971, 1.31906709482, 5223.69391980220},
  {0.00000018485, 1.42429748614, 1577.34354244780},
  {0.00000010078, 5.91378194648, 10977.07880469900},
  {0.00000008634, 0.27146150602, 5486.77784317500},
  {0.00000008654, 1.42046854427, 6275.96230299060}, //10
  {-82.0}
} ;

const struct VSOP_87 Earth_R2[] =
{
  {0.00004359385, 5.78455133738, 6283.07584999140},
  {0.00000123633, 5.57934722157, 12566.15169998280},
  {0.00000012341, 3.14159265359, 0.00000000000},
  {0.00000008792, 3.62777733395, 77713.77146812050},
  {0.00000005689, 1.86958905084, 5573.14280143310},
  {0.00000003301, 5.47027913302, 18849.22754997420},
  {-82.0}
} ;

const struct VSOP_87 Earth_R3[] =
{
  {0.00000144595, 4.27319435148, 6283.07584999140},
  {0.00000006729, 3.91697608662, 12566.15169998280},
  {-82.0}
} ;

const struct VSOP_87 Earth_R4[] =
{
  {0.00000003858, 2.56384387339, 6283.07584999140},
  {-82.0}
} ;

const struct VSOP_87 Mars_L0[] =
{
  {6.20347711583, 0.00000000000, 0.00000000000},
  {0.18656368100, 5.05037100303, 3340.61242669980},
  {0.01108216792, 5.40099836958, 6681.22485339960},
  {0.00091798394, 5.75478745111, 10021.83728009940},
  {0.00027744987, 5.97049512942, 3.52311834900},
  {0.00010610230, 2.93958524973, 2281.23049651060},
  {0.00012315897, 0.84956081238, 2810.92146160520},
  {0.00008926772, 4.15697845939, 0.01725365220},
  {0.00008715688, 6.11005159792, 13362.44970679920},
  {0.00006797552, 0.36462243626, 398.14900340820}, //10
  {0.00007774867, 3.33968655074, 5621.84292321040},
  {0.00003575079, 1.66186540141, 2544.31441988340},
  {0.00004161101, 0.22814975330, 2942.46342329160},
  {0.00003075250, 0.85696597082, 191.44826611160},
  {0.00002628122, 0.64806143570, 3337.08930835080},
  {0.00002937543, 6.07893711408, 0.06731030280},
  {0.00002389420, 5.03896401349, 796.29800681640},
  {0.00002579842, 0.02996706197, 3344.13554504880},
  {0.00001528140, 1.14979306228, 6151.53388830500},
  {0.00001798808, 0.65634026844, 529.69096509460}, //20
  {0.00001264356, 3.62275092231, 5092.15195811580},
  {0.00001286232, 3.06795924626, 2146.16541647520},
  {0.00001546408, 2.91579633392, 1751.53953141600},
  {0.00001024907, 3.69334293555, 8962.45534991020},
  {0.00000891567, 0.18293899090, 16703.06213349900},
  {0.00000858760, 2.40093704204, 2914.01423582380},
  {0.00000832718, 2.46418591282, 3340.59517304760},
  {0.00000832724, 4.49495753458, 3340.62968035200},
  {0.00000712899, 3.66336014788, 1059.38193018920},
  {0.00000748724, 3.82248399468, 155.42039943420}, //30
  {0.00000723863, 0.67497565801, 3738.76143010800},
  {0.00000635557, 2.92182704275, 8432.76438481560},
  {0.00000655163, 0.48864075176, 3127.31333126180},
  {0.00000550472, 3.81001205408, 0.98032106820},
  {0.00000552746, 4.47478863016, 1748.01641306700},
  {0.00000425972, 0.55365138172, 6283.07584999140},
  {0.00000415132, 0.49662314774, 213.29909543800},
  {0.00000472164, 3.62547819410, 1194.44701022460},
  {0.00000306552, 0.38052862973, 6684.74797174860},
  {0.00000312141, 0.99853322843, 6677.70173505060}, //40
  {0.00000293199, 4.22131277914, 20.77539549240},
  {0.00000302377, 4.48618150321, 3532.06069281140},
  {0.00000274028, 0.54222141841, 3340.54511639700},
  {0.00000281073, 5.88163372945, 1349.86740965880},
  {0.00000231185, 1.28240685294, 3870.30339179440},
  {0.00000283600, 5.76885494123, 3149.16416058820},
  {0.00000236114, 5.75504515576, 3333.49887969900},
  {0.00000274035, 0.13372501211, 3340.67973700260},
  {0.00000299396, 2.78323705697, 6254.62666252360},
  {0.00000204161, 2.82133266185, 1221.84856632140}, //50
  {0.00000238857, 5.37155471672, 4136.91043351620},
  {0.00000188639, 1.49103016486, 9492.14631500480},
  {0.00000221225, 3.50466672203, 382.89653222320},
  {0.00000179196, 1.00561112574, 951.71840625060},
  {0.00000172110, 0.43943041719, 5486.77784317500},
  {0.00000193126, 3.35715137745, 3.59042865180},
  {0.00000144305, 1.41874193418, 135.06508003540},
  {0.00000160011, 3.94854735192, 4562.46099302120},
  {0.00000174068, 2.41360332576, 553.56940284240},
  {0.00000130993, 4.04491720264, 12303.06777661000}, //60
  {0.00000138245, 4.30145176915, 7.11354700080},
  {0.00000128062, 1.80665643332, 5088.62883976680},
  {0.00000139897, 3.32592516164, 2700.71514038580},
  {0.00000128102, 2.20806651008, 1592.59601363280},
  {0.00000116945, 3.12805282207, 7903.07341972100},
  {0.00000110375, 1.05195079687, 242.72860397400},
  {0.00000113486, 3.70070798123, 1589.07289528380},
  {0.00000100090, 3.24343740861, 11773.37681151540},
  {0.00000095592, 0.53954181149, 20043.67456019880},
  {0.00000098947, 4.84558294740, 6681.24210705180}, //70
  {-76.0}
} ;

const struct VSOP_87 Mars_L1[] =
{
  {3340.85627474342, 0.00000000000, 0.00000000000},
  {0.01458227051, 3.60426053609, 3340.61242669980},
  {0.00164901343, 3.92631250962, 6681.22485339960},
  {0.00019963338, 4.26594061030, 10021.83728009940},
  {0.00003452399, 4.73210386365, 3.52311834900},
  {0.00002485480, 4.61277567318, 13362.44970679920},
  {0.00000841551, 4.45858256765, 2281.23049651060},
  {0.00000537566, 5.01589727492, 398.14900340820},
  {0.00000521041, 4.99422678175, 3344.13554504880},
  {0.00000432614, 2.56066402860, 191.44826611160}, //10
  {0.00000429656, 5.31646162367, 155.42039943420},
  {0.00000381747, 3.53881289437, 796.29800681640},
  {0.00000314129, 4.96335266049, 16703.06213349900},
  {0.00000282804, 3.15967518204, 2544.31441988340},
  {0.00000205664, 4.56891455660, 2146.16541647520},
  {0.00000168805, 1.32894813366, 3337.08930835080},
  {0.00000157587, 4.18501035954, 1751.53953141600},
  {0.00000133686, 2.23325104196, 0.98032106820},
  {0.00000116561, 2.21347652545, 1059.38193018920},
  {0.00000117591, 6.02407213861, 6151.53388830500}, //20
  {0.00000113595, 5.42803224317, 3738.76143010800},
  {0.00000133563, 5.97421903927, 1748.01641306700},
  {0.00000091098, 1.09627836591, 1349.86740965880},
  {0.00000083301, 5.29636626272, 6684.74797174860},
  {0.00000113876, 2.12869455089, 1194.44701022460},
  {0.00000080776, 4.42813405865, 529.69096509460},
  {0.00000079531, 2.24864266330, 8962.45534991020},
  {0.00000072505, 5.84208163240, 242.72860397400},
  {0.00000072946, 2.50189460554, 951.71840625060},
  {0.00000071487, 3.85636094435, 2914.01423582380}, //30
  {0.00000085342, 3.90854841008, 553.56940284240},
  {0.00000067582, 5.02327686473, 382.89653222320},
  {0.00000065089, 1.01802439311, 3340.59517304760},
  {0.00000065089, 3.04879603978, 3340.62968035200},
  {0.00000061508, 4.15183159800, 3149.16416058820},
  {0.00000048477, 4.87362121538, 213.29909543800},
  {0.00000046584, 1.31452419914, 3185.19202726560},
  {0.00000056520, 3.88813699320, 4136.91043351620},
  {0.00000047613, 1.18238046057, 3333.49887969900},
  {0.00000041343, 0.71385375517, 1592.59601363280}, //40
  {0.00000040055, 5.31611875491, 20043.67456019880},
  {0.00000040272, 2.72542480614, 7.11354700080},
  {0.00000032886, 5.41067411968, 6283.07584999140},
  {0.00000028244, 0.04534124888, 9492.14631500480},
  {0.00000022294, 5.88516997273, 3870.30339179440},
  {-76.0}
} ;

const struct VSOP_87 Mars_L2[] =
{
  {0.00058015791, 2.04979463279, 3340.61242669980},
  {0.00054187645, 0.00000000000, 0.00000000000},
  {0.00013908426, 2.45742359888, 6681.22485339960},
  {0.00002465104, 2.80000020929, 10021.83728009940},
  {0.00000398379, 3.14118428289, 13362.44970679920},
  {0.00000222022, 3.19436080019, 3.52311834900},
  {0.00000120957, 0.54325292454, 155.42039943420},
  {0.00000061517, 3.48529427371, 16703.06213349900},
  {0.00000053638, 3.54191121461, 3344.13554504880},
  {0.00000034268, 6.00188499119, 2281.23049651060}, //10
  {0.00000031665, 4.14015171788, 191.44826611160},
  {0.00000029839, 1.99870679845, 796.29800681640},
  {0.00000023168, 4.33403365928, 242.72860397400},
  {0.00000021659, 3.44532466378, 398.14900340820},
  {0.00000016044, 6.11000472441, 2146.16541647520},
  {0.00000020370, 5.42191375400, 553.56940284240},
  {0.00000014927, 6.09541783564, 3185.19202726560},
  {0.00000016227, 0.65678953303, 0.98032106820},
  {0.00000014317, 2.61851897591, 1349.86740965880},
  {0.00000014416, 4.01923812101, 951.71840625060}, //20
  {0.00000011934, 3.86122163021, 6684.74797174860},
  {0.00000015648, 1.22086121940, 1748.01641306700},
  {0.00000011260, 4.71822363671, 2544.31441988340},
  {0.00000013352, 0.60189008414, 1194.44701022460},
  {0.00000010396, 0.25038714677, 382.89653222320},
  {0.00000009468, 0.68170713564, 1059.38193018920},
  {0.00000009229, 3.83209092321, 20043.67456019880},
  {0.00000009005, 3.88271826102, 3738.76143010800},
  {0.00000007501, 5.46498630412, 1751.53953141600},
  {0.00000006497, 5.47773072872, 1592.59601363280}, //30
  {0.00000006311, 2.34104793674, 3097.88382272579},
  {0.00000006859, 2.57522504136, 3149.16416058820},
  {0.00000005870, 1.14783576679, 7.11354700080},
  {-76.0}
} ;

const struct VSOP_87 Mars_L3[] =
{
  {0.00001482423, 0.44434694876, 3340.61242669980},
  {0.00000662095, 0.88469178686, 6681.22485339960},
  {0.00000188268, 1.28799982497, 10021.83728009940},
  {0.00000041474, 1.64850786997, 13362.44970679920},
  {0.00000022661, 2.05267665262, 155.42039943420},
  {0.00000025994, 0.00000000000, 0.00000000000},
  {0.00000008024, 1.99858757687, 16703.06213349900},
  {0.00000010454, 1.58006906385, 3.52311834900},
  {0.00000004900, 2.82452457966, 242.72860397400},
  {0.00000003782, 2.01914272515, 3344.13554504880}, //10
  {0.00000003176, 4.59144897927, 3185.19202726560},
  {0.00000003134, 0.65044714325, 553.56940284240},
  {-76.0}
} ;

const struct VSOP_87 Mars_L4[] =
{
  {0.00000113969, 3.14159265359, 0.00000000000},
  {0.00000028725, 5.63662412043, 6681.22485339960},
  {0.00000024447, 5.13868481454, 3340.61242669980},
  {0.00000011187, 6.03161074431, 10021.83728009940},
  {0.00000003190, 3.56267988299, 155.42039943420},
  {0.00000003252, 0.13228350651, 13362.44970679920},
  {0.00000000787, 0.49340783377, 16703.06213349900},
  {0.00000000776, 1.31734531594, 242.72860397400},
  {-76.0}
} ;

const struct VSOP_87 Mars_L5[] =
{
  {0.00000000710, 4.04089996521, 6681.22485339960},
  {0.00000000868, 3.14159265359, 0.00000000000},
  {-76.0}
} ;

const struct VSOP_87 Mars_B0[] =
{
  {0.03197134986, 3.76832042432, 3340.61242669980},
  {0.00298033234, 4.10616996243, 6681.22485339960},
  {0.00289104742, 0.00000000000, 0.00000000000},
  {0.00031365538, 4.44651052853, 10021.83728009940},
  {0.00003484100, 4.78812547889, 13362.44970679920},
  {0.00000442999, 5.65233015876, 3337.08930835080},
  {0.00000443401, 5.02642620491, 3344.13554504880},
  {0.00000399109, 5.13056814700, 16703.06213349900},
  {0.00000292506, 3.79290644595, 2281.23049651060},
  {0.00000181982, 6.13648011704, 6151.53388830500}, //10
  {0.00000163159, 4.26399626634, 529.69096509460},
  {0.00000159678, 2.23194610246, 1059.38193018920},
  {0.00000139323, 2.41796344238, 8962.45534991020},
  {0.00000149297, 2.16501209917, 5621.84292321040},
  {0.00000142686, 1.18215016110, 3340.59517304760},
  {0.00000142685, 3.21292180820, 3340.62968035200},
  {-66.0}
} ;

const struct VSOP_87 Mars_B1[] =
{
  {0.00350068845, 5.36847836211, 3340.61242669980},
  {0.00014116030, 3.14159265359, 0.00000000000},
  {0.00009670755, 5.47877786506, 6681.22485339960},
  {0.00001471918, 3.20205766795, 10021.83728009940},
  {0.00000425864, 3.40843812875, 13362.44970679920},
  {0.00000102039, 0.77617286189, 3337.08930835080},
  {0.00000078848, 3.71768293865, 16703.06213349900},
  {0.00000026171, 2.48293558065, 2281.23049651060},
  {0.00000032708, 3.45803723682, 5621.84292321040},
  {-66.0}
} ;

const struct VSOP_87 Mars_B2[] =
{
  {0.00016726690, 0.60221392419, 3340.61242669980},
  {0.00004986799, 3.14159265359, 0.00000000000},
  {0.00000302141, 5.55871276021, 6681.22485339960},
  {0.00000025767, 1.89662673499, 13362.44970679920},
  {0.00000021452, 0.91749968618, 10021.83728009940},
  {0.00000011820, 2.24240738700, 3337.08930835080},
  {0.00000007985, 2.24892866611, 16703.06213349900},
  {-66.0}
} ;

const struct VSOP_87 Mars_B3[] =
{
  {0.00000606506, 1.98050633529, 3340.61242669980},
  {0.00000042611, 0.00000000000, 0.00000000000},
  {0.00000013652, 1.79588228800, 6681.22485339960},
  {0.00000002730, 3.45377082121, 10021.83728009940},
  {-66.0}
} ;

const struct VSOP_87 Mars_B4[] =
{
  {0.00000011334, 3.45724352586, 3340.61242669980},
  {0.00000013369, 0.00000000000, 0.00000000000},
  {0.00000000744, 0.50445805257, 6681.22485339960},
  {-66.0}
} ;

const struct VSOP_87 Mars_R0[] =
{
  {1.53033488276, 0.00000000000, 0.00000000000},
  {0.14184953153, 3.47971283519, 3340.61242669980},
  {0.00660776357, 3.81783442097, 6681.22485339960},
  {0.00046179117, 4.15595316284, 10021.83728009940},
  {0.00008109738, 5.55958460165, 2810.92146160520},
  {0.00007485315, 1.77238998069, 5621.84292321040},
  {0.00005523193, 1.36436318880, 2281.23049651060},
  {0.00003825160, 4.49407182408, 13362.44970679920},
  {0.00002306539, 0.09081742493, 2544.31441988340},
  {0.00001999399, 5.36059605227, 3337.08930835080}, //10
  {0.00002484385, 4.92545577893, 2942.46342329160},
  {0.00001960198, 4.74249386323, 3344.13554504880},
  {0.00001167115, 2.11261501155, 5092.15195811580},
  {0.00001102828, 5.00908264160, 398.14900340820},
  {0.00000899077, 4.40790433994, 529.69096509460},
  {0.00000992252, 5.83862401067, 6151.53388830500},
  {0.00000807348, 2.10216647104, 1059.38193018920},
  {0.00000797910, 3.44839026172, 796.29800681640},
  {0.00000740980, 1.49906336892, 2146.16541647520},
  {0.00000692340, 2.13378814785, 8962.45534991020}, //20
  {0.00000633144, 0.89353285018, 3340.59517304760},
  {0.00000725583, 1.24516913473, 8432.76438481560},
  {0.00000633140, 2.92430448169, 3340.62968035200},
  {0.00000574352, 0.82896196337, 2914.01423582380},
  {0.00000526187, 5.38292276228, 3738.76143010800},
  {0.00000629976, 1.28738135858, 1751.53953141600},
  {0.00000472776, 5.19850457873, 3127.31333126180},
  {0.00000348095, 4.83219198908, 16703.06213349900},
  {0.00000283702, 2.90692294913, 3532.06069281140},
  {0.00000279552, 5.25749247548, 6283.07584999140}, //30
  {0.00000233827, 5.10546492529, 5486.77784317500},
  {0.00000219428, 5.58340248784, 191.44826611160},
  {0.00000269891, 3.76394728622, 5884.92684658320},
  {0.00000208333, 5.25476080773, 3340.54511639700},
  {0.00000275224, 2.90818883832, 1748.01641306700},
  {0.00000275501, 1.21767967781, 6254.62666252360},
  {0.00000239133, 2.03669896238, 1194.44701022460},
  {0.00000223190, 4.19861593779, 3149.16416058820},
  {0.00000182686, 5.08062683355, 6684.74797174860},
  {0.00000186213, 5.69871555748, 6677.70173505060}, //40
  {0.00000175995, 5.95341786369, 3870.30339179440},
  {0.00000178613, 4.18423025538, 3333.49887969900},
  {0.00000208336, 4.84626442122, 3340.67973700260},
  {0.00000228128, 3.25529020620, 6872.67311951120},
  {0.00000144286, 0.21296012258, 5088.62883976680},
  {-82.0}
} ;

const struct VSOP_87 Mars_R1[] =
{
  {0.01107433340, 2.03250524950, 3340.61242669980},
  {0.00103175886, 2.37071845682, 6681.22485339960},
  {0.00012877200, 0.00000000000, 0.00000000000},
  {0.00010815880, 2.70888093803, 10021.83728009940},
  {0.00001194550, 3.04702182503, 13362.44970679920},
  {0.00000438579, 2.88835072628, 2281.23049651060},
  {0.00000395698, 3.42324611291, 3344.13554504880},
  {0.00000182572, 1.58428644001, 2544.31441988340},
  {0.00000135850, 3.38507017993, 16703.06213349900},
  {0.00000128204, 0.62991220570, 1059.38193018920}, //10
  {0.00000127068, 1.95389775740, 796.29800681640},
  {0.00000118443, 2.99761345074, 2146.16541647520},
  {0.00000128362, 6.04343360441, 3337.08930835080},
  {0.00000087537, 3.42052758979, 398.14900340820},
  {0.00000083026, 3.85574986653, 3738.76143010800},
  {0.00000075598, 4.45101839349, 6151.53388830500},
  {0.00000071999, 2.76442180680, 529.69096509460},
  {0.00000066542, 2.54892602695, 1751.53953141600},
  {0.00000054314, 0.67750943459, 8962.45534991020},
  {0.00000051035, 3.72585409207, 6684.74797174860}, //20
  {0.00000066430, 4.40597549957, 1748.01641306700},
  {0.00000047863, 2.28527896843, 2914.01423582380},
  {0.00000049428, 5.72959428364, 3340.59517304760},
  {0.00000049424, 1.47717922226, 3340.62968035200},
  {0.00000057518, 0.54354327916, 1194.44701022460},
  {0.00000048318, 2.58061691301, 3149.16416058820},
  {0.00000036384, 6.02728752344, 3185.19202726560},
  {-82.0}
} ;

const struct VSOP_87 Mars_R2[] =
{
  {0.00044242247, 0.47930603943, 3340.61242669980},
  {0.00008138042, 0.86998398093, 6681.22485339960},
  {0.00001274915, 1.22594050809, 10021.83728009940},
  {0.00000187387, 1.57298991982, 13362.44970679920},
  {0.00000040744, 1.97080175060, 3344.13554504880},
  {0.00000052396, 3.14159265359, 0.00000000000},
  {0.00000026616, 1.91665615762, 16703.06213349900},
  {0.00000017825, 4.43499505333, 2281.23049651060},
  {0.00000011713, 4.52510453730, 3185.19202726560},
  {0.00000010209, 5.39143469548, 1059.38193018920}, //10
  {0.00000009950, 0.41870577185, 796.29800681640},
  {-82.0}
} ;

const struct VSOP_87 Mars_R3[] =
{
  {0.00001113107, 5.14987350142, 3340.61242669980},
  {0.00000424446, 5.61343766478, 6681.22485339960},
  {0.00000100044, 5.99726827028, 10021.83728009940},
  {0.00000019606, 0.07633062094, 13362.44970679920},
  {0.00000003477, 0.42951907576, 16703.06213349900},
  {0.00000004693, 3.14159265359, 0.00000000000},
  {-82.0}
} ;

const struct VSOP_87 Mars_R4[] =
{
  {0.00000019552, 3.58211650473, 3340.61242669980},
  {0.00000016323, 4.05116076923, 6681.22485339960},
  {0.00000005848, 4.46383962094, 10021.83728009940},
  {0.00000001532, 4.84374321619, 13362.44970679920},
  {-82.0}
} ;

const struct VSOP_87 Jupiter_L0[] =
{
  {0.59954691495, 0.00000000000, 0.00000000000},
  {0.09695898711, 5.06191793105, 529.69096509460},
  {0.00573610145, 1.44406205976, 7.11354700080},
  {0.00306389180, 5.41734729976, 1059.38193018920},
  {0.00097178280, 4.14264708819, 632.78373931320},
  {0.00072903096, 3.64042909255, 522.57741809380},
  {0.00064263986, 3.41145185203, 103.09277421860},
  {0.00039806051, 2.29376744855, 419.48464387520},
  {0.00038857780, 1.27231724860, 316.39186965660},
  {0.00027964622, 1.78454589485, 536.80451209540}, //10
  {0.00013589738, 5.77481031590, 1589.07289528380},
  {0.00008246362, 3.58227961655, 206.18554843720},
  {0.00008768686, 3.63000324417, 949.17560896980},
  {0.00007368057, 5.08101125612, 735.87651353180},
  {0.00006263171, 0.02497643742, 213.29909543800},
  {0.00006114050, 4.51319531666, 1162.47470440780},
  {0.00004905419, 1.32084631684, 110.20632121940},
  {0.00005305283, 1.30671236848, 14.22709400160},
  {0.00005305457, 4.18625053495, 1052.26838318840},
  {0.00004647249, 4.69958109497, 3.93215326310}, //20
  {0.00003045009, 4.31675960318, 426.59819087600},
  {0.00002610001, 1.56667594850, 846.08283475120},
  {0.00002028191, 1.06376547379, 3.18139373770},
  {0.00001764768, 2.14148077766, 1066.49547719000},
  {0.00001722983, 3.88036008872, 1265.56747862640},
  {0.00001920959, 0.97168928755, 639.89728631400},
  {0.00001633217, 3.58201089758, 515.46387109300},
  {0.00001431997, 4.29683690269, 625.67019231240},
  {0.00000973278, 4.09764957065, 95.97922721780},
  {0.00000884439, 2.43701426123, 412.37109687440}, //30
  {0.00000732875, 6.08534113239, 838.96928775040},
  {0.00000731072, 3.80591233956, 1581.95934828300},
  {0.00000691928, 6.13368222939, 2118.76386037840},
  {0.00000709190, 1.29272573658, 742.99006053260},
  {0.00000614464, 4.10853496756, 1478.86657406440},
  {0.00000495224, 3.75567461379, 323.50541665740},
  {0.00000581902, 4.53967717552, 309.27832265580},
  {0.00000375657, 4.70299124833, 1368.66025284500},
  {0.00000389864, 4.89716105852, 1692.16566950240},
  {0.00000341006, 5.71452525783, 533.62311835770}, //40
  {0.00000330458, 4.74049819491, 0.04818410980},
  {0.00000440854, 2.95818460943, 454.90936652730},
  {0.00000417266, 1.03554430161, 2.44768055480},
  {0.00000244170, 5.22020878900, 728.76296653100},
  {0.00000261540, 1.87652461032, 0.96320784650},
  {0.00000256568, 3.72410724159, 199.07200143640},
  {0.00000261009, 0.82047246448, 380.12776796000},
  {0.00000220382, 1.65115015995, 543.91805909620},
  {0.00000201996, 1.80684574186, 1375.77379984580},
  {0.00000207327, 1.85461666594, 525.75881183150}, //50
  {0.00000197046, 5.29252149016, 1155.36115740700},
  {0.00000235141, 1.22693908124, 909.81873305460},
  {0.00000174809, 5.90973505276, 956.28915597060},
  {0.00000149368, 4.37745104275, 1685.05212250160},
  {0.00000175184, 3.22634903433, 1898.35121793960},
  {0.00000175191, 3.72966554761, 942.06206196900},
  {0.00000157909, 4.36483921766, 1795.25844372100},
  {0.00000137871, 1.31797920785, 1169.58825140860},
  {0.00000117495, 2.50022140890, 1596.18644228460},
  {0.00000150502, 3.90625022622, 74.78159856730}, //60
  {0.00000116757, 3.38920921041, 0.52126486180},
  {0.00000105895, 4.55439798236, 526.50957135690},
  {0.00000130531, 4.16867945489, 1045.15483618760},
  {0.00000141445, 3.13568357861, 491.55792945680},
  {-76.0}
} ;

const struct VSOP_87 Jupiter_L1[] =
{
  {529.93480757497, 0.00000000000, 0.00000000000},
  {0.00489741194, 4.22066689928, 529.69096509460},
  {0.00228918538, 6.02647464016, 7.11354700080},
  {0.00027655380, 4.57265956824, 1059.38193018920},
  {0.00020720943, 5.45938936295, 522.57741809380},
  {0.00012105732, 0.16985765041, 536.80451209540},
  {0.00006068051, 4.42419502005, 103.09277421860},
  {0.00005433924, 3.98478382565, 419.48464387520},
  {0.00004237795, 5.89009351271, 14.22709400160},
  {0.00002211854, 5.26771446618, 206.18554843720}, //10
  {0.00001295769, 5.55132765087, 3.18139373770},
  {0.00001745919, 4.92669378486, 1589.07289528380},
  {0.00001163411, 0.51450895328, 3.93215326310},
  {0.00001007216, 0.46478398551, 735.87651353180},
  {0.00001173129, 5.85647304350, 1052.26838318840},
  {0.00000847678, 5.75805850450, 110.20632121940},
  {0.00000827329, 4.80312015734, 213.29909543800},
  {0.00001003574, 3.15040301822, 426.59819087600},
  {0.00001098735, 5.30704981594, 515.46387109300},
  {0.00000816397, 0.58643054886, 1066.49547719000}, //20
  {0.00000725447, 5.51827471473, 639.89728631400},
  {0.00000567845, 5.98867049451, 625.67019231240},
  {0.00000474181, 4.13245269168, 412.37109687440},
  {0.00000412930, 5.73652891261, 95.97922721780},
  {0.00000335817, 3.73248749046, 1162.47470440780},
  {0.00000345249, 4.24159565410, 632.78373931320},
  {0.00000234066, 6.24302226646, 309.27832265580},
  {0.00000194784, 2.21879010911, 323.50541665740},
  {0.00000234340, 4.03469970332, 949.17560896980},
  {0.00000183938, 6.27963588822, 543.91805909620}, //30
  {0.00000198525, 1.50458442825, 838.96928775040},
  {0.00000186899, 6.08620565908, 742.99006053260},
  {0.00000171380, 5.41655983845, 199.07200143640},
  {0.00000130771, 0.62643377351, 728.76296653100},
  {0.00000107575, 4.49282760117, 956.28915597060},
  {0.00000115393, 0.68019050174, 846.08283475120},
  {0.00000115047, 5.28641699144, 2118.76386037840},
  {0.00000066824, 5.73365126533, 21.34064100240},
  {0.00000069618, 5.97263450278, 532.87235883230},
  {0.00000064850, 6.08803490288, 1581.95934828300}, //40
  {0.00000079686, 5.82412400273, 1045.15483618760},
  {0.00000057939, 0.99453087342, 1596.18644228460},
  {0.00000065635, 0.12924191430, 526.50957135690},
  {0.00000058509, 0.58626971028, 1155.36115740700},
  {0.00000056600, 1.41198438841, 533.62311835770},
  {0.00000071643, 5.34162650321, 942.06206196900},
  {0.00000057368, 5.96851304799, 1169.58825140860},
  {0.00000054935, 5.42806383723, 10.29494073850},
  {0.00000052016, 0.22981299129, 1368.66025284500},
  {0.00000052309, 5.72661448388, 117.31986822020}, //50
  {0.00000050418, 6.08075147811, 525.75881183150},
  {0.00000047418, 3.62611843241, 1478.86657406440},
  {0.00000039888, 4.16158013600, 1692.16566950240},
  {0.00000046678, 0.51144073175, 1265.56747862640},
  {0.00000032827, 5.03596689455, 220.41264243880},
  {0.00000033558, 0.09913904872, 302.16477565500},
  {0.00000029379, 3.35927241533, 4.66586644600},
  {0.00000029307, 0.75907909735, 88.86568021700},
  {0.00000032449, 5.37492530697, 508.35032409220},
  {0.00000029483, 5.42208897099, 1272.68102562720}, //60
  {0.00000021802, 6.15054054070, 1685.05212250160},
  {-76.0}
} ;

const struct VSOP_87 Jupiter_L2[] =
{
  {0.00047233598, 4.32148323554, 7.11354700080},
  {0.00030629053, 2.93021440216, 529.69096509460},
  {0.00038965550, 0.00000000000, 0.00000000000},
  {0.00003189317, 1.05504615595, 522.57741809380},
  {0.00002723358, 3.41411526638, 1059.38193018920},
  {0.00002729292, 4.84545481351, 536.80451209540},
  {0.00001721069, 4.18734385158, 14.22709400160},
  {0.00000383258, 5.76790714387, 419.48464387520},
  {0.00000367498, 6.05509120409, 103.09277421860},
  {0.00000377524, 0.76048964872, 515.46387109300}, //10
  {0.00000337386, 3.78644384244, 3.18139373770},
  {0.00000308200, 0.69356654052, 206.18554843720},
  {0.00000218408, 3.81389191353, 1589.07289528380},
  {0.00000198883, 5.33996443444, 1066.49547719000},
  {0.00000197445, 2.48356402053, 3.93215326310},
  {0.00000146230, 3.81373196838, 639.89728631400},
  {0.00000155862, 1.40642426467, 1052.26838318840},
  {0.00000129570, 5.83738872525, 412.37109687440},
  {0.00000141932, 1.63435169016, 426.59819087600},
  {0.00000117327, 1.41435462588, 625.67019231240}, //20
  {0.00000096733, 4.03383427887, 110.20632121940},
  {0.00000090823, 1.10630629042, 95.97922721780},
  {0.00000078769, 4.63726131329, 543.91805909620},
  {0.00000072392, 2.21716670026, 735.87651353180},
  {0.00000087292, 2.52235174825, 632.78373931320},
  {0.00000056910, 3.12292059854, 213.29909543800},
  {0.00000048622, 1.67283791618, 309.27832265580},
  {0.00000058475, 0.83216317444, 199.07200143640},
  {0.00000040150, 4.02485444740, 21.34064100240},
  {0.00000039784, 0.62416945827, 323.50541665740}, //30
  {0.00000035718, 2.32581247002, 728.76296653100},
  {0.00000025620, 2.51240623862, 1162.47470440780},
  {0.00000029255, 3.60838327799, 10.29494073850},
  {0.00000023591, 3.00532139306, 956.28915597060},
  {0.00000027814, 3.23992013743, 838.96928775040},
  {0.00000025993, 4.50118298290, 742.99006053260},
  {0.00000025194, 1.21868110687, 1045.15483618760},
  {0.00000019458, 4.29028644674, 532.87235883230},
  {0.00000017660, 0.80953941560, 508.35032409220},
  {0.00000015355, 5.81037986941, 1596.18644228460}, //40
  {0.00000017058, 4.20001977723, 2118.76386037840},
  {0.00000017040, 1.83402146640, 526.50957135690},
  {0.00000014661, 3.99989622586, 117.31986822020},
  {0.00000013639, 1.80336677963, 302.16477565500},
  {0.00000013230, 2.51856643603, 88.86568021700},
  {0.00000012756, 4.36856232414, 1169.58825140860},
  {0.00000015292, 0.68174165476, 942.06206196900},
  {0.00000010986, 4.43586634639, 525.75881183150},
  {0.00000013920, 5.95169568482, 316.39186965660},
  {0.00000009437, 2.17684563456, 1155.36115740700}, //50
  {0.00000008812, 3.29452783338, 220.41264243880},
  {0.00000007823, 5.75672228354, 846.08283475120},
  {0.00000007549, 2.70955516779, 533.62311835770},
  {0.00000009681, 1.71563161051, 1581.95934828300},
  {0.00000008690, 3.31924493607, 831.85574074960},
  {0.00000006285, 0.49939863541, 949.17560896980},
  {0.00000006685, 2.17560093281, 1265.56747862640},
  {-76.0}
} ;

const struct VSOP_87 Jupiter_L3[] =
{
  {0.00006501665, 2.59862880482, 7.11354700080},
  {0.00001356524, 1.34635886411, 529.69096509460},
  {0.00000470716, 2.47503977883, 14.22709400160},
  {0.00000416960, 3.24451243214, 536.80451209540},
  {0.00000352851, 2.97360159003, 522.57741809380},
  {0.00000154880, 2.07565585817, 1059.38193018920},
  {0.00000086771, 2.51431584316, 515.46387109300},
  {0.00000033538, 3.82633794497, 1066.49547719000},
  {0.00000044378, 0.00000000000, 0.00000000000},
  {0.00000022644, 2.98231326774, 543.91805909620}, //10
  {0.00000023737, 1.27667172313, 412.37109687440},
  {0.00000028457, 2.44754756058, 206.18554843720},
  {0.00000019798, 2.10099934005, 639.89728631400},
  {0.00000019740, 1.40255938973, 419.48464387520},
  {0.00000018768, 1.59368403500, 103.09277421860},
  {0.00000017033, 2.30214681202, 21.34064100240},
  {0.00000016774, 2.59821460673, 1589.07289528380},
  {0.00000016214, 3.14521117299, 625.67019231240},
  {0.00000016055, 3.36030126297, 1052.26838318840},
  {0.00000013392, 2.75973892202, 95.97922721780}, //20
  {0.00000013234, 2.53862244340, 199.07200143640},
  {0.00000012611, 6.26578110400, 426.59819087600},
  {0.00000008637, 2.26563256289, 110.20632121940},
  {0.00000006725, 3.42566433316, 309.27832265580},
  {0.00000008701, 1.76334960737, 10.29494073850},
  {0.00000006527, 4.03869562907, 728.76296653100},
  {0.00000005368, 5.25196153539, 323.50541665740},
  {0.00000005675, 2.52096417685, 508.35032409220},
  {0.00000005399, 2.91184687105, 1045.15483618760},
  {0.00000003996, 4.30290261177, 88.86568021700}, //30
  {0.00000003857, 3.52381361552, 302.16477565500},
  {0.00000003774, 4.09125315146, 735.87651353180},
  {0.00000003269, 1.43175991274, 956.28915597060},
  {0.00000002783, 4.35817507670, 1596.18644228460},
  {0.00000002661, 1.25276590759, 213.29909543800},
  {0.00000002553, 2.23785673285, 117.31986822020},
  {0.00000002371, 2.89662409244, 742.99006053260},
  {0.00000002656, 5.01505839848, 838.96928775040},
  {0.00000001948, 2.77248294666, 1169.58825140860},
  {-76.0}
} ;

const struct VSOP_87 Jupiter_L4[] =
{
  {0.00000669483, 0.85282421090, 7.11354700080},
  {0.00000099961, 0.74258947751, 14.22709400160},
  {0.00000114019, 3.14159265359, 0.00000000000},
  {0.00000050024, 1.65346208248, 536.80451209540},
  {0.00000043585, 5.82026386621, 529.69096509460},
  {0.00000031813, 4.85829986650, 522.57741809380},
  {0.00000014742, 4.29061635784, 515.46387109300},
  {0.00000008899, 0.71478520741, 1059.38193018920},
  {0.00000004957, 1.29502259434, 543.91805909620},
  {0.00000004484, 2.31715516627, 1066.49547719000}, //10
  {0.00000004251, 0.48326797501, 21.34064100240},
  {0.00000003100, 3.00245542678, 412.37109687440},
  {0.00000002055, 0.39858940218, 639.89728631400},
  {0.00000001762, 4.90536207307, 625.67019231240},
  {0.00000001902, 4.25925620271, 199.07200143640},
  {0.00000001695, 4.26147580803, 206.18554843720},
  {0.00000001375, 5.25546955667, 1052.26838318840},
  {0.00000001203, 4.71614633845, 95.97922721780},
  {0.00000001086, 1.28604571172, 1589.07289528380},
  {-76.0}
} ;

const struct VSOP_87 Jupiter_L5[] =
{
  {0.00000049577, 5.25658966184, 7.11354700080},
  {0.00000015761, 5.25126837478, 14.22709400160},
  {0.00000004343, 0.01461869263, 536.80451209540},
  {0.00000001526, 1.09739911439, 522.57741809380},
  {0.00000000728, 5.85949047619, 543.91805909620},
  {-76.0}
} ;

const struct VSOP_87 Jupiter_B0[] =
{
  {0.02268615703, 3.55852606718, 529.69096509460},
  {0.00109971634, 3.90809347389, 1059.38193018920},
  {0.00110090358, 0.00000000000, 0.00000000000},
  {0.00008101427, 3.60509573368, 522.57741809380},
  {0.00006043996, 4.25883108794, 1589.07289528380},
  {0.00006437782, 0.30627121409, 536.80451209540},
  {0.00001106880, 2.98534421928, 1162.47470440780},
  {0.00000941651, 2.93619072405, 1052.26838318840},
  {0.00000894088, 1.75447429921, 7.11354700080},
  {0.00000767280, 2.15473594060, 632.78373931320}, //10
  {0.00000944328, 1.67522288396, 426.59819087600},
  {0.00000684220, 3.67808770098, 213.29909543800},
  {0.00000629223, 0.64343282328, 1066.49547719000},
  {0.00000835861, 5.17881973234, 103.09277421860},
  {0.00000531670, 2.70305954352, 110.20632121940},
  {0.00000558524, 0.01354830508, 846.08283475120},
  {0.00000464449, 1.17337249185, 949.17560896980},
  {0.00000431072, 2.60825000494, 419.48464387520},
  {0.00000351433, 4.61062990714, 2118.76386037840},
  {0.00000123148, 3.34968181384, 1692.16566950240}, //20
  {0.00000115038, 5.04892295442, 316.39186965660},
  {0.00000132160, 4.77816990670, 742.99006053260},
  {0.00000103402, 2.31878999565, 1478.86657406440},
  {0.00000116379, 1.38688232033, 323.50541665740},
  {0.00000102420, 3.15293785436, 1581.95934828300},
  {0.00000103762, 3.70103838110, 515.46387109300},
  {-66.0}
} ;

const struct VSOP_87 Jupiter_B1[] =
{
  {0.00177351787, 5.70166488486, 529.69096509460},
  {0.00003230171, 5.77941619340, 1059.38193018920},
  {0.00003081364, 5.47464296527, 522.57741809380},
  {0.00002211914, 4.73477480209, 536.80451209540},
  {0.00001694232, 3.14159265359, 0.00000000000},
  {0.00000346445, 4.74595174109, 1052.26838318840},
  {0.00000234264, 5.18856099929, 1066.49547719000},
  {0.00000196154, 6.18554286642, 7.11354700080},
  {0.00000150468, 3.92721226087, 1589.07289528380},
  {0.00000114128, 3.43897271830, 632.78373931320}, //10
  {0.00000096667, 2.91426304090, 949.17560896980},
  {0.00000076599, 2.50522188662, 103.09277421860},
  {0.00000081671, 5.07666097497, 1162.47470440780},
  {0.00000076572, 0.61288981445, 419.48464387520},
  {0.00000073875, 5.49958292155, 515.46387109300},
  {0.00000049915, 3.94799616572, 735.87651353180},
  {0.00000060544, 5.44740084359, 213.29909543800},
  {0.00000036561, 4.69828392839, 543.91805909620},
  {0.00000046032, 0.53850360901, 110.20632121940},
  {0.00000045123, 1.89516645239, 846.08283475120}, //20
  {0.00000036019, 6.10952578764, 316.39186965660},
  {0.00000031975, 4.92452714629, 1581.95934828300},
  {-66.0}
} ;

const struct VSOP_87 Jupiter_B2[] =
{
  {0.00008094051, 1.46322843658, 529.69096509460},
  {0.00000742415, 0.95691639003, 522.57741809380},
  {0.00000813244, 3.14159265359, 0.00000000000},
  {0.00000398951, 2.89888666447, 536.80451209540},
  {0.00000342226, 1.44683789727, 1059.38193018920},
  {0.00000073948, 0.40724675866, 1052.26838318840},
  {0.00000046151, 3.48036895772, 1066.49547719000},
  {0.00000029314, 0.99088831805, 515.46387109300},
  {0.00000029717, 1.92504171329, 1589.07289528380},
  {0.00000022753, 4.27124052435, 7.11354700080}, //10
  {0.00000013916, 2.92242387338, 543.91805909620},
  {0.00000012067, 5.22168932482, 632.78373931320},
  {0.00000010703, 4.88024222475, 949.17560896980},
  {0.00000006078, 6.21089108431, 1045.15483618760},
  {-66.0}
} ;

const struct VSOP_87 Jupiter_B3[] =
{
  {0.00000251624, 3.38087923084, 529.69096509460},
  {0.00000121738, 2.73311837200, 522.57741809380},
  {0.00000048694, 1.03689996685, 536.80451209540},
  {0.00000010988, 2.31463561347, 1052.26838318840},
  {0.00000008067, 2.76729757621, 515.46387109300},
  {0.00000006205, 1.78115827370, 1066.49547719000},
  {0.00000007287, 4.25268318975, 1059.38193018920},
  {0.00000003627, 1.13028917221, 543.91805909620},
  {0.00000002798, 3.14159265359, 0.00000000000},
  {-66.0}
} ;

const struct VSOP_87 Jupiter_B4[] =
{
  {0.00000015050, 4.52956999637, 522.57741809380},
  {0.00000005370, 4.47427159142, 529.69096509460},
  {0.00000004456, 5.43908581047, 536.80451209540},
  {0.00000003422, 0.00000000000, 0.00000000000},
  {0.00000001833, 4.51807036227, 515.46387109300},
  {0.00000001322, 4.20117611581, 1052.26838318840},
  {-66.0}
} ;

const struct VSOP_87 Jupiter_B5[] =
{
  {0.00000001445, 0.09198554072, 522.57741809380},
  {-66.0}
} ;

const struct VSOP_87 Jupiter_R0[] =
{
  {5.20887429471, 0.00000000000, 0.00000000000},
  {0.25209327020, 3.49108640015, 529.69096509460},
  {0.00610599902, 3.84115365602, 1059.38193018920},
  {0.00282029465, 2.57419879933, 632.78373931320},
  {0.00187647391, 2.07590380082, 522.57741809380},
  {0.00086792941, 0.71001090609, 419.48464387520},
  {0.00072062869, 0.21465694745, 536.80451209540},
  {0.00065517227, 5.97995850843, 316.39186965660},
  {0.00029134620, 1.67759243710, 103.09277421860},
  {0.00030135275, 2.16132058449, 949.17560896980}, //10
  {0.00023453209, 3.54023147303, 735.87651353180},
  {0.00022283710, 4.19362773546, 1589.07289528380},
  {0.00023947340, 0.27457854894, 7.11354700080},
  {0.00013032600, 2.96043055741, 1162.47470440780},
  {0.00009703346, 1.90669572402, 206.18554843720},
  {0.00012749004, 2.71550102862, 1052.26838318840},
  {0.00009161431, 4.41352618935, 213.29909543800},
  {0.00007894539, 2.47907551404, 426.59819087600},
  {0.00007057978, 2.18184753111, 1265.56747862640},
  {0.00006137755, 6.26417542514, 846.08283475120}, //20
  {0.00005477093, 5.65729325169, 639.89728631400},
  {0.00003502519, 0.56531297394, 1066.49547719000},
  {0.00004136890, 2.72219979684, 625.67019231240},
  {0.00004170012, 2.01605033912, 515.46387109300},
  {0.00002499966, 4.55182055941, 838.96928775040},
  {0.00002616955, 2.00993967129, 1581.95934828300},
  {0.00001911876, 0.85621927419, 412.37109687440},
  {0.00002127644, 6.12751461750, 742.99006053260},
  {0.00001610549, 3.08867789275, 1368.66025284500},
  {0.00001479484, 2.68026191372, 1478.86657406440}, //30
  {0.00001230708, 1.89042979701, 323.50541665740},
  {0.00001216810, 1.80171561024, 110.20632121940},
  {0.00000961072, 4.54876989805, 2118.76386037840},
  {0.00000885708, 4.14785948471, 533.62311835770},
  {0.00000776700, 3.67696954690, 728.76296653100},
  {0.00000998579, 2.87208940110, 309.27832265580},
  {0.00001014959, 1.38673237666, 454.90936652730},
  {0.00000727162, 3.98824686402, 1155.36115740700},
  {0.00000655289, 2.79065604219, 1685.05212250160},
  {0.00000821465, 1.59342534396, 1898.35121793960}, //40
  {0.00000620798, 4.82284338962, 956.28915597060},
  {0.00000653981, 3.38150775269, 1692.16566950240},
  {0.00000812036, 5.94091899141, 909.81873305460},
  {0.00000562120, 0.08095987241, 543.91805909620},
  {0.00000542221, 0.28360266386, 525.75881183150},
  {0.00000457859, 0.12722694510, 1375.77379984580},
  {-82.0}
} ;

const struct VSOP_87 Jupiter_R1[] =
{
  {0.01271801596, 2.64937511122, 529.69096509460},
  {0.00061661771, 3.00076251018, 1059.38193018920},
  {0.00053443592, 3.89717644226, 522.57741809380},
  {0.00031185167, 4.88276663526, 536.80451209540},
  {0.00041390257, 0.00000000000, 0.00000000000},
  {0.00011847190, 2.41329588176, 419.48464387520},
  {0.00009166360, 4.75979408587, 7.11354700080},
  {0.00003175763, 2.79297987071, 103.09277421860},
  {0.00003203446, 5.21083285476, 735.87651353180},
  {0.00003403605, 3.34688537997, 1589.07289528380}, //10
  {0.00002600003, 3.63435101622, 206.18554843720},
  {0.00002412207, 1.46947308304, 426.59819087600},
  {0.00002806064, 3.74223693580, 515.46387109300},
  {0.00002676575, 4.33052878699, 1052.26838318840},
  {0.00002100507, 3.92762682306, 639.89728631400},
  {0.00001646182, 5.30953510947, 1066.49547719000},
  {0.00001641257, 4.41628669824, 625.67019231240},
  {0.00001049866, 3.16113622955, 213.29909543800},
  {0.00001024802, 2.55432643018, 412.37109687440},
  {0.00000740996, 2.17094630558, 1162.47470440780}, //20
  {0.00000806404, 2.67750801380, 632.78373931320},
  {0.00000676928, 6.24953479790, 838.96928775040},
  {0.00000468895, 4.70973463481, 543.91805909620},
  {0.00000444683, 0.40281181402, 323.50541665740},
  {0.00000567076, 4.57655414712, 742.99006053260},
  {0.00000415894, 5.36836018215, 728.76296653100},
  {0.00000484689, 2.46882793186, 949.17560896980},
  {0.00000337555, 3.16781951120, 956.28915597060},
  {0.00000401738, 4.60528841541, 309.27832265580},
  {0.00000347378, 4.68148808722, 14.22709400160}, //30
  {0.00000260753, 5.34290306101, 846.08283475120},
  {0.00000220084, 4.84210964963, 1368.66025284500},
  {0.00000203217, 5.59995425432, 1155.36115740700},
  {0.00000246603, 3.92313823537, 942.06206196900},
  {0.00000183504, 4.26526769703, 95.97922721780},
  {0.00000180134, 4.40165491159, 532.87235883230},
  {0.00000197134, 3.70551461394, 2118.76386037840},
  {0.00000196005, 3.75877587139, 199.07200143640},
  {0.00000200190, 4.43888814441, 1045.15483618760},
  {0.00000170225, 4.84647488867, 526.50957135690}, //40
  {0.00000146335, 6.12958365535, 533.62311835770},
  {0.00000133483, 1.32245735855, 110.20632121940},
  {0.00000132076, 4.51187950811, 525.75881183150},
  {-82.0}
} ;

const struct VSOP_87 Jupiter_R2[] =
{
  {0.00079644833, 1.35865896596, 529.69096509460},
  {0.00008251618, 5.77773935444, 522.57741809380},
  {0.00007029864, 3.27476965833, 536.80451209540},
  {0.00005314006, 1.83835109712, 1059.38193018920},
  {0.00001860833, 2.97682139367, 7.11354700080},
  {0.00000836267, 4.19889881718, 419.48464387520},
  {0.00000964466, 5.48031822015, 515.46387109300},
  {0.00000406453, 3.78250730354, 1066.49547719000},
  {0.00000426570, 2.22753101795, 639.89728631400},
  {0.00000377316, 2.24248352873, 1589.07289528380}, //10
  {0.00000497920, 3.14159265359, 0.00000000000},
  {0.00000339043, 6.12690864038, 625.67019231240},
  {0.00000362943, 5.36761847267, 206.18554843720},
  {0.00000342048, 6.09922969324, 1052.26838318840},
  {0.00000279920, 4.26162555827, 412.37109687440},
  {0.00000332578, 0.00328961161, 426.59819087600},
  {0.00000229777, 0.70530766213, 735.87651353180},
  {0.00000200783, 3.06850623368, 543.91805909620},
  {0.00000199807, 4.42884165317, 103.09277421860},
  {0.00000257290, 0.96295364983, 632.78373931320}, //20
  {0.00000138606, 2.93235671606, 14.22709400160},
  {0.00000113535, 0.78713911289, 728.76296653100},
  {0.00000086025, 5.14434751994, 323.50541665740},
  {0.00000094565, 1.70498041073, 838.96928775040},
  {0.00000083469, 0.05834873484, 309.27832265580},
  {0.00000075198, 1.60495195911, 956.28915597060},
  {0.00000070451, 1.50988357484, 213.29909543800},
  {0.00000080328, 2.98122361797, 742.99006053260},
  {0.00000056203, 0.95534810533, 1162.47470440780},
  {0.00000061649, 6.10137889854, 1045.15483618760}, //30
  {0.00000066572, 5.47307178077, 199.07200143640},
  {0.00000050057, 2.72063162317, 532.87235883230},
  {0.00000051904, 5.58435625607, 942.06206196900},
  {0.00000039833, 5.94566506227, 95.97922721780},
  {0.00000044548, 5.52445621411, 508.35032409220},
  {0.00000044282, 0.27118152557, 526.50957135690},
  {-82.0}
} ;

const struct VSOP_87 Jupiter_R3[] =
{
  {0.00003519257, 6.05800633846, 529.69096509460},
  {0.00001073239, 1.67321345760, 536.80451209540},
  {0.00000915666, 1.41329676116, 522.57741809380},
  {0.00000341593, 0.52296542656, 1059.38193018920},
  {0.00000254893, 1.19625473533, 7.11354700080},
  {0.00000221512, 0.95225226237, 515.46387109300},
  {0.00000069078, 2.26885282314, 1066.49547719000},
  {0.00000089729, 3.14159265359, 0.00000000000},
  {0.00000057827, 1.41389745339, 543.91805909620},
  {0.00000057653, 0.52580117593, 639.89728631400}, //10
  {0.00000051079, 5.98016364677, 412.37109687440},
  {0.00000046935, 1.57864237959, 625.67019231240},
  {0.00000042824, 6.11689609099, 419.48464387520},
  {0.00000037477, 1.18262762330, 14.22709400160},
  {0.00000033816, 1.66671706951, 1052.26838318840},
  {0.00000031195, 1.04290245896, 1589.07289528380},
  {0.00000030023, 4.63236245032, 426.59819087600},
  {0.00000033531, 0.84784977903, 206.18554843720},
  {0.00000020804, 2.50071243814, 728.76296653100},
  {0.00000014466, 0.96040197071, 508.35032409220}, //20
  {0.00000012969, 1.50233788550, 1045.15483618760},
  {0.00000011654, 3.55513510121, 323.50541665740},
  {0.00000012319, 2.60952614503, 735.87651353180},
  {0.00000015023, 0.89136998434, 199.07200143640},
  {0.00000011160, 1.79041437555, 309.27832265580},
  {0.00000010554, 6.27845112678, 956.28915597060},
  {0.00000009812, 6.26016859519, 103.09277421860},
  {0.00000009301, 3.45126812476, 838.96928775040},
  {-82.0}
} ;

const struct VSOP_87 Jupiter_R4[] =
{
  {0.00000128628, 0.08419309557, 536.80451209540},
  {0.00000113458, 4.24858855779, 529.69096509460},
  {0.00000082650, 3.29754909408, 522.57741809380},
  {0.00000037883, 2.73326611144, 515.46387109300},
  {0.00000026694, 5.69142588558, 7.11354700080},
  {0.00000017650, 5.40012536918, 1059.38193018920},
  {0.00000012612, 6.01560416057, 543.91805909620},
  {0.00000009287, 0.76813946494, 1066.49547719000},
  {0.00000008107, 5.68228065707, 14.22709400160},
  {0.00000006271, 5.12286932534, 639.89728631400}, //10
  {0.00000006978, 1.42751292055, 412.37109687440},
  {0.00000005377, 3.33501947275, 625.67019231240},
  {0.00000002911, 3.40334805052, 1052.26838318840},
  {0.00000002593, 4.16090412984, 728.76296653100},
  {0.00000002562, 2.89802035072, 426.59819087600},
  {-82.0}
} ;

const struct VSOP_87 Jupiter_R5[] =
{
  {0.00000011188, 4.75249399945, 536.80451209540},
  {0.00000004255, 5.91516229170, 522.57741809380},
  {0.00000002079, 5.56781555864, 515.46387109300},
  {0.00000001908, 4.29659647286, 543.91805909620},
  {0.00000001875, 3.69357495838, 7.11354700080},
  {0.00000001590, 5.49312796166, 1066.49547719000},
  {0.00000001612, 4.13222808529, 1059.38193018920},
  {-82.0}
} ;

const struct VSOP_87 Saturn_L0[] =
{
  {0.87401354029, 0.00000000000, 0.00000000000},
  {0.11107659780, 3.96205090194, 213.29909543800},
  {0.01414150958, 4.58581515873, 7.11354700080},
  {0.00398379386, 0.52112025957, 206.18554843720},
  {0.00350769223, 3.30329903015, 426.59819087600},
  {0.00206816296, 0.24658366938, 103.09277421860},
  {0.00079271288, 3.84007078530, 220.41264243880},
  {0.00023990338, 4.66976934860, 110.20632121940},
  {0.00016573583, 0.43719123541, 419.48464387520},
  {0.00014906995, 5.76903283845, 316.39186965660}, //10
  {0.00015820300, 0.93808953760, 632.78373931320},
  {0.00014609562, 1.56518573691, 3.93215326310},
  {0.00013160308, 4.44891180176, 14.22709400160},
  {0.00015053509, 2.71670027883, 639.89728631400},
  {0.00013005305, 5.98119067061, 11.04570026390},
  {0.00010725066, 3.12939596466, 202.25339517410},
  {0.00005863207, 0.23657028777, 529.69096509460},
  {0.00005227771, 4.20783162380, 3.18139373770},
  {0.00006126308, 1.76328499656, 277.03499374140},
  {0.00005019658, 3.17787919533, 433.71173787680}, //20
  {0.00004592541, 0.61976424374, 199.07200143640},
  {0.00004005862, 2.24479893937, 63.73589830340},
  {0.00002953815, 0.98280385206, 95.97922721780},
  {0.00003873696, 3.22282692566, 138.51749687070},
  {0.00002461172, 2.03163631205, 735.87651353180},
  {0.00003269490, 0.77491895787, 949.17560896980},
  {0.00001758143, 3.26580514774, 522.57741809380},
  {0.00001640183, 5.50504966218, 846.08283475120},
  {0.00001391336, 4.02331978116, 323.50541665740},
  {0.00001580641, 4.37266314120, 309.27832265580}, //30
  {0.00001123515, 2.83726793572, 415.55249061210},
  {0.00001017258, 3.71698151814, 227.52618943960},
  {0.00000848643, 3.19149825839, 209.36694217490},
  {0.00001087237, 4.18343232481, 2.44768055480},
  {0.00000956752, 0.50740889886, 1265.56747862640},
  {0.00000789205, 5.00745123149, 0.96320784650},
  {0.00000686965, 1.74714407827, 1052.26838318840},
  {0.00000654470, 1.59889331515, 0.04818410980},
  {0.00000748811, 2.14398149298, 853.19638175200},
  {0.00000633980, 2.29889903023, 412.37109687440}, //40
  {0.00000743584, 5.25276954625, 224.34479570190},
  {0.00000852677, 3.42141350697, 175.16605980020},
  {0.00000579857, 3.09259007048, 74.78159856730},
  {0.00000624904, 0.97046831256, 210.11770170030},
  {0.00000529861, 4.44938897119, 117.31986822020},
  {0.00000542643, 1.51824320514, 9.56122755560},
  {0.00000474279, 5.47527185987, 742.99006053260},
  {0.00000448542, 1.28990416161, 127.47179660680},
  {0.00000546358, 2.12678554211, 350.33211960040},
  {0.00000478054, 2.96488054338, 137.03302416240}, //50
  {0.00000354944, 3.01286483030, 838.96928775040},
  {0.00000451827, 1.04436664241, 490.33408917940},
  {0.00000347413, 1.53928227764, 340.77089204480},
  {0.00000343475, 0.24604039134, 0.52126486180},
  {0.00000309001, 3.49486734909, 216.48048917570},
  {0.00000322185, 0.96137456104, 203.73786788240},
  {0.00000372308, 2.27819108625, 217.23124870110},
  {0.00000321543, 2.57182354537, 647.01083331480},
  {0.00000330196, 0.24715617844, 1581.95934828300},
  {0.00000249116, 1.47010534421, 1368.66025284500}, //60
  {0.00000286688, 2.37043745859, 351.81659230870},
  {0.00000220225, 4.20422424873, 200.76892246580},
  {0.00000277775, 0.40020408926, 211.81462272970},
  {0.00000204500, 6.01082206600, 265.98929347750},
  {0.00000207663, 0.48349820488, 1162.47470440780},
  {0.00000208655, 1.34516255304, 625.67019231240},
  {0.00000182454, 5.49122292426, 2.92076130680},
  {0.00000226609, 4.91003163138, 12.53017297220},
  {0.00000207659, 1.28302218900, 39.35687591520},
  {0.00000173914, 1.86305806814, 0.75075952540}, //70
  {0.00000184690, 3.50344404958, 149.56319713460},
  {0.00000183511, 0.97254952728, 4.19278569400},
  {0.00000146068, 6.23102544071, 195.13984817330},
  {0.00000164541, 0.44005517520, 5.41662597140},
  {0.00000147526, 1.53529320509, 5.62907429250},
  {0.00000139666, 4.29450260069, 21.34064100240},
  {0.00000131283, 4.06828961903, 10.29494073850},
  {0.00000117283, 2.67920400584, 1155.36115740700},
  {0.00000149299, 5.73594349789, 52.69019803950},
  {0.00000122373, 1.97588777199, 4.66586644600}, //80
  {0.00000113747, 5.59427544714, 1059.38193018920},
  {0.00000102702, 1.19748124058, 1685.05212250160},
  {0.00000118156, 5.34072933900, 554.06998748280},
  {0.00000109275, 3.43812715686, 536.80451209540},
  {0.00000110399, 0.16604024090, 1.48447270830},
  {0.00000124969, 6.27737805832, 1898.35121793960},
  {0.00000089949, 5.80392934702, 114.13847448250},
  {0.00000103956, 2.19210363069, 88.86568021700},
  {0.00000112437, 1.10502663534, 191.20769491020},
  {0.00000106570, 4.01156608514, 956.28915597060}, //90
  {-76.0}
} ;

const struct VSOP_87 Saturn_L1[] =
{
  {213.54295595986, 0.00000000000, 0.00000000000},
  {0.01296855005, 1.82820544701, 213.29909543800},
  {0.00564347566, 2.88500136429, 7.11354700080},
  {0.00098323030, 1.08070061328, 426.59819087600},
  {0.00107678770, 2.27769911872, 206.18554843720},
  {0.00040254586, 2.04128257090, 220.41264243880},
  {0.00019941734, 1.27954662736, 103.09277421860},
  {0.00010511706, 2.74880392800, 14.22709400160},
  {0.00006939233, 0.40493079985, 639.89728631400},
  {0.00004803325, 2.44194097666, 419.48464387520}, //10
  {0.00004056325, 2.92166618776, 110.20632121940},
  {0.00003768630, 3.64965631460, 3.93215326310},
  {0.00003384684, 2.41694251653, 3.18139373770},
  {0.00003302200, 1.26256486715, 433.71173787680},
  {0.00003071382, 2.32739317750, 199.07200143640},
  {0.00001953036, 3.56394683300, 11.04570026390},
  {0.00001249348, 2.62803737519, 95.97922721780},
  {0.00000921683, 1.96089834250, 227.52618943960},
  {0.00000705587, 4.41689249330, 529.69096509460},
  {0.00000649654, 6.17418093659, 202.25339517410}, //20
  {0.00000627603, 6.11088227167, 309.27832265580},
  {0.00000486843, 6.03998200305, 853.19638175200},
  {0.00000468377, 4.61707843907, 63.73589830340},
  {0.00000478501, 4.98776987984, 522.57741809380},
  {0.00000417010, 2.11708169277, 323.50541665740},
  {0.00000407630, 1.29949556676, 209.36694217490},
  {0.00000343826, 3.95854178574, 412.37109687440},
  {0.00000339724, 3.63396398752, 316.39186965660},
  {0.00000335936, 3.77173072712, 735.87651353180},
  {0.00000331933, 2.86077699882, 210.11770170030}, //30
  {0.00000352489, 2.31707079463, 632.78373931320},
  {0.00000289429, 2.73263080235, 117.31986822020},
  {0.00000265801, 0.54344631312, 647.01083331480},
  {0.00000230493, 1.64428879621, 216.48048917570},
  {0.00000280911, 5.74398845416, 2.44768055480},
  {0.00000191667, 2.96512946582, 224.34479570190},
  {0.00000172891, 4.07695221044, 846.08283475120},
  {0.00000167131, 2.59745202658, 21.34064100240},
  {0.00000136328, 2.28580246629, 10.29494073850},
  {0.00000131364, 3.44108355646, 742.99006053260}, //40
  {0.00000127838, 4.09533471247, 217.23124870110},
  {0.00000108862, 6.16141072262, 415.55249061210},
  {0.00000093909, 3.48397279899, 1052.26838318840},
  {0.00000092482, 3.94755499926, 88.86568021700},
  {0.00000097584, 4.72845436677, 838.96928775040},
  {0.00000086600, 1.21951325061, 440.82528487760},
  {0.00000083463, 3.11269504725, 625.67019231240},
  {0.00000077588, 6.24408938835, 302.16477565500},
  {0.00000061557, 1.82789612597, 195.13984817330},
  {0.00000061900, 4.29344363385, 127.47179660680}, //50
  {0.00000067106, 0.28961738595, 4.66586644600},
  {0.00000056919, 5.01889578112, 137.03302416240},
  {0.00000054160, 5.12628572382, 490.33408917940},
  {0.00000054585, 0.28356341456, 74.78159856730},
  {0.00000051425, 1.45766406064, 536.80451209540},
  {0.00000065843, 5.64757042732, 9.56122755560},
  {0.00000057780, 2.47630552035, 191.95845443560},
  {0.00000044444, 2.70873627665, 5.41662597140},
  {0.00000046799, 1.17721211050, 149.56319713460},
  {0.00000040380, 3.88870105683, 728.76296653100}, //60
  {0.00000037768, 2.53379013859, 12.53017297220},
  {0.00000046649, 5.14818326902, 515.46387109300},
  {0.00000045891, 2.23198878761, 956.28915597060},
  {0.00000040400, 0.41281520440, 269.92144674060},
  {0.00000037191, 3.78239026411, 2.92076130680},
  {0.00000033778, 3.21070688046, 1368.66025284500},
  {0.00000037969, 0.64665967180, 422.66603761290},
  {0.00000032857, 0.30063884563, 351.81659230870},
  {0.00000033050, 5.43038091186, 1066.49547719000},
  {0.00000030276, 2.84067004928, 203.00415469950}, //70
  {0.00000035116, 6.08421794089, 5.62907429250},
  {0.00000029667, 3.39052569135, 1059.38193018920},
  {0.00000033217, 4.64063092111, 277.03499374140},
  {0.00000031876, 4.38622923770, 1155.36115740700},
  {0.00000028913, 2.02614760507, 330.61896365820},
  {0.00000028264, 2.74178953996, 265.98929347750},
  {0.00000030089, 6.18684614308, 284.14854074220},
  {0.00000031329, 2.43455855525, 52.69019803950},
  {0.00000026493, 4.51214170121, 340.77089204480},
  {-76.0}
} ;

const struct VSOP_87 Saturn_L2[] =
{
  {0.00116441181, 1.17987850633, 7.11354700080},
  {0.00091920844, 0.07425261094, 213.29909543800},
  {0.00090592251, 0.00000000000, 0.00000000000},
  {0.00015276909, 4.06492007503, 206.18554843720},
  {0.00010631396, 0.25778277414, 220.41264243880},
  {0.00010604979, 5.40963595885, 426.59819087600},
  {0.00004265368, 1.04595556630, 14.22709400160},
  {0.00001215527, 2.91860042123, 103.09277421860},
  {0.00001164684, 4.60942128971, 639.89728631400},
  {0.00001081967, 5.69130351670, 433.71173787680}, //10
  {0.00001020079, 0.63369182642, 3.18139373770},
  {0.00001044754, 4.04206453611, 199.07200143640},
  {0.00000633582, 4.38825410036, 419.48464387520},
  {0.00000549329, 5.57303134242, 3.93215326310},
  {0.00000456914, 1.26840971349, 110.20632121940},
  {0.00000425100, 0.20935499279, 227.52618943960},
  {0.00000273739, 4.28841011784, 95.97922721780},
  {0.00000161571, 1.38139149420, 11.04570026390},
  {0.00000129494, 1.56586884170, 309.27832265580},
  {0.00000117008, 3.88120915956, 853.19638175200}, //20
  {0.00000105415, 4.90003203599, 647.01083331480},
  {0.00000100967, 0.89270493100, 21.34064100240},
  {0.00000095227, 5.62561150598, 412.37109687440},
  {0.00000081948, 1.02477558315, 117.31986822020},
  {0.00000074857, 4.76178468163, 210.11770170030},
  {0.00000082727, 6.05030934786, 216.48048917570},
  {0.00000095659, 2.91093561539, 316.39186965660},
  {0.00000063696, 0.35179804917, 323.50541665740},
  {0.00000084860, 5.73472777961, 209.36694217490},
  {0.00000060647, 4.87517850190, 632.78373931320}, //30
  {0.00000066459, 0.48297940601, 10.29494073850},
  {0.00000067184, 0.45648612616, 522.57741809380},
  {0.00000053281, 2.74730541387, 529.69096509460},
  {0.00000045827, 5.69296621745, 440.82528487760},
  {0.00000045293, 1.66856699796, 202.25339517410},
  {0.00000042330, 5.70768187703, 88.86568021700},
  {0.00000032140, 0.07050050346, 63.73589830340},
  {0.00000031573, 1.67190022213, 302.16477565500},
  {0.00000031150, 4.16379537691, 191.95845443560},
  {0.00000024631, 5.65564728570, 735.87651353180}, //40
  {0.00000026558, 0.83256214407, 224.34479570190},
  {0.00000020108, 5.94364609981, 217.23124870110},
  {0.00000017511, 4.90014736798, 625.67019231240},
  {0.00000017130, 1.62593421274, 742.99006053260},
  {0.00000013744, 3.76497167300, 195.13984817330},
  {0.00000012236, 4.71789723976, 203.00415469950},
  {0.00000011940, 0.12620714199, 234.63973644040},
  {0.00000016040, 0.57886320845, 515.46387109300},
  {0.00000011154, 5.92216844780, 536.80451209540},
  {0.00000014068, 0.20675293700, 838.96928775040}, //50
  {0.00000011013, 5.60207982774, 728.76296653100},
  {0.00000011718, 3.12098483554, 846.08283475120},
  {0.00000009962, 4.15472049127, 860.30992875280},
  {0.00000010601, 3.20327613035, 1066.49547719000},
  {0.00000010072, 0.25709351996, 330.61896365820},
  {0.00000009490, 0.46379969328, 956.28915597060},
  {0.00000010240, 4.98736656070, 422.66603761290},
  {0.00000008287, 2.13990364272, 269.92144674060},
  {0.00000007238, 5.39724715258, 1052.26838318840},
  {0.00000007730, 5.24602742309, 429.77958461370}, //60
  {0.00000006353, 4.46211130731, 284.14854074220},
  {0.00000005935, 5.40967847103, 149.56319713460},
  {0.00000007550, 4.03401153929, 9.56122755560},
  {-76.0}
} ;

const struct VSOP_87 Saturn_L3[] =
{
  {0.00016038734, 5.73945377424, 7.11354700080},
  {0.00004249793, 4.58539675603, 213.29909543800},
  {0.00001906524, 4.76082050205, 220.41264243880},
  {0.00001465687, 5.91326678323, 206.18554843720},
  {0.00001162041, 5.61973132428, 14.22709400160},
  {0.00001066581, 3.60816533142, 426.59819087600},
  {0.00000239377, 3.86088273439, 433.71173787680},
  {0.00000236975, 5.76826451465, 199.07200143640},
  {0.00000165641, 5.11641150216, 3.18139373770},
  {0.00000131409, 4.74327544615, 227.52618943960}, //10
  {0.00000151352, 2.73594641861, 639.89728631400},
  {0.00000061630, 4.74287052463, 103.09277421860},
  {0.00000063365, 0.22850089497, 419.48464387520},
  {0.00000040437, 5.47298059144, 21.34064100240},
  {0.00000040205, 5.96420266720, 95.97922721780},
  {0.00000038746, 5.83386199529, 110.20632121940},
  {0.00000028025, 3.01235311514, 647.01083331480},
  {0.00000025029, 0.98808170740, 3.93215326310},
  {0.00000018101, 1.02506397063, 412.37109687440},
  {0.00000017879, 3.31913418974, 309.27832265580}, //20
  {0.00000016208, 3.89825272754, 440.82528487760},
  {0.00000015763, 5.61667809625, 117.31986822020},
  {0.00000019014, 1.91614237463, 853.19638175200},
  {0.00000018262, 4.96738415934, 10.29494073850},
  {0.00000012947, 1.18068953942, 88.86568021700},
  {0.00000017919, 4.20376505349, 216.48048917570},
  {0.00000011453, 5.57520615096, 11.04570026390},
  {0.00000010548, 5.92906266269, 191.95845443560},
  {0.00000010389, 3.94838736947, 209.36694217490},
  {0.00000008650, 3.39335369698, 302.16477565500}, //30
  {0.00000007580, 4.87736913157, 323.50541665740},
  {0.00000006697, 0.38198725552, 632.78373931320},
  {0.00000005864, 1.05621157685, 210.11770170030},
  {0.00000005449, 4.64268475485, 234.63973644040},
  {0.00000006327, 2.25492722762, 522.57741809380},
  {0.00000003602, 2.30677010956, 515.46387109300},
  {0.00000003229, 2.20309400066, 860.30992875280},
  {0.00000003701, 3.14159265359, 0.00000000000},
  {0.00000002583, 4.93447677059, 224.34479570190},
  {0.00000002543, 0.42393884183, 625.67019231240}, //40
  {0.00000002213, 3.19814958289, 202.25339517410},
  {0.00000002421, 4.76621391814, 330.61896365820},
  {0.00000002850, 0.58604395010, 529.69096509460},
  {0.00000001965, 4.39525359412, 124.43341522100},
  {0.00000002154, 1.35488209144, 405.25754987360},
  {0.00000002296, 3.34809165905, 429.77958461370},
  {0.00000002018, 3.06693569701, 654.12438031560},
  {0.00000001979, 1.02981005658, 728.76296653100},
  {-76.0}
} ;

const struct VSOP_87 Saturn_L4[] =
{
  {0.00001661894, 3.99826248978, 7.11354700080},
  {0.00000257107, 2.98436499013, 220.41264243880},
  {0.00000236344, 3.90241428075, 14.22709400160},
  {0.00000149418, 2.74110824208, 213.29909543800},
  {0.00000109598, 1.51515739251, 206.18554843720},
  {0.00000113953, 3.14159265359, 0.00000000000},
  {0.00000068390, 1.72120953337, 426.59819087600},
  {0.00000037699, 1.23795458356, 199.07200143640},
  {0.00000040060, 2.04644897412, 433.71173787680},
  {0.00000031219, 3.01094184090, 227.52618943960}, //10
  {0.00000015111, 0.82897064529, 639.89728631400},
  {0.00000009444, 3.71485300868, 21.34064100240},
  {0.00000005690, 2.41995290633, 419.48464387520},
  {0.00000004470, 1.45120818748, 95.97922721780},
  {0.00000005608, 1.15607095740, 647.01083331480},
  {0.00000004463, 2.11783225176, 440.82528487760},
  {0.00000003229, 4.09278077834, 110.20632121940},
  {0.00000002871, 2.77203153866, 412.37109687440},
  {0.00000002796, 3.00730249564, 88.86568021700},
  {0.00000002638, 0.00255721254, 853.19638175200}, //20
  {0.00000002574, 0.39246854091, 103.09277421860},
  {0.00000001862, 5.07955457727, 309.27832265580},
  {0.00000002225, 3.77689198137, 117.31986822020},
  {0.00000001769, 5.19176876406, 302.16477565500},
  {0.00000001921, 2.82884328662, 234.63973644040},
  {0.00000001805, 2.23816036743, 216.48048917570},
  {0.00000001211, 1.54685246534, 191.95845443560},
  {-76.0}
} ;

const struct VSOP_87 Saturn_L5[] =
{
  {0.00000123615, 2.25923345732, 7.11354700080},
  {0.00000034190, 2.16250652689, 14.22709400160},
  {0.00000027546, 1.19868150215, 220.41264243880},
  {0.00000005818, 1.21584270184, 227.52618943960},
  {0.00000005318, 0.23550400093, 433.71173787680},
  {0.00000003677, 6.22669694355, 426.59819087600},
  {0.00000003057, 2.97372046322, 199.07200143640},
  {0.00000002861, 4.28710932685, 206.18554843720},
  {0.00000001617, 6.25265362286, 213.29909543800},
  {0.00000001279, 5.27612561266, 639.89728631400}, //10
  {0.00000000932, 5.56741549127, 647.01083331480},
  {0.00000000756, 6.17716234487, 191.95845443560},
  {-76.0}
} ;

const struct VSOP_87 Saturn_B0[] =
{
  {0.04330678040, 3.60284428399, 213.29909543800},
  {0.00240348303, 2.85238489390, 426.59819087600},
  {0.00084745939, 0.00000000000, 0.00000000000},
  {0.00030863357, 3.48441504465, 220.41264243880},
  {0.00034116063, 0.57297307844, 206.18554843720},
  {0.00014734070, 2.11846597870, 639.89728631400},
  {0.00009916668, 5.79003189405, 419.48464387520},
  {0.00006993564, 4.73604689179, 7.11354700080},
  {0.00004807587, 5.43305315602, 316.39186965660},
  {0.00004788392, 4.96512927420, 110.20632121940}, //10
  {0.00003432125, 2.73255752123, 433.71173787680},
  {0.00001506129, 6.01304536144, 103.09277421860},
  {0.00001060298, 5.63099292414, 529.69096509460},
  {0.00000969071, 5.20434966103, 632.78373931320},
  {0.00000942050, 1.39646678088, 853.19638175200},
  {0.00000707645, 3.80302329547, 323.50541665740},
  {0.00000552313, 5.13149109045, 202.25339517410},
  {0.00000399675, 3.35891413961, 227.52618943960},
  {0.00000316063, 1.99716764199, 647.01083331480},
  {0.00000319380, 3.62571550980, 209.36694217490}, //20
  {0.00000284494, 4.88648481625, 224.34479570190},
  {0.00000314225, 0.46510272410, 217.23124870110},
  {0.00000236442, 2.13887472281, 11.04570026390},
  {0.00000215354, 5.94982610103, 846.08283475120},
  {0.00000208522, 2.12003893769, 415.55249061210},
  {0.00000178958, 2.95361514672, 63.73589830340},
  {0.00000207213, 0.73021462851, 199.07200143640},
  {0.00000139140, 1.99821990940, 735.87651353180},
  {0.00000134884, 5.24500819605, 742.99006053260},
  {0.00000140585, 0.64417620299, 490.33408917940}, //30
  {0.00000121669, 3.11537140876, 522.57741809380},
  {0.00000139240, 4.59535168021, 14.22709400160},
  {0.00000115524, 3.10891547171, 216.48048917570},
  {0.00000114218, 0.96261442133, 210.11770170030},
  {-66.0}
} ;

const struct VSOP_87 Saturn_B1[] =
{
  {0.00397554998, 5.33289992556, 213.29909543800},
  {0.00049478641, 3.14159265359, 0.00000000000},
  {0.00018571607, 6.09919206378, 426.59819087600},
  {0.00014800587, 2.30586060520, 206.18554843720},
  {0.00009643981, 1.69674660120, 220.41264243880},
  {0.00003757161, 1.25429514018, 419.48464387520},
  {0.00002716647, 5.91166664787, 639.89728631400},
  {0.00001455309, 0.85161616532, 433.71173787680},
  {0.00001290595, 2.91770857090, 7.11354700080},
  {0.00000852630, 0.43572078997, 316.39186965660}, //10
  {0.00000284386, 1.61881754773, 227.52618943960},
  {0.00000292185, 5.31574251270, 853.19638175200},
  {0.00000275090, 3.88864137336, 103.09277421860},
  {0.00000297726, 0.91909206723, 632.78373931320},
  {0.00000172359, 0.05215146556, 647.01083331480},
  {0.00000127731, 1.20711452525, 529.69096509460},
  {0.00000166237, 2.44351613165, 199.07200143640},
  {0.00000158220, 5.20850125766, 110.20632121940},
  {0.00000109839, 2.45695551627, 217.23124870110},
  {0.00000081759, 2.75839171353, 210.11770170030}, //20
  {0.00000081010, 2.86038377187, 14.22709400160},
  {0.00000068658, 1.65537623146, 202.25339517410},
  {0.00000059281, 1.82410768234, 323.50541665740},
  {0.00000065161, 1.25527521313, 216.48048917570},
  {0.00000061024, 1.25273412095, 209.36694217490},
  {0.00000046386, 0.81534705304, 440.82528487760},
  {0.00000036163, 1.81851057689, 224.34479570190},
  {0.00000034041, 2.83971297997, 117.31986822020},
  {0.00000032164, 1.18676132343, 846.08283475120},
  {0.00000033114, 1.30557080010, 412.37109687440}, //30
  {0.00000027282, 4.64744847591, 1066.49547719000},
  {0.00000022805, 4.12923703368, 415.55249061210},
  {-66.0}
} ;

const struct VSOP_87 Saturn_B2[] =
{
  {0.00020629977, 0.50482422817, 213.29909543800},
  {0.00003719555, 3.99833475829, 206.18554843720},
  {0.00001627158, 6.18189939500, 220.41264243880},
  {0.00001346067, 0.00000000000, 0.00000000000},
  {0.00000705842, 3.03914308836, 419.48464387520},
  {0.00000365042, 5.09928680706, 426.59819087600},
  {0.00000329632, 5.27899210039, 433.71173787680},
  {0.00000219335, 3.82841533795, 639.89728631400},
  {0.00000139393, 1.04272623499, 7.11354700080},
  {0.00000103980, 6.15730992966, 227.52618943960}, //10
  {0.00000092961, 1.97994412845, 316.39186965660},
  {0.00000071242, 4.14754353431, 199.07200143640},
  {0.00000051927, 2.88364833898, 632.78373931320},
  {0.00000048961, 4.43390206741, 647.01083331480},
  {0.00000041373, 3.15927770079, 853.19638175200},
  {0.00000028602, 4.52978327558, 210.11770170030},
  {0.00000023969, 1.11595912146, 14.22709400160},
  {0.00000020511, 4.35095844197, 217.23124870110},
  {0.00000019532, 5.30779711223, 440.82528487760},
  {0.00000018263, 0.85391476786, 110.20632121940}, //20
  {0.00000015742, 4.25767226302, 103.09277421860},
  {0.00000016840, 5.68112084135, 216.48048917570},
  {0.00000013613, 2.99904334066, 412.37109687440},
  {0.00000011567, 2.52679928410, 529.69096509460},
  {0.00000007963, 3.31512423920, 202.25339517410},
  {0.00000006599, 0.28766025146, 323.50541665740},
  {0.00000006312, 1.16121321336, 117.31986822020},
  {0.00000005891, 3.58260177246, 309.27832265580},
  {0.00000006648, 5.55714129949, 209.36694217490},
  {-66.0}
} ;

const struct VSOP_87 Saturn_B3[] =
{
  {0.00000666252, 1.99006340181, 213.29909543800},
  {0.00000632350, 5.69778316807, 206.18554843720},
  {0.00000398051, 0.00000000000, 0.00000000000},
  {0.00000187838, 4.33779804809, 220.41264243880},
  {0.00000091884, 4.84104208217, 419.48464387520},
  {0.00000042369, 2.38073239056, 426.59819087600},
  {0.00000051548, 3.42149490328, 433.71173787680},
  {0.00000025661, 4.40167213109, 227.52618943960},
  {0.00000020551, 5.85313509872, 199.07200143640},
  {0.00000018081, 1.99321433229, 639.89728631400}, //10
  {0.00000010874, 5.37344546547, 7.11354700080},
  {0.00000009590, 2.54901825866, 647.01083331480},
  {0.00000007085, 3.45518372721, 316.39186965660},
  {0.00000006002, 4.80055225135, 632.78373931320},
  {0.00000005778, 0.01680378777, 210.11770170030},
  {0.00000004881, 5.63719730884, 14.22709400160},
  {0.00000004501, 1.22424419010, 853.19638175200},
  {0.00000005542, 3.51756747774, 440.82528487760},
  {0.00000003548, 4.71299370890, 412.37109687440},
  {0.00000002851, 0.62679207578, 103.09277421860}, //20
  {0.00000002173, 3.71982274459, 216.48048917570},
  {-66.0}
} ;

const struct VSOP_87 Saturn_B4[] =
{
  {0.00000080384, 1.11918414679, 206.18554843720},
  {0.00000031660, 3.12218745098, 213.29909543800},
  {0.00000017143, 2.48073200414, 220.41264243880},
  {0.00000011844, 3.14159265359, 0.00000000000},
  {0.00000009005, 0.38441424927, 419.48464387520},
  {0.00000006164, 1.56186379537, 433.71173787680},
  {0.00000004660, 1.28235639570, 199.07200143640},
  {0.00000004775, 2.63498295487, 227.52618943960},
  {0.00000001487, 1.43096671616, 426.59819087600},
  {0.00000001424, 0.66988083613, 647.01083331480}, //10
  {0.00000001075, 6.18092274059, 639.89728631400},
  {0.00000001145, 1.72041928134, 440.82528487760},
  {-66.0}
} ;

const struct VSOP_87 Saturn_B5[] =
{
  {0.00000007895, 2.81927558645, 206.18554843720},
  {0.00000001014, 0.51187210270, 220.41264243880},
  {-66.0}
} ;

const struct VSOP_87 Saturn_R0[] =
{
  {9.55758135801, 0.00000000000, 0.00000000000},
  {0.52921382465, 2.39226219733, 213.29909543800},
  {0.01873679934, 5.23549605091, 206.18554843720},
  {0.01464663959, 1.64763045468, 426.59819087600},
  {0.00821891059, 5.93520025371, 316.39186965660},
  {0.00547506899, 5.01532628454, 103.09277421860},
  {0.00371684449, 2.27114833428, 220.41264243880},
  {0.00361778433, 3.13904303264, 7.11354700080},
  {0.00140617548, 5.70406652991, 632.78373931320},
  {0.00108974737, 3.29313595577, 110.20632121940}, //10
  {0.00069007015, 5.94099622447, 419.48464387520},
  {0.00061053350, 0.94037761156, 639.89728631400},
  {0.00048913044, 1.55733388472, 202.25339517410},
  {0.00034143794, 0.19518550682, 277.03499374140},
  {0.00032401718, 5.47084606947, 949.17560896980},
  {0.00020936573, 0.46349163993, 735.87651353180},
  {0.00020839118, 1.52102590640, 433.71173787680},
  {0.00020746678, 5.33255667599, 199.07200143640},
  {0.00015298457, 3.05943652881, 529.69096509460},
  {0.00014296479, 2.60433537909, 323.50541665740}, //20
  {0.00011993314, 5.98051421881, 846.08283475120},
  {0.00011380261, 1.73105746566, 522.57741809380},
  {0.00012884128, 1.64892310393, 138.51749687070},
  {0.00007752769, 5.85191318903, 95.97922721780},
  {0.00009796061, 5.20475863996, 1265.56747862640},
  {0.00006465967, 0.17733160145, 1052.26838318840},
  {0.00006770621, 3.00433479284, 14.22709400160},
  {0.00005850443, 1.45519636076, 415.55249061210},
  {0.00005307481, 0.59737534050, 63.73589830340},
  {0.00004695746, 2.14919036956, 227.52618943960}, //30
  {0.00004043988, 1.64010323863, 209.36694217490},
  {0.00003688132, 0.78016133170, 412.37109687440},
  {0.00003376457, 3.69528478828, 224.34479570190},
  {0.00002885348, 1.38764077631, 838.96928775040},
  {0.00002976033, 5.68467931117, 210.11770170030},
  {0.00003419551, 4.94549148887, 1581.95934828300},
  {0.00003460943, 1.85088802878, 175.16605980020},
  {0.00003400616, 0.55386747515, 350.33211960040},
  {0.00002507630, 3.53851863255, 742.99006053260},
  {0.00002448325, 6.18412386316, 1368.66025284500}, //40
  {0.00002406138, 2.96559220267, 117.31986822020},
  {0.00002881181, 0.17960757891, 853.19638175200},
  {0.00002173959, 0.01508587396, 340.77089204480},
  {0.00002024483, 5.05411271271, 11.04570026390},
  {-82.0}
} ;

const struct VSOP_87 Saturn_R1[] =
{
  {0.06182981282, 0.25843515034, 213.29909543800},
  {0.00506577574, 0.71114650941, 206.18554843720},
  {0.00341394136, 5.79635773960, 426.59819087600},
  {0.00188491375, 0.47215719444, 220.41264243880},
  {0.00186261540, 3.14159265359, 0.00000000000},
  {0.00143891176, 1.40744864239, 7.11354700080},
  {0.00049621111, 6.01744469580, 103.09277421860},
  {0.00020928189, 5.09245654470, 639.89728631400},
  {0.00019952612, 1.17560125007, 419.48464387520},
  {0.00018839639, 1.60819563173, 110.20632121940}, //10
  {0.00012892827, 5.94330258435, 433.71173787680},
  {0.00013876565, 0.75886204364, 199.07200143640},
  {0.00005396699, 1.28852405908, 14.22709400160},
  {0.00004869308, 0.86793894213, 323.50541665740},
  {0.00004247455, 0.39299384543, 227.52618943960},
  {0.00003252084, 1.25853470491, 95.97922721780},
  {0.00002856006, 2.16731405366, 735.87651353180},
  {0.00002909411, 4.60679154788, 202.25339517410},
  {0.00003081408, 3.43662557418, 522.57741809380},
  {0.00001987689, 2.45054204795, 412.37109687440}, //20
  {0.00001941309, 6.02393385142, 209.36694217490},
  {0.00001581446, 1.29191789712, 210.11770170030},
  {0.00001339511, 4.30801821806, 853.19638175200},
  {0.00001315590, 1.25296446023, 117.31986822020},
  {0.00001203085, 1.86654673794, 316.39186965660},
  {0.00001091088, 0.07527246854, 216.48048917570},
  {0.00000954403, 5.15173410519, 647.01083331480},
  {0.00000966012, 0.47991379141, 632.78373931320},
  {0.00000881827, 1.88471724478, 1052.26838318840},
  {0.00000874215, 1.40224683864, 224.34479570190}, //30
  {0.00000897512, 0.98343776092, 529.69096509460},
  {0.00000784866, 3.06377517461, 838.96928775040},
  {0.00000739892, 1.38225356694, 625.67019231240},
  {0.00000612961, 3.03307306767, 63.73589830340},
  {0.00000658210, 4.14362930980, 309.27832265580},
  {0.00000649600, 1.72489486160, 742.99006053260},
  {0.00000599236, 2.54924174765, 217.23124870110},
  {0.00000502886, 2.12958819475, 3.93215326310},
  {-82.0}
} ;

const struct VSOP_87 Saturn_R2[] =
{
  {0.00436902464, 4.78671673044, 213.29909543800},
  {0.00071922760, 2.50069994874, 206.18554843720},
  {0.00049766792, 4.97168150870, 220.41264243880},
  {0.00043220894, 3.86940443794, 426.59819087600},
  {0.00029645554, 5.96310264282, 7.11354700080},
  {0.00004141650, 4.10670940823, 433.71173787680},
  {0.00004720909, 2.47527992423, 199.07200143640},
  {0.00003789370, 3.09771025067, 639.89728631400},
  {0.00002963990, 1.37206248846, 103.09277421860},
  {0.00002556363, 2.85065721526, 419.48464387520}, //10
  {0.00002208457, 6.27588858707, 110.20632121940},
  {0.00002187621, 5.85545832218, 14.22709400160},
  {0.00001956896, 4.92448618045, 227.52618943960},
  {0.00002326801, 0.00000000000, 0.00000000000},
  {0.00000923840, 5.46392422737, 323.50541665740},
  {0.00000705936, 2.97081280098, 95.97922721780},
  {0.00000546115, 4.12854181522, 412.37109687440},
  {0.00000373838, 5.83435991809, 117.31986822020},
  {0.00000360882, 3.27703082368, 647.01083331480},
  {0.00000356350, 3.19152043942, 210.11770170030}, //20
  {0.00000390627, 4.48106176893, 216.48048917570},
  {0.00000431485, 5.17825414612, 522.57741809380},
  {0.00000325598, 2.26867601656, 853.19638175200},
  {0.00000405018, 4.17294157872, 209.36694217490},
  {0.00000204494, 0.08774848590, 202.25339517410},
  {0.00000206854, 4.02188336738, 735.87651353180},
  {0.00000178474, 4.09716541453, 440.82528487760},
  {0.00000180143, 3.59704903955, 632.78373931320},
  {0.00000153656, 3.13470530382, 625.67019231240},
  {0.00000147779, 0.13614300541, 302.16477565500}, //30
  {0.00000123189, 4.18895309647, 88.86568021700},
  {0.00000133076, 2.59350469420, 191.95845443560},
  {-82.0}
} ;

const struct VSOP_87 Saturn_R3[] =
{
  {0.00020315005, 3.02186626038, 213.29909543800},
  {0.00008923581, 3.19144205755, 220.41264243880},
  {0.00006908677, 4.35174889353, 206.18554843720},
  {0.00004087129, 4.22406927376, 7.11354700080},
  {0.00003879041, 2.01056445995, 426.59819087600},
  {0.00001070788, 4.20360341236, 199.07200143640},
  {0.00000907332, 2.28344368029, 433.71173787680},
  {0.00000606121, 3.17458570534, 227.52618943960},
  {0.00000596639, 4.13455753351, 14.22709400160},
  {0.00000483181, 1.17345973258, 639.89728631400}, //10
  {0.00000393174, 0.00000000000, 0.00000000000},
  {0.00000229472, 4.69838526383, 419.48464387520},
  {0.00000188250, 4.59003889007, 110.20632121940},
  {0.00000149508, 3.20199444400, 103.09277421860},
  {0.00000121442, 3.76831374104, 323.50541665740},
  {0.00000101215, 5.81884137755, 412.37109687440},
  {0.00000102146, 4.70974422803, 95.97922721780},
  {0.00000093078, 1.43531270909, 647.01083331480},
  {0.00000072601, 4.15395598507, 117.31986822020},
  {0.00000084347, 2.63462379693, 216.48048917570}, //20
  {0.00000062198, 2.31239345505, 440.82528487760},
  {0.00000045145, 4.37317047297, 191.95845443560},
  {0.00000049536, 2.38854232908, 209.36694217490},
  {0.00000054829, 0.30526468471, 853.19638175200},
  {0.00000040498, 1.83836569765, 302.16477565500},
  {0.00000038089, 5.94455115525, 88.86568021700},
  {0.00000032243, 4.01146349387, 21.34064100240},
  {0.00000040671, 0.68845183210, 522.57741809380},
  {-82.0}
} ;

const struct VSOP_87 Saturn_R4[] =
{
  {0.00001202050, 1.41499446465, 220.41264243880},
  {0.00000707796, 1.16153570102, 213.29909543800},
  {0.00000516121, 6.23973568330, 206.18554843720},
  {0.00000426664, 2.46924890293, 7.11354700080},
  {0.00000267736, 0.18659206741, 426.59819087600},
  {0.00000170171, 5.95926972384, 199.07200143640},
  {0.00000145113, 1.44211060143, 227.52618943960},
  {0.00000150339, 0.47970167140, 433.71173787680},
  {0.00000121033, 2.40527320817, 14.22709400160},
  {0.00000047332, 5.56857488676, 639.89728631400}, //10
  {0.00000015745, 2.90112466278, 110.20632121940},
  {0.00000016668, 0.52920774279, 440.82528487760},
  {0.00000018954, 5.85626429118, 647.01083331480},
  {0.00000014074, 1.30343550656, 412.37109687440},
  {0.00000012708, 2.09349305926, 323.50541665740},
  {0.00000014724, 0.29905316786, 419.48464387520},
  {0.00000011133, 2.46304825990, 117.31986822020},
  {0.00000011320, 0.21785507019, 95.97922721780},
  {0.00000009233, 2.28127318068, 21.34064100240},
  {0.00000009246, 1.56496312830, 88.86568021700}, //20
  {0.00000008970, 0.68301278041, 216.48048917570},
  {0.00000007674, 3.59367715368, 302.16477565500},
  {0.00000007823, 4.48688804175, 853.19638175200},
  {-82.0}
} ;

const struct VSOP_87 Saturn_R5[] =
{
  {0.00000128612, 5.91282565136, 220.41264243880},
  {0.00000032273, 0.69256228602, 7.11354700080},
  {0.00000026698, 5.91428528629, 227.52618943960},
  {0.00000019923, 0.67370653385, 14.22709400160},
  {0.00000020223, 4.95136801768, 433.71173787680},
  {0.00000013537, 1.45669521408, 199.07200143640},
  {0.00000014097, 2.67074280191, 206.18554843720},
  {0.00000013364, 4.58826996370, 426.59819087600},
  {0.00000007257, 4.62966127155, 213.29909543800},
  {0.00000004876, 3.61448275002, 639.89728631400}, //10
  {0.00000003136, 4.65661021909, 191.95845443560},
  {0.00000002917, 0.48665273315, 323.50541665740},
  {0.00000003759, 4.89624165044, 440.82528487760},
  {0.00000003303, 4.07190859545, 647.01083331480},
  {0.00000002883, 3.18003019204, 419.48464387520},
  {0.00000002338, 3.69553554327, 88.86568021700},
  {0.00000001950, 5.32729247780, 302.16477565500},
  {0.00000002052, 3.31663577368, 95.97922721780},
  {-82.0}
} ;

const struct VSOP_87 Uranus_L0[] =
{
  {5.48129294299, 0.00000000000, 0.00000000000},
  {0.09260408252, 0.89106421530, 74.78159856730},
  {0.01504247826, 3.62719262195, 1.48447270830},
  {0.00365981718, 1.89962189068, 73.29712585900},
  {0.00272328132, 3.35823710524, 149.56319713460},
  {0.00070328499, 5.39254431993, 63.73589830340},
  {0.00068892609, 6.09292489045, 76.26607127560},
  {0.00061998592, 2.26952040469, 2.96894541660},
  {0.00061950714, 2.85098907565, 11.04570026390},
  {0.00026468869, 3.14152087888, 71.81265315070}, //10
  {0.00025710505, 6.11379842935, 454.90936652730},
  {0.00021078897, 4.36059465144, 148.07872442630},
  {0.00017818665, 1.74436982544, 36.64856292950},
  {0.00014613471, 4.73732047977, 3.93215326310},
  {0.00011162535, 5.82681993692, 224.34479570190},
  {0.00010997934, 0.48865493179, 138.51749687070},
  {0.00009527487, 2.95516893093, 35.16409022120},
  {0.00007545543, 5.23626440666, 109.94568878850},
  {0.00004220170, 3.23328535514, 70.84944530420},
  {0.00004051850, 2.27754158724, 151.04766984290}, //20
  {0.00003354607, 1.06549008887, 4.45341812490},
  {0.00002926671, 4.62903695486, 9.56122755560},
  {0.00003490352, 5.48305567292, 146.59425171800},
  {0.00003144093, 4.75199307603, 77.75054398390},
  {0.00002922410, 5.35236743380, 85.82729883120},
  {0.00002272790, 4.36600802756, 70.32818044240},
  {0.00002051209, 1.51773563459, 0.11187458460},
  {0.00002148599, 0.60745800902, 38.13303563780},
  {0.00001991726, 4.92437290826, 277.03499374140},
  {0.00001376208, 2.04281409054, 65.22037101170}, //30
  {0.00001666910, 3.62744580852, 380.12776796000},
  {0.00001284183, 3.11346336879, 202.25339517410},
  {0.00001150416, 0.93344454002, 3.18139373770},
  {0.00001533223, 2.58593414266, 52.69019803950},
  {0.00001281641, 0.54269869505, 222.86032299360},
  {0.00001372100, 4.19641615561, 111.43016149680},
  {0.00001220998, 0.19901396193, 108.46121608020},
  {0.00000946195, 1.19249463066, 127.47179660680},
  {0.00001150993, 4.17898207045, 33.67961751290},
  {0.00001244342, 0.91612680579, 2.44768055480}, //40
  {0.00001072008, 0.23564502877, 62.25142559510},
  {0.00001090461, 1.77501638912, 12.53017297220},
  {0.00000707875, 5.18285226584, 213.29909543800},
  {0.00000653401, 0.96586909116, 78.71375183040},
  {0.00000627562, 0.18210181975, 984.60033162190},
  {0.00000524495, 2.01276706996, 299.12639426920},
  {0.00000559370, 3.35776737704, 0.52126486180},
  {0.00000606827, 5.43209728952, 529.69096509460},
  {0.00000404891, 5.98689011389, 8.07675484730},
  {0.00000467211, 0.41484068933, 145.10977900970}, //50
  {0.00000471288, 1.40664336447, 184.72728735580},
  {0.00000483219, 2.10553990154, 0.96320784650},
  {0.00000395614, 5.87039580949, 351.81659230870},
  {0.00000433532, 5.52142978255, 183.24281464750},
  {0.00000309885, 5.83301304674, 145.63104387150},
  {0.00000378609, 2.34975805006, 56.62235130260},
  {0.00000398996, 0.33810765436, 415.55249061210},
  {0.00000300379, 5.64353974146, 22.09140052780},
  {0.00000249229, 4.74617120584, 225.82926841020},
  {0.00000239334, 2.35045874708, 137.03302416240}, //60
  {0.00000294172, 5.83916826225, 39.61750834610},
  {0.00000216480, 4.77847481363, 340.77089204480},
  {0.00000251792, 1.63696775578, 221.37585028530},
  {0.00000219621, 1.92212987979, 67.66805156650},
  {0.00000201963, 1.29693040865, 0.04818410980},
  {0.00000224097, 0.51574863468, 84.34282612290},
  {0.00000216549, 6.14211862702, 5.93789083320},
  {0.00000222588, 2.84309380331, 0.26063243090},
  {0.00000207828, 5.58020570040, 68.84370773410},
  {0.00000187474, 1.31924326253, 0.16005869440}, //70
  {0.00000158028, 0.73811997211, 54.17467074780},
  {0.00000199146, 0.95634155010, 152.53214255120},
  {0.00000168648, 5.87874000882, 18.15924726470},
  {0.00000170300, 3.67717520688, 5.41662597140},
  {0.00000193652, 1.88800122606, 456.39383923560},
  {0.00000192998, 0.91616058506, 453.42489381900},
  {0.00000181934, 3.53624029238, 79.23501669220},
  {0.00000173145, 1.53860728054, 160.60889739850},
  {0.00000164588, 1.42379714838, 106.97674337190},
  {0.00000171968, 5.67952685533, 219.89137757700}, //80
  {0.00000162792, 3.05029377666, 112.91463420510},
  {0.00000146653, 1.26300172265, 59.80374504030},
  {0.00000139453, 5.38597723400, 32.19514480460},
  {0.00000138585, 4.25994786673, 909.81873305460},
  {0.00000143058, 1.29995487555, 35.42472265210},
  {0.00000123840, 1.37359990336, 7.11354700080},
  {0.00000104414, 5.02820888813, 0.75075952540},
  {0.00000103277, 0.68095301267, 14.97785352700},
  {0.00000094741, 0.90674090409, 74.66972398270},
  {0.00000082978, 2.92828718445, 265.98929347750}, //90
  {-76.0}
} ;

const struct VSOP_87 Uranus_L1[] =
{
  {75.02543121646, 0.00000000000, 0.00000000000},
  {0.00154458244, 5.24201658072, 74.78159856730},
  {0.00024456413, 1.71255705309, 1.48447270830},
  {0.00009257828, 0.42844639064, 11.04570026390},
  {0.00008265977, 1.50220035110, 63.73589830340},
  {0.00007841715, 1.31983607251, 149.56319713460},
  {0.00003899105, 0.46483574024, 3.93215326310},
  {0.00002283777, 4.17367533997, 76.26607127560},
  {0.00001926600, 0.53013080152, 2.96894541660},
  {0.00001232727, 1.58634458237, 70.84944530420}, //10
  {0.00000791206, 5.43641224143, 3.18139373770},
  {0.00000766954, 1.99555409575, 73.29712585900},
  {0.00000481671, 2.98401996914, 85.82729883120},
  {0.00000449798, 4.13826237508, 138.51749687070},
  {0.00000445600, 3.72300400331, 224.34479570190},
  {0.00000426554, 4.73126059388, 71.81265315070},
  {0.00000347735, 2.45372261286, 9.56122755560},
  {0.00000353752, 2.58324496886, 148.07872442630},
  {0.00000317084, 5.57855232072, 52.69019803950},
  {0.00000179920, 5.68367730922, 12.53017297220}, //20
  {0.00000171084, 3.00060075287, 78.71375183040},
  {0.00000205585, 2.36263144251, 2.44768055480},
  {0.00000158029, 2.90931969498, 0.96320784650},
  {0.00000189068, 4.20242881378, 56.62235130260},
  {0.00000154670, 5.59083925605, 4.45341812490},
  {0.00000183762, 0.28371004654, 151.04766984290},
  {0.00000143464, 2.59049246726, 62.25142559510},
  {0.00000151984, 2.94217326890, 77.75054398390},
  {0.00000153515, 4.65186885939, 35.16409022120},
  {0.00000121452, 4.14839204920, 127.47179660680}, //30
  {0.00000115546, 3.73224603791, 65.22037101170},
  {0.00000102022, 4.18754517993, 145.63104387150},
  {0.00000101718, 6.03385875009, 0.11187458460},
  {0.00000088202, 3.99035787994, 18.15924726470},
  {0.00000087549, 6.15520787584, 202.25339517410},
  {0.00000080530, 2.64124743934, 22.09140052780},
  {0.00000072047, 6.04545933578, 70.32818044240},
  {0.00000068570, 4.05071895264, 77.96299230500},
  {0.00000059173, 3.70413919082, 67.66805156650},
  {0.00000047267, 3.54312460519, 351.81659230870}, //40
  {0.00000042534, 5.72357370899, 5.41662597140},
  {0.00000044339, 5.90865821911, 7.11354700080},
  {0.00000035605, 3.29197259183, 8.07675484730},
  {0.00000035524, 3.32784616138, 71.60020482960},
  {0.00000036116, 5.89964278801, 33.67961751290},
  {0.00000030608, 5.46414592601, 160.60889739850},
  {0.00000031454, 5.62015632303, 984.60033162190},
  {0.00000038544, 4.91519003848, 222.86032299360},
  {0.00000034996, 5.08034112149, 38.13303563780},
  {0.00000030811, 5.49591403863, 59.80374504030}, //50
  {0.00000028947, 4.51867390414, 84.34282612290},
  {0.00000026627, 5.54127301037, 131.40394986990},
  {0.00000029866, 1.65980844667, 447.79581952650},
  {0.00000029206, 1.14722640419, 462.02291352810},
  {0.00000025753, 4.99362028417, 137.03302416240},
  {0.00000025373, 5.73584678604, 380.12776796000},
  {0.00000021672, 2.80556379586, 69.36497259590},
  {-76.0}
} ;

const struct VSOP_87 Uranus_L2[] =
{
  {0.00053033277, 0.00000000000, 0.00000000000},
  {0.00002357636, 2.26014661705, 74.78159856730},
  {0.00000769129, 4.52561041823, 11.04570026390},
  {0.00000551533, 3.25814281023, 63.73589830340},
  {0.00000541532, 2.27573907424, 3.93215326310},
  {0.00000529473, 4.92348433826, 1.48447270830},
  {0.00000257521, 3.69059216858, 3.18139373770},
  {0.00000238835, 5.85806638405, 149.56319713460},
  {0.00000181904, 6.21763603405, 70.84944530420},
  {0.00000049401, 6.03101301723, 56.62235130260}, //10
  {0.00000053504, 1.44225240953, 76.26607127560},
  {0.00000038222, 1.78467827781, 52.69019803950},
  {0.00000044753, 3.90904910523, 2.44768055480},
  {0.00000044530, 0.81152639478, 85.82729883120},
  {0.00000037403, 4.46228598032, 2.96894541660},
  {0.00000033029, 0.86388149962, 9.56122755560},
  {0.00000024292, 2.10702559049, 18.15924726470},
  {0.00000029423, 5.09818697708, 73.29712585900},
  {0.00000022135, 4.81730808582, 78.71375183040},
  {0.00000022491, 5.99320728691, 138.51749687070}, //20
  {0.00000017226, 2.53537183199, 145.63104387150},
  {0.00000021392, 2.39880709309, 77.96299230500},
  {0.00000020578, 2.16918786539, 224.34479570190},
  {0.00000016777, 3.46631344086, 12.53017297220},
  {0.00000012012, 0.01941361902, 22.09140052780},
  {0.00000010466, 4.45556032593, 62.25142559510},
  {0.00000011010, 0.08496274370, 127.47179660680},
  {0.00000008668, 4.25550086984, 7.11354700080},
  {0.00000010476, 5.16453084068, 71.60020482960},
  {0.00000007160, 1.24903906391, 5.41662597140}, //30
  {0.00000008387, 5.50115930045, 67.66805156650},
  {0.00000006087, 5.44611674384, 65.22037101170},
  {0.00000006013, 4.51836836347, 151.04766984290},
  {0.00000005718, 1.82933915340, 202.25339517410},
  {0.00000006109, 3.36320161279, 447.79581952650},
  {-76.0}
} ;

const struct VSOP_87 Uranus_L3[] =
{
  {0.00000120936, 0.02418789918, 74.78159856730},
  {0.00000068064, 4.12084267733, 3.93215326310},
  {0.00000052828, 2.38964061260, 11.04570026390},
  {0.00000043754, 2.95965039734, 1.48447270830},
  {0.00000045300, 2.04423798410, 3.18139373770},
  {0.00000045806, 0.00000000000, 0.00000000000},
  {0.00000024969, 4.88741307918, 63.73589830340},
  {0.00000021061, 4.54511486862, 70.84944530420},
  {0.00000019897, 2.31320314136, 149.56319713460},
  {0.00000008901, 1.57548871761, 56.62235130260}, //10
  {0.00000004271, 0.22777319552, 18.15924726470},
  {0.00000003613, 5.39244611308, 76.26607127560},
  {0.00000003488, 4.97622811775, 85.82729883120},
  {0.00000003479, 4.12969359977, 52.69019803950},
  {0.00000003572, 0.95052448578, 77.96299230500},
  {0.00000002328, 0.85770961794, 145.63104387150},
  {0.00000002696, 0.37287796344, 78.71375183040},
  {0.00000001946, 2.67997393431, 7.11354700080},
  {-76.0}
} ;

const struct VSOP_87 Uranus_L4[] =
{
  {0.00000113855, 3.14159265359, 0.00000000000},
  {0.00000005599, 4.57882424417, 74.78159856730},
  {0.00000003203, 0.34623003207, 11.04570026390},
  {0.00000001217, 3.42199121826, 56.62235130260},
  {-76.0}
} ;

const struct VSOP_87 Uranus_B0[] =
{
  {0.01346277639, 2.61877810545, 74.78159856730},
  {0.00062341405, 5.08111175856, 149.56319713460},
  {0.00061601203, 3.14159265359, 0.00000000000},
  {0.00009963744, 1.61603876357, 76.26607127560},
  {0.00009926151, 0.57630387917, 73.29712585900},
  {0.00003259455, 1.26119385960, 224.34479570190},
  {0.00002972318, 2.24367035538, 1.48447270830},
  {0.00002010257, 6.05550401088, 148.07872442630},
  {0.00001522172, 0.27960386377, 63.73589830340},
  {0.00000924055, 4.03822927853, 151.04766984290}, //10
  {0.00000760624, 6.14000431923, 71.81265315070},
  {0.00000420265, 5.21279984788, 11.04570026390},
  {0.00000430668, 3.55445034854, 213.29909543800},
  {0.00000436843, 3.38082524317, 529.69096509460},
  {0.00000522309, 3.32085194770, 138.51749687070},
  {0.00000434625, 0.34065281858, 77.75054398390},
  {0.00000462630, 0.74256727574, 85.82729883120},
  {0.00000232649, 2.25716421383, 222.86032299360},
  {0.00000215838, 1.59121704940, 38.13303563780},
  {0.00000244698, 0.78795150326, 2.96894541660}, //20
  {0.00000179935, 3.72487952673, 299.12639426920},
  {0.00000174895, 1.23550262213, 146.59425171800},
  {0.00000173667, 1.93654269131, 380.12776796000},
  {0.00000160368, 5.33635436463, 111.43016149680},
  {0.00000144064, 5.96239326415, 35.16409022120},
  {0.00000102049, 2.61876256513, 78.71375183040},
  {0.00000116363, 5.73877190007, 70.84944530420},
  {0.00000106441, 0.94103112994, 70.32818044240},
  {-66.0}
} ;

const struct VSOP_87 Uranus_B1[] =
{
  {0.00206366162, 4.12394311407, 74.78159856730},
  {0.00008563230, 0.33819986165, 149.56319713460},
  {0.00001725703, 2.12193159895, 73.29712585900},
  {0.00001368860, 3.06861722047, 76.26607127560},
  {0.00001374449, 0.00000000000, 0.00000000000},
  {0.00000399847, 2.84767037795, 224.34479570190},
  {0.00000450639, 3.77656180977, 1.48447270830},
  {0.00000307214, 1.25456766737, 148.07872442630},
  {0.00000154336, 3.78575467747, 63.73589830340},
  {0.00000110888, 5.32888676461, 138.51749687070}, //10
  {0.00000112432, 5.57299891505, 151.04766984290},
  {0.00000083493, 3.59152795558, 71.81265315070},
  {0.00000055573, 3.40135416354, 85.82729883120},
  {0.00000041377, 4.45476669141, 78.71375183040},
  {0.00000053690, 1.70455769943, 77.75054398390},
  {0.00000041912, 1.21476607434, 11.04570026390},
  {0.00000031959, 3.77446207748, 222.86032299360},
  {0.00000030297, 2.56371683644, 2.96894541660},
  {0.00000026977, 5.33695500294, 213.29909543800},
  {0.00000026222, 0.41620628369, 380.12776796000}, //20
  {-66.0}
} ;

const struct VSOP_87 Uranus_B2[] =
{
  {0.00009211656, 5.80044305785, 74.78159856730},
  {0.00000556926, 0.00000000000, 0.00000000000},
  {0.00000286265, 2.17729776353, 149.56319713460},
  {0.00000094969, 3.84237569809, 73.29712585900},
  {0.00000045419, 4.87822046064, 76.26607127560},
  {0.00000020107, 5.46264485369, 1.48447270830},
  {0.00000014793, 0.87983715652, 138.51749687070},
  {0.00000013963, 5.07234043994, 63.73589830340},
  {0.00000014261, 2.84517742687, 148.07872442630},
  {0.00000010122, 5.00290894862, 224.34479570190}, //10
  {0.00000008299, 6.26655615197, 78.71375183040},
  {-66.0}
} ;

const struct VSOP_87 Uranus_B3[] =
{
  {0.00000267832, 1.25097888291, 74.78159856730},
  {0.00000011048, 3.14159265359, 0.00000000000},
  {0.00000006154, 4.00663614486, 149.56319713460},
  {0.00000003361, 5.77804694935, 73.29712585900},
  {-66.0}
} ;

const struct VSOP_87 Uranus_B4[] =
{
  {0.00000005719, 2.85499529315, 74.78159856730},
  {-66.0}
} ;

const struct VSOP_87 Uranus_R0[] =
{
  {19.21264847881, 0.00000000000, 0.00000000000},
  {0.88784984055, 5.60377526994, 74.78159856730},
  {0.03440835545, 0.32836098991, 73.29712585900},
  {0.02055653495, 1.78295170028, 149.56319713460},
  {0.00649321851, 4.52247298119, 76.26607127560},
  {0.00602248144, 3.86003820462, 63.73589830340},
  {0.00496404171, 1.40139934716, 454.90936652730},
  {0.00338525522, 1.58002682946, 138.51749687070},
  {0.00243508222, 1.57086595074, 71.81265315070},
  {0.00190521915, 1.99809364502, 1.48447270830}, //10
  {0.00161858251, 2.79137863469, 148.07872442630},
  {0.00143705902, 1.38368574483, 11.04570026390},
  {0.00093192359, 0.17437193645, 36.64856292950},
  {0.00071424265, 4.24509327405, 224.34479570190},
  {0.00089805842, 3.66105366329, 109.94568878850},
  {0.00039009624, 1.66971128869, 70.84944530420},
  {0.00046677322, 1.39976563936, 35.16409022120},
  {0.00039025681, 3.36234710692, 277.03499374140},
  {0.00036755160, 3.88648934736, 146.59425171800},
  {0.00030348875, 0.70100446346, 151.04766984290}, //20
  {0.00029156264, 3.18056174556, 77.75054398390},
  {0.00020471584, 1.55588961500, 202.25339517410},
  {0.00025620360, 5.25656292802, 380.12776796000},
  {0.00025785805, 3.78537741503, 85.82729883120},
  {0.00022637152, 0.72519137745, 529.69096509460},
  {0.00020473163, 2.79639811626, 70.32818044240},
  {0.00017900561, 0.55455488605, 2.96894541660},
  {0.00012328151, 5.96039150918, 127.47179660680},
  {0.00014701566, 4.90434406648, 108.46121608020},
  {0.00011494701, 0.43774027872, 65.22037101170}, //30
  {0.00015502809, 5.35405037603, 38.13303563780},
  {0.00010792699, 1.42104858472, 213.29909543800},
  {0.00011696085, 3.29825599114, 3.93215326310},
  {0.00011959355, 1.75044072173, 984.60033162190},
  {0.00012896507, 2.62154018241, 111.43016149680},
  {0.00011852996, 0.99342814582, 52.69019803950},
  {0.00009111446, 4.99638600045, 62.25142559510},
  {0.00008420550, 5.25350716616, 222.86032299360},
  {0.00007449125, 0.79491905956, 351.81659230870},
  {0.00008402147, 5.03877516489, 415.55249061210}, //40
  {0.00006046370, 5.67960948357, 78.71375183040},
  {0.00005524133, 3.11499484161, 9.56122755560},
  {0.00007329454, 3.97277527840, 183.24281464750},
  {0.00005444878, 5.10575635361, 145.10977900970},
  {0.00005238103, 2.62960141797, 33.67961751290},
  {0.00004079167, 3.22064788674, 340.77089204480},
  {0.00003801606, 6.10985558505, 184.72728735580},
  {0.00003919476, 4.25015288873, 39.61750834610},
  {0.00002940492, 2.14637460319, 137.03302416240},
  {0.00003781219, 3.45840272873, 456.39383923560}, //50
  {0.00002942239, 0.42393808854, 299.12639426920},
  {0.00003686787, 2.48718116535, 453.42489381900},
  {0.00003101743, 4.14031063896, 219.89137757700},
  {0.00002962641, 0.82977991995, 56.62235130260},
  {0.00002937799, 3.67657450930, 140.00196957900},
  {0.00002865128, 0.30996903761, 12.53017297220},
  {0.00002538032, 4.85457831993, 131.40394986990},
  {0.00001962510, 5.24342224065, 84.34282612290},
  {0.00002363550, 0.44253328372, 554.06998748280},
  {-82.0}
} ;

const struct VSOP_87 Uranus_R1[] =
{
  {0.01479896370, 3.67205705317, 74.78159856730},
  {0.00071212085, 6.22601006675, 63.73589830340},
  {0.00068626972, 6.13411265052, 149.56319713460},
  {0.00020857262, 5.24625494219, 11.04570026390},
  {0.00021468152, 2.60176704270, 76.26607127560},
  {0.00024059649, 3.14159265359, 0.00000000000},
  {0.00011405346, 0.01848461561, 70.84944530420},
  {0.00007496775, 0.42360033283, 73.29712585900},
  {0.00004243800, 1.41692350371, 85.82729883120},
  {0.00003505936, 2.58354048851, 138.51749687070}, //10
  {0.00003228835, 5.25499602896, 3.93215326310},
  {0.00003926694, 3.15513991323, 71.81265315070},
  {0.00003060010, 0.15321893225, 1.48447270830},
  {0.00003578446, 2.31160668309, 224.34479570190},
  {0.00002564251, 0.98076846352, 148.07872442630},
  {0.00002429445, 3.99440122468, 52.69019803950},
  {0.00001644719, 2.65349313124, 127.47179660680},
  {0.00001583766, 1.43045619196, 78.71375183040},
  {0.00001413112, 4.57461892062, 202.25339517410},
  {0.00001489525, 2.67559167316, 56.62235130260}, //20
  {0.00001403237, 1.36985349744, 77.75054398390},
  {0.00001228220, 1.04703640149, 62.25142559510},
  {0.00001508028, 5.05996325425, 151.04766984290},
  {0.00000992085, 2.17168865909, 65.22037101170},
  {0.00001032731, 0.26459059027, 131.40394986990},
  {0.00000861867, 5.05530802218, 351.81659230870},
  {0.00000744445, 3.07640148939, 35.16409022120},
  {0.00000604362, 0.90717667985, 984.60033162190},
  {0.00000646851, 4.47290422910, 70.32818044240},
  {0.00000574710, 3.23070708457, 447.79581952650}, //30
  {0.00000687470, 2.49912565674, 77.96299230500},
  {0.00000623602, 0.86253073820, 9.56122755560},
  {0.00000527794, 5.15136007084, 2.96894541660},
  {0.00000561839, 2.71778158980, 462.02291352810},
  {0.00000530364, 5.91655309045, 213.29909543800},
  {-82.0}
} ;

const struct VSOP_87 Uranus_R2[] =
{
  {0.00022439904, 0.69953118760, 74.78159856730},
  {0.00004727037, 1.69901641488, 63.73589830340},
  {0.00001681903, 4.64833551727, 70.84944530420},
  {0.00001433755, 3.52119917947, 149.56319713460},
  {0.00001649559, 3.09660078980, 11.04570026390},
  {0.00000770188, 0.00000000000, 0.00000000000},
  {0.00000461009, 0.76676632849, 3.93215326310},
  {0.00000500429, 6.17229032223, 76.26607127560},
  {0.00000390371, 4.49605283502, 56.62235130260},
  {0.00000389945, 5.52673426377, 85.82729883120}, //10
  {0.00000292097, 0.20389012095, 52.69019803950},
  {0.00000272898, 3.84707823651, 138.51749687070},
  {0.00000286579, 3.53357683270, 73.29712585900},
  {0.00000205449, 3.24758017121, 78.71375183040},
  {0.00000219674, 1.96418942891, 131.40394986990},
  {0.00000215788, 0.84812474187, 77.96299230500},
  {0.00000128834, 2.08146849515, 3.18139373770},
  {0.00000148554, 4.89840863841, 127.47179660680},
  {-82.0}
} ;

const struct VSOP_87 Uranus_R3[] =
{
  {0.00001164382, 4.73453291602, 74.78159856730},
  {0.00000212367, 3.34255734999, 63.73589830340},
  {0.00000196408, 2.98004616318, 70.84944530420},
  {0.00000104527, 0.95807937648, 11.04570026390},
  {0.00000071681, 0.02528455665, 56.62235130260},
  {0.00000072540, 0.99701907912, 149.56319713460},
  {0.00000054875, 2.59436811267, 3.93215326310},
  {0.00000034029, 3.81553325635, 76.26607127560},
  {0.00000032081, 3.59825177840, 131.40394986990},
  {0.00000029641, 3.44111535957, 85.82729883120}, //10
  {-82.0}
} ;

const struct VSOP_87 Uranus_R4[] =
{
  {0.00000052996, 3.00838033088, 74.78159856730},
  {0.00000009887, 1.91399083603, 56.62235130260},
  {-82.0}
} ;

const struct VSOP_87 Neptune_L0[] =
{
  {5.31188633047, 0.00000000000, 0.00000000000},
  {0.01798475509, 2.90101273050, 38.13303563780},
  {0.01019727662, 0.48580923660, 1.48447270830},
  {0.00124531845, 4.83008090682, 36.64856292950},
  {0.00042064450, 5.41054991607, 2.96894541660},
  {0.00037714589, 6.09221834946, 35.16409022120},
  {0.00033784734, 1.24488865578, 76.26607127560},
  {0.00016482741, 0.00007729261, 491.55792945680},
  {0.00009198582, 4.93747059924, 39.61750834610},
  {0.00008994249, 0.27462142569, 175.16605980020}, //10
  {0.00004216235, 1.98711914364, 73.29712585900},
  {0.00003364818, 1.03590121818, 33.67961751290},
  {0.00002284800, 4.20606932559, 4.45341812490},
  {0.00001433512, 2.78340432711, 74.78159856730},
  {0.00000900240, 2.07606702418, 109.94568878850},
  {0.00000744996, 3.19032530145, 71.81265315070},
  {0.00000506206, 5.74785370252, 114.39910691340},
  {0.00000399552, 0.34972342569, 1021.24889455140},
  {0.00000345195, 3.46186210169, 41.10198105440},
  {0.00000306338, 0.49684039897, 0.52126486180}, //20
  {0.00000287322, 4.50523446022, 0.04818410980},
  {0.00000323004, 2.24815188609, 32.19514480460},
  {0.00000340323, 3.30369900416, 77.75054398390},
  {0.00000266605, 4.88932609483, 0.96320784650},
  {0.00000227079, 1.79713054538, 453.42489381900},
  {0.00000244722, 1.24693337933, 9.56122755560},
  {0.00000232887, 2.50459795017, 137.03302416240},
  {0.00000282170, 2.24565579693, 146.59425171800},
  {0.00000251941, 5.78166597292, 388.46515523820},
  {0.00000150180, 2.99706110414, 5.93789083320}, //30
  {0.00000170404, 3.32390630650, 108.46121608020},
  {0.00000151401, 2.19153094280, 33.94024994380},
  {0.00000148295, 0.85948986145, 111.43016149680},
  {0.00000118672, 3.67706204305, 2.44768055480},
  {0.00000101821, 5.70539236951, 0.11187458460},
  {0.00000097873, 2.80518260528, 8.07675484730},
  {0.00000103054, 4.40441222000, 70.32818044240},
  {0.00000103305, 0.04078966679, 0.26063243090},
  {-76.0}
} ;

const struct VSOP_87 Neptune_L1[] =
{
  {38.37687716731, 0.00000000000, 0.00000000000},
  {0.00016604187, 4.86319129565, 1.48447270830},
  {0.00015807148, 2.27923488532, 38.13303563780},
  {0.00003334701, 3.68199676020, 76.26607127560},
  {0.00001305840, 3.67320813491, 2.96894541660},
  {0.00000604832, 1.50477747549, 35.16409022120},
  {0.00000178623, 3.45318524147, 39.61750834610},
  {0.00000106537, 2.45126138334, 4.45341812490},
  {0.00000105747, 2.75479326550, 33.67961751290},
  {0.00000072684, 5.48724732699, 36.64856292950}, //10
  {0.00000057069, 5.21649804970, 0.52126486180},
  {0.00000057355, 1.85767603384, 114.39910691340},
  {0.00000035368, 4.51676827545, 74.78159856730},
  {0.00000032216, 5.90411489680, 77.75054398390},
  {0.00000029871, 3.67043294114, 388.46515523820},
  {0.00000028866, 5.16877529164, 9.56122755560},
  {0.00000028742, 5.16732589024, 2.44768055480},
  {0.00000025507, 5.24526281928, 168.05251279940},
  {-76.0}
} ;

const struct VSOP_87 Neptune_L2[] =
{
  {0.00053892649, 0.00000000000, 0.00000000000},
  {0.00000281251, 1.19084538887, 38.13303563780},
  {0.00000295693, 1.85520292248, 1.48447270830},
  {0.00000270190, 5.72143228148, 76.26607127560},
  {0.00000023023, 1.21035596452, 2.96894541660},
  {0.00000007333, 0.54033306830, 2.44768055480},
  {0.00000009057, 4.42544992035, 35.16409022120},
  {-76.0}
} ;

const struct VSOP_87 Neptune_L3[] =
{
  {0.00000031254, 0.00000000000, 0.00000000000},
  {0.00000012461, 6.04431418812, 1.48447270830},
  {0.00000014541, 1.35337075856, 76.26607127560},
  {0.00000011547, 6.11257808366, 38.13303563780},
  {-76.0}
} ;

const struct VSOP_87 Neptune_L4[] =
{
  {0.00000113998, 3.14159265359, 0.00000000000},
  {-76.0}
} ;

const struct VSOP_87 Neptune_B0[] =
{
  {0.03088622933, 1.44104372626, 38.13303563780},
  {0.00027780087, 5.91271882843, 76.26607127560},
  {0.00027623609, 0.00000000000, 0.00000000000},
  {0.00015355490, 2.52123799481, 36.64856292950},
  {0.00015448133, 3.50877080888, 39.61750834610},
  {0.00001999919, 1.50998669505, 74.78159856730},
  {0.00001967540, 4.37778195768, 1.48447270830},
  {0.00001015137, 3.21561035875, 35.16409022120},
  {0.00000605767, 2.80246601405, 73.29712585900},
  {0.00000594878, 2.12892708114, 41.10198105440}, //10
  {0.00000588805, 3.18655882497, 2.96894541660},
  {0.00000401830, 4.16883287237, 114.39910691340},
  {0.00000254333, 3.27120499438, 453.42489381900},
  {0.00000261647, 3.76722704749, 213.29909543800},
  {0.00000279964, 1.68165309699, 77.75054398390},
  {0.00000205590, 4.25652348864, 529.69096509460},
  {0.00000140455, 3.52969556376, 137.03302416240},
  {-66.0}
} ;

const struct VSOP_87 Neptune_B1[] =
{
  {0.00227279214, 3.80793089870, 38.13303563780},
  {0.00001803120, 1.97576485377, 76.26607127560},
  {0.00001385733, 4.82555548018, 36.64856292950},
  {0.00001433300, 3.14159265359, 0.00000000000},
  {0.00001073298, 6.08054240712, 39.61750834610},
  {0.00000147903, 3.85766231348, 74.78159856730},
  {0.00000136448, 0.47764957338, 1.48447270830},
  {0.00000070285, 6.18782052139, 35.16409022120},
  {0.00000051899, 5.05221791891, 73.29712585900},
  {0.00000037273, 4.89476629246, 41.10198105440}, //10
  {0.00000042568, 0.30721737205, 114.39910691340},
  {0.00000037104, 5.75999349109, 2.96894541660},
  {0.00000026399, 5.21566335936, 213.29909543800},
  {-66.0}
} ;

const struct VSOP_87 Neptune_B2[] =
{
  {0.00009690766, 5.57123750291, 38.13303563780},
  {0.00000078815, 3.62705474219, 76.26607127560},
  {0.00000071523, 0.45476688580, 36.64856292950},
  {0.00000058646, 3.14159265359, 0.00000000000},
  {0.00000029915, 1.60671721861, 39.61750834610},
  {0.00000006472, 5.60736756575, 74.78159856730},
  {-66.0}
} ;

const struct VSOP_87 Neptune_B3[] =
{
  {0.00000273423, 1.01688979072, 38.13303563780},
  {0.00000002274, 2.36805657126, 36.64856292950},
  {0.00000002029, 5.33364321342, 76.26607127560},
  {0.00000002393, 0.00000000000, 0.00000000000},
  {-66.0}
} ;

const struct VSOP_87 Neptune_B4[] =
{
  {0.00000005728, 2.66872693322, 38.13303563780},
  {-66.0}
} ;

const struct VSOP_87 Neptune_R0[] =
{
  {30.07013206102, 0.00000000000, 0.00000000000},
  {0.27062259490, 1.32999458930, 38.13303563780},
  {0.01691764281, 3.25186138896, 36.64856292950},
  {0.00807830737, 5.18592836167, 1.48447270830},
  {0.00537760613, 4.52113902845, 35.16409022120},
  {0.00495725642, 1.57105654815, 491.55792945680},
  {0.00274571970, 1.84552256801, 175.16605980020},
  {0.00135134095, 3.37220607384, 39.61750834610},
  {0.00121801825, 5.79754444303, 76.26607127560},
  {0.00100895397, 0.37702748681, 73.29712585900}, //10
  {0.00069791722, 3.79617226928, 2.96894541660},
  {0.00046687838, 5.74937810094, 33.67961751290},
  {0.00024593778, 0.50801728204, 109.94568878850},
  {0.00016939242, 1.59422166991, 71.81265315070},
  {0.00014229686, 1.07786112902, 74.78159856730},
  {0.00012011825, 1.92062131635, 1021.24889455140},
  {0.00008394731, 0.67816895547, 146.59425171800},
  {0.00007571800, 1.07149263431, 388.46515523820},
  {0.00005720852, 2.59059512267, 4.45341812490},
  {0.00004839672, 1.90685991070, 41.10198105440}, //20
  {0.00004483492, 2.90573457534, 529.69096509460},
  {0.00004270202, 3.41343865825, 453.42489381900},
  {0.00004353790, 0.67985662370, 32.19514480460},
  {0.00004420804, 1.74993796503, 108.46121608020},
  {0.00002881063, 1.98600105123, 137.03302416240},
  {0.00002635535, 3.09755943422, 213.29909543800},
  {0.00003380930, 0.84810683275, 183.24281464750},
  {0.00002878942, 3.67415901855, 350.33211960040},
  {0.00002306293, 2.80962935724, 70.32818044240},
  {0.00002530149, 5.79839567009, 490.07345674850}, //30
  {0.00002523132, 0.48630800015, 493.04240216510},
  {0.00002087303, 0.61858378281, 33.94024994380},
  {-82.0}
} ;

const struct VSOP_87 Neptune_R1[] =
{
  {0.00236338502, 0.70498011235, 38.13303563780},
  {0.00013220279, 3.32015499895, 1.48447270830},
  {0.00008621863, 6.21628951630, 35.16409022120},
  {0.00002701740, 1.88140666779, 39.61750834610},
  {0.00002153150, 5.16873840979, 76.26607127560},
  {0.00002154735, 2.09431198086, 2.96894541660},
  {0.00001463924, 1.18417031047, 33.67961751290},
  {0.00001603165, 0.00000000000, 0.00000000000},
  {0.00001135773, 3.91891199655, 36.64856292950},
  {0.00000897650, 5.24122933533, 388.46515523820}, //10
  {0.00000789908, 0.53315484580, 168.05251279940},
  {0.00000760030, 0.02051033644, 182.27960680100},
  {0.00000607183, 1.07706500350, 1021.24889455140},
  {0.00000571622, 3.40060785432, 484.44438245600},
  {0.00000560790, 2.88685815667, 498.67147645760},
  {-82.0}
} ;

const struct VSOP_87 Neptune_R2[] =
{
  {0.00004247412, 5.89910679117, 38.13303563780},
  {0.00000217570, 0.34581829080, 1.48447270830},
  {0.00000163025, 2.23872947130, 168.05251279940},
  {0.00000156285, 4.59414467342, 182.27960680100},
  {0.00000117940, 5.10295026024, 484.44438245600},
  {-82.0}
} ;

const struct VSOP_87 Neptune_R3[] =
{
  {0.00000166297, 4.55243893489, 38.13303563780},
  {-82.0}
} ;

const struct PLANET Planets[] =
{
  {"Mercury",
    {Mercury_L0, Mercury_L1, Mercury_L2, Mercury_L3, Mercury_L4, Mercury_L5},
    {Mercury_B0, Mercury_B1, Mercury_B2, Mercury_B3, Mercury_B4},
    {Mercury_R0, Mercury_R1, Mercury_R2, Mercury_R3}
  },
  {"Venus",
    {Venus_L0, Venus_L1, Venus_L2, Venus_L3, Venus_L4, Venus_L5},
    {Venus_B0, Venus_B1, Venus_B2, Venus_B3, Venus_B4},
    {Venus_R0, Venus_R1, Venus_R2, Venus_R3, Venus_R4}
  },
  {"Earth",
    {Earth_L0, Earth_L1, Earth_L2, Earth_L3, Earth_L4, Earth_L5},
    {Earth_B0, Earth_B1},
    {Earth_R0, Earth_R1, Earth_R2, Earth_R3, Earth_R4}
  },
  {"Mars",
    {Mars_L0, Mars_L1, Mars_L2, Mars_L3, Mars_L4, Mars_L5},
    {Mars_B0, Mars_B1, Mars_B2, Mars_B3, Mars_B4},
    {Mars_R0, Mars_R1, Mars_R2, Mars_R3, Mars_R4}
  },
  {"Jupiter",
    {Jupiter_L0, Jupiter_L1, Jupiter_L2, Jupiter_L3, Jupiter_L4, Jupiter_L5},
    {Jupiter_B0, Jupiter_B1, Jupiter_B2, Jupiter_B3, Jupiter_B4, Jupiter_B5},
    {Jupiter_R0, Jupiter_R1, Jupiter_R2, Jupiter_R3, Jupiter_R4, Jupiter_R5}
  },
  {"Saturn",
    {Saturn_L0, Saturn_L1, Saturn_L2, Saturn_L3, Saturn_L4, Saturn_L5},
    {Saturn_B0, Saturn_B1, Saturn_B2, Saturn_B3, Saturn_B4, Saturn_B5},
    {Saturn_R0, Saturn_R1, Saturn_R2, Saturn_R3, Saturn_R4, Saturn_R5}
  },
  {"Uranus",
    {Uranus_L0, Uranus_L1, Uranus_L2, Uranus_L3, Uranus_L4},
    {Uranus_B0, Uranus_B1, Uranus_B2, Uranus_B3, Uranus_B4},
    {Uranus_R0, Uranus_R1, Uranus_R2, Uranus_R3, Uranus_R4}
  },
  {"Neptune",
    {Neptune_L0, Neptune_L1, Neptune_L2, Neptune_L3, Neptune_L4},
    {Neptune_B0, Neptune_B1, Neptune_B2, Neptune_B3, Neptune_B4},
    {Neptune_R0, Neptune_R1, Neptune_R2, Neptune_R3}
  },
} ;
